//-*-c++-*-
//-----------------------------------------------------------------------------
// arcSynth.cpp:  
// 
// SynthFile manages the reading of synth fileset
// scripts, creates an nsr tree, and writes patterned data to optical.
// SynthFile is intended for test stimulation.
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//-----------------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msTest/arcSynth.cpp,v $
// $Revision: 1.6 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/04/30 23:31:49 $
//-----------------------------------------------------------------------------


#include "arcSynth.h"

static int printFiles( CHAR_PTR dir,TreeNode& root );
SynthObjectStatus SynthFile::object_status_;
RWBoolean ScopeTrace::enabled;

SynthFile::SynthFile (TreeDB&   tree,
               TreeNode&        root,
               hpACDev*         acdev,
               FreeSpaceAllocator& fsa,
               UINT32           sectorSize,
               CHAR_PTR         synthScriptName)
        : tree_(tree), root_(root), acdev_(acdev),
          allocator_(fsa), sectorSize_(sectorSize)
{
    // Open the script file.
    printf ("file> %s\n\n", synthScriptName);
    synthFile_ = fopen (synthScriptName, "r");
    if (synthFile_ == NULL) {
        fprintf (stderr, "Unable to open synth script file %s.\n",
                 (char*)synthScriptName);
        fprintf (stderr,
                 "SynthFile Object is now invalid, parsing will not continue.\n");
        object_status_ = synth_stat_invalid;
        // Something hangs when object status is invalidated.  Since we're interacting
        // with computer generated code here, I'm punting and exiting the whole
        // executable.  Hope this isn't causing a problem.
        fprintf (stderr, "\nThe executable will now abort (per %s, line %d).\n\n",
                 __FILE__, __LINE__);
        exit (1);
    }else
        synthin = synthFile_;

    actionRepeat_ = 0;
    parseState = parse_default;
    buildTreeRoot();
    setAttrDefaults();
    fileSize_ = 1024;

    synthExtent_ = new SynthExtent (fsa, sectorSize);

    // Debugging trace switches.  You can add on/off methods if you want.
    trace_buildComponents = FALSE;
    trace_allocDetails = FALSE;
    trace_allocations = FALSE;
    trace_file_summary = FALSE;
    trace_buildCompletions = FALSE;
    trace_memberCalls = FALSE;

    if (trace_memberCalls)
        ScopeTrace::enable();
    else
        ScopeTrace::disable();

//    synthData = new SynthData;
    erased_ = new ExtentList;
    unErased_ = new ExtentList;
}

SynthFile::~SynthFile()
{
    
//    delete synthData;
    delete synthExtent_;
    actionList_.clearAndDestroy();

    erased_->clear();
    delete erased_;
    unErased_->clear();
    delete unErased_;
}


void SynthFile::parse() {
    synthparse();
    processPriorAction();
}

// Since we grab the largest free extent from the current surface, then hack
// off the ends until we're done, usually there's some space unused.  This
// method gives it back to the freespace allocator.
void SynthFile::returnUnusedSpace()
{
   CHECK_OBJECT;

   ScopeTrace trace("returnUnusedSpace");

    if (trace_allocations) {
    fflush(stderr);
    fprintf (stderr, "\n****\n\n");
        fprintf (stderr, "\nreturning erased\n-------------------\n");
        erased_->fprint(stderr);
    }
    allocator_.erasedDealloc(sectorSize_, *erased_);

    if (trace_allocations) {
        fprintf (stderr, "\nreturning unErased\n----------------------\n");
        unErased_->fprint(stderr);
    }
    allocator_.dealloc(sectorSize_, *unErased_);
    fflush(stderr);
}

// Ahh, ideals.  In a kinder world, all my classes would have extensive
// self testing methods, to be fed into a regressive framework.  To tell
// you the truth, just about all the self testing methods you'll find
// in my current code body have degenerated into unsupported junk.
void SynthFile::test()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("test");
    
    printf ("SyntFile Test\n-----------------\n");
    printf ("testing allocator...\n");
    allocator_.test();
    
    printf ("Parsing...\n\n");
    synthparse();
    printf ("Done parsing\n");
}

// Except for the "cd" command, more than one methods are used to carry
// out a command.  dir, symlink and file are started here.  When a new
// file/symlink/dir is started, that method attempts to process the prior
// command.  In other words, command completion is delimited by either the
// next command or the end of file.  If you think on this for a moment,
// you'll see why a syntax error on a command will be reported as a
// parse error on the subsequent command.
void SynthFile::synth_dir (const char* dir)
{
    CHECK_OBJECT;
    
    ScopeTrace trace("synth_dir");
    if (trace_buildComponents)
        printf ("synth_dir %s\n", dir);

    processPriorAction();
    parseState = parse_dir;
    name_ = dir;
}
void SynthFile::synth_slink (const char* source, const char* target)
{
    CHECK_OBJECT;
    
    ScopeTrace trace("synth_slink");
    if (trace_buildComponents)
        printf ("synth_slink %s, %s\n", source, target);
    processPriorAction();
    parseState = parse_slink;
    name_ = source;
    pelt_ = target;
}

// THIS JUST USES A COPY OF THE pathToNode FUNCTION AS USED IN
// ARCBROWSE.  
void SynthFile::synth_cd (const char* dir)
{
    CHECK_OBJECT;
    
    ScopeTrace trace("synth_cd");
    if (trace_buildComponents)
        printf ("synth_cd %s\n", dir);
    processPriorAction();
    parseState = parse_cd;

    TreeNode tempNode;
    int rval = pathToNode(dir, tempNode);
    if (rval == 0) {
        curTreeDir_ = tempNode;
        curTreeDir_.refresh();
    }
}

void SynthFile::synth_file (const char* name, const char * size)
{
    CHECK_OBJECT;
    
    ScopeTrace trace("synth_file, with size");
    if (trace_buildComponents)
        printf ("synth_file %s\n", name, size);
    processPriorAction();
    parseState = parse_file;
    name_ = name;
    fileSize_ = atol(size);
    if (fileSize_ == ULONG_MAX) {
        fprintf (stderr, "Warning: Bad synth file size: %s\n", size);
        fileSize_ = 0;
    }
}
void SynthFile::synth_file (const char* name)
{
    CHECK_OBJECT;
    
    ScopeTrace trace("synth_file, without size");
    if (trace_buildComponents)
        printf ("synth_file %s\n", name);
    processPriorAction();
    parseState = parse_file;
    name_ = name;
}

// The _k and _m signatures of synth_file carry out the K(1024)
// and Meg (1024*1024) allocations.
void SynthFile::synth_file_k (const char* name, const char * size)
{
    CHECK_OBJECT;
    
    ScopeTrace trace("synth_file, with size");
    if (trace_buildComponents)
        printf ("synth_file %s\n", name, size);
    processPriorAction();
    parseState = parse_file;
    name_ = name;
    fileSize_ = atol(size);
    fileSize_ *= 1024; // K units
    if (fileSize_ == ULONG_MAX) {
        fprintf (stderr, "Warning: Bad synth file size: %s\n", size);
        fileSize_ = 0;
    }
}
void SynthFile::synth_file_m (const char* name, const char * size)
{
    CHECK_OBJECT;
    
    ScopeTrace trace("synth_file, with size");
    if (trace_buildComponents)
        printf ("synth_file %s\n", name, size);
    processPriorAction();
    parseState = parse_file;
    name_ = name;
    fileSize_ = atol(size);
    fileSize_ *= 1048576; // meg units
    if (fileSize_ == ULONG_MAX) {
        fprintf (stderr, "Warning: Bad synth file size: %s\n", size);
        fileSize_ = 0;
    }
}


void SynthFile::synth_attr (const char* attr)
{
    CHECK_OBJECT;
    
    ScopeTrace trace("synth_attr");

    if (trace_buildComponents)
        printf ("synth_attr %s\n", attr);
    switch(parseState) {
    case parse_file:
        fileAttr_.setAttributes(attr);
        break;
    case parse_dir:
        dirAttr_.setAttributes(attr);
        break;
    case parse_slink:
        linkAttr_.setAttributes(attr);
        break;
    default:
        fprintf (stderr, "Warning: attributes declared out of scope.\n");
    }
}

// Here are the bookend allocatoin methods.  A series of allocations is
// processed into the current action list, which then gets associated
// with the file in process.  It may help to think of the allocation
// stuff as a "mini" language encapsulated in the synth language.
void SynthFile::synth_begin_alloc ()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("synth_begin_alloc");
    if (trace_allocDetails)
        printf ("synth_begin_alloc\n");

    actionList_.clearAndDestroy();
    actionIdentity_ = 0;
}
void SynthFile::synth_end_alloc ()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("synth_end_alloc");
    if (trace_allocDetails)
        printf ("synth_end_alloc\n");
    insertAction();

    if (trace_allocDetails)
        printActionList();
}
void SynthFile::synth_alloc_separator()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("synth_alloc_separator");
    if (trace_allocDetails)
        printf ("synth_alloc_separator\n");
    insertAction();
}
void SynthFile::alloc_first ()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("alloc_first");
    if (trace_allocDetails)
        printf ("alloc_first\n");
    action_.actionType_ = alloc_action_first;
}
void SynthFile::alloc_last ()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("alloc_last");
    if (trace_allocDetails)
        printf ("alloc_last\n");
    action_.actionType_ = alloc_action_last;
}
void SynthFile::alloc_erased ()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("alloc_erased");
    if (trace_allocDetails)
        printf ("alloc_erased\n");
    action_.allocationType_ = allocation_erased;
}
void SynthFile::alloc_unerased ()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("alloc_unerased");
    if (trace_allocDetails)
        printf ("alloc_unerased\n");
    action_.allocationType_ = allocation_unerased;
}
void SynthFile::alloc_increment ()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("alloc_increment");
    if (trace_allocDetails)
        printf ("alloc_increment\n");
    action_.actionType_ = alloc_action_increment;
}
void SynthFile::alloc_repeat (const char* count)
{
    CHECK_OBJECT;
    
    ScopeTrace trace("alloc_repeat");
    if (trace_allocDetails)
        printf ("alloc_repeat %s\n", count);
    actionRepeat_ = atol(count);
    if (actionRepeat_ == ULONG_MAX) {
        fprintf (stderr, "Warning: Bad SynthFile repeat value: %s\n", count);
        actionRepeat_ = 0;
    }
}
void SynthFile::alloc_all ()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("alloc_all");
    if (trace_allocDetails)
        printf ("alloc_all\n");
    action_.allocateAll_ = TRUE;
}
void SynthFile::alloc_everything ()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("alloc_everything");
    if (trace_allocDetails)
        printf ("alloc_everything\n");
    action_.takeEveryThing_ = TRUE;
}
void SynthFile::alloc_sectors (const char* numSectors)
{
    CHECK_OBJECT;
    
    ScopeTrace trace("alloc_sectors");
    if (trace_allocDetails)
        printf ("alloc_sectors %s\n", numSectors);
    action_.numSectors_ = atol(numSectors);
    if (action_.numSectors_ == ULONG_MAX) {
        fprintf (stderr, "Warning: Bad SynthFile num sectors value: %s\n", numSectors);
        action_.numSectors_ = 0;
    }    
}

// Add the just defined allocation action into the current list.
void SynthFile::insertAction()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("insertAction");

    action_.identity_ = actionIdentity_++;
    if (trace_allocDetails) {
        fprintf (stderr, "file %s, insertAction inserting: ", (char*)name_);
        action_.dump();
    }
        
    AllocAction* tempAction = new AllocAction;
    *tempAction = action_;
    if (actionRepeat_ > 0)
        for (int i = 0; i < actionRepeat_; i++)
            actionList_.append(tempAction);
    else
        actionList_.append(tempAction);

    // Set defaults, for next action.
    actionRepeat_ = 0;
    action_.allocationType_ = allocation_file;
    action_.allocateAll_ = FALSE;
    action_.takeEveryThing_ = FALSE;
}

void SynthFile::setAttrDefaults() {
    CHECK_OBJECT;
    
    ScopeTrace trace("setAttrDefaults");

    dirAttr_.setAttributes("uid 0x5f4 gid 0x14 times 0x3082ab62 0x3082a65c \
0x3082a65c links 2 -1--..---------- ---RWX--R-X--R-X ");

    fileAttr_.setAttributes("uid 0x5f4 gid 0x14 times 0x3082a6b5 0x307ab789 \
0x307df98e links 1 -1-1..---------- ---R----R----R-- ");

    linkAttr_.setAttributes("uid 0x5f4 gid 0x14 times 0x3082a6bd 0x307abc09 \
0x307abc09 links 1 11--..---------- ---RWX--R-X--R-X ");
}

void SynthFile::printActionList ()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("printActionList");

    RWSlistCollectablesIterator actionIter(actionList_);
    AllocAction*            action;

    actionIter.reset();
    action = (AllocAction*) actionIter();
    while (action != NULL) {
        action->dump();
        action = (AllocAction*) actionIter();
    }
}

// Carry out the just-defined command (not alloc, but file system level).
// The methods that carry out these commands do tree population & modification.
void SynthFile::processPriorAction()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("processPriorAction");

    switch (parseState) {
    case parse_dir:
        mkdir();
        break;
    case parse_file:
        mkfile();
        break;
    case parse_slink:
        mklink();
        break;
    case parse_cd:
        // Do nothing.  It's done immediately from the parser call.
        break;
    default:
        // Do nothing.  This is a valid case at the beginning.  Other
        // cases should be caught as a syntax error.
        break;
    }
}
void  SynthFile::mkdir()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("mkdir");

    if (trace_buildCompletions)
        printf ("mkdir %s\n", (char*)name_);
    
    if (tree_.allocNode(curNode_) != NSR::error_None) {
        fprintf(stderr,"Tree allocation error, parser object is now invalid.\n");
        invalidateObject();
    }

    curNode_.attr_ = dirAttr_;
    curNode_.filename_ = name_;
    curNode_.setParent(curTreeDir_);
    curTreeDir_.addChild(curNode_);
    curNode_.sync();
    curTreeDir_.sync();
    root_.refresh();
}

void  SynthFile::mkfile() 
{
    CHECK_OBJECT;
    
    ScopeTrace trace("mkfile");

    if (trace_buildCompletions)
        printf ("mkfile %s\n", (char*)name_);

    if (tree_.allocNode(curNode_) != NSR::error_None) {
        fprintf(stderr,"Tree allocation error, parser object is now invalid.\n");
        invalidateObject();
    }

    //REMOVED: The tree should do this now on allocNode.
    // Clear the elist and plist, to avoid growing across files in swap file.
//    curNode_.plist_.clear();
//    curNode_.elist_.clear();
    
    curNode_.attr_ = fileAttr_;
    curNode_.filename_ = name_;
    curNode_.setParent(curTreeDir_);
    curTreeDir_.addChild(curNode_);
    curNode_.sync();
    curTreeDir_.sync();
    root_.refresh();

    fileData();
    curNode_.sync();
}
void  SynthFile::mklink()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("mklink");

    if (trace_buildCompletions)
        printf ("mklink %s\n", (char*)name_);

    if (tree_.allocNode(curNode_) != NSR::error_None) {
        fprintf(stderr,"Tree allocation error, parser object is now invalid.\n");
        invalidateObject();
    }

    curNode_.attr_ = linkAttr_;
    curNode_.filename_ = name_;
    formPlist ((char*)pelt_, curNode_.plist_);
    curNode_.setParent(curTreeDir_);
    curTreeDir_.addChild(curNode_);
    curNode_.sync();
    curTreeDir_.sync();
    root_.refresh();
}

void  SynthFile::buildTreeRoot()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("buildTreeRoot");

    if (tree_.allocNode(root_) != NSR::error_None) {
        fprintf(stderr,"Tree allocation error, parser object is now invalid.\n");
        invalidateObject();
    }

    root_.attr_.fileType(NSR::file_Directory);
    root_.attr_.unixMode(S_IRWXU);
    root_.attr_.uid(getuid());
    root_.attr_.gid(getgid());
    root_.attr_.linkCount(1);
    root_.attr_.accessTime(time((time_t *) NULL));
    root_.attr_.attributeTime(time((time_t *) NULL));
    root_.attr_.modificationTime(time((time_t *) NULL));
    root_.filename_ = "/";
    root_.setParent(root_);
    root_.sync();
    curTreeDir_ = root_;
    curTreeDir_.refresh();
}

//-----------------------------------------------------------------------------
// METHOD:  browseShell::pathToNode()
//          (shamelessly ripped from bsh.cpp)
//
// Parses a path string and finds the node corresponding to it.
//
// ****** This routine does NOT follow symbolic links! ******
//
// ARGUMENTS:
//
// CHAR_PTR  path  IN   The path to parse
// TreeNode& node  OUT  Node corresponding to last element of path
//
// RETURNS:
//
// Returns 0 on success, -1 if a component of the path does not exist or is
// not a directory.
//
// PRECONDITIONS:
// POSTCONDITIONS:
//-----------------------------------------------------------------------------

INT32
SynthFile::pathToNode(CHAR_PTR path,TreeNode& node)
{
    CHECK_OBJECT_RET(-1);
    
    ScopeTrace trace("pathToNode");

    TreeNode    cur,child;
    CHAR_PTR    cptr;
    char        curelt[256];
    INT32       index;

    cptr = path;
    if(*cptr == '/') {          // Absolute path
        cur = root_;
        while(*cptr == '/') cptr++;             // Removing preceding /'s
    }
    else {
        cur = curTreeDir_;
    }

    // Start consuming path elements.

    while(1) {
        index = 0;
        while((!isspace(*cptr)) && (*cptr != 0) && (*cptr != '/')) {
            curelt[index] = *cptr;
            cptr++;
            index++;
        }
        if(index == 0) {
            break;
        }
        curelt[index] = 0;              // null-terminate

        // We should now have a path element in curelt.

        if(cur.search(curelt,child) != NSR::error_None) {
            if(strcmp(curelt,".") == 0) {
                child = cur;
            }
            else {
                if(strcmp(curelt,"..") == 0) {
                    cur.getParent(child);
                }
                else {
                    return -1;
                }
            }
        }

        cur = child;
        if((*cptr == 0) || isspace(*cptr)) {            // We're done.
            break;
        }

        if(child.attr_.fileType() != NSR::file_Directory) {
            printf("Path component %s is not a directory!\n",curelt);
            return -1;
        }

        // We found a directory node, good.
        while(*cptr == '/') cptr++;             // Remove trailing /'s
    }

    node = cur;
    return 0;
}

// Stolen from arcWrite, rather than passing in a function pointer.
void SynthFile::formPlist(CHAR_PTR buf,PathElementList& plist)
{
    CHECK_OBJECT;
    
    ScopeTrace trace("formPlist");

    int         i,index,count;
    char        element[MAX_NAMELENGTH];
    PathElement pelt;

    i = 0;
    count = strlen(buf);
    while(i < count) {
        index = 0;
        if(buf[i] == '/') {                     // Rooted path
            pelt.type = NSR::path_Standard_Root;
            pelt.name = "";
            plist.append(pelt);
            i++;
            continue;
        }

        while((buf[i] != '/') && (buf[i] != 0)) {
            element[index] = buf[i];
            index++,i++;
        }
        i++;
        element[index] = 0;
        if(!element[0]) {       // Empty path element; assume we're done
            break;
        }
        if(strcmp(element,".") == 0) {          // Current dir path element
            pelt.type = NSR::path_Same;
            pelt.name = "";
            plist.append(pelt);
            continue;
        }
        if(strcmp(element,"..") == 0) {         // Parent dir path element
            pelt.type = NSR::path_Parent;
            pelt.name = "";
            plist.append(pelt);
            continue;
        }
        // If we got here, it's a regular path element.

        pelt.type = NSR::path_Name;
        pelt.name = element;
        plist.append(pelt);
    }
}

//-------------------------------------------------------------------------
// For the current file, iterate through the current allocation list until
// either the file is entirely allocated to disk, or an error occurs.
//-------------------------------------------------------------------------
void
SynthFile::fileData()
{
    CHECK_OBJECT;
    
    ScopeTrace trace("fileData");

    RWSlistCollectablesIterator actionIter(actionList_);
    AllocAction*            action;

    if (trace_allocations) {
        printf ("file %s, size %ld\n===============\n", (char*)name_, fileSize_);
        system("date");
    }
    
    // Keep track of how much file data has been allocated within the loop.
    INT32 dataCopied = 0;

    // Loop through the allocation actions until either the file is entirely
    // allocated, or an error is found.
    // Barring some error, The allocation action list is
    // iterated through repeatedly until the file is allocated.
    
    // Note that the increment surface action is performed within this loop.
    // The surface is also implicitly incremented when a sufficiently large
    // extent is not found on the current surface.  Unfortunately, this can
    // lead to more increments than the script writer intends, if the script
    // is not written with available freespace in mind.
    while (dataCopied < fileSize_ &&
           object_status_ == synth_stat_normal
           && (object_status_ == synth_stat_normal) ) {
        
        actionIter.reset();
        action = (AllocAction*) actionIter();
        while (dataCopied < fileSize_ &&
               action != NULL &&
               object_status_ == synth_stat_normal
               && (object_status_ == synth_stat_normal) ) {

            if (action->actionType_ == alloc_action_increment) {
                if (trace_allocations)
                    printf ("---- incrementing surface, in SynthFile::FileData\n");
                allocator_.incrementSurface();
            }
            
            else
                performAllocAction(*action, dataCopied);
        
            action = (AllocAction*) actionIter();
        }
    }

    if (trace_allocations)
        system("date");
}

//-------------------------------------------------------------------------
//  To clarify, here is this functions plan.
// 1) Per amount of data left uncopied, and allocation requested sectors,
//    arrive at the number of sectors to allocate.
// 2) Attempt to do the allocation.  If the allocation fails, invalidate
//    the object and return.
// 3) Optionally trim the extent length down, excluding unused portion of
//    last sector.
// 4) Assign the allocation to a file, erased or unerased space.
//-------------------------------------------------------------------------
void
SynthFile::performAllocAction (const AllocAction action, INT32& dataCopied)
{
    CHECK_OBJECT;
    
    ScopeTrace trace("performAllocAction");
    if (trace_allocations)
        action.dump();

    INT32 allocSectors;
    Extent allocExtent;
    INT32 dataUncopied, sectorsUncopied;

    // 1) Per amount of data left uncopied, and allocation requested sectors,
    //    arrive at the number of sectors to allocate.
    dataUncopied = fileSize_ - dataCopied;
    sectorsUncopied = (INT32) ceil ( ((float)dataUncopied) / (float) sectorSize_);

    // For files, don't ask for more than we can use.
    if (action.allocationType_ == allocation_file &&
        ( (action.numSectors_ * sectorSize_) > dataUncopied) )
        allocSectors = sectorsUncopied;
    else if (action.allocateAll_)
        allocSectors = sectorsUncopied;
    else
        allocSectors = action.numSectors_;

    // 2) Attempt to do the allocation.  If the allocation fails, invalidate
    //    the object and return.
    if (0 == getAllocation (action.actionType_ == alloc_action_first,
                            allocSectors, allocExtent, action.allocateAll_,
                            (action.allocationType_ == allocation_file &&
                             action.takeEveryThing_))
        )
    {
        fprintf (stderr, "\nERROR: Unable to allocate %ld sectors for file %s\n\n",
                 allocSectors, (char*)name_);
        invalidateObject();
        return;
    }
    
    // 3) Optionally trim the extent length down, excluding unused portion of
    //    last sector.
    if (!action.takeEveryThing_)
        if (allocExtent.len > dataUncopied)
            allocExtent.len = dataUncopied;

    // 4) Assign the allocation to a file, erased or unerased space.
    switch (action.allocationType_) {
    case allocation_file:
        if (trace_allocations)
            fprintf (stderr, "\nfile: %s\n", (char*)name_);
        
        writeSynthFile(allocExtent, dataCopied);
        
        // Add the extent to the tree node.
        curNode_.elist_.append(allocExtent);

        // add copied sectors to the file total.
        dataCopied += allocExtent.len;
        break;
    case allocation_erased:
        erased_->append(allocExtent);
        if (trace_allocations) {
            fprintf (stderr, "\nAdded erased\n");
            allocExtent.fprint(stderr);
            fprintf (stderr, "erased_.fprint:-----------------------\n");
            erased_->fprint(stderr);
        }
        
        break;
    case allocation_unerased:
        unErased_->append(allocExtent);
        if (trace_allocations) {
            fprintf (stderr, "\nAdded un erased\n");
            allocExtent.fprint(stderr);
            fprintf (stderr, "unerased.fprint:-----------------------\n");
            unErased_->fprint(stderr);
        }
         
        break;
    default:
        fprintf (stderr, "Error, invalid allocation action.");
        invalidateObject();
    }
}

//-------------------------------------------------------------------------
// Attempt to allocate requested sectors from a surface.  Try the current
// surface first, then the others.  If all fail, return zero allocated.
//  Special Cases:
//    allocateAll    - Use as much of the extent as can be used.
//    takeEveryThing - Ignore the file size, and allocate the entire extent.
//-------------------------------------------------------------------------
INT32
SynthFile::getAllocation (RWBoolean first, INT32 requestedSectors, Extent& extent,
                          RWBoolean allocateAll, RWBoolean takeEveryThing)
{
    CHECK_OBJECT_RET(0);
    
    ScopeTrace trace("getAllocation");

    // incrementSurface returns TRUE if we've reached the end of the list
    // and are starting back at the beginning.  If we start at the beginning
    // twice, we did not find a suitable extent.  Clearly this could be optimized,
    // but I'm not taking the extra time for test stimulation code.
    int failed = -1;
    RWBoolean found = FALSE;
    INT32 allocated = 0;

    // Attempt to allocate.  If unsuccessful, increment to the next suface and
    // try again.  After trying all sufaces and failing on each, fail.
    while (!found && (failed < 1) && (object_status_ == synth_stat_normal)) {

        // Attempt to allocate sectors from current surface.  Return value will be
        // zero if unsuccessful.
        if (first) {
            allocated = synthExtent_->head(requestedSectors, extent, allocateAll, takeEveryThing);
            if (trace_allocations)
                printf ("Requested %ld for file %s. Allocated from head: %d, %d, %d (%d sectors)\n",
                        requestedSectors, (char*)name_, extent.start.sid, extent.start.sector,
                        extent.len, extent.len / sectorSize_);
        }
        else {
            allocated = synthExtent_->tail(requestedSectors, extent, allocateAll, takeEveryThing);
            if (trace_allocations)
                printf ("Requested %ld for file %s. Allocated from tail: %d, %d, %d (%d sectors)\n",
                        requestedSectors, (char*)name_, extent.start.sid, extent.start.sector,
                        extent.len, extent.len / sectorSize_);
        }
        

        // If we got less than expected, increment surface and free
        // any allocated space, to try again.
        INT32 allocatedSectors = allocated / sectorSize_;
        if (allocatedSectors < requestedSectors) {
            if (trace_allocations)
                printf ("-----Incrementing surface, in getAllocation.\n");
            if (allocator_.incrementSurface())
                failed++;
            
            if (allocated != 0)
                allocator_.dealloc(sectorSize_, extent);
        }
        else {
            if (trace_allocations)
                printf ("allocated: %d, %d, %d\n",
                        extent.start.sid, extent.start.sector, extent.len);
            found = TRUE;
        }
        
    }

    // If no suface provides the requested sectors, shut down the object.
    if (failed > 0) {
        fprintf (stderr,
                 "SynthFile::getAllocation: A sufficiently large extent was not found.\n");
        invalidateObject();
        extent.len = 0;
        allocated = 0;
    }
    
    return allocated;
}


void
SynthFile::writeSynthFile (Extent& extent, INT32 extentOffset)
{
    CHECK_OBJECT;
    
    ScopeTrace trace("writeSynthFile");

    Extent written;
    Extent unWritten = extent;
    INT32 offset = extentOffset;
    INT32 maxIo = 262144; // MAX_IOSIZE
    
    while (unWritten.len > 0 && (object_status_ == synth_stat_normal) ) {
        
        INT32 thisChunk = (unWritten.len < maxIo ?
                           unWritten.len : maxIo);

        written = unWritten;
        written.len = thisChunk;

        unWritten.len -= thisChunk;
        unWritten.start.sector += thisChunk / sectorSize_;
        
        // Create a data buffer.
        SynthData synthData((const char*)name_,
                           (const)offset,
                           (const)written.len);
        
        if (!writeExtent(written, synthData.data())) {
            break;
        }
        
        offset += written.len;
    }
}

int
SynthFile::writeExtent (Extent& extent, unsigned char* buf)
{
    CHECK_OBJECT_RET(0);
    
    ScopeTrace trace("writeExtent");

    if (trace_allocations)
        printf ("Writing file %s: sid:%d, sector:%d, len:%d\n",
                (char*)name_, extent.start.sid, extent.start.sector, extent.len);

    // Write the data to optical via the autochanger device object.
    if(acdev_->write(extent.start.sid,
                     extent.start.sector,
                     extent.len,
                     buf) < extent.len) {
        fprintf(stderr,"Error in SynthFile::writeExtent.");
        fprintf(stderr,"Got a short count writing file %s\n",(char*)name_);
        fprintf(stderr,"acdev error is %s, aborting data copy\n",
                (acdev_->error()).message());
        invalidateObject();
        return 0;
    }
    return extent.len;
}

static int
getSymlinkPath( CHAR_PTR path,TreeNode& node )
{
    int                 i;
    int                 len;
    PathElementList     *plistp;

    // Get the PathElementList for the symbolic link

    plistp = &(node.plist_);
    if(plistp == NULL) {
        fprintf(stderr,"getSymlinkPath():  no PathElementList!\n");
        return -1;
    }

    // Now we have to parse through the PathElementList and construct a
    // standard Unix path from it.

    path[0] = 0;
    for(i = 0; i < plistp->getCount(); i++) {
        PathElement& pelt = (*plistp)[i];

        len = strlen(pelt.name);
        switch(pelt.type) {
        case NSR::path_Custom_Root:             // Root if nonzero length
            if(len == 0) {      // Reset
                path[0] = 0;
                strcat(path,"/");
            } 
            else {              // This is an unusual case!
                fprintf(stderr,"WARNING:  Symbolic link path component of \
type 1 with nonzero length!\n");
                path[0] = 0;
                strcat(path,"/");
            }
            break;
        case NSR::path_Standard_Root:           // Root
            path[0] = 0;
            strcat(path,"/");
            break;
        case NSR::path_Parent:                  // ..
            strcat(path,"../");
            break;
        case NSR::path_Same:                    // .
            strcat(path,"./");
            break;
        case NSR::path_Name:                    // Just a path element
            strcat(path,pelt.name);
            strcat(path,"/");
            break;
        default:
            fprintf(stderr,"WARNING:  Symbolic link contains unknown path \
component type!\n");
            strcat(path,"unknown_type/");
            break;
        }
    }
    len = strlen(path);
    if(path[len - 1] == '/') {          // Remove trailing /'s.
        path[len - 1] = 0;
    }
    return 0;
}

static void
printFile( CHAR_PTR path,Attributes& attr,UINT32 filesize )
{
    mode_t      mode;

    if(path == NULL) {
        printf("ERROR:  NULL path passed to printFile()\n");
        return;
    }

    // First print the mode bits.  

    mode = attr.unixMode();
    if(attr.fileType() == NSR::file_Directory) {
        printf("d");
    }
    else {
        if(attr.fileType() == NSR::file_SymbolicLink) {
            printf("l");
        }
        else {
            printf("-");
        }
    }
        
    // Print out the rwx bits.  Very kludgy, I'm sure there's a better way.

    if(mode & S_IRUSR) {
        printf("r");
    } 
    else {
        printf("-");
    }
    if(mode & S_IWUSR) {
        printf("w");
    } 
    else {
        printf("-");
    }
    switch((mode & S_IXUSR) | (mode & S_ISUID)) {
    case S_IXUSR | S_ISUID:
        printf("s");
        break;
    case S_IXUSR:
        printf("x");
        break;
    case S_ISUID:
        printf("S");
        break;
    default:
        printf("-");
        break;
    }
    if(mode & S_IRGRP) {
        printf("r");
    } 
    else {
        printf("-");
    }
    if(mode & S_IWGRP) {
        printf("w");
    } 
    else {
        printf("-");
    }
    switch((mode & S_IXGRP) | (mode & S_ISGID)) {
    case S_IXGRP | S_ISGID:
        printf("s");
        break;
    case S_IXGRP:
        printf("x");
        break;
    case S_ISGID:
        printf("S");
        break;
    default:
        printf("-");
        break;
    }
    if(mode & S_IROTH) {
        printf("r");
    } 
    else {
        printf("-");
    }
    if(mode & S_IWOTH) {
        printf("w");
    } 
    else {
        printf("-");
    }
    if(mode & S_IXOTH) {
        printf("x");
    } 
    else {
        printf("-");
    }
    
    // Now print out the length of the file, in bytes

    printf("  %8d  ",filesize);

    // Print out the modify time, in some reasonable format

    time_t mtime = (time_t) attr.modificationTime();

    struct tm *tm = localtime(&mtime);
    char date[64];
    if(tm->tm_year < 95) {                      // Horrible. . .
        strcpy(date,nl_asctime(tm,"%b %d %Y",0));
    }
    else {
        strcpy(date,nl_asctime(tm,"%b %d %H:%M",0));
    }
    // Remove trailing \n
    date[strlen(date) - 1] = 0;
    printf("%s ",date);

    // Print out the pathname.

    printf("%s\n",path);
}

static int
printFiles( CHAR_PTR dir,TreeNode& root )
{
    int                 i;
    int                 rval;
    char                path[MAX_NAMELENGTH];
    TreeNode            tnode;
    Error               err;

    // Print out information for the current file.
    if((root.filename_)[0] == 0) {      // Parent entry, skip
        return 0;
    }
    if(root.address_.sector == ROOT_ICB_ADDRESS) {
        sprintf(path,"%s",(CHAR_PTR) root.filename_);
    }
    else {
        if(strcmp(dir,"/") == 0) {
            sprintf(path,"%s%s",dir,(CHAR_PTR) root.filename_);
        }
        else {
            sprintf(path,"%s/%s",dir,(CHAR_PTR) root.filename_);
        }
    }
    UINT32 filesize;

    if(root.attr_.fileType() == NSR::file_SymbolicLink) {
        char    buffer[MAX_NAMELENGTH];

        getSymlinkPath(buffer,root);
        filesize = strlen(buffer);
    }
    else {
        filesize = root.elist_.totalBytes();
    }
    printFile(path,root.attr_,filesize);

    // If we're a directory, recur on our children.

    if(root.attr_.fileType() == NSR::file_Directory) {
        for(i = 0; i < root.getNumChildren(); i++) {
            err = root.getChild(i,tnode);
            if(err != NSR::error_None) {                // Parent entry
                continue;
            }
            rval = printFiles(path,tnode);
            if(rval) {
                return rval;
            }
        }
    }

    return 0;
}


//-------------------------------------------------------------------------
//  Grab sectors from the beginning of the current working extent.
//-------------------------------------------------------------------------
INT32 SynthExtent::head (INT32 sectors, Extent& headExtent,
                         RWBoolean allocateAll, RWBoolean takeEveryThing)
{
    ScopeTrace trace ("SynthExtent::head");

    // Make sure a sufficiently large extent is available on the current
    // surface before proceding.
    INT32 requestedBytes = sectors * sectorSize_;
    if (requestedBytes > allocator_.largestFreeExtent()) {
        if (!allocateAll) {
            fprintf (stderr,
                     "Error in SynthExtent::head, requested extent size not avaliable\n");
            return 0;
        }
    }

    // Attempt to allocate the largest extent on the current surface.
    Extent largestFreeExtent;
    if (allocator_.takeLargestFreeExtent(largestFreeExtent) == 0) {
        fprintf (stderr, "Error in SynthExtent::head: Zero sectors allocted.\n");
        return 0;
    }

    // If takeEverything, use the whole extent, otherwise remove as much as can
    // be used by the current sector request.
    if (takeEveryThing) {
        // Take the whole extent, and don't dealloc back anything.
        headExtent = largestFreeExtent;
    }
    else if (largestFreeExtent.len > requestedBytes) {

        // Take out only what is needed from the start.  The start sector of
        // the largestFreeExtent becomes 'sectors' greater.
        headExtent.start.sid = largestFreeExtent.start.sid;
        headExtent.start.sector = largestFreeExtent.start.sector;
        headExtent.len = requestedBytes;
        
        // return the rest to the allocator (start moves back 'sectors'.)
        largestFreeExtent.len -= requestedBytes;
        largestFreeExtent.start.sector += sectors;
        allocator_.dealloc (sectorSize_, largestFreeExtent);
    }
    else if (largestFreeExtent.len < requestedBytes) {
        if (!allocateAll) {
            fprintf (stderr,
                     "Error in SynthExtent::head, Less than requested sectors avail.\n");
            allocator_.dealloc (sectorSize_, largestFreeExtent);
            return 0;
        }

        headExtent = largestFreeExtent;
    }
    else {
        headExtent = largestFreeExtent;
    }
    
    INT32 size = headExtent.len;
    return size;
}

//-------------------------------------------------------------------------
//  Grab sectors from the end of the current working extent.
//-------------------------------------------------------------------------
INT32 SynthExtent::tail (INT32 sectors, Extent& tailExtent,
                         RWBoolean allocateAll, RWBoolean takeEveryThing)
{
    ScopeTrace trace("SynthExtent::tail");
    
    INT32 requestedBytes = sectors * sectorSize_;
    if (requestedBytes > allocator_.largestFreeExtent()) {
        if (!allocateAll) {
            fprintf (stderr,
                     "Error in SynthExtent::tail, requested extent size not avaliable\n");
            return 0;
        }
    }
    
    Extent largestFreeExtent;
    if (allocator_.takeLargestFreeExtent(largestFreeExtent) == 0) {
        fprintf (stderr, "Error in SynthExtent::tail: largest extent not found.\n");
        return 0;
    }
    
    if (takeEveryThing) {
        // Take the whole extent, and don't dealloc back anything.
        tailExtent = largestFreeExtent;
    }
    
    else if (largestFreeExtent.len > requestedBytes) {
        // Take out only what is needed from the end.  The start sector of the
        // tailExtent must be 'sectors' greater than the largest free extent start
        // sector.
        tailExtent.start.sid = largestFreeExtent.start.sid;
        INT32 freeSectors = largestFreeExtent.len / sectorSize_;
        tailExtent.start.sector = largestFreeExtent.start.sector + freeSectors  - sectors;
        tailExtent.len = requestedBytes;
        
        // return the rest to the allocator (start doesn't move.)
        largestFreeExtent.len -= requestedBytes;
        allocator_.dealloc (sectorSize_, largestFreeExtent);
    }
    else if (largestFreeExtent.len < requestedBytes) {
        if (!allocateAll) {
            fprintf (stderr,
                     "Error in SynthExtent::tail, requested extent size not avaliable\n");
            allocator_.dealloc (sectorSize_, largestFreeExtent);
            return 0;
        }

        tailExtent = largestFreeExtent;
    }
    else
        tailExtent = largestFreeExtent;

    INT32 size = tailExtent.len;
    return size;
}


