//-*-c++-*-
//-----------------------------------------------------------------------------
// arcSynth.h:  
// 
// This header file provides definitions and declarations for SynthFile,
// and supporting classes.  SynthFile manages the reading of synth fileset
// scripts, creates an nsr tree, and writes patterned data to optical.
// SynthFile is intended for test stimulation.
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
// CLASSES:
//      ScopeTrace   - For debugging only. Would be a good toolbox candidate.
//      SynthExtent  - Provides head/tail extent partitioning.
//      AllocAction  - Each instance defines one allocation action, such as
//                     take 20 sectors from beginning of extent. 
//      SynthFile    - Encapsulates the parser, carries out the tree
//                     population, posesses the allocation action container
//                     and drives that processing.
//
//-----------------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msTest/arcSynth.h,v $
// $Revision: 1.7 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/03/22 15:32:32 $
//-----------------------------------------------------------------------------

#ifndef _ARCSYNTH_H_                            // Allow multiple inclusion
#define _ARCSYNTH_H_
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <assert.h>
#include <dirent.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <rw/gslist.h>
#include <rw/cstring.h>
#include <rw/ctoken.h>
#include <rw/collect.h>
#include <rw/colclass.h>

#include <rw/slistcol.h>
#include "nsrLib.h"
#include "dlist.h"                      // Doubly-linked list toolkit
#include "dqueue.h"                     // Queue macro package based on dlist
#include "arcImEx.h"
#include "fsalloc.h"
#include "arcImEx.h"
#include "syndata.h"

//-----------------------------------------------------------------------------
// Class:  ScopeTrace                                   --TRP
//  Developer support class which prints a given string upon both entering and
//  exiting scope.  All class instance printing can be turn on or off.
//
// PUBLIC DATA MEMBERS:
//
// PUBLIC METHODS:
//  ScopeTrace()
//      constructor.
//  ~ScopeTrace()
//      destructor.
//  enable()
//      Turn on printing - for all instances of the class.
//  disable()
//      Turn off printing - for all instances of the class.
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// PRIVATE DATA MEMBERS:        None.
//      static RWBoolean enabled
//      RWCString tag_
//
// PRIVATE METHODS:             None.
//
// IMPLEMENTATION NOTES:        None.
//-----------------------------------------------------------------------------
class ScopeTrace
{
public:
    ScopeTrace(char* tag) {
        tag_ = tag;
        if (enabled) {
            printf (">%s>\n", (char*)tag_);
            entryTime = time(NULL);
        }
    }
    ~ScopeTrace() {
        if (enabled) {
            exitTime = time(NULL);
            printf ("<%s(%ld)<\n",
                    (char*)tag_,
                    exitTime - entryTime);
        }
        
    }
    static void enable() {
        enabled = TRUE;
    }
    static void disable() {
        enabled = FALSE;
    }
    
private:
    static RWBoolean enabled;
    RWCString tag_;
    time_t entryTime, exitTime;
};


typedef enum
{
    synth_stat_invalid,
    synth_stat_normal,
} SynthObjectStatus;

//-----------------------------------------------------------------------------
// Class: SynthExtent                                   --TRP
//  Given a reference to a free space allocator, take the largest available
//  extent for the current sid (state is in the allocator).  head and tail
//  methods are available to partition that extent.
//
// PUBLIC DATA MEMBERS:
// PUBLIC METHODS:
//  SynthExtent()
//      Constructor - Given fsalloc allocator, gets the largest extent for
//      current sid.
//  ~SynthExtent()
//      Destructor
//  head
//      Remove and provide sectors from beginning of extent.
//  tail
//      Remove and provide sectors from end of extent.
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// PRIVATE DATA MEMBERS:
//  FreeSpaceAllocator& allocator_
//  INT32               sectorSize_
//
// PRIVATE METHODS:             None.
//
// IMPLEMENTATION NOTES:        None.
//-----------------------------------------------------------------------------
class SynthExtent
{
public:
    //-------------------------------------------------------------------------
    // Construction, destruction
    //-------------------------------------------------------------------------
    SynthExtent(FreeSpaceAllocator& allocator, INT32 sectorSize)
            : allocator_(allocator), sectorSize_(sectorSize)
    {};
    ~SynthExtent(){};
    
    //-------------------------------------------------------------------------
    // Access methods.
    //-------------------------------------------------------------------------
    INT32 head (INT32 sectors, Extent& extent, RWBoolean allocateAll,
                RWBoolean takeEveryThing);
    INT32 tail (INT32 sectors, Extent& extent, RWBoolean allocateAll,
                RWBoolean takeEveryThing);
    
private:
    //-------------------------------------------------------------------------
    // Allocation state.
    //-------------------------------------------------------------------------
    FreeSpaceAllocator& allocator_;
    INT32               sectorSize_;
};

typedef enum
{
    alloc_action_invalid,
    alloc_action_first,
    alloc_action_last,
    alloc_action_increment
} AllocActionType;

typedef enum
{
    allocation_invalid,
    allocation_erased,
    allocation_unerased,
    allocation_file
} AllocationType;

//-----------------------------------------------------------------------------
// Class: AllocAction                                   --TRP
//  One Allocation action.  The allocation syntax of a synth script may
//  contain  a set of alloc actions per file.
//
// PUBLIC DATA MEMBERS:
//  INT16       identity_
//  AllocActionType     actionType_
//  AllocationType      allocationType_
//  INT32               numSectors_
//  RWBoolean           allocateAll_
//
// PUBLIC METHODS:
//  AllocAction()
//  ~AllocAction()
//  setValuesToDefault()
//  isEqual()
//  operator=()
//  dump()
//      Debugging method - prints out object contents.
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// PRIVATE DATA MEMBERS:        None.
//
// PRIVATE METHODS:             None.
//
// IMPLEMENTATION NOTES:        None.
//-----------------------------------------------------------------------------
class AllocAction : public RWCollectable
{
public:
    INT16               identity_;
    AllocActionType     actionType_;
    AllocationType      allocationType_;
    INT32               numSectors_;
    RWBoolean           allocateAll_;     //(!) Use as much of the extent as possible.
    RWBoolean           takeEveryThing_;  //(~) Take the entire extent, regardless of
                                          // file size.
    
    AllocAction() {
        setValuesToDefault();
    }
    ~AllocAction() {};

    setValuesToDefault()
    {
        identity_ = -1;
        actionType_  = alloc_action_first;
        allocationType_ = allocation_file;
        numSectors_ = 10;
        allocateAll_ = FALSE;
        takeEveryThing_ = FALSE;
    }
    
    RWBoolean           isEqual(const RWCollectable* c) const {
        return (identity_ == ((AllocAction*)c)->identity_);
    }
    void                operator=(const AllocAction& action) {
        identity_       = action.identity_;
        actionType_     = action.actionType_;
        allocationType_ = action.allocationType_;
        numSectors_     = action.numSectors_;
        allocateAll_    = action.allocateAll_;
        takeEveryThing_ = action.takeEveryThing_;
    }
    void                dump() const {
        RWCString sectors;
        if (takeEveryThing_)
            sectors = "take everything";
        else if (allocateAll_)
            sectors = "all that can be used";
        else
            sprintf ((char*)sectors, "%d", numSectors_);
        
        RWCString allocType;
        switch (allocationType_) {
        case allocation_file:
            allocType = "a file";
            break;
        case allocation_erased:
            allocType = "erased";
            break;
        case allocation_unerased:
            allocType = "unerased";
            break;
        default:
            allocType = "<INVALID>";
        }

        switch (actionType_) {
        case alloc_action_first:
            printf ("Action %d: %s sectors to %s, from first available.\n",
                    identity_, (char*)sectors, (char*)allocType);
            break;
        case alloc_action_last:
            printf ("Action %d: %s sectors to %s, from last available.\n",
                    identity_, (char*)sectors, (char*)allocType);
            break;
        case alloc_action_increment:
            printf ("Action %d: increment surface.\n",
                    identity_);
            break;
        default:
            printf ("Action %d: <INVALID ACTION>\n");
            break;
        }
    }
};



//-----------------------------------------------------------------------------
// Class: SynthFile                                     --TRP
//  Synthetic fileset creation class.  Encapsulates the synth language lex/yacc
// parser, carries out tree building, performs high level allocation control,
// writes data to optical via the autochanger device object.  SynthFile posesses
// a FreeSpaceAllocator object, a tree w/root, and an ac device.
//
// PUBLIC DATA MEMBERS:
//
// PUBLIC METHODS:
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// PRIVATE DATA MEMBERS:        None.
//
// PRIVATE METHODS:             None.
//
// IMPLEMENTATION NOTES:
// Uses a copy (gasp) of the arcBrowse pathToNode to carry out "cd" commands.
//-----------------------------------------------------------------------------
typedef enum {parse_default, parse_dir, parse_file, parse_slink, parse_cd} ParseState;
#define CHECK_OBJECT if (object_status_ != synth_stat_normal) return
#define CHECK_OBJECT_RET(A) if (object_status_ != synth_stat_normal) return (A)
extern FILE *synthin;
class SynthFile
{
public:
    
    //-------------------------------------------------------------------------
    // Construction, destruction, parse commencement.
    //-------------------------------------------------------------------------
    SynthFile (TreeDB&          tree,
               TreeNode&        root,
               hpACDev*         acdev,
               FreeSpaceAllocator& fsa,
               UINT32           sectorSize,
               CHAR_PTR         synthScriptName);
    ~SynthFile();
    
    void parse();

    //-------------------------------------------------------------------------
    // Parser related stuff.
    //-------------------------------------------------------------------------
    void synth_dir (const char* dir);
    void synth_slink (const char* source, const char* target);
    void synth_cd (const char* dir);
    void synth_file (const char* name, const char * size);
    void synth_file (const char* name);
    void synth_file_k (const char* name, const char * size);
    void synth_file_m (const char* name, const char * size);
    
    void synth_attr (const char* attr);

    // All these are used only for allocation specification, from the parser.
    void synth_begin_alloc ();
    void synth_end_alloc ();
    void synth_alloc_separator();
    void alloc_first ();
    void alloc_last ();
    void alloc_erased ();
    void alloc_unerased ();
    void alloc_increment ();
    void alloc_repeat (const char* count);
    void alloc_all ();
    void alloc_everything();
    void alloc_sectors (const char* numSectors);
        

    //-------------------------------------------------------------------------
    // Allocation Specification stuff.
    //-------------------------------------------------------------------------
    RWSlistCollectables         actionList_;
    INT32                       actionRepeat_;
    INT32                       actionIdentity_;
    void insertAction ();
    void fileData();
    ExtentList*                  erased_;
    ExtentList*                  unErased_;
    void performAllocAction (AllocAction action, INT32& dataCopied);
    INT32 getAllocation (RWBoolean first, INT32 sectors, Extent& extent,
                         RWBoolean allocateAll, RWBoolean takeEveryThing);
    int writeExtent (Extent& extent, unsigned char* buf);
    void writeSynthFile (Extent& extent, INT32 extentOffset);
    void returnUnusedSpace();
    
    //------------------------------------------------------------------------
    // Defaults - both initial and carry-over within a script.
    //-------------------------------------------------------------------------
    AllocAction                 action_;
    Attributes                  fileAttr_;
    Attributes                  dirAttr_;
    Attributes                  linkAttr_;
    void setAttrDefaults();

    INT32                       fileSize_;
    TreeNode                    curTreeDir_;
    TreeNode                    curNode_;
    INT32                       pathToNode(CHAR_PTR path,TreeNode& node);
    
    //-------------------------------------------------------------------------
    // Carry out actions defined by the parser, once actions have been
    // completely defined.
    //-------------------------------------------------------------------------
    RWCString name_;
    RWCString pelt_;
    ParseState parseState;
    void processPriorAction();
    void buildTreeRoot();
    void formPlist(CHAR_PTR buf,PathElementList& plist);
    void allocateFile(RWCollectable* action, void* owner);
    
    //-------------------------------------------------------------------------
    // Tree building.
    //-------------------------------------------------------------------------
    void mkdir ();
    void mkfile ();
    void mklink ();
    
    //-------------------------------------------------------------------------
    // Programming support
    //-------------------------------------------------------------------------
    void print();
//    void validate();
    void test();
    void printActionList ();
    RWBoolean trace_buildComponents;
    RWBoolean trace_allocDetails;
    RWBoolean trace_allocations;
    RWBoolean trace_file_summary;
    RWBoolean trace_buildCompletions;
    RWBoolean trace_memberCalls;
    
    static SynthObjectStatus   object_status_ = synth_stat_normal;
    static invalidateObject() {object_status_ = synth_stat_invalid;}
    static int objectIsValid() 
    {
        if (object_status_ == synth_stat_invalid)
            return 0;
        else
            return 1;
    }
    

    //-------------------------------------------------------------------------
    // Private data
    //-------------------------------------------------------------------------
private:
    // synthparse is implemented by yacc generated code.  The central activity
    // of this class is all driven by calling synthparse.
    int synthparse ();

    // References to things passed in.
    FreeSpaceAllocator& allocator_;
    TreeDB&             tree_;
    TreeNode&            root_;
    hpACDev*            acdev_;
    INT32               sectorSize_;
    FILE*               synthFile_;
    
    SynthExtent*         synthExtent_;
};

#endif






