#ifndef _CPIO_H_ //-*-c++-*-
#define _CPIO_H_

//----------------------------------------------------------------------
// cpio.h
// 
// This file contains definitions and declarations for a cpio-archive
// reading class.  
//
//-----------------------------------------------------------------------------
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//  CLASSES:
//   cpioHeader
//   cpioReader
//   cpioWriter
//-----------------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msTest/cpio.h,v $
// $Revision: 1.3 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 95/10/25 17:27:12 $
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Macros
//-----------------------------------------------------------------------------

#define CPIO_MAXNAME            512
#define CPIO_ARCHIVE_EMPTY      1
#define CPIO_BAD_HEADER         2

#define ASCII_MAGIC_LENGTH      6
#define ASCII_HEADER_LENGTH     76

#define CPIO_MAGIC              070707 

#define LAST_RECORD_NAME        "TRAILER!!!"

// The constant BINARY_RECORD_LEN is equal to the number of bytes following
// the magic number up to but not including the filename in the header.

#define BINARY_RECORD_LEN       24

#define CPIO_ASCII              0
#define CPIO_BINARY             1

//-----------------------------------------------------------------------------
// CLASS: cpioHeader                                            --JKH
//
// This class is a simple data structure containing the header from a cpio
// archive record.
//
// PUBLIC DATA MEMBERS:
//
// short          c_magic:  cpio record magic number
// short          c_dev:  device on which file resides
// unsigned short c_ino:  inode number of file
// unsigned short c_mode:  mode and type of file
// unsigned short c_uid:  owner's uid
// unsigned short c_gid:  file's gid
// short          c_nlink:  number of links
// short          c_rdev:  device number for special device files
// unsigned short c_mtime[2]:  modify time
// short          c_namesize:  size of name in bytes including null byte
// unsigned short c_filesize[2]:  size of file in bytes
// char           c_filename[CPIO_MAXNAME]:  null-terminated filename
//
// PUBLIC METHODS:
// 
// cpioHeader()
//      Constructor.  Does nothing.
// ~cpioHeader()
//      Destructor.  Does nothing.
// cpioHeader(const cpioHeader&)
//      Copy constructor.
// operator=()
//      Assignment operator between two cpioHeaders.
// modifyTime()
//      Returns the modify time.
// modifyTime(int)
//      Sets the modify time.
// fileSize()
//      Returns the filesize.
// fileSize(int)
//      Sets the filesize.
//
// USAGE NOTES:  None.
//-----------------------------------------------------------------------------
// COLLABORATORS:  
//    None.
//
// PRIVATE METHODS:
//    None.
//
// IMPLEMENTATION NOTES:
//
// The modify time and filesize fields are represented as arrays of two
// unsigned shorts rather than ints for purposes of word alignment and padding 
// for the case of reading binary (versus ASCII) cpio archive headers.  
// Please see the cpio(4) man page for details.
//-----------------------------------------------------------------------------

struct cpioHeader
{
    //-------------------------------------------------------------------------
    // Public Data
    //-------------------------------------------------------------------------

    short               c_magic;                                /*  0 */
    short               c_dev;                                  /*  2 */
    unsigned short      c_ino;                                  /*  4 */
    unsigned short      c_mode;                                 /*  6 */
    unsigned short      c_uid;                                  /*  8 */
    unsigned short      c_gid;                                  /* 10 */
    short               c_nlink;                                /* 12 */
    short               c_rdev;                                 /* 14 */
    unsigned short      c_mtime[2];                             /* 16 */
    short               c_namesize;                             /* 20 */
    unsigned short      c_filesize[2];                          /* 22 */
    char                c_filename[CPIO_MAXNAME];               /* 28 */

    //-------------------------------------------------------------------------
    // C++isms:  constructors, destructors, assignment operators
    //-------------------------------------------------------------------------

    cpioHeader() {}
    cpioHeader(const cpioHeader& hdr) {
        this->operator=(hdr);
    }

    ~cpioHeader() {}
    cpioHeader&         operator=(const cpioHeader& hdr);

    //-------------------------------------------------------------------------
    // Other public methods
    //-------------------------------------------------------------------------

    int                 modifyTime(void) const;
    void                modifyTime(int newMtime);
    int                 fileSize(void) const;
    void                fileSize(int newFilesize); 
}; 

//----------------------------------------------------------------------------- // CLASS: cpioReader                                            --JKH 
//
// This class represents a cpio archive, and allows easy access to the files
// in the archive in a sequential fashion.
//
// PUBLIC DATA MEMBERS:
//  
// None.
//
// PUBLIC METHODS:
//
// cpioReader()
//    Constructor.  By default, opens stdin.
// ~cpioReader()
//    Destructor.  Does nothing.
// nextFile()
//    Reads the next cpio header and fills in the user's cpioHeader.  Note
//    that this method calls skipData(), so the previous file's data, if any,
//    will be skipped.
// isValid()
//    Returns 1 if the current cpioReader has a valid open file, 0 otherwise.
// reInit()
//    Closes the current file and opens a new one.
// readData()
//    Reads file data from the archive.
// dataLeft()
//    Returns the number of bytes of file data remaining in the stream.
//
// USAGE NOTES:
//
// The normal way to use a cpioReader object to read a cpio archive is
// to instantiate one bound to the source cpio archive (whether stdin or
// a file), then iteratively call nextFile() until it returns 
// CPIO_ARCHIVE_EMPTY, calling readData() after each call to nextFile() to
// read the file data.  Pseudocode:
//
// cpioReader    cpioRdr;        // Bound to stdin
// cpioHeader    cpioHdr;
//
// while(cpioRdr.nextFile(cpioHdr) != CPIO_ARCHIVE_EMPTY) {
//     cpioRdr.readData(buffer);
//     // Use the header and data buffer for whatever purpose. . .
// }
// 
//-----------------------------------------------------------------------------
// COLLABORATORS:  
//  cpioHeader
//
// PRIVATE DATA:
//
// int  fd_        Open file descriptor to the cpio archive
// int  residue_   Bytes of file data remaining in current file
// int  odd_       Whether filesize of current file is odd (for binary records)
// int  isbinary_  Whether the current record is a binary record
//
// PRIVATE METHODS:
//
// copy()
//    Copies one cpioReader object to another.
// Read()
//    Pipe-cognizant read() utility; see implementation notes, below.
// skipData()
//    Seeks to the end of the current file's data.
//
// IMPLEMENTATION NOTES:
// 
// In the binary record case, the file data begins on an even byte boundary,
// so if the filename size is odd, it must be padded by one byte.   
// Additionally, for binary records, the headers begin on even byte boundaries,
// so if file data size is odd, it must be padded by one byte.
//
// In reading from stdin, as in a pipeline, the input stream may not yet
// be filled up with enough bytes to satisfy a given read(2) request, and
// hence a special Read() method is used to iteratively call read(2)
// to give the stream time to fill up from the pipeline.  Note that this
// implemenation is dangerous in that the process may potentially block
// forever trying to read stdin.  If stdin is empty (such as by the pipeline
// processes exiting), and the cpioReader is still trying to Read(), it will
// never exit.  In the future we may want to add a maximum retry to Read().
// Normally this shouldn't be a problem so long as the pipeline cpio has
// a TRAILER!!! record in it.
//-----------------------------------------------------------------------------

class cpioReader
{
    private:

    //-------------------------------------------------------------------------
    // Private Data
    //-------------------------------------------------------------------------

    int                 fd_;
    int                 residue_;
    int                 odd_;
    int                 isbinary_;

    //-------------------------------------------------------------------------
    // Private Methods
    //-------------------------------------------------------------------------

    void                copy(const cpioReader& rdr) { 
        fd_ = rdr.fd_;
        residue_ = rdr.residue_;
        odd_ = rdr.odd_;
        isbinary_ = rdr.isbinary_;
    }
    int                 Read(int fd,char *buf,int nbytes);
    void                skipData();

    public:

    //-------------------------------------------------------------------------
    // C++isms:  constructors, destructors, assignment operators
    //-------------------------------------------------------------------------

                        cpioReader(const char *cpioFile = NULL);
                        cpioReader(const cpioReader& rdr) { copy(rdr); }
                        ~cpioReader();
    cpioReader&         operator=(const cpioReader&rdr) {
        copy(rdr);
        return *this;
    }

    //-------------------------------------------------------------------------
    // Other public methods
    //-------------------------------------------------------------------------

    int                 nextFile(cpioHeader& hdr);
    int                 readData(char *buf,int nbytes);
    int                 reInit(const char *cpioFile);
    int                 isValid() { return (fd_ >= 0); }
    int                 dataLeft() { return residue_; }
};

//-----------------------------------------------------------------------------
// CLASS: cpioWriter                                            --JKH
//
// This class writes files to a cpio stream.
//
// PUBLIC DATA MEMBERS:
//  
// None.
//
// PUBLIC METHODS:
//
// cpioWriter()
//    Constructor.  By default, opens stdout.
// ~cpioReader()
//    Destructor.  If the current archive hasn't been closed, closes it.
// writeHeader()
//    Writes the next cpio header from the user's cpioHeader.  
// isValid()
//    Returns 1 if the current cpioWriter has a valid open file, 0 otherwise.
// reInit()
//    Closes the current file and opens a new one.
// writeData()
//    Writes file data to the archive.
// dataExpected()
//    Returns the number of bytes of file data expected for writeData()
// closeArchive()
//    Writes the trailer record and closes the archive.
//
// USAGE NOTES:
//
// The normal way to use a cpioWriter object to write a cpio archive is
// to instantiate one bound to the output file (whether stdout or
// a file), then iteratively call writeHeader() and writeData() for each
// file to be added to the archive, then calling closeArchive() when the
// last file has been added.  By default, the cpioWriter will write an
// ASCII cpio archive (for compatability).  Use the setBinary() and 
// setAscii() methods to change the mode.
//
// cpioWriter    cpioWtr;        // Bound to stdout
// cpioHeader    cpioHdr;
//
// while(getFile(cpioHdr) != SOURCE_EMPTY) {
//     cpioWtr.writeHeader(cpioHdr);
//     cpioWtr.writeData(file data);
// }
// 
//-----------------------------------------------------------------------------
// COLLABORATORS:  
//  cpioHeader
//
// PRIVATE DATA:
//
// int  fd_        Open file descriptor to the cpio archive
// int  residue_   Bytes of file data remaining in current file
// int  odd_       Whether filesize of current file is odd (for binary records)
// int  isbinary_  Whether the current archive is in binary format
// int  open_      Whether the archive is open
//
// PRIVATE METHODS:
//
// copy()
//    Copies one cpioReader object to another.
// Write()
//    Pipe-cognizant write() utility; see implementation notes, below.
//
// IMPLEMENTATION NOTES:
// 
// In the binary record case, the file data begins on an even byte boundary,
// so if the filename size is odd, it must be padded by one byte.   
// Additionally, for binary records, the headers begin on even byte boundaries,
// so if file data size is odd, it must be padded by one byte.
//
// In reading from stdin, as in a pipeline, the input stream may not yet
// be filled up with enough bytes to satisfy a given read(2) request, and
// hence a special Read() method is used to iteratively call read(2)
// to give the stream time to fill up from the pipeline.  Note that this
// implemenation is dangerous in that the process may potentially block
// forever trying to read stdin.  If stdin is empty (such as by the pipeline
// processes exiting), and the cpioReader is still trying to Read(), it will
// never exit.  In the future we may want to add a maximum retry to Read().
// Normally this shouldn't be a problem so long as the pipeline cpio has
// a TRAILER!!! record in it.
//-----------------------------------------------------------------------------

class cpioWriter
{
    private:

    //-------------------------------------------------------------------------
    // Private Data
    //-------------------------------------------------------------------------

    int                 fd_;
    int                 residue_;
    int                 odd_;
    int                 isbinary_;
    int                 open_;
    int                 count_;

    //-------------------------------------------------------------------------
    // Private Methods
    //-------------------------------------------------------------------------

    void                copy(const cpioWriter& writer) { 
        fd_ = writer.fd_;
        residue_ = writer.residue_;
        odd_ = writer.odd_;
        isbinary_ = writer.isbinary_;
        open_ = writer.open_;
    }
    int                 Write(int fd,const char *buf,int nbytes);


    //-------------------------------------------------------------------------
    // C++isms:  constructors, destructors, assignment operators
    //-------------------------------------------------------------------------

    public:

                        cpioWriter(int type = CPIO_ASCII,
                                   const char *cpioFile = NULL);
                        cpioWriter(const cpioWriter& writer) { copy(writer); }
                        ~cpioWriter();
    cpioWriter&         operator=(const cpioWriter& writer) {
        copy(writer);
        return *this;
    }

    //-------------------------------------------------------------------------
    // Other public methods
    //-------------------------------------------------------------------------

    int                 writeHeader(const cpioHeader& hdr);
    int                 writeData(const char *buf,int nbytes);
    int                 reInit(int type = CPIO_ASCII,
                               const char *cpioFile = NULL);
    int                 isValid() { return (fd_ >= 0); }
    int                 dataExpected() { return residue_; }
    int                 closeArchive(void);
};


#endif // _CPIO_H_
