//-----------------------------------------------------------------------------
// cpioRead.cpp                                                 --JKH
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//  CLASSES:
//   cpioReader
//-----------------------------------------------------------------------------
//
// $Source: /projects/HMS/nsr/udfAncil/msTest/cpioRead.cpp,v $
// $Revision: 1.3 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 95/10/25 17:27:14 $
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include "cpio.h"

//-----------------------------------------------------------------------------
// METHOD:  cpioReader::cpioReader()                            --JKH
//
// Constructor for the cpioReader class.
//
// ARGUMENTS:
//
// const char *cpioFile  IN  Pathname of cpio archive to read
//
// RETURN VALUES:   None.
//
// PRECONDITIONS:
//
// --cpioFile points to a valid pathname of a cpio archive, or is NULL,
//   which indicates that the cpioReader object is to read from stdin.
//
// POSTCONDITIONS:
//
// --The argument archive file will be opened for read.
//-----------------------------------------------------------------------------

cpioReader::cpioReader(const char *cpioFile) : fd_(-1),
                                               residue_(0),
                                               odd_(0),
                                               isbinary_(0)
{
    if(cpioFile == NULL) {                      // Read from stdin
        fd_ = fileno(stdin);
    }
    else {
        fd_ = open(cpioFile,O_RDONLY);
    }
}

//-----------------------------------------------------------------------------
// METHOD:  cpioReader::~cpioReader()                           --JKH
//
// Destructor for the cpioReader class.
//
// ARGUMENTS:  None.
//
// RETURN VALUES:   None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --The archive file descriptor, if open, will be closed.
//-----------------------------------------------------------------------------

cpioReader::~cpioReader()
{
    if(fd_ > 0) {                               // Don't close stdin
        close(fd_);
    }
}

//-----------------------------------------------------------------------------
// METHOD:  cpioReader::nextFile()                              --JKH
//
// Fills in the user's cpioHeader struct with the header of the next file
// in the archive.
//
// ARGUMENTS:
//
// cpioHeader&  hdr  OUT   User's struct to be filled in
//
// RETURN VALUES: 
//
// Returns an integer with value 0 upon success, CPIO_BAD_HEADER if there
// was an error reading the header, and CPIO_ARCHIVE_EMPTY if the last
// record has been read.
//
// PRECONDITIONS:
//
// --The cpioReader object was successfully opened.
//
// POSTCONDITIONS:
//
// --The previous file's data, if any, will be skipped.
// --The user's struct will be filled in with the next file's metadata.
// --The archive file position will point to the first byte of data.
//-----------------------------------------------------------------------------

int
cpioReader::nextFile(cpioHeader& hdr)
{
    int         rval;
    int         len;
    int         index = 0;
    int         mtime,filesize;
    char        buffer[CPIO_MAXNAME];
    short       *sptr;

    if(fd_ < 0) {
        return CPIO_ARCHIVE_EMPTY;
    }

    if(residue_ > 0) {
        skipData();
    }

    memset((char *) &hdr,0,sizeof(hdr));
    odd_ = 0;

    // First read in the magic number.  We check for a binary header, then for
    // an ASCII header.

    len = sizeof(short);
    rval = Read(fd_,buffer + index,len);
    if(rval < len) {
        if(rval < 0) {
            perror("read");
        }
        return CPIO_BAD_HEADER;
    }
    sptr = (short *) (buffer + index);
    index += rval;
    if(*sptr == CPIO_MAGIC) {                   // We have a binary header.
        isbinary_ = 1;
        hdr.c_magic = *sptr;
        len = BINARY_RECORD_LEN;
        rval = Read(fd_,buffer + index,len);
        if(rval < len) {
            if(rval < 0) {
                perror("read");
            }
            return CPIO_BAD_HEADER;
        }

        sptr = (short *) (buffer + index);
        hdr.c_dev = *sptr++;
        hdr.c_ino = *sptr++;
        hdr.c_mode = *sptr++;
        hdr.c_uid = *sptr++;
        hdr.c_gid = *sptr++;
        hdr.c_nlink = *sptr++;
        hdr.c_rdev = *sptr++;
        hdr.c_mtime[0] = *sptr++;
        hdr.c_mtime[1] = *sptr++;
        hdr.c_namesize = *sptr++;
        hdr.c_filesize[0] = *sptr++;
        hdr.c_filesize[1] = *sptr++;

        index += rval;

        // Now read the filename.

        len = ((hdr.c_namesize + 1)/2) * 2;
        rval = Read(fd_,hdr.c_filename,len);
        if(rval < len) {
            if(rval < 0) {
                perror("read");
            }
            return CPIO_BAD_HEADER;
        }

        // At this point, we've filled in the header, and should be pointing
        // to the beginning of file data.  Check to see if we just read the
        // last record.  If so, close the file descriptor.

        if(strcmp(hdr.c_filename,LAST_RECORD_NAME) == 0) {
            rval = CPIO_ARCHIVE_EMPTY;
            close(fd_);
            fd_ = -1;
        }
        else {
            rval = 0;
        }
        residue_ = hdr.fileSize();
        odd_ = residue_ % 2;
        return rval;
    }
    else {                              // We MIGHT have an ASCII header.
        len = ASCII_MAGIC_LENGTH - sizeof(short);
        rval = Read(fd_,buffer + index,len);
        if(rval < len) {
            if(rval < 0) {
                perror("read");
            }
            return CPIO_BAD_HEADER;
        }
        if(sscanf(buffer,"%06ho",&(hdr.c_magic)) != 1)
        {
            return CPIO_BAD_HEADER;
        }
        if(hdr.c_magic != CPIO_MAGIC) {
            return CPIO_BAD_HEADER;
        }

        // We have a valid ASCII header.  Keep going.

        isbinary_ = 0;

        len = ASCII_HEADER_LENGTH - ASCII_MAGIC_LENGTH;
        rval = Read(fd_,buffer,len);
        if(rval < len) {
            if(rval < 0) {
                perror("read");
            }
            return CPIO_BAD_HEADER;
        }

        buffer[rval] = 0;               // Null-terminate
        if(sscanf(buffer,"%06ho%06ho%06ho%06ho%06ho%06ho%06ho%011lo%06ho%011lo",
               &hdr.c_dev,&hdr.c_ino,&hdr.c_mode,
               &hdr.c_uid,&hdr.c_gid,&hdr.c_nlink,&hdr.c_rdev,
               &mtime,&hdr.c_namesize,&filesize) != 10)
        {
            return CPIO_BAD_HEADER;
        }
        hdr.modifyTime(mtime);
        hdr.fileSize(filesize);

        // Get the filename

        len = hdr.c_namesize;
        rval = Read(fd_,hdr.c_filename,len);
        if(rval < len) {
            if(rval < 0) {
                perror("read");
            }
            return CPIO_BAD_HEADER;
        }

        // At this point, we've filled in the header, and should be pointing
        // to the beginning of file data.  Check to see if we just read the
        // last record.  If so, close the file descriptor.

        if(strcmp(hdr.c_filename,LAST_RECORD_NAME) == 0) {
            rval = CPIO_ARCHIVE_EMPTY;
            close(fd_);
            fd_ = -1;
        }
        else {
            rval = 0;
        }
        residue_ = hdr.fileSize();
        odd_ = residue_ % 2;
        return rval;
    }

    // Should never get here.
}

//-----------------------------------------------------------------------------
// METHOD:  cpioReader::readData()                              --JKH
//
// Reads data from the file.  
//
// ARGUMENTS:
//
// char * buf     OUT  User's buffer.
// int    nbytes  IN   Maximum number of bytes to read.
//
// RETURN VALUES: 
//
// Returns the number of bytes read.
//
// PRECONDITIONS:
//
// --The cpioReader object was successfully opened.
// --The file header has already been read by a previous call to nextFile().
// --All of the file data has not already been read.
//
// POSTCONDITIONS:
//
// --The user's buffer will be filled in with file data.
// --The amount of data read will be nbytes or residue, whichever is less.
//-----------------------------------------------------------------------------

int
cpioReader::readData(char *buf,int nbytes)
{
    int         len,rval;

    if(fd_ < 0) {
        return -1;
    }
    if(nbytes < residue_) {
        len = nbytes;
    }
    else {
        len = residue_;
        if(isbinary_) {
            len += odd_;
        }
    }

    rval = Read(fd_,buf,len);
    if(rval > 0) {
        residue_ -= rval;
    }
    return rval;
}

//-----------------------------------------------------------------------------
// METHOD:  cpioReader::reInit()                                --JKH
//
// Reinitializes the reader to connect to a different file.
//
// ARGUMENTS:
//
// const char * cpioFile  IN  New file to open.
//
// RETURN VALUES: 
//
// Returns 0 on success, -1 on failure.
//
// PRECONDITIONS:
//
// --cpioFile points to a valid pathname of a cpio archive, or is NULL,
//   which indicates that the cpioReader object is to read from stdin.
//
// POSTCONDITIONS:
//
// --If the new file is opened successfully, the old one will be closed.
// --If the new file is not opened successfully, the old one will be left open.
//-----------------------------------------------------------------------------

int
cpioReader::reInit(const char *cpioFile)
{
    int         fd;

    if(cpioFile == NULL) {
        fd = fileno(stdin);
    }
    else {
        fd = open(cpioFile,O_RDONLY);
    }
    if(fd < 0) {
        return -1;
    }

    close(fd_);
    fd_ = fd;
    residue_ = 0;
    odd_ = 0;
    isbinary_ = 0;

    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  cpioReader::skipData()                              --JKH
//
// Skips over the current file's data.
//
// ARGUMENTS:  None.
// RETURN VALUES:  None.
// PRECONDITIONS:
//
// --The cpioReader object was successfully opened.
// --The file header has already been read by a previous call to nextFile().
//
// POSTCONDITIONS:
//
// --The current file's data will be skipped.
//-----------------------------------------------------------------------------

void
cpioReader::skipData()
{
    int    rval;

    if(fd_ < 0) {
        return;
    }

    if(residue_ == 0) {
        return;
    }

    if(fd_ != fileno(stdin)) {
        if(isbinary_ && odd_) {
            residue_++;
        }
        lseek(fd_,residue_,SEEK_CUR);
    }
    else {
        char    buffer[8196];

        while(residue_ > 0) {
            rval = readData(buffer,8196);
            if(rval < 0) {
                break;
            }
        }
    }
}

//-----------------------------------------------------------------------------
// METHOD:  cpioReader::Read()                                  --JKH
//
// Supplements the read(2) system call for reading from the stdin stream.
// If reading from stdin, this call iterates until all desired bytes have
// been read.  Note that this can lead to a blocked process if the stdin
// stream cannot provide as many bytes as requested.
//
// ARGUMENTS:
//
// int    fd      IN   File descriptor.
// char * buf     OUT  User's buffer.
// int    nbytes  IN   Number of bytes to read.
//
// RETURN VALUES: 
//
// Returns the number of bytes read.
//
// PRECONDITIONS:
//
// --The cpioReader object was successfully opened.
//
// POSTCONDITIONS:
//
// --The user's buffer will be filled in with data.
//-----------------------------------------------------------------------------

int
cpioReader::Read(int fd,char *buf,int nbytes)
{
    int    count = 0;
    int    rval;

    if(fd_ == fileno(stdin)) {
        while(count < nbytes) {
            rval = read(fd,buf + count,nbytes - count);
            count += rval;
        }
        return count;
    }
    else {
        return read(fd,buf,nbytes);
    }
}
