//-*-c++-*-
//-----------------------------------------------------------------------------
// fsalloc.C
//
// This file contains implementations for the FreeSpaceAllocator class and
// associated ancillary classes.
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
// CLASSES:
//
// SurfaceFreeSpaceList
// FreeSpaceAllocator
//
//-----------------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msTest/fsalloc.cpp,v $
// $Revision: 1.8 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/04/30 23:32:23 $
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <rw/gslist.h>
#include "fsalloc.h"

void
SurfaceFreeSpace::print(void)
{
    printf("SurfaceFreeSpace:  sid = %u, sector size = %u\n",sid_,sectorSize_);
    freeSpace_.print();
}


//-----------------------------------------------------------------------------
// METHOD:  SurfaceFreeSpaceList::search()                      --JKH
//
// This method searches the SurfaceFreeSpaceList for an entry with the given
// sid information.
//
// ARGS:
//
// Sid_t sid    IN      Which Sid_t
//
// RETURNS:
//
// Returns a pointer to the SurfaceFreeSpace object with the given sid
// information on success, NULL on failure.
//
// PRECONDITIONS:
//
// None.
//
// POSTCONDITIONS:
//
// See return values
//-----------------------------------------------------------------------------

SurfaceFreeSpace *
SurfaceFreeSpaceList::search(Sid_t sid) 
{
    INT32       i;

    for(i = 0; i < theList_.entries(); i++) {
        SurfaceFreeSpace        *sfsp = theList_.at(i);
        if(sfsp->sid_ == sid) {
            return sfsp;
        }
    }
    return NULL;
}

//-----------------------------------------------------------------------------
// METHOD:  SurfaceFreeSpaceList::at()                          --JKH
//
// This method returns a pointer to the item in the list at the given index.
//
// ARGS:
//
// UINT32 index  IN   Index of desired item.
//
// RETURNS:
//
// Returns a pointer to the item in the list at the given index, or NULL if the
// index is out of range.
//
// PRECONDITIONS:
//
// None.
//
// POSTCONDITIONS:
//
// See return values
//-----------------------------------------------------------------------------

SurfaceFreeSpace *
SurfaceFreeSpaceList::at(UINT32 index) 
{
    if(index >= theList_.entries()) {
        return NULL;
    }

    return theList_.at(index);
}

const SurfaceFreeSpace *
SurfaceFreeSpaceList::at(UINT32 index) const
{
    if(index >= theList_.entries()) {
        return NULL;
    }

    return theList_.at(index);
}

//-----------------------------------------------------------------------------
// METHOD:  SurfaceFreeSpaceList::append()                      --JKH
//
// Adds a copy of the argument item to the list.
//
// ARGUMENTS:
//
// const SurfaceFreeSpace&  sfs  IN  Item to add to the list
//
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --A copy of the item will be added to the list.
//-----------------------------------------------------------------------------

void
SurfaceFreeSpaceList::append(const SurfaceFreeSpace& sfs)
{
    SurfaceFreeSpace *sfsp = new SurfaceFreeSpace;

    if(sfsp == NULL) {
        return;
    }

    *sfsp = sfs;
    theList_.append(sfsp);
}

//-----------------------------------------------------------------------------
// METHOD:  SurfaceFreeSpaceList::remove()                      --JKH
//
// Removes the item specified by the given index from the list.
//
// ARGUMENTS:
//
// UINT32  index   IN   Index of item to remove.
//
// RETURNS:  None.
//
// PRECONDITIONS:
//
// --index must be in the range [0,# of items in list)
//
// POSTCONDITIONS:
//
// --The specified item will be removed and its memory freed.
//-----------------------------------------------------------------------------

void
SurfaceFreeSpaceList::remove(UINT32 index)
{
    if(index >= entries()) {
        return;
    }
    SurfaceFreeSpace *sfsp = theList_.removeReference(theList_.at(index));
    delete sfsp;
}

void 
SurfaceFreeSpaceList::print(void)
{
    INT32               i;
    SurfaceFreeSpace    *sfsptr;

    for(i = 0; i < theList_.entries(); i++) {
        sfsptr = theList_.at(i);
        sfsptr->print();
    }
}

//-----------------------------------------------------------------------------
// METHOD:  SurfaceFreeSpaceList::copy()                        --JKH
//
// Performs a deep copy of one SurfaceFreeSpaceList to another.
//
// ARGUMENTS:
//
// const SurfaceFreeSpaceList&  obj   IN  Source list for copy.
//
// RETURNS:  None.
//
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be set equal to the source list.
//-----------------------------------------------------------------------------

void
SurfaceFreeSpaceList::copy(const SurfaceFreeSpaceList& obj)
{
    if(&obj == this) {          // Trying to overwrite myself!
        return;
    }

    clear();                    // Clear myself out.

    INT32       i;

    for(i = 0; i < obj.entries(); i++) {
        const SurfaceFreeSpace *sfsp = obj.at(i);

        if(sfsp == NULL) {     // This is an error case.  Bail out.
            clear();
            return;
        }
        append(*sfsp);
    }
}


//-----------------------------------------------------------------------------
// METHOD:  SurfaceFreeSpaceList::free()                        --JKH
//
// Clears out the contents of a SurfaceFreeSpaceList.
//
// ARGUMENTS:  None.
//
// RETURNS:  None.
//
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be freed.
//-----------------------------------------------------------------------------

void
SurfaceFreeSpaceList::free()
{
    SurfaceFreeSpace *sfsp;
    INT32       i;

    for(i = 0; i < entries(); i++) {
        sfsp = at(i);
        delete sfsp;
    }
    theList_.clear();
}

//-----------------------------------------------------------------------------
// METHOD:  FreeSpaceAllocator::alloc()                         --JKH
//
// Allocates the specified number of bytes and fills in the user's Extent
// or ExtentList.  In the third signature, this method allocates ALL bytes
// associated with a given Sid_t.
//
// ARGUMENTS:
//
// UINT32       sectorSize  IN  Sector size of space to allocate
// UINT32       nbytes      IN   Number of bytes to allocate.
// Extent&      extent      OUT  Allocated extent
//
// OR
//
// UINT32       sectorSize  IN  Sector size of space to allocate
// UINT32       nbytes      IN   Number of bytes to allocate.
// ExtentList&  extList     OUT  Allocated extent list.
//
// OR
// 
// UINT32       sectorSize  IN  Sector size of space to allocate
// Sid_t        sid         IN   Sid to allocate from
// ExtentList&  extList     OUT  Allocated extent list.
//
// RETURNS:
//
// Returns 0 on success, -1 if the allocation cannot be fulfilled.
//
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The specified number of bytes will be allocated and filled into the
//   user's Extent or ExtentList.
//
// IMPLEMENTATION NOTES:
//
// In the (UINT32,Extent&) signature, this method will only succeed if the
// requested number of bytes can be allocated contiguously.  In the
// (UINT32,ExtentList&) signature, the method will attempt to keep all of
// the allocated extents on one surface if possible.
//-----------------------------------------------------------------------------

INT32
FreeSpaceAllocator::alloc(UINT32 sectorSize,UINT32 nbytes,Extent& extent)
{
    INT32               i,j;
    UINT32              nsectors,ss;
    Extent              *extp;
    SurfaceFreeSpace    *sfsp;

    needToConsolidateSids_ = TRUE;

    // First find an extent list that has at least nbytes free bytes

    for(i = 0; i < sfsList_.entries(); i++) {
        sfsp = sfsList_.at(i);

        if(sfsp->sectorSize_ != sectorSize) {
            continue;
        }
        if(sfsp->freeSpace_.totalBytes() >= nbytes) {
            ss = sfsp->sectorSize_;

            // Go through the list and find an extent >= nbytes.

            for(j = 0; j < sfsp->freeSpace_.entries(); j++) {
                extp = sfsp->freeSpace_.at(j);
                if(extp->len >= nbytes) {
                    nsectors = ((nbytes + ss - 1) / ss);
                    extent.start = extp->start;
                    extent.len = nbytes;

                    extp->start.sector += nsectors;
                    extp->len -= nsectors*ss;

                    return 0;
                }
            }
        }
    }

    // If we got here, we couldn't find an extent big enough.

    return -1;
}

INT32
FreeSpaceAllocator::alloc(UINT32 sectorSize,UINT32 nbytes,ExtentList& extList)
{
    INT32               i,j;
    UINT32              residue;
    UINT32              nsectors,ss;
    Extent              *extp;
    Extent              extent;
    SurfaceFreeSpace    *sfsp;

    needToConsolidateSids_ = TRUE;

    if(totalBytes() < nbytes) {
        return 0;
    }

    // First try to find a list big enough.  If possible, we want to make
    // the allocations on one surface.

    for(i = 0; i < sfsList_.entries(); i++) {
        sfsp = sfsList_.at(i);

        if(sfsp->sectorSize_ != sectorSize) {
            continue;
        }

        if(sfsp->freeSpace_.totalBytes() >= nbytes) {
            ss = sfsp->sectorSize_;
            residue = nbytes;

            // Just start looping, copying extents until we've allocated
            // enough bytes.

            for(j = 0; j < sfsp->freeSpace_.entries(); j++) {
                extp = sfsp->freeSpace_.at(j);

                if(extp->len > residue) {
                    extent.start = extp->start;
                    nsectors = ((residue + ss - 1) / ss);
                    extent.len = residue;

                    extList.append(extent);
                    extp->start.sector += nsectors;
                    extp->len -= nsectors*ss;

                    residue = 0;
                    break;
                }
                else {                          // Just copy the extent
                    extList.append(*extp);
                    residue -= extp->len;
                    sfsp->freeSpace_.remove(j);
                    j--;
                }
            }

            // At this point, residue had better be 0!

            assert(residue == 0);
            return 0;
        }
    }

    // If we got here, we just start snarfing extents until we've had enough.

    residue = nbytes;
    extList.clear();
    for(i = 0; i < sfsList_.entries(); i++) {
        sfsp = sfsList_.at(i);

        if(sfsp->sectorSize_ != sectorSize) {
            continue;
        }
        ss = sfsp->sectorSize_;

        // Just start looping, copying extents until we've allocated
        // enough bytes.

        for(j = 0; j < sfsp->freeSpace_.entries(); j++) {
            extp = sfsp->freeSpace_.at(j);

            if(extp->len > residue) {
                extent.start = extp->start;
                nsectors = ((residue + ss - 1) / ss);
                extent.len = nsectors*ss;

                extList.append(extent);
                extp->start.sector += nsectors;
                extp->len -= nsectors*ss;

                residue = 0;
                break;
            }
            else {                              // Just copy the extent
                extList.append(*extp);
                residue -= extp->len;
                sfsp->freeSpace_.remove(j);
                j--;
            }
        }

        if(residue == 0) { 
            sfsList_.remove(i);
            i--;
            return 0;
        }
    }

    // If we got here, something's seriously wrong!

    return -1;
}

INT32
FreeSpaceAllocator::alloc(Sid_t sid,ExtentList& extList)
{
    INT32               i;
    SurfaceFreeSpace    *sfsp = sfsList_.search(sid);

    needToConsolidateSids_ = TRUE;

    if(sfsp == NULL) {
        return -1;
    }

    
    extList.clear();
    for(i = 0; i < sfsp->freeSpace_.entries(); i++) {
        extList.append(*(sfsp->freeSpace_.at(i)));
    }
    sfsp->freeSpace_.clear();
    for(i = 0; i < sfsList_.entries(); i++) {
        sfsp = sfsList_.at(i);
        if(sfsp->sid_ == sid) {
            sfsList_.remove(i);
            break;
        }
    }
    return 0;
}

INT32
FreeSpaceAllocator::erasedAlloc(Sid_t sid,ExtentList& extList)
{
    INT32               i;
    SurfaceFreeSpace    *sesp = erasedList_.search(sid);

    needToConsolidateSids_ = TRUE;

    if(sesp == NULL) {
        return -1;
    }

    extList.clear();
    for(i = 0; i < sesp->freeSpace_.entries(); i++) {
        extList.append(*(sesp->freeSpace_.at(i)));
    }
    sesp->freeSpace_.clear();
    for(i = 0; i < sfsList_.entries(); i++) {
        sesp = sfsList_.at(i);
        if(sesp->sid_ == sid) {
            erasedList_.remove(i);
            break;
        }
    }
    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  FreeSpaceAllocator::dealloc()                       --JKH
//
// Deallocates the specified extent(s) and marks the bytes as free.
//
// ARGUMENTS:
//
// UINT32               sectorSize  IN  Sector size of extent
// const Extent&        extent      IN  Extent to deallocate
//
// OR
//
// UINT32               sectorSize  IN  Sector size of extent
// const ExtentList&    extList     IN  Extent list to deallocate
//
// RETURNS:
//
// Returns 0 on success, -1 if the deallocation fails.
//
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The argument extent(s) will be marked as free and ready for subsequent
//   allocations.
//
// IMPLEMENTATION NOTES:
//-----------------------------------------------------------------------------

INT32
FreeSpaceAllocator::dealloc(UINT32 sectorSize,const Extent& extent)
{
    INT32               i;
    SurfaceFreeSpace    *sfsp;
    Extent              *extp;

//    needToConsolidateSids_ = TRUE;

    // Find the correct list, if there is one for that surface.

    sfsp = sfsList_.search(extent.start.sid);
    if(sfsp != NULL) {
        for(i = 0; i < sfsp->freeSpace_.entries(); i++) {
            extp = sfsp->freeSpace_.at(i);
            if(extent.start.sector < extp->start.sector) {
                if(i == 0) {
                    sfsp->freeSpace_.insertAt(0,extent);
                }
                else {
                    sfsp->freeSpace_.insertAt(i - 1,extent);
                }
                return 0;
            }
        }
        sfsp->freeSpace_.append(extent);
        return 0;
    }

    // We have to create a list.

    sfsp = new SurfaceFreeSpace;
    if(sfsp == NULL) {
        return -1;
    }

    sfsp->freeSpace_.append(extent);
    sfsp->sid_ = extent.start.sid;
    sfsp->sectorSize_ = sectorSize;
    sfsList_.append(*sfsp);
    delete sfsp;

    return 0;
}

INT32
FreeSpaceAllocator::dealloc(UINT32 sectorSize,const ExtentList& extList)
{
    INT32       i;
    INT32       rval;

//    needToConsolidateSids_ = TRUE;

    for(i = 0; i < extList.entries(); i++) {
        rval = dealloc(sectorSize,*(extList.at(i)));
        if(rval) {
            return rval;
        }
    }
    return 0;
}

INT32
FreeSpaceAllocator::erasedDealloc(UINT32 sectorSize,const Extent& extent)
{
    INT32               i;
    SurfaceFreeSpace    *sesp;
    Extent              *extp;

//    needToConsolidateSids_ = TRUE;

    // Find the correct list, if there is one for that surface.

    sesp = erasedList_.search(extent.start.sid);
    if(sesp != NULL) {
        for(i = 0; i < sesp->freeSpace_.entries(); i++) {
            extp = sesp->freeSpace_.at(i);
            if(extent.start.sector < extp->start.sector) {
                if(i == 0) {
                    sesp->freeSpace_.insertAt(0,extent);
                }
                else {
                    sesp->freeSpace_.insertAt(i - 1,extent);
                }
                return 0;
            }
        }
        sesp->freeSpace_.append(extent);
        return 0;
    }

    // We have to create a list.

    sesp = new SurfaceFreeSpace;
    if(sesp == NULL) {
        return -1;
    }

    sesp->freeSpace_.append(extent);
    sesp->sid_ = extent.start.sid;
    sesp->sectorSize_ = sectorSize;
    erasedList_.append(*sesp);
    delete sesp;

    return 0;
}

INT32
FreeSpaceAllocator::erasedDealloc(UINT32 sectorSize,const ExtentList& extList)
{
    INT32       i;
    INT32       rval;

//    needToConsolidateSids_ = TRUE;

    for(i = 0; i < extList.entries(); i++) {
        rval = erasedDealloc(sectorSize,*(extList.at(i)));
        if(rval) {
            return rval;
        }
    }
    return 0;
}



//-----------------------------------------------------------------------------
// METHOD:  FreeSpaceAllocator::largestFreeExtent()             --JKH
//
// Returns the size of the largest free extent.
//
// ARGUMENTS:  None.
//
// RETURNS:  See above.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:  None.
//-----------------------------------------------------------------------------

UINT32
FreeSpaceAllocator::largestFreeExtent(void)
{
    UINT32              largest = 0;
    INT32               i,j;
    SurfaceFreeSpace    *sfsptr;
    Extent              *extptr;

    for(i = 0; i < sfsList_.entries(); i++) {
        sfsptr = sfsList_.at(i);
        for(j = 0; j < sfsptr->freeSpace_.entries(); j++) {
            extptr = sfsptr->freeSpace_.at(j);
            if(extptr->len > largest) {
                largest = extptr->len;
            }
        }
    }
    return largest;
}
UINT32
FreeSpaceAllocator::takeLargestFreeExtent(Extent& extent)
{
    if (needToConsolidateSids_) {
        consolidateSids();
        needToConsolidateSids_ = FALSE;
    }

    return takeLargestFreeExtent(curSurfaceIndex_, extent);
}

UINT32
FreeSpaceAllocator::takeLargestFreeExtent(INT32 surfaceIndex, Extent& extent)
{
    UINT32              sizeOfLargest = 0;
    INT32               i,j;
    SurfaceFreeSpace    *sfsptr;
    Extent              *extptr;
    UINT32              largestIndex = 0;

    if (needToConsolidateSids_) {
        consolidateSids();
        needToConsolidateSids_ = FALSE;
    }

    if (surfaceIndex < 0 ||
        surfaceIndex >= sfsList_.entries()) {
        fprintf (stderr, "Error in FreeSpaceAllocator::takeLargestFreeExtent:\n");
        fprintf (stderr, "\tattempt to access an invalid surface (%ld of %ld).\n",
                surfaceIndex, sfsList_.entries());
        return 0;
    }
    
    sfsptr = sfsList_.at(surfaceIndex);
    for(j = 0; j < sfsptr->freeSpace_.entries(); j++) {
        extptr = sfsptr->freeSpace_.at(j);
        if(extptr->len > sizeOfLargest) {
            largestIndex = j;
            sizeOfLargest = extptr->len;
        }
    }

    if (sizeOfLargest != 0) {
        extent = *extptr;
        sfsptr->freeSpace_.remove(largestIndex);
        if (objectTrace) {
            fprintf (stderr, "Largest extent found for surface index %d:\n",
                     surfaceIndex);
            extent.fprint(stderr);
        }
    }
    else if (objectTrace)
        fprintf (stderr, "No largest extent was found for surface index %d.\n",
                 surfaceIndex);

    return sizeOfLargest;
}

void
FreeSpaceAllocator::print(void)
{
    fprintf (stderr, "FreeSpaceAllocator freespace\n");
    sfsList_.print();
    fprintf (stderr, "FreeSpaceAllocator erasedpace\n");
    erasedList_.print();
}
void
FreeSpaceAllocator::copy(const FreeSpaceAllocator& fsa)
{
        sfsList_ = fsa.sfsList_;
        objectTrace = fsa.objectTrace;
        objectTrace = fsa.objectTrace;
        curSurface_ = fsa.curSurface_;
        curSurfaceIndex_ = fsa.curSurfaceIndex_;
        needToConsolidateSids_ = fsa.needToConsolidateSids_;
}

RWBoolean
FreeSpaceAllocator::incrementSurface()
{
    if (needToConsolidateSids_) {
        consolidateSids();
        needToConsolidateSids_ = FALSE;
    }
    
    if ((++curSurfaceIndex_) >= sidList_.entries())
    {
        curSurfaceIndex_ = 0;
/**/        curSurface_ = *(sidList_.at(curSurfaceIndex_));
        fprintf (stderr, "incremented, back to zero\n");
        return TRUE;
    }
    else {
        curSurface_ = *(sidList_.at(curSurfaceIndex_));
/**/        fprintf (stderr, "incremented, from %d to %d, within %d entries\n",
                 curSurfaceIndex_ -1, curSurfaceIndex_, (int)sidList_.entries());
        return FALSE;
    }
}


// Create a unique list of sids, from sfsList_.
INT16
FreeSpaceAllocator::consolidateSids()
{
    INT32 i;
    Sid_t sid;
    for(i = 0; i < sfsList_.entries(); i++)  {
        sid = (sfsList_.at(i))->sid_;
        if (sidList_.search(sid) == NULL) {
            sidList_.append(sid);
        }
    }

    needToConsolidateSids_ = FALSE;
    curSurfaceIndex_ = 0;
    return sidList_.entries();
}


void
FreeSpaceAllocator::test()
{
    printf ("Freespace allocator test\n");
    
//    FreeSpaceAllocator allocator_;
    Extent extent;
    
    /*allocator_.*/turnOnTrace();
        
    // Make an extent list.
    extent.start.sid = 0;
    extent.start.sector = 0;
    extent.len = 1000;
    /*allocator_.*/dealloc(512, extent);

    extent.start.sector = 50;
    extent.len = 5000;
    /*allocator_.*/dealloc(512, extent);

    extent.start.sid = 0;
    extent.start.sector = 100;
    extent.len = 8000;
    /*allocator_.*/dealloc(512, extent);

    extent.start.sid = 1;
    extent.start.sector = 100;
    extent.len = 3000;
    /*allocator_.*/dealloc(512, extent);

    extent.start.sid = 1;
    extent.start.sector = 0;
    extent.len = 7000;
    /*allocator_.*/dealloc(512, extent);

    printf ("\nBefore takeLargest\n");
    /*allocator_.*/print();
    /*allocator_.*/takeLargestFreeExtent(0, extent);
        
    printf ("\nAfter takeLargest from surface 0\n");
    /*allocator_.*/print();

    printf ("\nThe largest Extent for surface 0:\n");
    extent.fprint(stdout);

    /*allocator_.*/takeLargestFreeExtent(1, extent);
    printf ("\nAfter takeLargest from surface 1\n");
    /*allocator_.*/print();

    printf ("\nThe largest Extent for surface 1:\n");
    extent.fprint(stdout);

    printf ("\nAfter deallocating the largest extent back.\n");
    /*allocator_.*/dealloc(512, extent);
    /*allocator_.*/print();
}
