//-*-c++-*-
//-----------------------------------------------------------------------------
// fsalloc.h:  
// 
// This header file provides definitions and declarations for the 
// FreeSpaceAllocator class, which provides free space allocation 
// functionality.
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
// CLASSES:
//
// SurfaceFreeSpace
// SurfaceFreeSpaceList
// FreeSpaceAllocator
//
//-----------------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msTest/fsalloc.h,v $
// $Revision: 1.7 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/03/22 14:35:27 $
//-----------------------------------------------------------------------------

#ifndef _FSALLOC_H_                             // Allow multiple inclusion
#define _FSALLOC_H_

//-----------------------------------------------------------------------------
// CLASS:  SurfaceFreeSpace                                     --JKH
//
// The SurfaceFreeSpace class is a simple container for the free space of 
// a single surface.
// 
// PUBLIC DATA MEMBERS:
//
// Sid_t        sid;
// ExtentList   freeSpace_;
//
// PUBLIC METHODS:
//
// SurfaceFreeSpace()
//      Constructor.
// ~SurfaceFreeSpace()
//      Destructor
// SurfaceFreeSpace(const SurfaceFreeSpace& fsa)
//      Copy constructor.
// operator=()
//      Assignment operator.
// takeOutLargestExtent(Extent* largest)
//      Finds and REMOVES the largest extent from freeSpace_.  Returns the
//      length of that extent, or zero if no extent is returned due to error
//      or inavailability.
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// Extent
// ExtentList
//
// PRIVATE DATA MEMBERS:        None.
//
// PRIVATE METHODS:             None.
//
// IMPLEMENTATION NOTES:        None.
//-----------------------------------------------------------------------------

class SurfaceFreeSpace
{
    public:

    Sid_t               sid_;
    UINT32              sectorSize_;
    ExtentList          freeSpace_;

    SurfaceFreeSpace() : sid_(SID_T_INVALID),sectorSize_(0) {}
    ~SurfaceFreeSpace() {}
    SurfaceFreeSpace(const SurfaceFreeSpace& sfs) {
        this->operator=(sfs);
    }
    SurfaceFreeSpace&   operator=(const SurfaceFreeSpace& sfs) {
        sid_ = sfs.sid_;
        sectorSize_ = sfs.sectorSize_;
        freeSpace_ = sfs.freeSpace_;
    }
    UINT32 takelargestExtent(Extent* largest);
    void print(void);
};

//-----------------------------------------------------------------------------
// CLASS:  SurfaceFreeSpaceList                                 --JKH
//
// The SurfaceFreeSpaceList class is a simple container for a list of
// SurfaceFreeSpace objects.
// 
// PUBLIC DATA MEMBERS:  None.
//
// PUBLIC METHODS:
//
// SurfaceFreeSpaceList()
//      Constructor.
// ~SurfaceFreeSpaceList()
//      Destructor
// SurfaceFreeSpaceList(const SurfaceFreeSpaceList& fsa)
//      Copy constructor.
// operator=()
//      Assignment operator.
// clear()
//      Clears the list; removes all items.
// totalBytes()
//      Returns the total number of free bytes in the list.
// search()
//      Searches for a SurfaceFreeSpace entry with the given sid.
// at()
//      Returns a pointer to the item at the given index.
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// Extent
// ExtentList
// SurfaceFreeSpace
// RWGSlist
//
// PRIVATE DATA MEMBERS:        
//
// RWGSlist(SurfaceFreeSpace)   theList_        List of SurfaceFreeSpace objs.
//
// PRIVATE METHODS:             
//
// copy()
//      Copies one SurfaceFreeSpaceList object to another.
// free()
//      Frees memory associated with a SurfaceFreeSpaceList.
//
// IMPLEMENTATION NOTES:        None.
//-----------------------------------------------------------------------------

declare(RWGSlist,SurfaceFreeSpace)

class SurfaceFreeSpaceList
{
    public:

    //-------------------------------------------------------------------------
    // C++isms
    //-------------------------------------------------------------------------

    SurfaceFreeSpaceList() {}
    ~SurfaceFreeSpaceList() { free(); }
    SurfaceFreeSpaceList(const SurfaceFreeSpaceList& sfsl) { copy(sfsl); }
    SurfaceFreeSpaceList&       operator=(const SurfaceFreeSpaceList& sfsl) {
        copy(sfsl);
        return *this;
    }

    //-------------------------------------------------------------------------
    // List access methods
    //-------------------------------------------------------------------------

    void                        clear(void) { free(); }
    UINT32                      totalBytes() {
        INT32 i;
        UINT32 total = 0;

        for(i = 0; i < entries(); i++)  {
            total += (at(i))->freeSpace_.totalBytes();
        }
        return total;
    }

    SurfaceFreeSpace *          search(Sid_t sid);
    SurfaceFreeSpace *          at(UINT32 index);
    const SurfaceFreeSpace *    at(UINT32 index) const;
    void                        append(const SurfaceFreeSpace& sfs);
    void                        remove(UINT32 index);
    UINT32                      entries(void) const { 
        return theList_.entries(); 
    }

    void                        print(void);        

    private:

    //-------------------------------------------------------------------------
    // Private data
    //-------------------------------------------------------------------------

    RWGSlist(SurfaceFreeSpace)  theList_;

    //-------------------------------------------------------------------------
    // Private methods
    //-------------------------------------------------------------------------

    void                copy(const SurfaceFreeSpaceList& sfsl);
    void                free(void);
};

//-----------------------------------------------------------------------------
// CLASS:  FreeSpaceAllocator                                   --JKH
//
// The FreeSpaceAllocator class provides multi-surface freespace allocation
// and deallocation.
// 
// PUBLIC DATA MEMBERS:
//
// PUBLIC METHODS:
//
// FreeSpaceAllocator()
//      Constructor.
// ~FreeSpaceAllocator()
//      Destructor
// FreeSpaceAllocator(const FreeSpaceAllocator& fsa)
//      Copy constructor.
// operator=()
//      Assignment operator.
// alloc()
//      Allocates the specified number of bytes.
// dealloc()
//      Deallocates the given Extent or ExtentList.
// clear()
//      Deallocates everything in the FreeSpaceAllocator
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// Extent
// ExtentList
// SurfaceFreeSpace
// SurfaceFreeSpaceList
//
// PRIVATE DATA MEMBERS:
//
// SurfaceFreeSpaceList         sfsList_
//
// PRIVATE METHODS:
//
// copy()
//      Copies one FreeSpaceAllocator object to another
// free()
//      Frees memory associated with the FreeSpaceAllocator
//
// IMPLEMENTATION NOTES: 
//-----------------------------------------------------------------------------

class FreeSpaceAllocator 
{
    public:

    //-------------------------------------------------------------------------
    // C++isms:  constructors, destructors, etc.
    //-------------------------------------------------------------------------

    FreeSpaceAllocator() {
        objectTrace = FALSE;
        curSurface_ = 0;
        curSurfaceIndex_ = 0;
        needToConsolidateSids_ = TRUE;
    }
    ~FreeSpaceAllocator() { free(); sidList_.clear(); }
    FreeSpaceAllocator(const FreeSpaceAllocator& fsa) {
        copy(fsa);
    }
    FreeSpaceAllocator& operator=(const FreeSpaceAllocator& fsa) {
        copy(fsa);
        return *this;
    }

    //-------------------------------------------------------------------------
    // Allocation methods
    //-------------------------------------------------------------------------

    INT32       alloc(UINT32 sectorSize,UINT32 nbytes,Extent& extent);
    INT32       alloc(UINT32 sectorSize,UINT32 nbytes,ExtentList& extentList);
    INT32       alloc(Sid_t sid,ExtentList& extentList);
    INT32       dealloc(UINT32 sectorSize,const Extent& extent);
    INT32       dealloc(UINT32 sectorSize,const ExtentList& extentList);
    
    UINT32      totalBytes(void) { return sfsList_.totalBytes(); }
    UINT32      largestFreeExtent();

    //-------------------------------------------------------------------------
    // Developer support methods
    //-------------------------------------------------------------------------
    void        turnOnTrace()   {objectTrace = TRUE;}
    void        print(void);
    void        test();

    private:

    //-------------------------------------------------------------------------
    // Private data
    //-------------------------------------------------------------------------

    SurfaceFreeSpaceList        sfsList_;
    SurfaceFreeSpaceList        erasedList_;
    RWBoolean                   objectTrace;

    //-------------------------------------------------------------------------
    // Private methods
    //-------------------------------------------------------------------------

    void        copy(const FreeSpaceAllocator& fsa);
    void        free(void) {
    };


    //-------------------------------------------------------------------------
    // Test stimulation ONLY for synth
    //-------------------------------------------------------------------------

public:
    // First, the methods
    
    INT32       erasedAlloc(Sid_t sid,ExtentList& extentList);
    INT32       erasedDealloc(UINT32 sectorSize,const ExtentList& extentList);
    INT32       erasedDealloc(UINT32 sectorSize,const Extent& extent);
    UINT32      takeLargestFreeExtent(INT32 surfaceIndex, Extent& extent);
    UINT32      takeLargestFreeExtent(Extent& extent);

    RWBoolean   incrementSurface(void);
    INT16       consolidateSids(void);
                
    // State members
                
    int         curSurfaceIndex_;
    Sid_t       curSurface_;
    RWBoolean   needToConsolidateSids_;
    SidList     sidList_;
};

#endif  // _FSALLOC_H_
