/*****************************************************************************
** The following routines implement chaining (linked-list) hash tables.
**
**      Unpublished Confidential Information of Hewlett-Packard
**      Company. Do not disclose.  Copyright (c) Hewlett-Packard
**      Company 1995. All rights reserved
** 
** The contents of this software are proprietary and confidential to
** the Hewlett- Packard Company, and are limited in distribution to
** those with a direct need to know.  Individuals having access to
** this software are responsible for maintaining the confidentiality
** of the content and for keeping the software secure when not in use.
** Transfer to any party is strictly forbidden other than as expressly
** permitted in writing by Hewlett-Packard Company.  Unauthorized
** transfer to or possession by any unauthorized party may be a
** criminal offense.
**
**                  RESTRICTED RIGHTS LEGEND
**
**      Use,  duplication,  or disclosure by the U. S. Government is
**      subject to restrictions as set forth in subparagraph (c) (1)
**      (ii) in the Rights in Technical Data and Computer Software
**      clause in DFARS 252.227-7013 or any other successor clause.
**
**                      HEWLETT-PACKARD COMPANY
**                      3000 Hanover St.
**                      Palo Alto, CA  94304 U.S.A.
**
**      Rights for non-DOD U.S. Government Departments and Agencies
**      are as set forth in FAR 52.227-19 (c) (1,2) or any other
**      successor clause.
**
** Programmer:  Jon Hirota
**              June 1993
******************************************************************************/

#define _HASH_C_

#include <stdio.h>
#include <stdlib.h>
#include "hash.h"

#ifndef lint
static char rcsid[] = "$Header: hash.c,v 1.3 95/10/25 17:27:19 hmgr Exp $";
#endif

#define HTABLE_CALLOC(nelem,elsize)     calloc(nelem,elsize)
#define HTABLE_FREE(ptr)                free(ptr)

/*****************************************************************************
** htableCreate():  creates a hash table.  Takes as parameters the number of
**                  slots to put in the table, a compare function for the
**                  data, and a hash function for the data.  The size must
**                  be greater than zero and the two function pointers must
**                  be non-NULL.
**
** preconditions:   see above.
** postconditions:  memory will be allocated for the hash table.  Each slot
**                  will be initialized by a call to dlistCreate().
** return values:   The function will return a pointer to the hash table on
**                  success, NULL on failure.
******************************************************************************/

#ifdef __STDC__
hash_table *
htableCreate
(
        int     size,
        int     (*cmp_func)(void *,void *),
        int     (*hash_func)(void *)
)
#else
hash_table *
htableCreate(size,cmp_func,hash_func)
int     size;
int     (*cmp_func)();
int     (*hash_func)();
#endif
{
        int             i;
        hash_table      *newtable;

        if(size < 1)
        {
                return (hash_table *) NULL;
        }
#ifdef __STDC__
        if((cmp_func == (int (*)(void *, void *)) NULL) ||
                (hash_func == (int (*)(void *)) NULL))
#else
        if((cmp_func == (int (*)()) NULL) || (hash_func == (int (*)()) NULL))
#endif
        {
                return (hash_table *) NULL;
        }

        /*
        ** Allocate memory for the table
        */

        newtable = (hash_table *) HTABLE_CALLOC(sizeof(*newtable),1);
        if(newtable == (hash_table *) NULL)
        {
                return (hash_table *) NULL;
        }

        newtable->table = (dlist **) HTABLE_CALLOC(sizeof(dlist *),
                                                   (size_t) size);
        if(newtable->table == (dlist **) NULL)
        {
                HTABLE_FREE((void *) newtable);
                return (hash_table *) NULL;
        }

        newtable->capacity = size;
        newtable->hash_func = hash_func;
        newtable->count = 0;
        newtable->hash_err = HERR_NOERR;

        /*
        ** Initialize lists
        */

        for(i = 0; i < newtable->capacity; i++)
        {
            newtable->table[i] = dlistCreate(cmp_func);
            if(newtable->table[i] == (dlist *) NULL)
            {
                int     j;

                for(j = i - 1; j >= 0; j--)
                {
                    dlistDestroy(newtable->table[j]);
                }
                HTABLE_FREE((void *) newtable->table);
                HTABLE_FREE((void *) newtable);
                return (hash_table *) NULL;
            }
        }
        return newtable;
}

/*****************************************************************************
** htableHash():    adds an item to a hash table.
**
** preconditions:   argument hashtable is non-NULL; argument data item is
**                  non-NULL.
** postconditions:  the data item will be added to the hash table.
** return values:   returns 0 on success, -1 on failure.
******************************************************************************/

#ifdef __STDC__
int
htableHash
(
        hash_table      *table,
        void            *datap
)
#else
int
htableHash(table,datap)
hash_table      *table;
void            *datap;
#endif
{
        int     hashvalue;

        if(table == (hash_table *) NULL)
        {
                return HERR_NULL;
        }
        if(datap == (void *) NULL)
        {
                table->hash_err = HERR_NULL;
                return HERR_NULL;
        }

        hashvalue = table->hash_func(datap);
        if(hashvalue < 0)
        {
                hashvalue = -hashvalue;
        }
        hashvalue %= table->capacity;

        if(dlistInsert(table->table[hashvalue],datap))
        {
                switch(dlistError(table->table[hashvalue]))
                {
                 case DLERR_NULL:       /* Fatal error! */
                        table->hash_err = HERR_NULL;
                        return HERR_NULL;
                 case DLERR_NOMEM:
                        table->hash_err = HERR_NOMEM;
                        return HERR_NOMEM;
                 default:
                        break;
                }
        }
        (table->count)++;
        table->hash_err = HERR_NOERR;
        return HERR_NOERR;
}

/*****************************************************************************
** htableFind():  Attempts to find a data item in a hash table.
**
** preconditions:   The arguments are non-NULL.
** postconditions:  See return values.
** return values:   A pointer to the data value is returned if found; otherwise
**                  NULL is returned.
******************************************************************************/

#ifdef __STDC__
void *
htableFind
(
        hash_table      *table,
        void            *datap
)
#else
void *
htableFind(table,datap)
hash_table      *table;
void            *datap;
#endif
{
        dlist_node      *found_node;
        int             hashvalue;

        if(table == (hash_table *) NULL)
        {
                return (void *) NULL;
        }
        if(datap == (void *) NULL)
        {
                table->hash_err = HERR_NULL;
                return (void *) NULL;
        }

        hashvalue = table->hash_func(datap);
        if(hashvalue < 0)
        {
                hashvalue = -hashvalue;
        }
        hashvalue %= table->capacity;

        found_node = dlistFind(table->table[hashvalue],datap);
        if(found_node == (dlist_node *) NULL)
        {
                table->hash_err = dlistError(table->table[hashvalue]);
                return (void *) NULL;
        }
        table->hash_err = HERR_NOERR;
        return found_node->data;
}

/*****************************************************************************
** htableDestroy():  frees all memory associated with the hash table.  Note:
**                 this function does NOT free user data!
**
** preconditions:  the hash_table argument is non-NULL.
** postconditions: see above
** return values:  
******************************************************************************/

#ifdef __STDC__
int
htableDestroy
(
        hash_table      *table
)
#else
int
htableDestroy(table)
hash_table      *table;
#endif
{
        int     i;

        if(table == (hash_table *) NULL)
        {
                return HERR_NULL;
        }

        for(i = 0; i < table->capacity; i++)
        {
                dlistDestroy(table->table[i]);
        }
        table->capacity = 0;
        table->count = 0;
#ifdef __STDC__
        table->hash_func = (int (*)(void *)) NULL;
#else
        table->hash_func = (int (*)()) NULL;
#endif
        HTABLE_FREE((void *) table->table);
        HTABLE_FREE((void *) table);
        return HERR_NOERR;
}


/*****************************************************************************
** htableApply():  applies a function to all data items in the table.
**
** If the failsafe parameter is 1, application of the apply function will stop
** at the first failure of the apply function (nonzero return value).  Hence,
** if failsafe = 1 and htableApply returns 0, then the apply function executed
** successfully on all data items in the table.  If failsafe is 0, the apply
** function will be applied to all data items in the table, regardless of
** whether any applications fail, and this function will return the return
** value of the last application of the apply function.
**
** preconditions:   arguments are non-NULL.
** postconditions:  dependent upon function applied.
** return values:   see above.
******************************************************************************/

#ifdef __STDC__
int
htableApply
(
        hash_table      *table,
        int             (*apply_func)(void *),
        int             failsafe
)
#else
int
htableApply(table,apply_func,failsafe)
hash_table      *table;
int             (*apply_func)();
int             failsafe;
#endif
{
        int     i,rc;

        if(table == (hash_table *) NULL)
        {
                return -1;
        }
#ifdef __STDC__
        if(apply_func == (int (*)(void *)) NULL)
#else
        if(apply_func == (int (*)()) NULL)
#endif
        {
                table->hash_err = HERR_NULL;
                return -1;
        }

        table->hash_err = HERR_NOERR;
        for(i = 0; i < table->capacity; i++)
        {
                rc = dlistApply(table->table[i],apply_func,failsafe);
                if(failsafe && rc)
                {
                    table->hash_err = dlistError(table->table[i]);
                    break;
                }
        }
        return rc;
}

/*****************************************************************************
** htableDelete():  Deletes a data item from a hash_table.
**
** preconditions:   arguments are non-NULL.
** postconditions:  the item will be deleted from the table, and its memory
**                  freed.
** return values:   a pointer to the deleted node's data will be returned on
**                  success; otherwise NULL is returned.
******************************************************************************/

#ifdef __STDC__
void *
htableDelete
(
        hash_table      *table,
        void            *datap
)
#else
void *
htableDelete(table,datap)
hash_table      *table;
void            *datap;
#endif 
{
        int             hashvalue;
        void            *old_data;

        if(table == (hash_table *) NULL)
        {
                return (void *) NULL;
        }
        if(datap == (void *) NULL)
        {
                table->hash_err = HERR_NULL;
                return (void *) NULL;
        }

        hashvalue = table->hash_func(datap);
        if(hashvalue < 0)
        {
                hashvalue = -hashvalue;
        }
        hashvalue %= table->capacity;

        old_data = dlistDatadelete(table->table[hashvalue],datap);
        if(old_data == (void *) NULL)
        {
                table->hash_err = dlistError(table->table[hashvalue]);
                return (void *) NULL;
        }

        (table->count)--;
        table->hash_err = HERR_NOERR;
        return old_data;
}
