//-*-c++-*-
//-----------------------------------------------------------------------------
// jbread.cpp:  Main source file for a program to read arbitrary sectors
//              or (UDF) logical blocks from a surface in a jukebox.
//
// Usage:
//
// jbread [-b] [-B bias_offset] [-a acrcFile] [-s jbStatFile] 
//        { { [-v volSetName] [-p partition] } | [-d surface] } sector_list
//
// Invoked as moread:
//
// moread [-b] [-B bias_offset] [-r rawDeviceFile] sector_list
//
// Environment variables:
//
// ACRC_FILE:         Contains the pathname of the acrc file to use.
// JBSTAT_FILE:       Contains the pathname of the jbstat file to use.
// UDF_SS_DEVICE:     Compatibility for the rs program.  Contains the pathname 
//                    of the raw disk device file to use if executing as moread.
// UDF_BLOCK_OFFSET:  If the -b option is specified, specifies the bias (offset)
//                    used to accomplish logical block addressing.
//
// Initial revision:  Jon Hirota, 04/96
// Modifications:
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <rw/cstring.h>
#include <rw/gslist.h>
#include "nsrLib.h"
#include "arcImEx.h"

//-----------------------------------------------------------------------------
// Constants used in this file.
//-----------------------------------------------------------------------------

const char *MOREAD_NAME = "moread";
const char *ACRC_ENV = "ACRC_FILE";
const char *JBSTAT_ENV = "JBSTAT_FILE";
const char *BIAS_ENV = "UDF_BLOCK_OFFSET";
const char *RAWDEV_ENV = "UDF_SS_DEVICE";

const INT32 MAX_SECTOR_SIZE = 1024;

//-----------------------------------------------------------------------------
// Some local data structures used in this file.
//-----------------------------------------------------------------------------

struct Surface
{
    UINT32      slot;           // Slot = 0 is invalid; we use 1-based indexing
    UINT32      side;           // 0 = A, anything else = B
    Sid_t       sid;

    // Some methods

                Surface() : slot(0),side(0),sid(0) { }
    INT32       operator()(const char *ptr)
                {
                    char c;
                        
                    if(ptr == NULL)
                    {
                        return -1;
                    }
                    if(sscanf(ptr,"%u%c",&slot,&c) != 2)
                    {
                        fprintf(stderr,"Unknown surface specifier <%s>\n",ptr);
                        return -1;
                    }
                    if(slot == 0)       // User specified 0[a|b]
                    {
                        slot = UINT_MAX;
                        fprintf(stderr,"Unknown surface specifier <%s>\n",ptr);
                        return -1;
                    }
                    switch(tolower(c))
                    {
                    case 'a':
                        side = 0;
                        break;
                    case 'b':
                        side = 1;
                        break;
                    default:
                        fprintf(stderr,"Unknown surface specifier <%s>\n",ptr);
                        return -1;
                    }
                    return 0;
                }
};

declare(RWGSlist,UINT32);       // Overkill, but heck. . .

//-----------------------------------------------------------------------------
// Some local function prototypes
//-----------------------------------------------------------------------------

static void     freeSectorList(RWGSlist(UINT32)& sectorList);
static INT32    doRead(UINT32 biasOffset,UINT32 sector,SSDevice *driveDev);
static void     Usage(INT32 jbreadMode);

//-----------------------------------------------------------------------------
// METHOD:  main()
//
// Main entry point of the jbread program.  Gets relevent environment variables
// and command-line options.
//
// ARGUMENTS:  
//
// int   argc    IN   Number of command line arguments
// char *argv[]  IN   Array of pointers to command line arguments
//
// RETURNS:
//
// Returns 0 on a successful run, a nonzero value on error.
//
// PRECONDITIONS:  See the command-line argument checking code below.
//
// POSTCONDITIONS:  See above.
//
// NOTES:
//
// main() is a pretty huge, monolithic function doing all the work to check
// the command-line arguments, then setting up for the actual reads.  My
// apologies.
//-----------------------------------------------------------------------------

INT32
main
(
    int         argc,
    char        *argv[]
)
{
    char                ch;
    INT32               c;
    INT32               i;
    INT32               sector;
    UINT32              *iptr;
    INT32               partitionNum = -1;
    INT32               bias = 0;               // For rb compatibility
    INT32               biasOffset = 0;
    INT32               usage = 0;
    INT32               jbreadMode = 1;         // Invoked as jbread or moread?
    INT32               rval = 0;
    char *              cptr;
    RWCString           acrcFile;
    RWCString           jbStatFile;
    RWCString           rawDevFile;
    RWCString           volSetName;
    Surface             surface;
    RWGSlist(UINT32)    sectorList;     // Generics strike again!

    extern int  optind;                 // For getopt(3)
    extern char *optarg;                // For getopt(3)
        
    // We first need to determine if we were invoked as moread.

    if(strcmp(argv[0],MOREAD_NAME) == 0)
    {
        jbreadMode = 0;
    }

    // Now we need to get environment variables, if any.

    if(jbreadMode)              // jbread-specific environment variables
    {
        if((cptr = getenv(JBSTAT_ENV)) != NULL)
        {
            jbStatFile = cptr;
        }
        if((cptr = getenv(ACRC_ENV)) != NULL)
        {
            acrcFile = cptr;
        }
    }
    else                        // moread-specific environment variables
    {
        if((cptr = getenv(RAWDEV_ENV)) != NULL)
        {
            rawDevFile = cptr;
        }
    }

    // Environment variable(s) common to both programs

    if((cptr = getenv(BIAS_ENV)) != NULL)
    {
        if(sscanf(cptr,"%d",&bias) != 1)
        {
            bias = -1;
        }
    }

    // Now we need to go through the command-line arguments.  This is a fairly
    // involved process, as both programs have multiple constraints.

    if(jbreadMode)
    {
        while((c = getopt(argc,argv,"a:s:d:v:p:bB:")) != EOF)
        {
            switch(c)
            {
            case 'a':
                acrcFile = optarg;
                break;

            case 's':
                jbStatFile = optarg;
                break;

            case 'd':
                // Add the surface to the list!
                if(surface.slot != 0)
                {
                    fprintf(stderr,"BAD ARGS:  only one -d option allowed!\n");
                    usage = 1;
                }
                else
                {
                    if(surface(optarg))
                    {
                        usage = 1;
                    }
                }
                break;

            case 'v':
                volSetName = optarg;
                break;

            case 'p':
                if(sscanf(optarg,"%d",&partitionNum) != 1)
                {
                    fprintf(stderr,"BAD ARGS:  unknown -p option value!\n");
                    usage = 1;
                }
                break;

            case 'b':
                bias = 1;
                break;

            case 'B':
                if(sscanf(optarg,"%d",&biasOffset) != 1)
                {
                    fprintf(stderr,"BAD ARGS:  unknown -B option value!\n");
                    usage = 1;
                }
                break;

            default:
                usage = 1;
                break;
            }
        }
    }           // End of jbread mode specific command-line options
    else        // Begin moread mode specific command-line options
    {
        while((c = getopt(argc,argv,"bB:r:")) != EOF)
        {
            switch(c)
            {
            case 'b':
                bias = 1;
                break;
        
            case 'B':
                if(sscanf(optarg,"%d",&biasOffset) != 1)
                {
                    fprintf(stderr,"Bad -B option!\n");
                    usage = 1;
                }
                break;
            
            case 'r':
                rawDevFile = optarg;
                break;

            default:
                usage = 1;
                break;
            }
        }
    }           // End of moread mode specific command line options 

    // Now we interpret any remaining arguments as sectors.

    for(i = optind; i < argc; i++)
    {
        if(sscanf(argv[i],"%u",&sector) != 1)
        {
            fprintf(stderr,"Bad sector value <%s>!\n",argv[i]);
            usage = 1;
        }
        else
        {
            iptr = new UINT32(sector);
            if(iptr == NULL)
            {
                fprintf(stderr,"Memory allocation failure, abort!\n");
                freeSectorList(sectorList);
                return 2;
            }
            if(sectorList.append(iptr) == NULL)
            {
                fprintf(stderr,"Memory allocation failure, abort!\n");
                freeSectorList(sectorList);
                return 2;
            }
        }
    }

    if(jbreadMode)
    {
        // Now we have gone through all of the arguments of jbread.  We need
        // to figure out if we have a valid set of arguments.  Valid arguments
        // satisfy the following conditions:
        //
        // 1.  Exactly one -d option OR -v option was used.
        // 2.  If the volSetName is set, the jbStatFile must be set.
        // 3.  If bias is set, biasOffset must be >= 0.
        // 4.  If biasOffset is > 0, bias must be set.
        // 5.  acrcFile must be set.
        // 6.  sector list must contain at least one entry.

        // condition 1
        if((!volSetName.isNull() && (surface.slot != 0)) ||
           (volSetName.isNull() && (surface.slot == 0)))
        {
            fprintf(stderr,"BAD ARGS:  You must specify exactly ONE of the -v \
or -d options!\n");
            usage = 1;
        }
        if(!volSetName.isNull())                // condition 2
        {
            if(jbStatFile.isNull())
            {
                fprintf(stderr,"BAD ARGS:  If you specify a volume set name, \
you must specify a jbStatFile\nor have the %s environment variable \
set.\n",JBSTAT_ENV);
                usage = 1;
            }
        }
        if(bias)                                // condition 3
        {
            if(biasOffset < 0)
            {
                usage = 1;
            }
        }
        if(biasOffset > 0)                      // condition 4
        {
            if(!bias)
            {
                fprintf(stderr,"BAD ARGS:  You must specify -b with a -B \
option.\n");
                usage = 1;
            }
        }
        if(acrcFile.isNull())                   // condition 5
        {
            fprintf(stderr,"BAD ARGS:  You must specify an acrc file or have \
the %s environment variable set.\n",ACRC_ENV);
            usage = 1;
        }
        if(sectorList.entries() < 1)            // condition 6
        {
            fprintf(stderr,"BAD ARGS:  You must specify at least one sector.\n");
            usage = 1;
        }
    }           // End of jbread mode validity check
    else        // Begin of moread mode validity check
    {
        // We have gone through all of the moread mode specific options.  We
        // must check to see if the arguments are valid.  They must satisfy
        // the following conditions:
        //
        // 1.  If bias is set, biasOffset must be >= 0.
        // 2.  If biasOffset is > 0, bias must be set.
        // 3.  Sector list must contain at least one entry.
        // 4.  rawDevFile must be set.

        if(bias)                                // condition 1
        {
            if(biasOffset < 0)
            {
                usage = 1;
            }
        }
        if(biasOffset > 0)                      // condition 2
        {
            if(!bias)
            {
                fprintf(stderr,"BAD ARGS:  You must specify -b with a -B \
option.\n");
                usage = 1;
            }
        }
        if(sectorList.entries() < 1)            // condition 3
        {
            fprintf(stderr,"BAD ARGS:  You must specify at least one sector.\n");
            usage = 1;
        }
        if(rawDevFile.isNull())                 // condition 4
        {
            fprintf(stderr,"BAD ARGS:  You must specify a raw device file, or \
have the %s\nenvironment variable set.\n",RAWDEV_ENV);
            usage = 1;
        }
    }           // End of moread mode validity check

    // If usage = 1, quit now.

    if(usage)
    {
        Usage(jbreadMode);
        freeSectorList(sectorList);
        return 2;
    }

    // Whew!  All that command-line checking was a lot of work.  We're
    // now ready to check to see that the various files are valid and exist
    // and that sort of thing.

    NSRJukeBox          *jbptr = NULL;
    hpACDev             *acdev = NULL;
    SSDevice            *driveDev = NULL;
    VolsetNameMap       *vsnmp = NULL;  // For NSRJukeBox database access
    VolsetNameMapList   vsnmList;       // For NSRJukebox database access
    SurfaceMap          *smp = NULL;
    SurfaceMapList      smList;
    Error               err;

    if(jbreadMode)
    {
        // We basically construct an NSRJukeBox object.  If the construction
        // fails, something was most likely wrong with the acrc file.  If
        // it works, and -v was set, we add the jbstat file, and check to
        // make sure it was valid.  Finally, we check to see if the -p option
        // was used, and check to see if the partition number was in range.
        // If not, we issue a warning and use the default partition (the last
        // one).

        jbptr = new NSRJukeBox(acrcFile);
        if(jbptr == NULL)
        {
            fprintf(stderr,"Memory allocation failure, aborting!\n");
            freeSectorList(sectorList);
            return 2;
        }
        if(jbptr->error() != NSR::error_None)
        {
            fprintf(stderr,"ERROR:  Construction of NSRJukeBox object failed, \
error is <%s>\n",jbptr->error().message());
            fprintf(stderr,"Please check your acrc file to make sure it is \
valid.\n");
            delete jbptr;
            freeSectorList(sectorList);
            return 2;
        }

        // We get the surface map list now, so that we can go between sid's
        // and slot/side mappings.

        smList = jbptr->surfaceMapList();

        if(!volSetName.isNull())
        {
            err = jbptr->readStatusFile(jbStatFile);
            if(err != NSR::error_None)
            {
                fprintf(stderr,"ERROR:  Unable to read jbstat file, error is \
<%s>\n",err.message());
                fprintf(stderr,"Please check you jbstat file to make sure it \
is valid.\n");
                delete jbptr;
                freeSectorList(sectorList);
                return 2;
            }

            // Make sure that volume set name was valid.  We don't want to
            // use the volume set name as a working volume set in the 
            // NSRJukeBox object, because that would cause it to swap each
            // volume.  We will access the NSRJukeBox's internal database
            // directly instead, and search for it.

            vsnmList = jbptr->volsetNameMapList();
            vsnmp = vsnmList.search(volSetName);
            if(vsnmp == NULL)           // Not found!
            {
                fprintf(stderr,"ERROR:  specified volume set name <%s> is \
not in the jbstat file.\n",(CONST_CHAR_PTR) volSetName);
                delete jbptr;
                freeSectorList(sectorList);
                return 2;
            }

            // It's valid; get the surface map list so that we can print
            // out which surfaces are involved here.

            fprintf(stderr,"Volume set <%s> = ",(CONST_CHAR_PTR) volSetName);
            for(i = 0; i < vsnmp->sidList_.entries(); i++)
            {
                smp = smList.search(*(vsnmp->sidList_.at(i)));
                if(smp == NULL)
                {
                    continue;
                }
                fprintf(stderr,"%d%c ",smp->slot_,smp->side_ ? 'B' : 'A');
            }
            fprintf(stderr,"\n");

            // If the partitionNum variable is set, make sure it's in range.

            if(partitionNum >= 0)
            {
                if(partitionNum >= vsnmp->sidList_.entries())
                {
                    fprintf(stderr,"WARNING:  partition number %d is out of \
range.\n");

                    // This rather nasty expression is simply saying:  search 
                    // the surface map list for the entry containing the last 
                    // sid in the volSetNameMap SidList, and return its entry.

                    smp = smList.search(*(vsnmp->sidList_.at(
                                        vsnmp->sidList_.entries() - 1)));
                }
                else
                {
                    smp = smList.search(*(vsnmp->sidList_.at(partitionNum)));
                }    
                  
            }
            else        // partitionNum isn't set
            {
                smp = smList.search(*(vsnmp->sidList_.at(
                                    vsnmp->sidList_.entries() - 1)));
            }

            // Set the surface variable.

            assert(smp != NULL);
            surface.slot = smp->slot_;
            surface.side = smp->side_;
            surface.sid = smp->sid_;
        }       // End of volset name handling
        else
        {
            smp = smList.search(surface.slot,surface.side);
            if(smp == NULL)
            {
                fprintf(stderr,"ERROR:  surface %d%c is not present in the \
jukebox!\n",surface.slot,surface.side ? 'B' : 'A');
                delete jbptr;
                freeSectorList(sectorList);
                return 2;
            }
            surface.sid = smp->sid_;
        }

        // At this point, we have a valid surface in our surface variable.

        fprintf(stderr,"Reading from surface %d%c\n",surface.slot,
                surface.side ? 'B' : 'A');

        // We allocate a SSDevice object.

        acdev = jbptr->acdev();
        assert(acdev != NULL);
        driveDev = acdev->allocDrive(surface.sid);
        if(driveDev == NULL)
        {
            fprintf(stderr,"ERROR:  Unable to allocate drive device in \
jukebox, error is\n<%s>\n",acdev->error().message());
            delete jbptr;
            freeSectorList(sectorList);
            return 2;
        }

        // Okay, at this point, we should be ready to do the read (finally!)
    }   // End of jbread mode setup
    else        // Begin moread mode setup
    {
        // We basically just need to allocate a single-surface device object.

        driveDev = (SSDevice *) new DriveDevice((CONST_CHAR_PTR) rawDevFile);
        if(driveDev == NULL)
        {
            fprintf(stderr,"ERROR:  memory allocation failure, aborting!\n");
            freeSectorList(sectorList);
            return 2;
        }
        if(driveDev->error() != NSR::error_None)
        {
            fprintf(stderr,"ERROR:  construction of DriveDevice failed, error \
is <%s>\n",driveDev->error().message());
            fprintf(stderr,"Please make sure the raw device file <%s> exists \
and is accessible.\n",(CONST_CHAR_PTR) rawDevFile);
            freeSectorList(sectorList);
            return 2;
        }
    }

    // Whew!  That was a lot of work too.  Autochangers really are sort of a
    // pain.  We're now ready to do the actual reads.  We use the SSDevice
    // object we were passed and iterator over the sector list.

    fflush(stderr);

    putw(driveDev->sectorSize(),stdout);
    putw(sectorList.entries(),stdout);
    for(i = 0; i < sectorList.entries(); i++)
    {
        iptr = sectorList.at(i);
        assert(iptr != NULL);
        if(doRead(biasOffset,*iptr,driveDev))
        {
            fprintf(stderr,"ERROR:  read failed on sector <%u>, aborting\n",
                    *iptr+biasOffset);
            break;
        }
    }
    fflush(stdout);
        
    // We're done, finally.  Clean up.

    if(jbptr != NULL)
    {
        acdev->freeDrive(driveDev);
        delete jbptr;                   
    }
    freeSectorList(sectorList);
    return rval;
}
                    
//-----------------------------------------------------------------------------
// METHOD:  freeSectorList()
//
// Frees memory associated with an RWGSlist(UINT32) list.
//
// ARGUMENTS:  
//
// RWGSlist(UINT32)&  sectorList  IN/OUT  Sector list to free.
//
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS: 
//
// --Memory associated with the sector list will be freed.
//-----------------------------------------------------------------------------

static void
freeSectorList
(
    RWGSlist(UINT32)& sectorList
)
{
    UINT32                      *iptr;

    for(INT32 i = 0; i < sectorList.entries(); i++)
    {
        iptr = sectorList.at(i);
        delete iptr;
    }
    sectorList.clear();
}

//-----------------------------------------------------------------------------
// METHOD:  doRead()
//
// Reads the specified sector from the argument SSDevice and displays it to
// screen in the following format:
//
// (word) sector address
// (word) sector + bias
// (bytes) The sector's bytes
//
// ARGUMENTS:  
//
// UINT32       biasOffset  IN  The bias offset.
// UINT32       sector      IN  Sector address.
// SSDevice *   driveDev    IN  Drive to read from.
//
// RETURNS:  0 on success, -1 on failure.
//
// PRECONDITIONS:
//
// --sector is in range
// --biasOffset is in range.
//
// POSTCONDITIONS:
//
// --See above.
//
// NOTES:
// 
//-----------------------------------------------------------------------------

static INT32
doRead
(
    UINT32      biasOffset,
    UINT32      sector,
    SSDevice    *driveDev
)
{
    INT32               i;
    UINT32              sectorSize;
    INT32               rval;
    Error               err;
    unsigned char       buffer[MAX_SECTOR_SIZE];

    sectorSize = driveDev->sectorSize();
    if(sectorSize == 0)
    {
        fprintf(stderr,"ERROR:  unable to get surface sector size!\n");
        return -1;
    }
    rval = driveDev->read(sector+biasOffset,sectorSize,(BYTE_PTR) buffer,err);
    if(rval < sectorSize)
    {
        fprintf(stderr,"ERROR:  read got short count %d, expected %d, error \
is\n<%s>\n",rval,sectorSize,err.message());
        return -1;
    }

    // Write header and output to screen.

    putw(sector,stdout);
    putw(sector+biasOffset,stdout);
    for(i = 0; i < sectorSize; i++)
    {
        putchar(buffer[i]);
    }
    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  Usage()
//
// Prints usage information to stdout.
//
// ARGUMENTS:  
//
// INT32  jbReadMode  IN  If nonzero, we're in jbread mode, otherwise moread.
//
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:  None.
//-----------------------------------------------------------------------------

static void
Usage
(
    INT32       jbReadMode
)
{
    char        **cptr;
    static char *jbreadUsage[] = {
"",
"jbread [-b] [-B bias_offset] [-a acrcFile] [-s jbStatFile]", 
"       { { [-v volSetName] [-p partition] } | [-d surface] } sector_list",
"",
"[-b]:              Use logical block addressing bias",
"[-B bias_offset]:  Specify the bias offset (in sectors)",
"[-a acrcFile]:     Specify the pathname of the acrc file",
"[-s jbStatFile]:   Specify the pathname of the jbstat file",
"[-v volSetName]:   Specify a volume set name to use for the read",
"[-p partition]:    The 0-based index of the partition within the volume set",
"                   specified by the -v option to read",
"[-d surface]:      Specify a surface, in the form of 1A, 3B, etc.  Note that",
"                   surfaces use 1-based indexing (i.e., 0A is NOT valid)",
"sector_list:       A list of sector (or logical block) addresses",
"",
"ENVIRONMENT VARIABLES",
"",
"ACRC_FILE:        Contains the pathname of the acrc file to use.  Overridden",
"                  by the -a option if specified.",
"JBSTAT_FILE:      Contains the pathname of the jbstat file to use.  Overriden",
"                  by the -s option if specified.",
"UDF_BLOCK_OFFSET: If the -b option is specified, specifies the bias (offset)",
"                  used to accomplish logical block addressing.  Overridden by",
"                  the -B option if specified.",
"",
"You must specify at least one sector, and either a volume set name or",
"the -d option.  You must specify the acrc file with the -a option",
"or have the ACRC_FILE variable set.  If you specify a volume set name, you",
"must specify the jbstat file with the -s option or have the JBSTAT_FILE ",
"variable set.  Additionally, you may specify the -p option to specify which",
"partition within the volume set to read.  If -p is not specified, the last",
"partition is read.",
"",
NULL
};
    static char *moreadUsage[] = {
"",
"moread [-b] [-B bias_offset] [-r rawDeviceFile] sector_list",
"",
"[-b]:               Use logical block addressing bias",
"[-B bias_offset]:   Specify the bias offset (in sectors)",
"[-r rawDeviceFile]: Specify the raw disk device file to use",
"sector_list:        A list of sector (or logical block) addresses",
"",
"ENVIRONMENT VARIABLES",
"",
"UDF_BLOCK_OFFSET: If the -b option is specified, specifies the bias (offset)",
"                  used to accomplish logical block addressing.  Overridden by",
"                  the -B option if specified.",
"UDF_SS_DEVICE:    Specifies the raw disk device file to use.  Overriden by",
"                  the -r option if specified.",
"",
"You must specify at least one sector, and either the -r option must be given",
"or the UDF_SS_DEVICE environment variable set.",
"",
NULL
};

    if(jbReadMode)
    {
        cptr = jbreadUsage;
    }
    else
    {
        cptr = moreadUsage;
    }
    while(*cptr != NULL)
    {
        printf("%s\n",*cptr++);
    }
}
