#ifndef __ARC_TEST_H__ //-*-c++-*-
#define __ARC_TEST_H__
// -----------------------------------------------------------------------------
//  acScript.h
//  Classes for reading and executing archive scripts.
//
// (c) Copyright 1995  Hewlett-Packard Company.
// The contents of this software are proprietary and confidential to
// the Hewlett-Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transter to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the Government  is
//      subject to restrictions as set forth in subdivision (b)
//      (3)  (ii)  of the Rights in Technical Data and Computer
//      Software clause at 52.227-7013.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304
//  CLASSES:
// TypedPath -    UNIX path, with Script purpose and path type (file, directory).
//                Is used to define paths stated at the beginning of arc scripts.
// TypedPathDictionary - A dictionary of, guess what, typed paths.
// Volume       - Defines one volume (density, type, name).
// ScriptAction - Defines one script action (label, unlabel, export, import, erase),
//                contains the command line.  Continue or stop special actions allow
//                for the conditional onPass and OnFail actions.
// Script-        Encapsulates the parsing of an arcTest script, then carries
//                out those actions.

//
//-----------------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msVerify/acScript.h,v $
// $Revision: 1.11 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/03/22 09:47:48 $
//-----------------------------------------------------------------------------

#include "cmr.h"
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <rw/regexp.h>

#define GENERATE_PROG_NAME      "gen"
#define EXPORT_PROG_NAME        "./arcWrite"
#define P_EXPORT_PROG_NAME      "./ParcWrite"
#define C_EXPORT_PROG_NAME      "./CarcWrite"
#define IMPORT_PROG_NAME        "./arcRead"
#define P_IMPORT_PROG_NAME      "./ParcRead"
#define C_IMPORT_PROG_NAME      "./CarcRead"
#define FSCOMP_PROG_NAME        "fscomp"
#define ERASE_PROG_NAME         "rm -Rf"

typedef enum 
{
    purpose_invalid,
    purpose_datacheck,
    purpose_native,
    purpose_cmr,
    purpose_free,
    purpose_synth,
    purpose_script,
    purpose_config,
    purpose_status
} PathPurpose;
typedef enum
{
    pathtype_invalid,
    pathtype_file,
    pathtype_directory
} PathType;

//----------------------------------------------------------------------
// CLASS: TypedPath     --trp
//
//  UNIX path, with Script purpose and path type (file, directory).
//
// PUBLIC DATA MEMBERS:
// PUBLIC METHODS:
//  isValid()
//      Returns TRUE if the object is in a valid state.
//  isEqual()
//      rogue wave required.  TRUE if this obj == other TypedPath obj.
//  operator=()
//  operator==()
//  operator!=()
//      Standard copy and equality operators.
//  setPathType()
//      Define path type as file or directory.
//  appendPurposeDesc()
//      Return text describing the path purpose.
//  appendTypeDesc()
//      Return text describing the path type (invalid, file, directory).
//
// USAGE NOTES:
//----------------------------------------------------------------------
// COLLABORATORS:
// PRIVATE METHODS:
// IMPLEMENTATION NOTES:
//----------------------------------------------------------------------
class TypedPath : public RWCollectable
{
public:
    PathType pathtype_;
    PathPurpose purpose_;
    RWCString path_;

    TypedPath(){};
    TypedPath(const PathPurpose& type)
    {
        purpose_ = type;
        setPathType();
    }
    TypedPath(PathPurpose type, RWCString path)
            :purpose_(type),
             path_(path)
    {
        setPathType();
    }
    ~TypedPath() {};
    RWBoolean isValid(){
        if (path_.isNull())
            return FALSE;
        return TRUE;
    }
    RWBoolean isEqual(const TypedPath& otherPath) const {
        if (purpose_ != otherPath.purpose_)
            return FALSE;
        return (path_ == otherPath.path_);
    }
    void operator= (const TypedPath& otherPath){
        purpose_ = otherPath.purpose_;
        path_ = otherPath.path_;
    }
    RWBoolean operator== (const TypedPath& otherPath) const {
        return isEqual(otherPath);
    }
    RWBoolean operator!= (const TypedPath& otherPath) const {
        return !(isEqual(otherPath));
    }
    virtual setPathType();
    void checkPath(PathType& actualType,
                   RWBoolean& foundTarget,
                   RWBoolean& foundParentWithoutTarget);
    char* appendPurposeDesc(RWCString& s);
    char* appendTypeDesc(RWCString& s);
   
#ifdef VERIFY
    void dump();
    void validate();
    void test();
#endif
};


//----------------------------------------------------------------------
// CLASS: TypedPathDictionary     --trp
//
//  Keyed container of typed path objects.
//
// PUBLIC DATA MEMBERS:
// PUBLIC METHODS:
//  insert()
//      insert a key and associated typed path object
//  retrieve()
//      given a key, return the typed path object
//  clearAndDestroy()
//      clear dictionary and free all contained objects
//  entries()
//      return number of type path entries
//
// USAGE NOTES:
//----------------------------------------------------------------------
// COLLABORATORS:
// PRIVATE METHODS:
// IMPLEMENTATION NOTES:
//----------------------------------------------------------------------
class TypedPathDictionary
{
    RWCollectableString key_;

public:
    RWHashDictionary dictionary_;
    TypedPathDictionary() {};
    ~TypedPathDictionary() {};

    void        insert(const RWCString& pathId, const TypedPath& path);
    RWBoolean   retrieve(const RWCString& pathId, TypedPath& path);
    void        clearAndDestroy() {dictionary_.clearAndDestroy();}
    int         entries() {return (int) dictionary_.entries();}
    
#ifdef VERIFY
    void dump();
    void validate() {};
    void test();
#endif

};

typedef enum 
{
    action_invalid,
    action_label,
    action_unlabel,
    action_export,
    action_import,
    action_erase,
    action_special
} ActionType;
typedef enum 
{
    voltype_invalid,
    voltype_worm,
    voltype_mo
} VolType;

typedef enum
{
    density_invalid,
    density_512,
    density_1024
} Density;

class Volume : public RWCollectable
{
    Density     density_;
    VolType     volType_;
    RWCollectableString volName_;
    RWBoolean   hasBeenAllocated_;

    RWBoolean isEqual(Volume& othervol) {
        return (volName_ == othervol.volName_);
    }
    void operator=(Volume& othervol)
    {
        density_ = othervol.density_;
        volType_ = othervol.volType_;
        volName_ = othervol.volName_;
    }
    unsigned hash() const {
        return volName_.hash();
    }
    
};

typedef enum {
    execute_invalid,
    execute_always,
    execute_ifOk,
    execute_ifNotOk
} ExecuteOption;

typedef enum 
{
    special_notSpecial,
    special_continue,
    special_stop
} ExecuteSpecial;

typedef enum
{
    impact_none,
    impact_toFailOnly,
    impact_absolute // to pass OR to fail
} StatusImpact;


//----------------------------------------------------------------------
// CLASS: ScriptAction     --trp
//  Defines one script action (label, unlabel, export, import, erase),
//  contains the command line.  Continue or stop special actions allow
//  for the conditional onPass and OnFail actions.
//
// PUBLIC DATA MEMBERS:
// PUBLIC METHODS:
//  isValid()
//
//  isEqual()
//  verifyFailure()
//      Posts to stderr, and marks object status as invalid.
//  verifyAction()
//      Verify that the action is intrinsically valid.
//  buildCommandLine()
//      Create the command line string.
//
// USAGE NOTES:
//----------------------------------------------------------------------
// COLLABORATORS:
// PRIVATE METHODS:
// IMPLEMENTATION NOTES:
//----------------------------------------------------------------------
class ScriptAction : public RWCollectable//, public BaseCmr
{
public:
    RWCString           commandLine_;
    RWCString           executableName_;
    ObjectStatus        object_status;
    int                 addedVolumes;
    ExecuteOption       execute_option;
    ExecuteSpecial      specialAction;
    StatusImpact        statusImpact;
    ActionType          actionType;
    
    ScriptAction (char* executableName) :
            executableName_(executableName),
            object_status (status_normal),
            addedVolumes (0),
            execute_option(execute_invalid)
    {}
    ~ScriptAction () {};

    RWBoolean isValid() {};
    virtual RWBoolean isEqual (ScriptAction& otherAction);
    void verifyFailure(char* reason);
};

typedef enum 
{
    actionstat_ok,
    actionstat_abort,
    actionstat_pass,
    actionstat_fail
} ActionStatus;

void showCommandLine(RWCollectable *c, void *dummy);
void tallyAddedVolumes(RWCollectable *c, void *volListObject);
void executeActions(RWCollectable *c, void *volListObject);
    
#define MAX_ACTIONS 10
extern FILE* scriptin;

//----------------------------------------------------------------------
// CLASS: Script     --trp
//  Encapsulates the parsing of an arcTest script, then carries out those
//  actions.
//  
//
// PUBLIC DATA MEMBERS:
// PUBLIC METHODS:
//
// USAGE NOTES:
//----------------------------------------------------------------------
// COLLABORATORS:
// PRIVATE METHODS:
// IMPLEMENTATION NOTES:
//
//
//----------------------------------------------------------------------
class Script //: public BaseCmr
{
    
public:
    Script(char* scriptFileName);
    ~Script() {fclose(inputFile_);}
    RWBoolean isValid() {return TRUE;}
    static ActionStatus lastActionStatus;
    void dump() {
        printf ("\n Script %s\n----------------------------\n",
                (char*) scriptId_);
        printf ("\n Path Dictionary \n----------------\n");
        pathDict_.dump();
        printf ("\n Actions\n--------\n");
        actionQueue.apply(showCommandLine, NULL);

        printf ("\nTotal volumes to add\n------------------\n");
//        actionQueue.apply(tallyAddedVolumes, volList_);
    }

    RWBoolean loadScript() {return scriptparse();}
    RWBoolean verifyScript();
    RWBoolean executeScript() {
        executeExImOnly = FALSE;
/**/        actionQueue.apply(showCommandLine, NULL);
        actionQueue.apply(executeActions, this);
    }

//  A special state is available for WORM testing, which allows
// export and import commands only to be performed until failure.  This
// state is reflected in "executeExImOnly".  The "executeScriptUntilFailure"
// method runs through the actions once normally, then a second time
// with exImOnly set to true.
    RWBoolean executeExImOnly;
    int repeatFileSetCounter;
    RWCString repeatFileSetName;
    RWBoolean executeScriptUntilFailure() 
    {
        executeExImOnly = FALSE;
        actionQueue.apply(executeActions, this);
        executeExImOnly = TRUE;
        repeatFileSetCounter = 1;
        while (lastActionStatus == actionstat_ok ||
               lastActionStatus == actionstat_pass)
            actionQueue.apply(executeActions, this);
        executeExImOnly = FALSE;
    }
    
    RWBoolean scriptStatus();
    int analyzeSystemReturn(int systemReturn, StatusImpact impact);
    static ObjectStatus object_status;
    static invalidateObject() {
        object_status = status_invalid;
        lastActionStatus = actionstat_abort;
    }
    
    void setRedirectOutputFile(char* outFileName)
    {
        outputFileName_ = outFileName;
        redirectOutput = TRUE;
    }
    RWCString outputFileName_;
    RWBoolean redirectOutput;
    char* scriptId()
    {
        return (char*)scriptId_;
    }
    RWBoolean trace_;
    void turnOnTrace() {trace_ = TRUE;}
    void turnOnArcLibTrace() {arcLibTrace_ = TRUE;}
    
    RWBoolean debug;
    void turnOnDebug() {debug = TRUE;}
    void usePurifyVersion() { usePurify = TRUE;}
    void useCoverageVersion() { useCoverage = TRUE;}

private:
    RWBoolean arcLibTrace_;
    RWBoolean usePurify;
    RWBoolean useCoverage;
    RWCString arcTestRoot;
    RWCString scriptId_;
    int scriptparse ();
    FILE *inputFile_;
    ScriptAction *action;
    RWCString fscompDataCheck_;
    RWCString fscompNative_;


    TypedPathDictionary pathDict_;
    RWSlistCollectablesQueue actionQueue;

    // I don't use this any more.
    void addArcTestRootIfRelative(RWCString& path)
    {
        if (path.length() == 0) {
            path.prepend(arcTestRoot);
            return;
        }
        if (path[0] != '/')
            path.prepend(arcTestRoot);
    }

    // If an action does not contain config and/or status, add what
    // that was stated first in the paths section.  This simplifies
    // script writing, since each label/unlable/export/import generally
    // states the same config & status files.
    void ensureConfigAndStatus()
    {
        // ensure Config
        if (!action->commandLine_.contains("-a"))
            addFirstFoundConfig();
        
        // ensure Status
        if (!action->commandLine_.contains("-s"))
            addFirstFoundStatus();
    }
    

    void addFirstFoundConfig()
    {
        RWHashDictionaryIterator iter(pathDict_.dictionary_);
        TypedPath* path;
        RWCollectableString* str;
        RWBoolean haveConfig = FALSE;

        iter.reset();
        str = (RWCollectableString*) iter();
        while (str != NULL && haveConfig == FALSE) {
            path = (TypedPath*)iter.value();
            if (path->purpose_ == purpose_config) {
                haveConfig = TRUE;
                addStrOption ("a", (char*)path->path_, action->commandLine_);
            }
            str = (RWCollectableString*) iter();
        }
    }
    void addFirstFoundStatus()
    {
        RWHashDictionaryIterator iter(pathDict_.dictionary_);
        TypedPath* path;
        RWCollectableString* str;
        RWBoolean haveStatus = FALSE;

        iter.reset();
        str = (RWCollectableString*) iter();
        while (str != NULL) {
            path = (TypedPath*)iter.value();
            if (path->purpose_ == purpose_status) {
                haveStatus = TRUE;
                addStrOption ("s", (char*)path->path_, action->commandLine_);
            }
            str = (RWCollectableString*) iter();
        }
    }
    
    
    void addStrOption (char* command, char* arg, RWCString& appendToThis);
    void addIdOption (char* command, char* id, RWCString& appendToThis);
    void addOption (char* command, RWCString& appendToThis);

    
    // Methods called by the parser from within the yacc(bison) source.

    // test id
    void scriptId (char* id);

    // file declaration
    void file_start() {}
    void file_datacheck (char* id, char* path);
    void file_native (char* id, char* path);
    void file_cmr (char* id, char* path);
    void file_synth (char* id, char* path);
    void file_script (char* id, char* path);
    void file_config (char* id, char* path);
    void file_status (char* id, char* path);
    void file_free (char* id, char* path);
    void file_end();

    // generate declaration
    void generate_start();
    void generate_native (char* id);
    void generate_script (char* id);
    void generate_datacheck (char* id);
    void generate_end();
    
    // export declaration
    void export_start();
    void export_fileset (char* filesetName);
    void export_volset (char* volName);
    void export_native (char* id);
    void export_cmr (char* id);
    void export_config(char* id);
    void export_status(char* id);
    void export_synth(char* id);
    void export_free (char* id);
    void export_end();

    // label declaration
    void label_start();
    void label_config (char* id);
    void label_status (char* id);
    void label_media (char* mediaType);
    void label_density (char* density);
    void label_volumeSet (char* volSetName);
    void label_fileset (char* fileSetName);
    void label_addVolume (char* volName);
    void label_end();
    
    // grow declaration
    void grow_start();
    void grow_config (char* id);
    void grow_status (char* id);
    void grow_volumeSet (char* id);
    void grow_fileset (char* id);
    void grow_addVolume (char* id);
    void grow_end();

    // import declaration
    void import_start();
    void import_fileset (char* filesetName);
    void import_volset (char* volName);
    void import_native (char* id);
    void import_cmr (char* id);
    void import_config(char* id);
    void import_status(char* id);
    void import_synth();
    void import_end();
    void import_free(char* id);

    // fscomp declaration
    void fscomp_start();
    void fscomp_datacheck(char* id);
    void fscomp_native(char* id);
    void fscomp_end();

    // on pass declarations
    void onPass_start();
    void onPass_erase (char* id);
    void onPass_unLabel (char* id);
    void onPass_stop ();
    void onPass_end();
    
    // on fail declaration
    void onFail_start();
    void onFail_erase (char* id);
    void onFail_unLabel (char* id);
    void onFail_continue ();
    void onFail_stop ();
    void onFail_end();
};

#endif // __ARC_TEST_H__





