#ifndef _NSR_CMRDIFF_H_ //-*-c++-*-
#define _NSR_CMRDIFF_H_

// -----------------------------------------------------------------------------
//  cmrDiff.h
//      CmrDiff is used for comparing two UDF compliant representations.  One
//      of the representations must control the process, and is referred to as
//      the "driver".  The other representation is passive relative to CmrDiff,
//      and is referred to as the "rider".  CmrDiff provides four functions to
//      be called by the "driver" code (begin dir, end dir, file, symlink),
//      and requires two function pointers to be provided by the "rider" code
//      (root directory access, general directory access).  A log file is
//      created, which contains a hierarchical list of differences between the
//      "driver" and "rider" representations.
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//  CLASSES:
//  ValNode             Base class for validating a fileset node.
//  FileNode            File validation node, derived from ValNode.
//  SymlinkNode         Symlink validation node, derived from ValNode.
//  ValDirectory        Owns/manages sorted directory entries.
//  Log                 Manages numeric formatting and log file writing CmrDiff
//                      throught the LineBuffer class.
//  CmrDiff             Owns a stack of ValDirectory objects, manages fileset
//                      validation.
//
// -----------------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msVerify/cmrDiff.cpp,v $
// $Revision: 1.11 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/01/10 11:17:09 $
//-----------------------------------------------------------------------------

#include <stdlib.h>
#include "cmr.h"


//----------------------------------------------------------------------
// METHOD: ValNode::compareTo  --trp
//
//  Sorting function used by the rogue wave containers.
//
// ARGS:
//  RWCollectable * c  IN      Node to be compared to.
//
// RETURNS:
//   1 if other node is prior in sort
//  -1 if other node is latter in sort
//
// PRE-CONDITIONS:
//   Class local node has a defined name and type.
//   Comparison node has a valid name and type.
//  
// POST-CONDITIONS:
//
// ERRORS:
//
// NOTES:
//
//----------------------------------------------------------------------
int
ValNode::compareTo(const RWCollectable* c) const
{
    const ValNode* v = (const ValNode*)c;

    if (type_ != v->type_)
        return type_ > v->type_ ? 1 : -1;
    
    return name_ > v->name_ ? 1 : -1;
}


//----------------------------------------------------------------------
// METHOD: ValNode::isCommonEqualWithLog  --trp
//
//  
//
// ARGS:
//  RWCollectable* c    IN      node for comparison
//  Log*           log  OUT     Logging object, to receive and print
//                              difference info, if not NULL.
//
// RETURNS:
//  TRUE if the other node is equal.
//  FALSE if the other node is not equal.
//
// PRE-CONDITIONS:
//  The RWCollectable* points to a valid ValNode, FileNode or SymlinkNode
//  object.
//  The Log* is either NULL, or points to a valid logging object.
//  
// POST-CONDITIONS:
//  If the Log* is not NULL, and differences are found, the log file will
//  contain difference information.
//
// ERRORS:
//
// NOTES:
//
//----------------------------------------------------------------------
// Determine type, name and attribute equality, logging any differences to
// the log file IF the Log object is not NULL.
RWBoolean
ValNode::isCommonEqualWithLog (const RWCollectable* c, Log *log)
{
    ValNode* v = (ValNode*)c;
    RWBoolean isEqual = TRUE;

    if (type_ != v->type_) {
        if (log != NULL) {
            RWCString typeDesc;
            log->beginSection("type");
            log->beginSection((char*)(RWCString)*(RWCollectableString*)
                              log->referenceLabels[0]);
            getTypeDesc(type_, typeDesc);
            log->addChunk(typeDesc);
            log->endLine();
            log->endSection();

            log->beginSection((char*)(RWCString)*(RWCollectableString*)
                              log->referenceLabels[1]);
            getTypeDesc(v->type_, typeDesc);
            log->addChunk(typeDesc);
            log->endLine();
            log->endSection();
            log->endSection();

            isEqual = FALSE;
        }
    }
    
    if (name_ != v->name_) {
        if (log != NULL) {
            log->beginSection("name");
            log->beginSection((char*)(RWCString)*(RWCollectableString*)
                              log->referenceLabels[0]);
            log->addChunk(name_);
            log->endLine();
            log->endSection();
            log->beginSection((char*)(RWCString)*(RWCollectableString*)
                              log->referenceLabels[1]);
            log->addChunk(v->name_);
            log->endLine();
            log->endSection();
            log->endSection();

            isEqual = FALSE;
        }
    }

    if (attr_ != v->attr_) {
        RWCString attrDesc = "";
        log->beginSection("attributes");
        log->beginSection((char*)(RWCString)*(RWCollectableString*)
                          log->referenceLabels[0]);
        log->addChunk(attr_.appendAttributes(attrDesc=""));
        log->endLine();
        log->endSection();
        
        log->beginSection((char*)(RWCString)*(RWCollectableString*)
                          log->referenceLabels[1]);
        log->addChunk(v->attr_.appendAttributes(attrDesc=""));
        log->endLine();
        log->endSection();
        log->endSection();
    }
    
    return isEqual;
}

//----------------------------------------------------------------------
// METHOD: ValNode::isEqual  --trp
//
//  Use isCommonEqualWithLog to provide uniqueness info to the roguewave
//  container.
//
// ARGS:
//  RWCollectable*      c       IN      comparison object.
//
// RETURNS:
//  TRUE if the nodes are equal.
//  FALSE if the nodes are different.
//
// PRE-CONDITIONS:
//  The RWCollectable* points to a valid ValNode, FileNode or SymlinkNode
//  object.
//  
// POST-CONDITIONS:
//
// ERRORS:
//
// NOTES:
//
//----------------------------------------------------------------------
RWBoolean
ValNode::isEqual(const RWCollectable* c)
{
    return isEqualWithLog(c, NULL);
}

//----------------------------------------------------------------------
// METHOD: ValNode::isEqualWithLog  --trp
//
//  Compares to Valnodes, logging differences and returning equality.
//
// ARGS:
//  void*       parent  IN      driver dir parent, or NULL for root
//  Log*           log  OUT     Logging object, to receive and print
//                              difference info, if not NULL.
//
// RETURNS:
//
// PRE-CONDITIONS:
//  The RWCollectable* points to a valid ValNode, FileNode or SymlinkNode
//  object.
//  
// POST-CONDITIONS:
//
// ERRORS:
//
// NOTES:
//
//----------------------------------------------------------------------
// This isEqual does log differences.  It is not used by roguewave.
RWBoolean
ValNode::isEqualWithLog(const RWCollectable* c, Log *log)
{
    if (!isCommonEqualWithLog(c, NULL))
        return FALSE;

    if (type_ == nodetype_file) {
        ValNode* v = (ValNode*) c;
        return CmrTools::elistCompare (elist_, v->elist_, log);
    }
    if (type_ == nodetype_symlink) {
        ValNode* v = (ValNode*) c;
        return CmrTools::plistCompare (plist_, v->plist_, log);
    }

    return TRUE;
}

//----------------------------------------------------------------------
// METHOD:  ValNode::operator<<  --trp
//
//  Populate Valnode with name or attributes.
//
// ARGS:
//  char *      name    IN      string, defining node name
//  Attributes  attr    IN      node defining attr struct
//
// RETURNS:
//  void
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
// The "<<" operator is provided for a terse way to populate ValNodes.
void
ValNode::operator<<(char *name)
{
    name_ = (const char *) name;
}
void
ValNode::operator<<(Attributes attr)
{
    attr_ = attr;
}
void
ValNode::operator<<(ExtentList elist)
{
    elist_ = elist;
}

void
ValNode::operator<<(PathElementList plist)
{
    plist_ = plist;
}

//----------------------------------------------------------------------
// METHOD: ValNode::getTypeDesc  --trp
//
//  Given a node type, provide a description string.
//
// ARGS:
//  NodeType&   node     IN      type of node (dir, symlink, file)
//  RWCString&  typeDesc OUT     description of node type
//
// RETURNS:
//  void
//
// PRE-CONDITIONS:
//  
// POST-CONDITIONS:
//
// ERRORS:
//
// NOTES:
//
//----------------------------------------------------------------------
void
ValNode::getTypeDesc(NodeType& type, RWCString &typeDesc)
{
    switch (type) {
    case nodetype_file:
        typeDesc = "FILE";
        break;
        
    case nodetype_symlink:
        typeDesc =  "SYMLINK";
        break;
        
    case nodetype_dir:
        typeDesc =  "DIRECTORY";
        break;
        
    case nodetype_unknown:
    default:
        typeDesc =  "UNKNOWN TYPE";
    }
}

//----------------------------------------------------------------------
// METHOD: ValDirectory::ValDirectory  --trp
//
//  ValDirectory contructor.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
//  
// POST-CONDITIONS:
//  identity is registered.
//  name is initialized to a debugging value
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------

ValDirectory::ValDirectory():
        name_("invalid name"),
        BaseCmr("ValDirectory")
{
}

//----------------------------------------------------------------------
// METHOD: ValDirectory::~ValDirectory  --trp
//
//  ValDirectory destructor.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  The riderEntries_ and driverEntries_ sorted lists and contents are
//  destroyed.
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
ValDirectory::~ValDirectory()
{
    riderEntries_.clearAndDestroy();
    driverEntries_.clearAndDestroy();
}

//----------------------------------------------------------------------
// METHOD: ValDirectory::isEqual  --trp
//
//  Equality for ValDirectories - used only by rogue wave.
//
// ARGS:
//  RWCollectable c     IN      node for comparison
//
// RETURNS:
//  TRUE if the  are equal.
//  FALSE if the nodes are not equal.
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
// ERRORS:
//
// NOTES:
//
//----------------------------------------------------------------------
RWBoolean
ValDirectory::isEqual (const RWCollectable* c)
{
    const ValDirectory* v = (const ValDirectory*)c;

    return (name_ == v->name_ &&
            appDirPointer_ == v->appDirPointer_);
}

//----------------------------------------------------------------------
// METHOD: CmrDiff::CmrDiff  --trp
//
//  CmrDiff constructor.
//
// ARGS:
//
//  RiderAccessFunc     riderAccess        IN App node access.
//  RiderRootAccessFunc riderRootAccess    IN App root access.
//  RiderEndDirectory   riderEndDirectory  IN Notifies app of dir end.
//  char*               driverDesc         IN Logging file driver desc.
//  char*               riderDesc          IN Logging file rider desc.
//  char*               logFileName        IN file to receive diff desc.
//
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  Class identity is registered.
//  App access function pointers are saved.
//  Rider and driver desc strings are saved for logging description.
//  Log file is opened.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
CmrDiff::CmrDiff(RiderAccessFunc riderAccess,
                 RiderRootAccessFunc riderRootAccess,
                 RiderEndDirectory riderEndDirectory,
                 char *driverDesc,
                 char *riderDesc,
                 char *logFileName):
        BaseCmr("CmrDiff"),
        riderAccess_ (riderAccess),
        riderRootAccess_ (riderRootAccess),
        riderEndDirectory_ (riderEndDirectory),
        isFirstDirectory(TRUE)
{
    // Set up the difference logging object.
    log_ = new Log(logFileName);

    // Set up driver and rider reference labels in the log object.  The order
    // of insertion defines the access index.
    log_->insertReferenceLabel(driverDesc);
    log_->insertReferenceLabel(riderDesc);

    declareErrorFile ("CmrDiff.err");
}

//----------------------------------------------------------------------
// METHOD: CmrDiff::~CmrDiff  --trp
//
//  CmrDiff desctructor.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  The logging object is deleted.
//  The directory stack and contents are deleted.
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
CmrDiff::~CmrDiff()
{
    delete log_;
    dirStack_.clearAndDestroy();
}

//----------------------------------------------------------------------
// METHOD: CmrDiff::beginDir  --trp
//
//  Sets up difference work for a new directory.  In the case of root,
//  the name & attributes are immediately checked.
//
// ARGS:
//  void*       parent  IN      driver dir parent, or NULL for root
//  char*       dirName IN      driver dir name, or "/" for root
//  Attributes& attr    IN      driver dir attributes
//
// RETURNS:
//  void data pointer to application domain directory node.
//
// PRE-CONDITIONS:
//  the member is called for root on the first instance only
//  parent is non null for non-root
//  dirName is "/" for root, and root only
//  
// POST-CONDITIONS:
//  if root, root is validated against the rider root node
//  if not root, the directory is inserted into the top dirStack_
//      ValDirectory riderEntries_ sorted list
//  all rider child entries are inserted into a new ValDirectory
//      riderEntries_ sorted list
//  the new ValDirectory object is pushed into the member dirStack_ 
//
// ERRORS:
//
// NOTES:
//
//----------------------------------------------------------------------
void
buildFullPath (RWCollectable* c, void *str)
{
    RWCString *s = (RWCString*) str;
    if (strlen(((ValDirectory *)c)->name()) > 1) // skip root "/"
    {
        s->prepend(((ValDirectory *)c)->name());
        s->prepend("/");
    }
}
void*
CmrDiff::beginDir(void *parent, char *dirName, Attributes& attr)
{
    // Root (assumed to be first directory - multiple filesets not supported)
    // is validated immediately here.
    if (isFirstDirectory == 1) {
        isFirstDirectory = 0;
        
        // Get the name, attributes and data pointer for the application
        // root node.
        Attributes riderRootAttr;
        RWCString riderRootName;
        riderRootAccess_ (riderRootName, riderRootAttr, &riderRoot_);

        // If the name or attributes are not equal, log the difference.
        log_->beginSection("<ROOT>");
        if (strcmp (dirName, riderRootName) != 0) {
            log_->beginSection("name");
            log_->beginSection((char*) (RWCString) *(RWCollectableString *)
                           log_->referenceLabels[0]);
            log_->addChunk(dirName);
            log_->endLine();
            log_->endSection();
            
            log_->beginSection((char*) (RWCString) *(RWCollectableString *)
                           log_->referenceLabels[1]);
            log_->addChunk(riderRootName);
            log_->endLine();
            log_->endSection();
            log_->endSection();
        }

        if (attr != riderRootAttr){
            RWCString attrDesc = "";
            log_->beginSection("attributes");
            log_->beginSection((char*)(RWCString)*(RWCollectableString*)
                              log_->referenceLabels[0]);
            log_->addChunk(attr.appendAttributes(attrDesc=""));
            log_->endLine();
            log_->endSection();
        
            log_->beginSection((char*)(RWCString)*(RWCollectableString*)
                              log_->referenceLabels[1]);
            log_->addChunk(riderRootAttr.appendAttributes(attrDesc=""));
            log_->endLine();
            log_->endSection();
            log_->endSection();
        }
        
        log_->endSection(); // "<ROOT>"
    }
    
    // Not root
    else
    {
        // Make a new directory stack object, with current dir name.
        ValNode *dirNode = new ValNode(nodetype_dir, (char*)dirName, attr);

        // Insert the current directory into the parent driver entries list.
        ValDirectory *valDir;
        (RWCollectable *) valDir = dirStack_.top();
        valDir->driverEntries_.insert (dirNode);
    }
    
    // Create a ValDirectory object for this directory, assign the name and
    // rider entries, then push it onto the stack.
    ValDirectory *dir = new ValDirectory;
    *dir << dirName;
    RWBoolean targetIsRoot = dirStack_.entries() == 0 ? TRUE : FALSE;

//    fprintf (stdout, "CmrDiff::beginDir, dirName: %s\n", dirName);

    // Build the full path string (e.g. "/users/you/foo/").
    RWCString path;
    dirStack_.apply(buildFullPath, &path);
    if (path.length() < 1)
        path = "/";
    else
        path += "/";

    riderAccess_ (path,
                  &((*dir).riderEntries_),
                  &(dir->appDirPointer_));
    dirStack_.push(dir);

    // The application domain data pointer for this directory is returned to
    // the CmrReader class.  CmrReader will pass this pointer back as the
    // parent of child objects.
    return (void*)(dir->appDirPointer_);
}

//----------------------------------------------------------------------
// METHOD: CmrDiff::endDir  --trp
//
//  Performs difference work on the current completed directory.
//
// ARGS:
//  void*       parent  IN      driver dir parent, or NULL for root
//  char*       dirName IN      driver dir name, or "/" for root
//  Attributes& attr    IN      driver dir attributes
//
// RETURNS:
//  void data pointer to application domain directory node.
//
// PRE-CONDITIONS:
//  the member is called for root on the first instance only
//  parent is non null for non-root
//  dirName is "/" for root, and root only
//  
// POST-CONDITIONS:
//  if root, root is validated against the rider root node
//  if not root, the directory is inserted into the top dirStack_
//      ValDirectory riderEntries_ sorted list
//  all rider child entries are inserted into a new ValDirectory
//      riderEntries_ sorted list
//  the new ValDirectory object is pushed into the member dirStack_ 
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------

void
CmrDiff::endDir(void *closingDir)
{
    // Validate each driver/rider pair for the closing directory.
    ValDirectory *valDir = (ValDirectory*) dirStack_.pop();
    //(RWCollectable *) valDir = dirStack_.top();
    
    RWBinaryTreeIterator driverIter(valDir->driverEntries_);
    RWBinaryTreeIterator riderIter(valDir->riderEntries_);
    ValNode *driverNode, *riderNode;

    (RWCollectable*) driverNode = driverIter();
    (RWCollectable*) riderNode = riderIter();

    // Build the full path string (e.g. "/users/you/foo/").
    RWCString path;
    dirStack_.apply(buildFullPath, &path);
    if (path.length() < 1)
        path = "/";
    else
        path += "/";

    if (closingDir != NULL) {
        while (driverNode && riderNode)
        {
            RWCString pathWithNode(path);
            pathWithNode += driverNode->name();
            pathWithNode.prepend(": ");
            RWCString typeDesc;
            ValNode::getTypeDesc(driverNode->type_, typeDesc);
            pathWithNode.prepend(typeDesc);
            log_->beginSection(pathWithNode);
            if (!driverNode->isEqualWithLog(riderNode, log_)) {
                ;
            }
            log_->endSection();
        
            (RWCollectable*) driverNode = driverIter();
            (RWCollectable*) riderNode = riderIter();
        }
    
        // Make sure there is a 1 to 1 relationship to the directory entries.
        if (driverNode)
        {
            log_->addChunk("file counts not equal for directory");
            log_->addChunk((char*) path);
            log_->endLine();
            log_->addChunk((char*) (RWCString) *(RWCollectableString *)
                           log_->referenceLabels[0]);
            log_->addChunk("has more entries");
            log_->endLine();
        }

        if (riderNode)
        {
            log_->addChunk("file counts not equal for directory");
            log_->addChunk((char*) path);
            log_->endLine();
            log_->addChunk((char*) (RWCString) *(RWCollectableString *)
                           log_->referenceLabels[1]);
            log_->addChunk("has more entries");
            log_->endLine();
        }
    
        // Clear out the validated directory.
        if (riderEndDirectory_ != NULL)
            riderEndDirectory_(valDir->appDirPointer_);
    }

    delete valDir;
}

//----------------------------------------------------------------------
// METHOD: CmrDiff::file  --trp
//
//  Collect file information for difference checking.  
//
// ARGS:
//  char*       fileName IN     file name, from the reader
//  Attributes& attr     IN     file attributes, from the reader
//  ExtentList& elist    IN     file extent list, from the reader
//
// RETURNS:
//  void
// PRE-CONDITIONS:
//  
// POST-CONDITIONS:
//  The current file is insert into the directory stack top ValDirectory
//  object sorted driverEntries list.
//
// ERRORS:
//
// NOTES:
//  The actual difference check is performed in endDir.
//
//----------------------------------------------------------------------
void
CmrDiff::file(char *fileName, Attributes& attr, ExtentList& elist)
{
    // Create a new file node, and add it to the current directories driver
    //list.
    ValNode *file = new ValNode (nodetype_file, fileName, attr);
    *file << elist;
    
    ValDirectory *valDir;
    (RWCollectable *) valDir = dirStack_.top();
    valDir->driverEntries_.insert (file);
}

//----------------------------------------------------------------------
// METHOD: CmrDiff::symlink  --trp
//
//  Collect difference checking information for a symlink.
//
// ARGS:
//  char*               symlinkName     name of symlink, from reader
//  Attributes&         attr            symlink attributes, from reader
//  PathElementList&    plist           symlink plist, from reader
//
// RETURNS:
//  void
//
// PRE-CONDITIONS:
//  
// POST-CONDITIONS:
//  The current symlink is insert into the directory stack top ValDirectory
//  object sorted driverEntries list.
//
// ERRORS:
//
// NOTES:
//
//----------------------------------------------------------------------
void
CmrDiff::symlink(char *symlinkName, Attributes& attr,
                      PathElementList& plist)
{
    // Create a new symlink node, and add it to the current directories
    // driver list.
    ValNode *symlink = new ValNode (nodetype_symlink, symlinkName, attr);
    *symlink << plist;

    ValDirectory *valDir;
    (RWCollectable *) valDir = dirStack_.top();
    valDir->driverEntries_.insert (symlink);
}

#endif
