// CmrFormat.C

// -----------------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msVerify/cmrFormt.cpp,v $
// $Revision: 1.4 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 95/10/25 17:27:28 $
//-----------------------------------------------------------------------------
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// CmrFormat       Reads a CMR file and re-writes CMR to a second file, using
//                 format options specified on the command line.
//
// This program was written primarily as test and example code for the
// CmrReader and CmrWriter classes.
// You can also use this program to reformat a CMR file.  Only the indent tab
// size and short/long reserved words are reformatted.
//
// usage
//     CmrFormat -i <input file> -o <output file> -t <tab size> -w <word size>
//
//-----------------------------------------------------------------------------

#include "../nsrLib.h"
#include "cmr.h"

// Command line collection.
typedef struct
{
    RWCString infile;
    RWCString outfile;
    int tabSize;
    ReservedWordsType wordSize; // type is defined in cmr.h
} cmdVal;
static int getCommandLine (int argc, char *argv[], cmdVal &cmdValues);
static void Usage (void);

// These functions are passed to the CmrReader object.  They must
// agree in form with the function pointer typedefs defined for CmrReader
// in cmr.h (just prior to the class definition).
VoidPtr myStartDir (VoidPtr parentDir, char *name, Attributes& attr);
void myEndDir (VoidPtr closedDir);
void myFile (char *name, Attributes& attr, ExtentList& elist);
void mySymlink (char *name, Attributes& attr, PathElementList& plist);

// trace must be defined when you use the archive library.
int trace = 0;

// I declared this global writer pointer, since it is called in each
// of my four reader callback functions.  In a more substantial program,
// passing the object to each of the functions would perhaps be more
// appropriate.
CmrWriter *cmrWrite;

//-----------------------------------------------------------------------------
// main         Gets command line, instantiates a writer, then instantiates
//              a reader and intitializes the parse.
//-----------------------------------------------------------------------------
main (int argc, char *argv[])
{
    // Get command line info.
    cmdVal cmdValues;
    getCommandLine (argc, argv, cmdValues);

    // The CmrWriter is instantiated with info from the command line.
    cmrWrite = new CmrWriter (cmdValues.outfile);
    cmrWrite->setTabSize(cmdValues.tabSize);
    if (cmdValues.wordSize == shortWords)
        cmrWrite->useShortReservedWords();

    // Once our callback functions are defined, we merely need to
    // instantiate a reader and initiate the parse.
    // The parser is generated by flex/bison, hence the odd "cmrparse"
    // member name.
    CmrReader cmrReader (cmdValues.infile, &myStartDir,
                         &myEndDir, &myFile, &mySymlink);
    cmrReader.cmrparse ();

    // Since we instantiated the writer with "new", if we don't
    // explicitly delete it here, the destructor won't be of
    // use to us.  If you are missing the fileset termination
    // token "end-fileset" or "es", you probably are not accessing
    // the destructor.
    delete cmrWrite;
}

// -----------------------------------------------------------------------------
// The four reader callback functions are defined here ("callback" merely means
// that we pass pointers to our functions to the reader, and it "calls back"
// into these functions).
//
// myStartDir and myEndDir use a void pointer, typed to "VoidPtr" to keep track
// of parent directories. This trivial example merely passes NULL to the reader,
// but your application would most likely pass a real pointer.
//
// When the reader class calls the "myStartDir" function pointer, the function
// returns a pointer to the directory node in the application data structure.
// The reader saves that pointer on an internal stack.  When the reader calls
// your "myStartDir" function on a directory which is a child of that directory,
// it returns the original (parent) directory pointer in the "parentDir"
// parameter.  Refer to the other examples for a more realistic use of the
// directory pointers.
// -----------------------------------------------------------------------------

VoidPtr myStartDir (VoidPtr parentDir, char *name, Attributes& attr){
  cmrWrite->beginDir (name, attr);
  return NULL;
}

void myEndDir (VoidPtr closedDir){
  cmrWrite->endDir ();
}

void myFile (char *name, Attributes& attr, ExtentList& elist){
  cmrWrite->file (name, attr, elist);
}

void mySymlink (char *name, Attributes& attr, PathElementList& plist){
  cmrWrite->symlink (name, attr, plist);
}

// -----------------------------------------------------------------------------
// Only command line stuff from here to end of file.
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// FUNCTION:  getCommandLine()
//
// Process command line arguements, and pass back either a populated cmdVal
// struct, or print usage and return -1.  Added args must be added to cmdVal
// typedef.
//-----------------------------------------------------------------------------

static int
getCommandLine (int argc, char *argv[], cmdVal & cmdValues)
{
    int c;
    int haveInFile = 0, haveOutFile = 0, haveInvalidInput = 0;
    int haveTabSize = 0, haveWordSize = 0;

    extern char *optarg;                // For getopt()

    // Parse command-line arguments.
    while ((c = getopt (argc, argv, "i:o:t:w:")) != EOF)
    {
        switch (c)
        {
        case 'i':
            cmdValues.infile = optarg;
            haveInFile = 1;
            break;
        case 'o':
            cmdValues.outfile = optarg;
            haveOutFile = 1;
            break;
        case 't':
            cmdValues.tabSize = atoi(optarg);
            if (cmdValues.tabSize == 0)
                break;
            haveTabSize = 1;
            break;
        case 'w':
            if (strcmp(optarg, "s") == 0)
            {
                cmdValues.wordSize = shortWords;
                haveWordSize = 1;
                break;
            }
            else if (strcmp(optarg, "l") == 0)
            {
                cmdValues.wordSize = longWords;
                haveWordSize = 1;
                break;
            }
            else
            {
                break;
            }
            
        default:
            haveInvalidInput = 1;
            break;
        }
    }

    if (!haveInFile || !haveOutFile || haveInvalidInput ||
        !haveTabSize || !haveWordSize)
    {
        Usage ();
        exit (2);
    }
    return 0;
}

// -----------------------------------------------------------------------------
// FUNCTION:  Usage()
//
// Prints usage information to screen.
//-----------------------------------------------------------------------------

static void
Usage (void)
{
    static char *usage_str[] =
    {
        "Usage:\n",
        "CmrFormat -i infile -o outfile -t tab size -w word size\n",
        "-i infile:     original CMR file",
        "-o outfile:    new CMR file, with short names and tab size of 1",
        "-t tab size:   number of characters to shift upon directory indent",
        "-w word size:  must be 'l' or 's' for long or short reserved words",
        "",
        "Example:\n",
        "./CmrFormat -i my_cmr_file -o new_cmr_file -t 4 -w s\n",
        NULL
    };
    int i = 0;

    while (usage_str[i] != NULL)
    {
        printf ("%s\n", usage_str[i]);
        i++;
    }
}



