// -*-c++-*-
//----------------------------------------------------------------------
//  cmrReader.C
//   Parses CMR files, providing dir, file & symlink info via function
//   pointers.
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//  CLASSES:
//   CmrReader
//  
//----------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msVerify/cmrRead.cpp,v $
// $Revision: 1.10 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 95/10/30 12:26:27 $
//-----------------------------------------------------------------------------

#include "cmr.h"
#include <rw/cstring.h>

#define VERIFY_CMR_STATE(A) \
if (CmrReader::cmrState_ != (A)) \
CmrReader::error.logError(errorWarning, "syntax or parser error");\
if (CmrReader::cmrState_ != (A)) \
object_status = status_invalid

//----------------------------------------------------------------------
// METHOD: CmrReader::CmrReader  --trp
//      
//  CmrReader contructor.
//
// ARGS:
//  char*            inputFileName  IN  input CMR file to be read
//  StartDirFunc     userStartDir   IN  app code func to be called with
//                                      dir start info
//  EndDirFunc       userEndDir     IN  app code func to be called with
//                                      dir end info
//  FileFunc         userFile       IN  app code func to be called with
//                                      file info
//  SymlinkFun       userSymlink    IN  app code func to be called with
//                                      symlink info
//
// RETURNS:
// 
// PRE-CONDITIONS:
//  All parameter info is saved.
//  Identity is registered.
//  Input file is opened.
// 
// POST-CONDITIONS:
// ERRORS:
//  Object disabling error if file cannot be opened write only.
//
// NOTES:
//  Second signiture constructor takes no arguments, and aquires info
//  via setInFil and setCallbackFunctions.
//----------------------------------------------------------------------
CmrReader::CmrReader (char *inputFileName,
           StartDirFunc userStartDir, EndDirFunc userEndDir,
           FileFunc userFile, SymlinkFunc userSymlink)
        :BaseCmr("CmrReader")
{
    object_status = status_uninitialized;
    
  // Open the input file read only.
  inputFile_ = fopen (inputFileName, "r");
  if (inputFile_ == NULL) {
      error.logError (errorWarning, "input file cannot be opened for reading");
      object_status = status_invalid;
      return;
  }
  
  cmrin = inputFile_;

  // Save user function pointers.
  userStartDir_ = userStartDir;
  userEndDir_ = userEndDir;
  userFile_ = userFile;
  userSymlink_ = userSymlink;

  // Initialize state
  cmrState_ = cmr_default;
  plist_ = NULL;
  elist_ = NULL;
  object_status = status_normal;
  
}
CmrReader::CmrReader ()
        :BaseCmr("CmrReader")
{
  // Initialize state
    cmrState_ = cmr_default;
    plist_ = NULL;
    elist_ = NULL;
}

//----------------------------------------------------------------------
// METHOD: CmrReader::~CmrReader  --trp
//      
//  CmrReader destructor.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  If a non-NULL plist or elist are present, they are deleted.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
CmrReader::~CmrReader ()
{
    if (elist_ != NULL)
        delete elist_;
    if (plist_ != NULL)
        delete plist_;
}

//----------------------------------------------------------------------
// METHOD: CmrReader::setInFile  --trp
//      
//  Declare the input CMR file.
//
// ARGS:
//  char*  inputFileName  IN   Name of input CMR file.
//
// RETURNS:
//  void
// 
// PRE-CONDITIONS:
//  The input filename points to a file which can be opened write only.
// 
// POST-CONDITIONS:
//  The input file is opened, or the program is terminted.
//
// ERRORS:
//  The input file can be opened write only.
//
// NOTES:
//
//----------------------------------------------------------------------
void CmrReader::setInFile(char *inputFileName)
{
    if (inputFile_ == NULL) {
        error.logError (errorWarning, "input file cannot be opened for reading");
        object_status = status_invalid;
        return;
    }

    cmrin = inputFile_;
}

//----------------------------------------------------------------------
// METHOD: CmrReader::setCallBackFunctions  --trp
//      
//  Means (other than one of the constructors) to define app
//  callback functions.
//
// ARGS:
//  StartDirFunc     userStartDir   IN  app code func to be called with
//                                      dir start info
//  EndDirFunc       userEndDir     IN  app code func to be called with
//                                      dir end info
//  FileFunc         userFile       IN  app code func to be called with
//                                      file info
//  SymlinkFun       userSymlink    IN  app code func to be called with
//                                      symlink info
//
// RETURNS:
//  void
// 
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  The app function pointers are saved.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrReader::setCallBackFunctions(
    StartDirFunc userStartDir, EndDirFunc userEndDir,
    FileFunc userFile, SymlinkFunc userSymlink)
{
  // Save user function pointers.
  userStartDir_ = userStartDir;
  userEndDir_ = userEndDir;
  userFile_ = userFile;
  userSymlink_ = userSymlink;
}

//----------------------------------------------------------------------
// METHOD: CmrReader::startDir  --trp
//      
//  Save starting directory information.
//
// ARGS:
//  char*       dirName IN      Name of directory.
//
// RETURNS:
// PRE-CONDITIONS:
//  Prior node parsing has been completed.
//
// POST-CONDITIONS:
//  The parser state is set to file processing.
//  Parser state is checked, and the new directory information is
//  initialized.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrReader::startDir (const char *dirName)
{
    VERIFY_OBJECT_STATUS();

  VERIFY_CMR_STATE (cmr_default);
  cmrState_ = cmr_dir;

  attr_.fileType(NSR::file_Directory);
  name_ = dirName;
}
//----------------------------------------------------------------------
// METHOD: CmrReader::sendDir  --trp
//      
//  Send current directory information to the application.
//
// ARGS:
//
// RETURNS:
// PRE-CONDITIONS:
//  Directory name & common data have been collected.
// 
// POST-CONDITIONS:
//  The app start directory function is called.
//  The application directory data pointer is pushed onto the stack.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrReader::sendDir ()
{
    VERIFY_OBJECT_STATUS();
    VERIFY_CMR_STATE (cmr_dir);

    cmrState_ = cmr_default;

    thisDir = userStartDir_ (parentDirStack.top (),
                           name_, attr_);
    parentDirStack.push ((void*)thisDir);
}

//----------------------------------------------------------------------
// METHOD: CmrReader::endDir  --trp
//      
//  Send end directory info to application.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
//  The directory stack is not empty.
//  Current parser mode is directory.
// 
// POST-CONDITIONS:
//  The app is notified of current directory end, and given back a
//  pointer to this directory in their application data.
//  This directory is removed from the stack.
//  Parser mode is set to default.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrReader::endDir ()
{
    VERIFY_OBJECT_STATUS();
    VERIFY_CMR_STATE (cmr_default);

  userEndDir_ (parentDirStack.top ());
  parentDirStack.pop ();
}

//----------------------------------------------------------------------
// METHOD: CmrReader::startFile  --trp
//      
//  File processing is started.
//
// ARGS:
//  char*  fileName   IN   name of new file.
//
// RETURNS:
// PRE-CONDITIONS:
//  Prior node parsing has been completed.
// 
// POST-CONDITIONS:
//  The parser state is set to file processing.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void CmrReader::startFile (const char *fileName)
{
    VERIFY_OBJECT_STATUS();
    VERIFY_CMR_STATE (cmr_default);

    cmrState_ = cmr_file;

  attr_.fileType(NSR::file_Raw);
  name_ = fileName;
}

//----------------------------------------------------------------------
// METHOD: CmrReader::endFile  --trp
//      
//  Call the app file function with completed file info.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
//  Parser mode is file.
// 
// POST-CONDITIONS:
//  Parser mode is set to default.
//  The application "file" function is called.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrReader::endFile ()
{
    VERIFY_OBJECT_STATUS();
    VERIFY_CMR_STATE (cmr_file);

    cmrState_ = cmr_default;

  userFile_ (name_, attr_, *elist_);
}

//----------------------------------------------------------------------
// METHOD: CmrReader::startSymlink  --trp
//      
//  Start symlink processing.
//  <long_desc_if_needed>
//
// ARGS:
//  char*  symLinkName  IN  name of the new symlink.
//
// RETURNS:
//  <explanation of different values>
// 
// PRE-CONDITIONS:
//  Prior node parsing has been completed.
//  List_preconditions  (E.g., assumes)
// 
// POST-CONDITIONS:
//  Parser mode is set to symlink.
//  Symlink information is saved.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrReader::startSymlink (const char *symLinkName)
{
    VERIFY_OBJECT_STATUS();
    VERIFY_CMR_STATE (cmr_default);

    cmrState_ = cmr_symlink;

  attr_.fileType(NSR::file_SymbolicLink);
  name_ = symLinkName;
}

//----------------------------------------------------------------------
// METHOD: CmrReader::endSymLink  --trp
//      
//  End processing of the current symlink, call application function
//  with symlink info.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
//  Parser mode is symlink.
// 
// POST-CONDITIONS:
//  Parser mode is set to default.
//  The user symlink function is called.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrReader::endSymLink ()
{
    VERIFY_OBJECT_STATUS();
    VERIFY_CMR_STATE (cmr_symlink);

    cmrState_ = cmr_default;

  userSymlink_ (name_, attr_, *plist_);
}

//----------------------------------------------------------------------
// METHOD: CmrReader::commonInfo  --trp
//      
//  Set the attr_ attributes member, by feeding the Attributes object
//  written string to the Attributes object reader.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrReader::commonInfo(const char *common)
{
    VERIFY_OBJECT_STATUS();

    attr_.setAttributes(common);
}

//----------------------------------------------------------------------
// METHOD: CmrReader::beginPath  --trp
//      
//  Begin processing of a path element list.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  If an existing non-NULL plist was present, it is cleared & deleted.
//  A new empty plist is created.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrReader::beginPath ()
{
    VERIFY_OBJECT_STATUS();

    if (plist_ == NULL) {
        plist_ = new PathElementList;
        return;
    }
    
    if (plist_->getCount() > 1)
        plist_->clear ();
}

//----------------------------------------------------------------------
// METHOD: CmrReader::pathEntry  --trp
//      
//  Add this path entry to the current path entry list.
//
// ARGS:
//  char*  entryName  IN   Path entry string.
//
// RETURNS:
// PRE-CONDITIONS:
//  A valid plist_ exists.
// 
// POST-CONDITIONS:
//  An additional path entry is appended to the current plist_.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrReader::pathEntry (const char *entryName)
{
    VERIFY_OBJECT_STATUS();

    PathElement pathElement;

    if (strcmp (entryName, "PATH_UNKNOWN") == 0)
    {
        pathElement.type = NSR::path_Unknown;
        pathElement.name = entryName;
    }
    else if (strcmp (entryName, "PATH_CUSTOM_ROOT") == 0)
    {
        pathElement.type = NSR::path_Custom_Root;
        pathElement.name = entryName;
    }
    else if (strcmp (entryName, "PATH_STANDARD_ROOT") == 0)
    {
        pathElement.type = NSR::path_Standard_Root;
        pathElement.name = entryName;
    }
    else if (strcmp (entryName, "PATH_PARENT") == 0)
    {
        pathElement.type = NSR::path_Parent;
        pathElement.name = entryName;
    }
    else if (strcmp (entryName, "PATH_SAME") == 0)
    {
        pathElement.type = NSR::path_Same;
        pathElement.name = entryName;
    }
    else
    {
        pathElement.type = NSR::path_Name;
        pathElement.name = entryName;
    }

    plist_->append (pathElement);
}

//----------------------------------------------------------------------
// METHOD: CmrReader::beginAllocation  --trp
//      
//  Begin a new extent list.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  If an existing non-NULL elist was present, it is cleared & deleted.
//  A new empty elist is created.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrReader::beginAllocation ()
{
    VERIFY_OBJECT_STATUS();

    if (elist_ == NULL) {
        elist_ = new ExtentList;
        return;
    }
    
    if (elist_->getCount() > 0)
        elist_->clear ();
}

//----------------------------------------------------------------------
// METHOD: CmrReader::extent  --trp
//      
//  Append this extent to the current elist_.
//
// ARGS:
//  char*  sid          IN  Extent side id.
//  char*  startSector  IN  Extent start sector.
//  char*  length       IN  Extent length.
//
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  This extent is appended to the current elist_.
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrReader::extent (const char *sid, const char *startSector,
                        const char *length)
{
    VERIFY_OBJECT_STATUS();

  Extent extent;

  extent.start.sid = strtoul (sid, NULL, 0);
  extent.start.sector = strtoul (startSector, NULL, 0);
  extent.len = strtoul (length, NULL, 0);

  elist_->append (extent);
}






