//-*-c++-*-
//----------------------------------------------------------------------
//  cmrTools.C
//   Collection of somewhat generic tools written for and used by the cmr
//   classes cmrReader, cmrWriter, cmrDiff.
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//  CLASSES:
//   cmrTools
//   LineBuffer
//   Log
//
//----------------------------------------------------------------------
//
// $Source: /projects/HMS/nsr/udfAncil/msVerify/cmrTools.cpp,v $
// $Revision: 1.7 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/01/10 11:17:57 $
//

#include "cmr.h"

//----------------------------------------------------------------------
// METHOD: cmrError::CmrError  --trp
//      
//  CmrError constructor.
//
// ARGS:
//
// RETURNS:
//  void
// 
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  errFile_ is initialized to NULL
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
CmrError::CmrError()
        :errFile_ (NULL)
{
    fileIsOpen = FALSE;
}
//----------------------------------------------------------------------
// METHOD: CmrError::isValid  --trp
//      
//  Indicates whether the class is in a valid state.
//
// ARGS:
// RETURNS:
//  TRUE if the error file appears valid (non NULL).
// 
// PRE-CONDITIONS:
// POST-CONDITIONS:
// 
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
RWBoolean CmrError::isValid()
{
    return !(errFile_ == NULL);
}

//----------------------------------------------------------------------
// METHOD: CmrError::declareErrorFile  --trp
//      
//  Define the error output file.
//
// ARGS:
//  char*  errFileName  IN  Name of the error file name.
//
// RETURNS:
//  void
// 
// PRE-CONDITIONS:
// POST-CONDITIONS:
//   Error file is opened, or the program is terminated.
//
// ERRORS:
//  "unable to open error file"
// NOTES:
//
//----------------------------------------------------------------------
void CmrError::declareErrorFile(char *errFileName)
{
    fileName_ = errFileName;
}

//----------------------------------------------------------------------
// METHOD: CmrError::logError  --trp
//      
//  Means of entering an error in the error log file.  Three severities
//  are supported (Information, Warning, Fatal), with fatal errors
//  causing a program abort.
//
// ARGS:
//  ErrorSeverity  severity  IN  One of Information, Warning, Fatal.
//  char*          errorText IN  Text of error to be logged.
//
// RETURNS:
//  void
// 
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  An error will be logged to the error file.
//  IF the severity is fatal, the program will abort.
// 
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void CmrError::logError(ErrorSeverity severity, char *errorText)
{
    if (!fileIsOpen) {
        errFile_ = fopen (fileName_, "w");
        if (errFile_ == NULL) {
            fprintf (stderr, "unable to open error file \"%s\"\n",
                     (char*)fileName_);
            fprintf (stderr, "ERROR TEXT: %s\n", errorText);
            fileIsOpen = FALSE;
        }
        else
            fileIsOpen = TRUE;
    }
    

    switch (severity)
    {
    case errorInform:
        if (errFile_ == NULL)
            fprintf (stderr, "Information: %s.\n", errorText);
        else
            fprintf (errFile_, "Information: %s.\n", errorText);
        break;
    case errorWarning:
        if (errFile_ == NULL)
            fprintf (stderr, "Warning: %s.\n", errorText);
        else
            fprintf (errFile_, "Warning: %s.\n", errorText);
        break;
    case errorFatal:
    default:
        if (errFile_ == NULL) {
            fprintf (stderr, "Fatal error: %s.\n", errorText);
            fprintf (stderr, "Program will now terminate.\n");
        }
        else {
            fprintf (errFile_, "Fatal error: %s.\n", errorText);
            fprintf (errFile_, "Program will now terminate.\n");
        }
            
        exit (2);
        break;
    }
}

//----------------------------------------------------------------------
// METHOD: BaseCmr::BaseCmr  --trp
//      
//  BaseCmr constructor.
//
// ARGS:
//  char*  identity  IN    Name of class inheriting from BaseCmr.
//
// RETURNS:
//  void
// 
// PRE-CONDITIONS:
// POST-CONDITIONS:
//   The identity is saved.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------

//RWCString BaseCmr::errFileName_;
//CmrError BaseCmr::error;
BaseCmr::BaseCmr(char *identity)
{
    identity_ = identity;
}


//----------------------------------------------------------------------
// METHOD: CmrTools::elistCompare  --trp
//
//  Compare two elist objects for equality, optionally logging
//  the differences.
//
// ARGS:
//  ExtentList&  elist1  IN  First comparison extent list.
//  ExtentList&  elist2  IN  Second comparison extent list.
//  Log*         log     IN  NULL, or logging object.
//
// RETURNS:
//  TRUE if the extent lists are equal.
//  FALSE if the extent lists are not equal.
//
// PRE-CONDITIONS:
//  The Log* is NULL, or points to a valid Log object.
//  
// POST-CONDITIONS:
//  If differences have been found, and a non-null Log object has
//  been passed via the Log* parameter, the log file will contain
//  text defining those differences.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
RWBoolean CmrTools::elistCompare(ExtentList &elist1,
                                 ExtentList &elist2,
                                 Log *log)
{
    if (elist1.getCount() != elist2.getCount()) {
        if (log != NULL) {
            log->beginSection("number of extents");
            log->beginSection((char*)(RWCString)*(RWCollectableString*)
                              log->referenceLabels[0]);
            log->addNumber((long unsigned)elist1.getCount());
            log->endLine();
            log->endSection();
            log->beginSection((char*)(RWCString)*(RWCollectableString*)
                              log->referenceLabels[1]);
            log->addNumber((long unsigned)elist2.getCount());
            log->endSection();
            log->endSection();
        }
        return FALSE;
    }
    
    for (int i = 0; i < elist1.getCount (); i++)
    {
        RWCString extentName;
        sprintf (extentName, "extent %d (0 - (n-1))", i);
        if (log != NULL) {
            log->beginSection((char*) extentName);
        }
        
        if (elist1[i].start.sid != elist2[i].start.sid) {
            if (log != NULL) {
                log->beginSection("surface");
                log->beginSection((char*)(RWCString)*(RWCollectableString*)
                                  log->referenceLabels[0]);
                log->addNumber(elist1[i].start.sid);
                log->endLine();
                log->endSection();
                log->beginSection((char*)(RWCString)*(RWCollectableString*)
                                  log->referenceLabels[1]);
                log->addNumber(elist2[i].start.sid);
                log->endLine();
                log->endSection();
                log->endSection();
            }
            return FALSE;
        }
        if (elist1[i].start.sector != elist2[i].start.sector) {
            if (log != NULL) {
                log->beginSection("start sector");
                log->beginSection((char*)(RWCString)*(RWCollectableString*)
                                  log->referenceLabels[0]);
                log->addNumber(elist1[i].start.sector);
                log->endLine();
                log->endSection();
                log->beginSection((char*)(RWCString)*(RWCollectableString*)
                                  log->referenceLabels[1]);
                log->addNumber(elist2[i].start.sector);
                log->endLine();
                log->endSection();
                log->endSection();
            }
            return FALSE;
        }
        if (elist1[i].len != elist2[i].len) {
            if (log != NULL) {
                log->beginSection("extent length");
                log->beginSection((char*)(RWCString)*(RWCollectableString*)
                                  log->referenceLabels[0]);
                log->addNumber(elist1[i].len);
                log->endLine();
                log->endSection();
                log->beginSection((char*)(RWCString)*(RWCollectableString*)
                                  log->referenceLabels[1]);
                log->addNumber(elist2[i].len);
                log->endLine();
                log->endSection();
                log->endSection();
            }
            return FALSE;
        }
        
         if (log != NULL)
             log->endSection();
    }

    return TRUE;
}

//----------------------------------------------------------------------
// METHOD: CmrTools::getPlistTypeDesc  --trp
//
//  Given a pathe element type, provide a descriptive string.
//
// ARGS:
//  int         type      IN   Path element type.
//  RWCString&  typeDesc  OUT  String populated with descriptive text.
//
// RETURNS:
//  void
//
// PRE-CONDITIONS:
//  "type" is a valid path element type.
//  
// POST-CONDITIONS:
//
// ERRORS:
//
// NOTES:
//
//----------------------------------------------------------------------
void CmrTools::getPlistTypeDesc(int type, RWCString& typeDesc)
{
  // path element types <- move to header, use standard types
    // or enum when available.
#define PATH_UNKNOWN        0
#define PATH_CUSTOM_ROOT    1
#define PATH_STANDARD_ROOT  2
#define PATH_PARENT         3
#define PATH_SAME           4
#define PATH_NAME           5

    switch (type) {
        case PATH_UNKNOWN:
            typeDesc = "UNKNOWN";
            break;
        case PATH_CUSTOM_ROOT:
            typeDesc = "CUSTOM ROOT";
            break;
        case PATH_STANDARD_ROOT:
            typeDesc = "STANDARD ROOT";
            break;
        case PATH_PARENT:
            typeDesc = "PARENT";
            break;
        case PATH_SAME:
            typeDesc = "SAME";
            break;
        case PATH_NAME:
            typeDesc = "NAME";
            break;
    }
}

//----------------------------------------------------------------------
// METHOD: CmrTools::plistCompare  --trp
//
//  Compare equality of two path element lists.
//
// ARGS:
//  PathElementList&   plist1  IN  First comparison plist.
//  PathElementList&   plist2  IN  Second comparison plist.
//  Log*               log     IN  NULL, or valid Log object.
//
// RETURNS:
//  TRUE if the path element lists are equal.
//  FALSE if the path element lists are not equal.
//
// PRE-CONDITIONS:
//  The Log* is NULL, or points to a valid Log object.
//  
// POST-CONDITIONS:
//  If differences have been found, and a non-null Log object has
//  been passed via the Log* parameter, the log file will contain
//  text defining those differences.
//
// NOTES:
//
//----------------------------------------------------------------------
RWBoolean CmrTools::plistCompare(PathElementList &plist1,
                                 PathElementList &plist2,
                                 Log *log)
{
    if (log != NULL)
        log->beginSection("path element list");
    
    if (plist1.getCount() != plist2.getCount()) {
        if (log != NULL) {
            log->beginSection("element count");
            log->beginSection((char*)(RWCString)*(RWCollectableString*)
                              log->referenceLabels[0]);
            log->addNumber((short) plist1.getCount());
            log->endLine();
            log->endSection();
            log->beginSection((char*)(RWCString)*(RWCollectableString*)
                              log->referenceLabels[1]);
            log->addNumber((short)plist2.getCount());
            log->endLine();
            log->endSection();
            log->endSection();            
        }
        return FALSE;
    }
    
    for (int i = 0; i < plist1.getCount (); i++)
    {
        RWCString whichElement;
        whichElement = "element " + i;
        if (log != NULL)
            log->beginSection(whichElement);
                
        log->beginSection("");
        if (plist1[i].type != plist2[i].type) {
            if (log != NULL) {
                log->beginSection("type");
                log->beginSection((char*)(RWCString)*(RWCollectableString*)
                                  log->referenceLabels[0]);
                RWCString typeDesc;
                getPlistTypeDesc((int) plist1[i].type, typeDesc);
                log->addChunk(typeDesc);
                log->endLine();
                log->endSection();
                log->beginSection((char*)(RWCString)*(RWCollectableString*)
                                  log->referenceLabels[1]);
                getPlistTypeDesc((int) plist2[i].type, typeDesc);
                log->addChunk(typeDesc);
                log->endLine();
                log->endSection();
                log->endSection();            
            } 
            return FALSE;
        }
        
        if (plist1[i].type == PATH_NAME)
            if (strcmp(plist1[i].name, plist2[i].name) != 0) {
                if (log != NULL) {
                    log->beginSection("name");
                    log->beginSection((char*)(RWCString)*(RWCollectableString*)
                                      log->referenceLabels[0]);
                    log->addChunk(plist1[i].name);
                    log->endLine();
                    log->endSection();
                    log->beginSection((char*)(RWCString)*(RWCollectableString*)
                                      log->referenceLabels[1]);
                    log->addChunk(plist2[i].name);
                    log->endLine();
                    log->endSection();
                    log->endSection();
                }
                return FALSE;
            }
        
        if (log != NULL)
            log->endSection();
    }

    if (log != NULL)
        log->endSection();

    return TRUE;
}

//----------------------------------------------------------------------
// METHOD: LineBuffer::LineBuffer  --trp
//
//  LineBuffer constructor.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  tab size, indent level, and identity are initialized
// ERRORS:
//
// NOTES:
//
//-----------------------------------------------------------------------------
LineBuffer::LineBuffer () :
        tabSize_ (DEFAULT_TABSIZE),
        indent_ ( 0),
        needIndent_ (TRUE),
        outputFilep_ (NULL),
        BaseCmr("LineBuffer")
{
}

//----------------------------------------------------------------------
// METHOD: LineBuffer::isValid  --trp
//
//  LineBuffer validation
//
// ARGS:
// RETURNS:
//  TRUE if the object is valid.
//  FALSE if the object is in an invalid state.
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
// ERRORS:
// NOTES:
//
//-----------------------------------------------------------------------------
RWBoolean LineBuffer::isValid()
{
    if (!outputFilep_->isValid())
        return FALSE;
    if (indent_ < 0)
        return FALSE;
    return TRUE;
}

//----------------------------------------------------------------------
// METHOD: LineBuffer::setTabSize  --trp
//
//  Sets indention tab size.
//
// ARGS:
//  int         tabSize         IN  tab size
//
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  tabSize_ is set.
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
// -----------------------------------------------------------------------------
// setTabSize   -       Setter for tabsize, the amount of rightshift on indent.
//
//  preconditions
//
//  postconditions
//-----------------------------------------------------------------------------
void LineBuffer::setTabSize (int tabSize)
{
  tabSize_ = tabSize;
}
//----------------------------------------------------------------------
// METHOD: LineBuffer::declareFlushFile  --trp
//
//  Define the file to receive linebuffer output.
//
// ARGS:
//  RWFile*  file  IN  Pointer to a file, already opened.
//
// RETURNS:
//  void.
// PRE-CONDITIONS:
//  RWFile* points to a valid RWFile object.
//
// POST-CONDITIONS:
//  A pointer to the RWFile object is saved.
//
// ERRORS:
// NOTES:
//  A file pointer is passed in rather than a filename, to allow the
//  using code to append to other files, etc.
//  file->isValid() error checking could be added.
//----------------------------------------------------------------------
void LineBuffer::declareFlushFile (RWFile * file)
{
    outputFilep_ = file;
}
//----------------------------------------------------------------------
// METHOD: LineBuffer::flushToOutput  --trp
//
//  Write the current buffer contents to the output file, and add a end
//  of line.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  If the buffer contained anything, it is written on a single line
//  to the output file.  The need indent state is set to yes.
// ERRORS:
// NOTES:
//
// ----------------------------------------------------------------------
void LineBuffer::flushToOutput ()
{
  // Write output 1 char at a time, to avoid writing
  // the null terminator (rogue wave writes the
  // '\0' to the file - won't break the parser, but
  // it looks messy).
    if (buffer_.length() < 1)
        return;
    
    for (int i = 0;
         i < buffer_.length () - 1 && buffer_[i] != '\0';
         i++)
    {
        outputFilep_->Write (buffer_[i]);
    }
    // End the line.
    outputFilep_->Write ('\n');
    // Re-init the buffer, and let the class know it
    // needs to add indention on the new line.
    buffer_ = "";
    needIndent_ = TRUE;
}

//----------------------------------------------------------------------
// METHOD: LineBuffer::indent  --trp
//
//  Append (tab size * indent) spaces to the current, empty line buffer.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  Spaces are appended to the buffer.
// ERRORS:
//
// NOTES:
//  If anyone ever wants to put in buffer contents before knowing the
//  indent level, the RWCString prepend member could be used instead.
//----------------------------------------------------------------------
void LineBuffer::indent ()
{
  for (int iSpace = 0; iSpace < tabSize_ * indent_; iSpace++)
    {
      buffer_.append (" ");
    }
}
//----------------------------------------------------------------------
// METHOD: LineBuffer::incrementIndent  --trp
//
//  Add one to the indention
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void LineBuffer::incrementIndent ()
{
    indent_++;
}

//----------------------------------------------------------------------
// METHOD: LineBuffer::decrementIndent  --trp
//
//  Subtract 1 from indent.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
// ERRORS:
//  If indent is already down to 1, an object-disabling error is thrown.
// NOTES:
//
//----------------------------------------------------------------------
void LineBuffer::decrementIndent ()
{
    if (indent_ < 1) {
        error.logError (errorFatal, "Attempt to decrement past left margin");
        object_status = status_invalid;
        return;
    }
    
    indent_--;
}

//----------------------------------------------------------------------
// METHOD: LineBuffer::append  --trp
//
//  Add a string to the buffer, without delimiting.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  A string is added to the buffer.
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void LineBuffer::append (const char *string)
{
  checkForIndent ();
  buffer_.append (string);
}
//----------------------------------------------------------------------
// METHOD: LineBuffer::addChunk  --trp
//
//  Add a string to the buffer, delimiting with one space.
//  Optionally stringize (from 'foo' to '"foo"').
//
// ARGS:
//  char*       string      IN   String to be added.
//  StringType  stringType  IN   If str_stringize, stringize the string.
//
// RETURNS:
// PRE-CONDITIONS:
//  string points to a valid string.
//
// POST-CONDITIONS:
//  The string, optionally encasing '"' chars, and a trailing blank
//  are added to the buffer.
// ERRORS:
// NOTES:
//  'Chunk' isn't the most descritptive name.  It's like a token, with
//  internal white space acceptable.
//----------------------------------------------------------------------
void LineBuffer::addChunk (const char *string, const StringType stringType)
{
  checkForIndent ();

  if (stringType == str_stringize)
    buffer_.append ("\"");

  buffer_.append (string);

  if (stringType == str_stringize)
    buffer_.append ("\"");

  buffer_.append (" ");
}

//----------------------------------------------------------------------
// METHOD: LineBuffer::addNumber  --trp
//
//  Add an number to the buffer, with option of base.
//
// ARGS:
// {int            val  IN    Number to add.
//  short          val  IN    Number to add.
//  long           val  IN    Number to add.
//  unsigned long  val  IN    Number to add.}
//  NumBase        base IN    One of num_decimal, num_octal, num_hex.
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  The number is added, with leading '0' in case of octal, and
//  leading '0x' in case of hex.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void LineBuffer::addNumber (int val, NumBase base)
{
  RWCString number;
  switch (base)
    {
    case num_decimal:
      sprintf (number, "%d", val);
      break;
    case num_octal:
      sprintf (number, "0%o", val);
      break;
    case num_hex:
      sprintf (number, "0x%x", val);
      break;
    }
  addChunk ((char *) number);
}
void LineBuffer::addNumber (short val, NumBase base)
{
  RWCString number;
  switch (base)
    {
    case num_decimal:
      sprintf (number, "%hd", val);
      break;
    case num_octal:
      sprintf (number, "0%ho", val);
      break;
    case num_hex:
      sprintf (number, "0x%hx", val);
      break;
    }
  addChunk ((char *) number);
}
void LineBuffer::addNumber (long val, NumBase base)
{
  RWCString number;
  switch (base)
    {
    case num_decimal:
      sprintf (number, "%ld", val);
      break;
    case num_octal:
      sprintf (number, "0%lo", val);
      break;
    case num_hex:
      sprintf (number, "0x%lx", val);
      break;
    }
  addChunk ((char *) number);
}
void LineBuffer::addNumber (unsigned long val, NumBase base)
{
  RWCString number;
  switch (base)
    {
    case num_decimal:
      sprintf (number, "%lu", val);
      break;
    case num_octal:
      sprintf (number, "0%lo", val);
      break;
    case num_hex:
      sprintf (number, "0x%lx", val);
      break;
    }
  addChunk ((char *) number);
}
//----------------------------------------------------------------------
// METHOD: LineBuffer::checkForIndent  --trp
//
//  If indent is needed, add it.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  Tab size * indent spaces are added to the buffer.
//  The needIndent_ boolean is set to FALSE.
// ERRORS:
// NOTES:
//  Indention is added once per line.  The need for indent is set to
//  TRUE for a new line, and checked whenever things are added to the
//  line.
//----------------------------------------------------------------------
void LineBuffer::checkForIndent ()
{
  if (needIndent_)
    {
      indent ();
      needIndent_ = FALSE;
    }
}

//----------------------------------------------------------------------
// METHOD: Log::Log  --trp
//
//  Log object constructor.
//
// ARGS:
//  char *    logFileName   IN  Name of file to be created & written to.
//
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  The log file is opened write only, or the program aborts.
//  The line buffer object receives the log file as it's flush file.
//  Tab size is initialized to 4.
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
RWDEFINE_COLLECTABLE(LogLevel, 300)
LogLevel::LogLevel()
        :needToPrint_ (TRUE)
{
}

//----------------------------------------------------------------------
// METHOD: Log::Log  --trp
//
//  Log object constructor.
//
// ARGS:
//  char *    logFileName   IN  Name of file to be created & written to.
//
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  The log file is opened write only, or the program aborts.
//  The line buffer object receives the log file as it's flush file.
//  Tab size is initialized to 4.
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
LogLevel::~LogLevel()
{
}

//----------------------------------------------------------------------
// METHOD: Log::Log  --trp
//
//  Log object constructor.
//
// ARGS:
//  char *    logFileName   IN  Name of file to be created & written to.
//
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  The log file is opened write only, or the program aborts.
//  The line buffer object receives the log file as it's flush file.
//  Tab size is initialized to 4.
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
Log::Log(char *logFileName)
        :BaseCmr("Log")
{
    /* Declare a log file.*/
    logFile_ = new RWFile (logFileName, "w");
    if (!logFile_->isValid()) {
        error.logError(errorWarning, "unable to open logfile");
        object_status = status_invalid;
        return;
    }

    haveDetail = FALSE;
    
    /* Set up the line buffer stuff. */
    line_.declareFlushFile(logFile_);
    line_.setTabSize(4);
}

//----------------------------------------------------------------------
// METHOD: Log::~Log  --trp
//
//  Log destructor.  Simply closes the log file.
//
// ARGS:
// RETURNS:
//
// PRE-CONDITIONS:
//  
// POST-CONDITIONS:
//  The log file object is closed and deleted.
//
// ERRORS:
//
// NOTES:
//
//----------------------------------------------------------------------
Log::~Log()
{
    delete logFile_;
}

//----------------------------------------------------------------------
// METHOD: Log::addChunk  --trp
//
//  Add a string to the current log line, with a trailing space
//  delimiter.  If any section lables need to be printed before hand,
//  print them.
//
// ARGS:
//  char *     string       IN  String to be inserted into the line.
//  StringType stringType   IN  Defines whether the inserted string will
//                              be "stringized".
//
// RETURNS:
// PRE-CONDITIONS:
//  string points to a valid string.
// 
// POST-CONDITIONS:
//  Any unprinted section labels are printed, and the string is added
//  to the current log line.
//
// ERRORS:
// NOTES:
//  Log uses LineBuffer pretty heavily.  This addChunk, like the
//  addNumber members, simply checks for section writes, then calls the
//  corresponding LineBuffer addChunk member with it's arguments.
//----------------------------------------------------------------------
void Log::addChunk (const char *string, const StringType stringType)
{
    haveDetail = TRUE;
    checkSectionWrites();
    
    line_.addChunk(string, stringType);
}

//----------------------------------------------------------------------
// METHOD: Log::addNumber  --trp
//
//  Check for section label writes, then add the number to the current
//  log line.  Four overloads accept int, short, log & unsigned long.
//
// ARGS:
// {int              val     IN   Number to be added.
//  short            val     IN   Number to be added.
//  long             val     IN   Number to be added.
//  unsigned long    val     IN   Number to be added.}
//  NumBase          base    IN   One of num_decimal, num_octal, num_hex.
//  
// RETURNS:
// PRE-CONDITIONS: 
// POST-CONDITIONS:
//  Unprinted section labels are printed and marked as printed, the
//  number is added to the current log line, in defined base format.
// ERRORS:
//
// NOTES:
//  Log uses LineBuffer pretty heavily.  These addNumber members, like
//  the addChunk member, simply checks for section writes, then calls
//  the corresponding LineBuffer addNumber member with it's arguments.
//----------------------------------------------------------------------
void Log::addNumber (int val, NumBase base)
{
    haveDetail = TRUE;
    checkSectionWrites();
    
    line_.addNumber(val, base);
}
void Log::addNumber (short val, NumBase base)
{
    haveDetail = TRUE;
    checkSectionWrites();
    
    line_.addNumber(val, base);
}

void Log::addNumber (long val, NumBase base)
{
    haveDetail = TRUE;
    checkSectionWrites();
    
    line_.addNumber (val, base);
}
void Log::addNumber (unsigned long val, NumBase base)
{
    haveDetail = TRUE;
    checkSectionWrites();
    
    line_.addNumber (val, base);
}
//----------------------------------------------------------------------
// METHOD: Log::beginSeciton  --trp
//
//  Begin a new section.  The section label will be printed, at one
//  indention level greater than the prior, only if an addChunk or
//  addNumber member is called, to print actual values.  The sections
//  are 'headings' to be used IF a detail line under those sections
//  is present.  If no detail line is present, the headings are deleted
//  without having been printed.
//
// ARGS:
//  char *     section    IN   String to be printed at the new section.
//
// RETURNS:
//
// PRE-CONDITIONS:
//  
// POST-CONDITIONS:
//
// ERRORS:
//
// NOTES:
//----------------------------------------------------------------------
void Log::beginSection(char * section)
{
    LogLevel *level = new LogLevel;
    //level->name_ = section;
    *level << section;
    level->indent_ = TRUE;
    levelStack_.push(level);
}

//----------------------------------------------------------------------
// METHOD: Log::endSection  --trp
//
//  End the current section, deleting that level from the stack
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  The top most level is removed from the stack, and the indention is
//  decremented.
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Log::endSection()
{
    LogLevel *level = (LogLevel*) levelStack_.pop();
    if ((!level->needToPrint_) && level->indent_)
        line_.decrementIndent();

    delete level;
}

//----------------------------------------------------------------------
// METHOD: Log::checkSectionWrites  --trp
//
//  If any section labels need printing, print them.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  Any unprinted section labels are printed and marked as already
//  printed.
// ERRORS:
// NOTES:
//  The section labels (in levelStack_) are copied into a temp string
//  stack, to reverse order.  Only unprinted labels are copied.  The
//  section label strings are then printed, at the appropriate indention
//  level and order, from the temp stack.
//  Two non-class functions, "orderNeededSections" and
//  "writeNeededSections" are used as rogue wave apply functions.  These
//  apply functions are presently not members of any class to simplify
//  their use as rw apply functions.
//----------------------------------------------------------------------
void orderNeededSections(RWCollectable *c, void *owner)
{
    Log *log = (Log*)owner;
    LogLevel *level = (LogLevel *)c;
    if (level->needToPrint_) {
        log->levelStringStack_->push(new RWCollectableString(level->name_));
        level->needToPrint_ = FALSE;
    }
}
void writeNeededSections(RWCollectable *c, void *owner)
{
    RWCString s = *((RWString*)(RWCollectableString*) c);
    
    Log *log = (Log*)owner;
    
    log->line_.addChunk( (char*) s);
    log->line_.flushToOutput();
    log->line_.incrementIndent();
}
void Log::checkSectionWrites()
{
    // Create a temporary stack of strings, to reverse the order of the
    // level stack, prior to printing.
    levelStringStack_ = new RWSlistCollectablesStack;
    levelStack_.apply(orderNeededSections, this);
    levelStringStack_->apply(writeNeededSections, this);
    levelStringStack_->clearAndDestroy();
    delete levelStringStack_;
}
//----------------------------------------------------------------------
// METHOD: Log::insertReferenceLabel  --trp
//
//  Add a reference label to the reference labels list "referenceLabels".
//  After adding references, they may be used in addChunk or beginSection
//  calls.
//
// ARGS:
//  char*  s   IN  Label string to be inserted.
//
// RETURNS:
// PRE-CONDITIONS:
//  s points to a valid string.
// POST-CONDITIONS:
//  A new string is inserted into referenceLabels.
// 
// ERRORS:
// NOTES:
//  This is offered to provide a convenient way to put strings into the
//  Log object scope, and use them later.  For instance, the following
//  code inserts two reference labels, and later uses the second one.
//    log_->insertReferenceLabel(progName);
//    log_->insertReferenceLabel(dataSource);
//    .
//    .
//    .
//    log->beginSection((char*)(RWCString)*(RWCollectableString*)
//                             log->referenceLabels[1]);
//
//----------------------------------------------------------------------
void Log::insertReferenceLabel(char *s)
{
    RWCollectableString *str = new RWCollectableString(s);
    referenceLabels.insert(str);
}


