//-*-c++-*-
//----------------------------------------------------------------------
//  cmrWriter.C
//   Writes CMR files, based on information provided through function
//   pointers.
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//  CLASSES:
//   CmrWriter
//  
// -----------------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msVerify/cmrWrite.cpp,v $
// $Revision: 1.13 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 95/10/30 12:26:34 $
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include "cmr.h"


//----------------------------------------------------------------------
// METHOD: CmrWriter::CmrWriter  --trp
//      
//  CmrWriter constructor.
//
// ARGS:
//  char*  outputFileName  IN  Name of output file to receive CMR.
//
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  The output file is opened writed only.
//  A new LineBuffer object is instantiated, with the parameter output
//  file declared as the flush file.
//  Option reservedWordsType is set to long words.
// ERRORS:
// A specified output file which cannot be created or appended to results
// in an error, resulting in CmrWriter status invalid (disabled).
// NOTES:
//
//----------------------------------------------------------------------
CmrWriter::CmrWriter (const char *outputFileName)
        :BaseCmr("CmrWriter")
{
    VERIFY_OBJECT_STATUS();
    
    // Open the output file, with the default mode of create/append.
    outputFile = new RWFile (outputFileName, "w");
    if (!outputFile->isValid ())
    {
        error.logError (errorWarning, "output file cannot be created or appended to");
        object_status = status_invalid;
        return;
    }
    
    line = new LineBuffer;
    line->declareFlushFile (outputFile);

    // Initialize defaults.
    reservedWordsType = longWords;
    firstEntry = NSR::True;

    object_status = status_normal;
}

//----------------------------------------------------------------------
// METHOD: CmrWriter::~CmrWriter  --trp
//      
//  CmrWrite destructor.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  The fileset terminator is appended to the output file and the output
//  file is closed.
//  The LineBuffer object is deleted.
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
CmrWriter::~CmrWriter ()
{
    // Close the fileset scope.
    if (line != NULL) {
        line->addChunk (getTok (tok_end_fileset));
        line->flushToOutput ();
    }

    if (outputFile != NULL)
        delete outputFile;

    if (line != NULL)
        delete line;
}

//----------------------------------------------------------------------
// METHOD: CmrWriter::beginDir  --trp
//      
//  Begin directory scope in the CMR file.
//  Output a new directory entry, flush the line to the output file,
//  and  shift indent right.
//
// ARGS:
//  char*        dirName  IN     Name of new directory.
//  Attributes&  attr     IN     Directory attributes.
//
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  A new directory entry is added to the output, and the indention
//  is incremented.
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrWriter::beginDir (const char *dirName, const Attributes& attr)
{
    VERIFY_OBJECT_STATUS();

    // Validations
    if (attr.fileType() != NSR::file_Directory)
    {
        RWCString err;
        sprintf (err, "non-directory entry %s sent as directory", dirName);
        error.logError (errorWarning, (char *) err);
        object_status = status_invalid;
        return;
    }

    // If this is the first directory, open file set scope.
    if (firstEntry)
    {
        line->addChunk (getTok (tok_begin_fileset));
        line->flushToOutput ();
        firstEntry = NSR::False;
    }

    // Write the directory name.
    line->addChunk (getTok (tok_begin_dir));
    line->addChunk (dirName, str_stringize);

    // Write the directory attributes.
    writeAttributes (attr);

    // Open the directory scope, for possible entries.
    line->addChunk (getTok (tok_begin_entries));

    // Flush the line, and ncrement indent for possible entry writing.
    line->flushToOutput ();
    line->incrementIndent ();
}

//----------------------------------------------------------------------
// METHOD: CmrWriter::endDir  --trp
//      
//  Close current directory scope.
//  Decrements indent, and writes dir closing tokens.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//   The indent is decremented.
//   The directory termination tokens are added to buffer
//   The buffer is flushed to output file
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrWriter::endDir ()
{
    VERIFY_OBJECT_STATUS();

    line->decrementIndent ();
    line->addChunk (getTok (tok_end_entries));
    line->addChunk (getTok (tok_end_dir));
    line->flushToOutput ();
}

//----------------------------------------------------------------------
// METHOD: CmrWriter::file  --trp
//      
//  Output a new file entry.
// 
// ARGS:
//  char*        fileName  IN  File name.
//  Attributes   attr      IN  File attributes.
//  ExtentList&  elist     IN  File extent list.
//
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  The file entry begin tokens are added to buffer.
//  The common attribute info is added to buffer.
//  The extent list is added to buffer.
//  The file entry end token is added to buffer.
//  The buffer is flushed to output file.
//  
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrWriter::file (const char *fileName, const Attributes& attr,
                 const ExtentList& elist)
{
    VERIFY_OBJECT_STATUS();

    // Validations
    if (attr.fileType() != NSR::file_Raw)
    {
        RWCString err;
        sprintf (err, "non-file entry %s sent as file", fileName);
        error.logError (errorWarning, (char *) err);
        object_status = status_invalid;
        return;
    }

    // Write the file name.
    line->addChunk (getTok (tok_begin_file));
    line->addChunk (fileName, str_stringize);

    // Write the file attributes.
    writeAttributes (attr);

    // Write the alloctation.
    writeExtentList (elist);

    // Write the end of file marker.
    line->addChunk (getTok (tok_end_file));

    // Flush the line
    line->flushToOutput ();
}

//----------------------------------------------------------------------
// METHOD: CmrWriter::symlink  --trp
//
//  Output a new symbolic link entry.
//
// ARGS:
//  char*            symlinkName  IN  Symbolic link name.
//  Attributes       attr         IN  Symbolic link attributes.
//  PathElementList  plist        IN  Symbolic link path element list.
//
// RETURNS:
// PRE-CONDITIONS:
//  The attribute type is symlink.
// 
// POST-CONDITIONS:
//  The symlink begin tokens are added to the buffer.
//  The common attributes are added to the buffer.
//  The path element list is added to the buffer.
//  The symlink end token added to buffer.
//  The buffer is flushed.
//
// ERRORS:
//  An invalid attributes type aborts the program.
//
// NOTES:
//
//----------------------------------------------------------------------
void
CmrWriter::symlink (const char *symlinkName, const Attributes& attr,
                    const PathElementList& plist)
{
    VERIFY_OBJECT_STATUS();

    // Validations
    if (attr.fileType() != NSR::file_SymbolicLink)
    {
      RWCString err;
      sprintf (err, "non-symlink entry %s sent as symlink", symlinkName);
      error.logError (errorWarning, (char *) err);
      object_status = status_invalid;
      return;
    }

    // Write the symlink name.
    line->addChunk (getTok (tok_begin_symlink));
    line->addChunk (symlinkName, str_stringize);

    // Write the symlink attributes.
    writeAttributes (attr);

    // Write the path element list.
    writePathElementList (plist);

    // Write the end of symlink marker.
    line->addChunk (getTok (tok_end_symlink));

    // Flush the line
    line->flushToOutput ();
}

//----------------------------------------------------------------------
// METHOD: CmrWriter::writeAttributes  --trp
//
//  Write attributes to the output.
//  Uses the Attributes::appendAttributes member to write attributes.
//  Note that attributes details are independent of CMR, and are
//  entirely encapsulated in Attributes - both the read and write
//  functionality.
//
// ARGS:
//  Attributes&  attr  IN  Node Attributes.
//
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  Common attributes tokens & numeric values appended to buffer.
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrWriter::writeAttributes (const Attributes & attr)
{
    VERIFY_OBJECT_STATUS();

    // Open common info scope.
    line->addChunk (getTok (tok_begin_common));

    // Use the attributes member to get all attributes text
    RWCString attrStr;
    line->addChunk(attr.appendAttributes(attrStr), str_stringize);

    // Close common info scope.
    line->addChunk (getTok (tok_end_common));
}

//----------------------------------------------------------------------
// METHOD: CmrWriter::writeExtentList  --trp
//      
//  Write an extent list to the output file.
//  Write the allocation open & close tokens, and the extent token
//  followed by an extent, for each extent in the list.
//
// ARGS:
//  ExtentList&  elist  IN  File extent list.
//
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  Begin allocation, extent list and end_allocation tokens and numeric 
//  values added to buffer.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrWriter::writeExtentList (const ExtentList & elist)
{
    VERIFY_OBJECT_STATUS();

    line->addChunk (getTok (tok_begin_allocation));

    for (int i = 0; i < elist.getCount (); i++)
    {
      line->addChunk (getTok (tok_extent));
      line->addNumber (elist[i].start.sid);
      line->addNumber (elist[i].start.sector);
      line->addNumber (elist[i].len);
    }

  line->addChunk (getTok (tok_end_allocation));
}

//----------------------------------------------------------------------
// METHOD: CmrWriter::  --trp
//      
//  Write a path element list to the output.
//  Write the path list open & close tokens, and the path element
//  strings.
//
// ARGS:
//  PathElementList&  plist  IN  Path element list of current symlink.
//
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  Path list tokens and values added to buffer.
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrWriter::writePathElementList (const PathElementList & plist)
{
    VERIFY_OBJECT_STATUS();

  // path element types
#define PATH_UNKNOWN        0
#define PATH_CUSTOM_ROOT    1
#define PATH_STANDARD_ROOT  2
#define PATH_PARENT         3
#define PATH_SAME           4
#define PATH_NAME           5

  line->addChunk (getTok (tok_begin_path));

  for (int i = 0; i < plist.getCount (); i++)
    {
      switch (plist[i].type)
        {

        case PATH_UNKNOWN:
          line->addChunk ("PATH_UNKNOWN");
          break;

        case PATH_CUSTOM_ROOT:
          line->addChunk ("PATH_CUSTOM_ROOT");
          break;

        case PATH_STANDARD_ROOT:
          line->addChunk ("PATH_STANDARD_ROOT");
          break;

        case PATH_PARENT:
          line->addChunk ("PATH_PARENT");
          break;

        case PATH_SAME:
          line->addChunk ("PATH_SAME");
          break;

        case PATH_NAME:
        default:
          line->addChunk (plist[i].name, str_stringize);
          break;
        }                       //switch

    }                           //for

  line->addChunk (getTok (tok_end_path));
}

//----------------------------------------------------------------------
// METHOD: CmrWriter::useShortReservedWords  --trp
//
//  Set reserved words option to short.
//
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  The reserved words option is set to short.
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrWriter::useShortReservedWords ()
{
    VERIFY_OBJECT_STATUS();

  reservedWordsType = shortWords;
}

//----------------------------------------------------------------------
// METHOD: CmrWriter::setTabSize  --trp
//      
//  Set indention tab size.
//
// ARGS:
//  int    tabSize  IN     Amount of spaces per indention.
//
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//  Indention tab size is set.
//
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
void
CmrWriter::setTabSize (const int tabSize)
{
    VERIFY_OBJECT_STATUS();

  line->setTabSize (tabSize);
}

//----------------------------------------------------------------------
// METHOD: CmrWriter::getTok  --trp
//      
//  Returns either short or long tokens per the enum list
//
// ARGS:
//  Token  tok     IN      Index into the token array.
//
// RETURNS:
//  char* of the requested token.  The token with be eith full text,
//  or a two character abbreviation, depending on the reserved words
//  length option.
// 
// PRE-CONDITIONS:
//  tok is a valid token index.
// 
// POST-CONDITIONS:
//  A pointer to one of the token strings defined in this functions static
//  array is returned.
// ERRORS:
// NOTES:
//
//----------------------------------------------------------------------
char*
CmrWriter::getTok (const Token tok)
{
    VERIFY_OBJECT_STATUS_R(NULL);

  typedef struct
  {
    char *shortName;
    char *longName;
  }
  TokenPair;
  static TokenPair token_str[] =
  {

  //short name long name
    {"bs", "begin_fileset"},
    {"es", "end_fileset"},
    {"bd", "begin_dir"},
    {"ed", "end_dir"},
    {"be", "begin_entries"},
    {"ee", "end_entries"},
    {"bf", "begin_file"},
    {"ef", "end_file"},
    {"pm", "permit"},
    {"ln", "links"},
    {"ow", "owners"},
//    {"sz", "size"},
    {"ti", "times"},
    {"ex", "extent"},
    {"bc", "begin_common"},
    {"ec", "end_common"},
    {"by", "begin_symlink"},
    {"ey", "end_symlink"},
    {"bp", "begin_path"},
    {"ep", "end_path"},
    {"bl", "begin_allocation"},
    {"el", "end_allocation"},
  };

  switch (reservedWordsType)
    {
    case shortWords:
      return token_str[tok].shortName;
    case longWords:
      return token_str[tok].longName;
    default:
      error.logError (errorWarning, "invalid reserved word type");
      object_status = status_invalid;
    }
}
