#ifndef _NSRIMP_ // -*-c++-*-
#define _NSRIMP_
//----------------------------------------------------------------------
//  nsrImp.h
//   Objects used only by the Archive object.
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//  CLASSES:
//   ICBInfo
//   VDSInfo
//   VolAreaCalc
//   PartitionAreaCalc
//   Volume
//   VolumeTable
//   ICB_Table
//   UniqueID
//   DirectoryEntry
//   DirectoryContents
//   Directory
//   DirectoryStack
//  
//----------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfCore/ms/nsrImp.h,v $
// $Revision: 1.25 $    $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/05/08 17:19:11 $

//----------------------------------------------------------------------
// CLASS: ICBInfo     --DVM
//
//  Holds ICBTag information.
//
//  This STRUCT holds the key pieces of the ICBTag descriptor field,
//  used in reading WORM ICB Hierarchies.
//
// PUBLIC DATA MEMBERS:
//  UINT32  priNoDirectEntries   See ICBTag
//  UINT16  maxEntries           See ICBTag
//  UINT16  strategyParm         See ICBTag
//
// USAGE NOTES:
//  This is a simple struct, for internal use only.
// 
//----------------------------------------------------------------------
//----------------------------------------------------------------------
struct ICBInfo {
    UINT32 priNoDirectEntries;
    UINT16 strategyParm;
    UINT16 maxEntries;
};

//----------------------------------------------------------------------
// CLASS: VDSInfo     --DVM
//
//  Holds information relevant to the VDS.
//
//  This holds all of the positional state information required by
//  simple reading and writing algorithms inside of a VDS.
//
// PUBLIC DATA MEMBERS:
//  UINT32  mainAddr     Address inside of main VDS.
//  UINT32  resvAddr     Address inside of reserve VDS.
//  UINT32  volDescSeqNo Associated Sequence number in VDS.
//
// USAGE NOTES:
//  This is a simple struct, for internal use only.
// 
//----------------------------------------------------------------------
//----------------------------------------------------------------------
struct VDSInfo {
    UINT32 mainAddr;
    UINT32 resvAddr;
    UINT32 volDescSeqNo;
};

//----------------------------------------------------------------------
// CLASS: VolAreaCalc     --DVM
//
//  Determines locations of descriptors and sequences in volume area.
//
// PUBLIC METHODS:
//  VolAreaCalc()    - Constructor requires sid,sectorSize,numSectors.
//  volRecogSeqLoc() - Returns the start address of the VRS.
//  mainAVDLoc()     - Returns first Address for AVD
//  resvAVDLoc()     - Returns second Address for AVD
//  mainVDSArea()    - Returns space for Main VDS
//  resvVDSArea()    - Returns space for Reserve VDS
//  integSeqArea()   - Returns space for Integrity Sequence
//  partitionArea()  - Returns space for Partition
//
// USAGE NOTES:
//  These represent mostly hard-coded assignments by HP. Some of these
//  are format-dictated, but others reflect HP's judgement.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  none.
//
// IMPLEMENTATION NOTES:
//  Holds the three constructor arguments in its state, but the rest
//  is functionally determined based on mostly hard-coded values.
//
//----------------------------------------------------------------------
class VolAreaCalc
{
    sid_t  sid_;
    UINT16 sectorSize_;
    UINT32 lastSector_;
public:
    VolAreaCalc(sid_t sid, UINT16 sectorSize, UINT32 numSectors ) ;
    Address volRecogSeqLoc() const ;
    Address mainAVDLoc()     const ;
    Address resvAVDLoc()     const ;
    Extent  mainVDSArea()    const ;
    Extent  resvVDSArea()    const ;
    Extent  integSeqArea()   const ;
    Extent  partitionArea()  const ;
};

//----------------------------------------------------------------------
// CLASS: PartitionAreaCalc     --DVM
//
//  Calculates locations of descriptors inside of partition.
//
//  This is used to determine the locations and areas of the FSDS, the
//  USEs, and the fileDataStart. Given the fileDataStart, it can
//  calculate all of the other areas. Additionally, it can be
//  given the addresses of the USEs and determine the fileDataStart,
//  and consequently all of the other locations.
// 
// PUBLIC METHODS:
//  PartitionAreaCalc() - constructor requiring sid, mediaType,
//                        blockLen, and either fileDataStart or
//                        the two USEExtents.
//  unerasedICBAddr()    - (Extent) unerased UST ICBAddr.
//  erasedICBAddr()      - (Extent) erased UST ICBAddr.
//  unerasedSpcTblArea() - Space reserved for unerased UST
//  erasedSpcTblArea()   - Space reserved for erased UST
//  fsdSeqArea()         - Space reserved for FSDS
//  fileDataStart()      - First sector not reserved in partition.
//
// USAGE NOTES:
//  This is not changeable - it must be reconstructed in order to
//  change the fileDataStart. Note that WORM uses the UnerasedSpace.
//  ALSO NOTE that the fileDataStart supplied in the constructor is
//  rounded down if necessary in order to guarantee that the same
//  value will be calculated from the ICBAddrs. Thus, users may wish
//  to instantiate with a desired value for it, and then call the
//  fileDataStart() method to determine the actual fileDataStart it
//  will be using for all of its calculations.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  None. Uses pre-defined fixed HP locations.
//
// IMPLEMENTATION NOTES:
//  This was placed into a class, so that the interdependencies of
//  this information could all be contained in a sinble place. The
//  only data member (besides the mediaType, BlockLen, and sid) is
//  fileDataStart. All other locations are functionally calculated
//  from that.
//
//  ----------------------------------------------------------------
//  |          (Rewritable)         |                               |
//  |---------------|---------------|                               |
//  |  |<-unerased  |  |<-erased    |     FSD Sequence              |
//  |  |  ICBAddr   |  |  ICBAddr   |                               |
//  |--             |--             |                               |
//  |<-unerasedUST  |<-erasedUST    |                               |
//  |  Area-------->|  Area-------->|                               |
//  |-------------------------------|                               |
//  |           (Worm)              |                               |
//  |---------------|---------------|                               |
//  |<------unerased UST Area------>|                               |
//  |-------------------------------|                               |
//  |<--unerased    |               |                               |
//  |   ICBAddr---->|               |                               |
//  |               |               |                               |
//  ----------------------------------------------------------------
//  ^               ^               ^                               ^
//  LOC_USE         |               |                               |
//                 1/4              |                               |
//                                 1/2                              |
//                                                    fileDataStart()
//
//----------------------------------------------------------------------
class PartitionAreaCalc 
{
    sid_t     sid_;
    UINT32    fileDataStart_;
    MediaType mediaType_;
    UINT16    blockLen_;

    UINT32 quarterPoint() const;
    UINT32 halfPoint() const;
    UINT32 q1Length() const;
    UINT32 q2Length() const;
    UINT32 h1Length() const;
    UINT32 h2Length() const;
public:
    PartitionAreaCalc(sid_t sid,
                      UINT16 blockLen,
                      MediaType mt,
                      UINT32 maxFileDataStart ) ;
    PartitionAreaCalc(UINT16 blockLen,
                      MediaType mt,
                      const Extent & unerasedICBAddr,
                      const Extent & erasedICBAddr ) ;
    UINT32 fileDataStart()      const { return fileDataStart_; }
    Extent unerasedICBAddr()    const ;
    Extent unerasedSpcTblArea() const ;
    Extent erasedICBAddr()      const ;
    Extent erasedSpcTblArea()   const ;
    Extent fsdSeqArea()         const ;
};

//----------------------------------------------------------------------
// CLASS: Volume     --JD/AP
//
//  Internally-used repository of information about a volume.
//
// PUBLIC DATA MEMBERS:
//   UINT32           fileDataStart
//   ExtentList       freeSpace
//   ExtentList       erasedFreeSpace
//   Extent           fsd
//   Address          lvid
//   UINT32           logicalBlockSize
//   UINT16           vid
//   UINT16           lvNumVols
//   WriteProtectType lvWriteProtect
//   IntegrityType    lvIntegrity
//
// PUBLIC METHODS:
//  methodName()
//      Description_of_method
//
// USAGE NOTES:
//  User notes go here.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  
//  <collaborating_class> (e.g., OWNS: and/or USES:)
//
// PRIVATE METHODS:
//  methodName()
//      Description_of_method
//
// IMPLEMENTATION NOTES:
//  Internal Notes go here (special cases, algorithm issues, etc.)
//
//----------------------------------------------------------------------
class Volume
{
    UINT16 partitionRefNo;
    UINT32 partitionStart;
    UINT32 partitionLength;
    sid_t  surf_id;
    String volumeName;

    Boolean dirty;

    Error err;             

public:

    // The following fields are filled in for each volume, and
    // expected to be valid.

    UINT16           vid;
    UINT32           logicalBlockSize;
    UINT32           fileDataStart;
    ExtentList       freeSpace;
    ExtentList       erasedFreeSpace;

    // The following fields are really only associated with root
    // volumes. However, they get filled out with whatever is found on
    // each surface. Therefore, only the root volume's fields should
    // be recognized as valid.

    Extent           fsdsExt;
    Extent           lvidsExt;

    UINT16           lvNumVols;
    WriteProtectType lvWriteProtect;
    IntegrityType    lvIntegrity;
    String           lvIntegImplementID;

    Volume() :
            dirty(True),
            fileDataStart(0),
            logicalBlockSize(0),
            lvNumVols(0),
            lvWriteProtect(wrprot_None),
            lvIntegrity(integ_Open),
            lvIntegImplementID("Unknown")
    {
        fsdsExt.len = 0;
    }

    // lvIntegImplementID holds the implementation of the current
    // integrity entry found on the disk. It gets set whenever we
    // write an integrity entry (see Archive::writeIntegrityEntry())
    // in the following public Archive methods:
    // 
    //   completeGrow()
    //   completeExport()
    //   markOpenIntegrity()
    // 
    // It also gets set when we read an integrity entry during the
    // adding of a labeled volume in:
    // 
    //   addLabeledVol()
    //
    Volume( sid_t sid, CONST_CHAR_PTR vname, UINT32 lbsz ) : 
            volumeName(vname), 
            logicalBlockSize(lbsz)
    {
        dirty = True; 
        surf_id = sid;
        lvNumVols = 0;
        fsdsExt.len = 0;
        lvWriteProtect = wrprot_None;
        lvIntegrity = integ_Open;
        lvIntegImplementID = "Uninitialized";
        fileDataStart = 0;
    }

    // JKH:  The operator= is explicitly overloaded here to accommodate a
    // compiler problem with the UnixWare CC compiler.  The method
    // VolumeTable::append() doesn't work correctly (has a problem with
    // const references as argument) without the explictly overloaded
    // operator.

    Volume&     operator=(const Volume& vol) {
        partitionRefNo     = vol.partitionRefNo;
        partitionStart     = vol.partitionStart;
        partitionLength    = vol.partitionLength;
        surf_id            = vol.surf_id;
        volumeName         = vol.volumeName;
        dirty              = vol.dirty;
        err                = vol.err;
        vid                = vol.vid;
        logicalBlockSize   = vol.logicalBlockSize;
        fileDataStart      = vol.fileDataStart;
        freeSpace          = vol.freeSpace;
        erasedFreeSpace    = vol.erasedFreeSpace;
        fsdsExt            = vol.fsdsExt;
        lvidsExt           = vol.lvidsExt;
        lvNumVols          = vol.lvNumVols;
        lvWriteProtect     = vol.lvWriteProtect;
        lvIntegrity        = vol.lvIntegrity;
        lvIntegImplementID = vol.lvIntegImplementID;
        return *this;
    }

    Error error() const { return err; }

    void free( const ExtentList &unerasedextlst,
               const ExtentList &erasedextlst );
    void addfree( ExtentList &toList,
                  const ExtentList &fromList );
    Extent alloc( UINT32 nBlocks,
                  Boolean contiguous );
    UINT32 totalFree() const;
    UINT32 largestExtent() const;

    void markClean() { dirty = False; }
    void markDirty() { dirty = True; }

    void setPartition( UINT16 prn, UINT32 pstart, UINT32 plen )
    {
        partitionRefNo  = prn;
        partitionStart  = pstart;
        partitionLength = plen;
    }

    UINT16 partRefNo()  const { return partitionRefNo;  }
    UINT32 partStart()  const { return partitionStart;  }
    UINT32 partLength() const { return partitionLength; }

    Boolean isDirty() const { return dirty; }

    sid_t sid() const { return surf_id; }
    void  sid( sid_t sid ) { surf_id = sid; }

    const String & name() const { return volumeName; }
    void name( CONST_CHAR_PTR s ) { volumeName = s; }

    void fprint( FILE * fp ) const {
        fprintf (fp,"lvNV= %d; lvWP= %s; lvI= %s by %s\n",
                 lvNumVols,
                 ((lvWriteProtect == wrprot_None)? "None" : "Hard/Soft"),
                 ((lvIntegrity == integ_Open)? "Open":"Closed"),
                 (CONST_CHAR_PTR)lvIntegImplementID);
    }
};

//----------------------------------------------------------------------
// CLASS: VolumeTable     --JD/AP
//
//  Simple list of volume information
//
// PUBLIC METHODS:
//  VolumeTable(size) - Constructor (defaults to room for ten)
//  error()           - returns the error state of the list
//  clear()           - shrinks and deallocates the list
//  append(ext)       - appends an entry to the list
//  operator[](i)     - accesses the ith entry in the list
//  remove(n)         - removes the nth entry in the list
//  getCount()        - returns the number of entries in the list
//
// USAGE NOTES:
//  fsdSeqExtents is for internal use only.
// 
//  Bounds are checked  on operator[], setting the state returned by
//  error(). If an out of bounds location is accessed, an empty
//  entry, with contents equal to the default Volume constructor) is
//  returned.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  
//  Volume
//
// PRIVATE METHODS:
//  copy()   - Paul Anderson boilerplate
//  extend() - Paul Anderson boilerplate
//  free()   - Paul Anderson boilerplate
//  full()   - Paul Anderson boilerplate
//  empty()  - Paul Anderson boilerplate
//
// IMPLEMENTATION NOTES:
//  Note that if a size of 0 is specified on construction, a size five
//  list will be instantiated.
//
//----------------------------------------------------------------------
class VolumeTable {

    Volume nullVolume;

    Volume *element;
    unsigned size, count;

    Error err;

    void copy(const VolumeTable &);
    void extend();
    void free() { delete [] element; element = 0; }
    int full()  const { return count == size; }
    int empty() const { return count == 0;    }

public:

    // Constructors
    VolumeTable( unsigned sz = 10 )
    {
        element = new Volume[ size = (sz) ? sz : 5 ];
        if (!element) {
            err = error_No_free_store;
            size = 0;
        }
        count = 0;
    }

    VolumeTable(const VolumeTable &obj) { copy(obj); }

    // Destructor
    ~VolumeTable() { free(); }

    Error error() const { return err; }

    // Selectors
    unsigned getCount() const { return count; }
    unsigned rootIndex() const { return count-1; }
    unsigned rootVid() const { return rootIndex(); }
    UINT16 vid( sid_t sid ) const;

    // Modifiers
    VolumeTable &operator=( const VolumeTable &obj )
    {
        if (this != &obj) {
            free();
            copy(obj);
        }
        return *this;
    }

    Volume &operator[]( UINT16 i)
    {
        if (i < count)
            return element[i];
        else {
            err = error_Index_out_of_bounds;
            return nullVolume;
        }
    }

    Volume &rootVol( void )
    {
        return element[rootIndex()];
    }

    void append( const Volume& vol )
    { 
        if ( full() ) extend();
        element[count++] = vol;
    }

    void remove( UINT16 );

    void clear( unsigned sz = 10 )
    {
        free();
        element = new Volume[ size = (sz) ? sz : 5 ];
        if (!element) {
            err = error_No_free_store;
            size = 0;
        }
        count = 0;
    }
};

//----------------------------------------------------------------------
// CLASS: ICB_Table     --JD
//
//  Internal-use-only table, holding all info about ICBs.
//
// PUBLIC DATA MEMBERS:
//  NONE
//
// PUBLIC METHODS:
//  !!! Undocumented class: ICB_Table
//  methodName()
//      Description_of_method
//
// USAGE NOTES:
//  User notes go here.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  
//  <collaborating_class> (e.g., OWNS: and/or USES:)
//
// PRIVATE METHODS:
//  methodName()
//      Description_of_method
//
// IMPLEMENTATION NOTES:
//  Internal Notes go here (special cases, algorithm issues, etc.)
//
//----------------------------------------------------------------------
class ICB_Table {

    const UINT32 logicalBlockSize;

    UINT32 extent;
    UINT32 relOffset;

    Error err;
    ExtentList table;

public:

    ICB_Table( const ExtentList &extlst, UINT32 lbsz ):
            table(extlst),
            logicalBlockSize(lbsz) 
    { 
        resetPtr(); 
    }

    ~ICB_Table() {}

    Error error() const { return err; }

    const Extent &operator[]( unsigned i) const { return table[i]; } //rvalue
    Extent &operator[]( unsigned i) { return table[i]; } //lvalue
    
    void bumpPtr( UINT32 n );

    void resetPtr()
    { 
        if (err == error_End_of_ICB_table) err = error_None;
        extent = 0; 
        relOffset = 0; 
    }

    ExtentList location() { return table; }

    // should be const 
    Address currentAddr();
    Address nextAddr();
    Address lastAddr();
    Address addrAtOffset(INT32 offset);
};

//----------------------------------------------------------------------
// CLASS: UniqueID     --JD
//
//  <terse_desc>
//  <long_desc_if_needed>
//
// PUBLIC DATA MEMBERS:
//  type   member   <descr>
//  type   member   <descr>
//
// PUBLIC METHODS:
//  methodName()
//      Description_of_method
//
// USAGE NOTES:
//  User notes go here.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  
//  <collaborating_class> (e.g., OWNS: and/or USES:)
//
// PRIVATE METHODS:
//  methodName()
//      Description_of_method
//
// IMPLEMENTATION NOTES:
//  Internal Notes go here (special cases, algorithm issues, etc.)
//
//----------------------------------------------------------------------
class UniqueID {

    UINT32 low_id;
    UINT32 high_id;

public:

    UniqueID( UINT32 start = 0 ) { low_id = start; high_id = 0; }
    void next() { if (++low_id == 0) high_id++; }
    UINT32 low() { return low_id; }
    UINT32 high() { return high_id; }
};

//----------------------------------------------------------------------
// CLASS: DirectoryEntry     --JD
//
//  <terse_desc>
//  <long_desc_if_needed>
//
// PUBLIC DATA MEMBERS:
//  type   member   <descr>
//  type   member   <descr>
//
// PUBLIC METHODS:
//  methodName()
//      Description_of_method
//
// USAGE NOTES:
//  User notes go here.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  
//  <collaborating_class> (e.g., OWNS: and/or USES:)
//
// PRIVATE METHODS:
//  methodName()
//      Description_of_method
//
// IMPLEMENTATION NOTES:
//  Internal Notes go here (special cases, algorithm issues, etc.)
//
//----------------------------------------------------------------------
class DirectoryEntry {

public:

    Boolean dir;
    Boolean hidden;
    String  name;
    Address icbAddr;

    DirectoryEntry() :
            dir(False),
            hidden(False)
    {}
};

//----------------------------------------------------------------------
// CLASS: DirectoryContents     --JD
//
//  <terse_desc>
//  <long_desc_if_needed>
//
// PUBLIC DATA MEMBERS:
//  type   member   <descr>
//  type   member   <descr>
//
// PUBLIC METHODS:
//  methodName()
//      Description_of_method
//
// USAGE NOTES:
//  User notes go here.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  
//  <collaborating_class> (e.g., OWNS: and/or USES:)
//
// PRIVATE METHODS:
//  methodName()
//      Description_of_method
//
// IMPLEMENTATION NOTES:
//  Internal Notes go here (special cases, algorithm issues, etc.)
//
//----------------------------------------------------------------------
class DirectoryContents {


    DirectoryEntry nullDirectoryEntry;

    DirectoryEntry *element;
    unsigned size, count;

    Error err;

    void copy(const DirectoryContents &);
    void extend();
    void free() { delete [] element; element = 0; }
    int full() const { return count == size; }
    int empty() const { return count == 0; }

public:

    // Constructors
    DirectoryContents( unsigned sz = 10 )
    {
        element = new DirectoryEntry[size = (sz) ? sz : 1 ];
        if (!element) { err = error_No_free_store; return; }
        count = 0;
    }

    DirectoryContents(const DirectoryContents &obj) { copy(obj); }

    // Destructor
    ~DirectoryContents() { free(); }

    Error error() const { return err; }

    // Modifiers
    DirectoryContents &operator=( const DirectoryContents &obj )
    {
        if (this != &obj) {
            free();
            copy(obj);
        }
        return *this;
    }

    DirectoryEntry &operator[]( unsigned i)
    {
        if (i < count)
            return element[i];
        else {
            err = error_Index_out_of_bounds;
            return nullDirectoryEntry;
        }
    }

    const DirectoryEntry &operator[]( unsigned i) const
    {
        if (i < count)
            return element[i];
        else {
            // err = error_Index_out_of_bounds;
            return nullDirectoryEntry;
        }
    }

    void append( const DirectoryEntry entry )
    { 
        if ( full() ) extend();
        element[count++] = entry;
    }

    void remove( unsigned );

    // Selectors
    unsigned getCount() const { return count; }
};

//----------------------------------------------------------------------
// CLASS: Directory     --JD
//
//  <terse_desc>
//  <long_desc_if_needed>
//
// PUBLIC DATA MEMBERS:
//  type   member   <descr>
//  type   member   <descr>
//
// PUBLIC METHODS:
//  methodName()
//      Description_of_method
//
// USAGE NOTES:
//  User notes go here.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  
//  <collaborating_class> (e.g., OWNS: and/or USES:)
//
// PRIVATE METHODS:
//  methodName()
//      Description_of_method
//
// IMPLEMENTATION NOTES:
//  Internal Notes go here (special cases, algorithm issues, etc.)
//
//----------------------------------------------------------------------
class Directory {

    Address icb;
    String dirName;
    Boolean dirty;

    Error err;

public:

    // Public data members

    DirectoryContents contents;

    // Public methods

    Directory() : icb(0,0) { dirty = True; }

    Directory( CONST_CHAR_PTR n, Address a ) : dirName(n), icb(a)
    { 
        dirty = True; 
        err = dirName.error();
    }

    // JKH:  The operator= is explicitly overloaded here to accommodate a
    // compiler problem with the UnixWare CC compiler.  The method
    // DirectoryStack::push() doesn't work correctly (has a problem with
    // const references as argument) without the explictly overloaded
    // operator.

    Directory&  operator=(const Directory& dir)
    {
        icb      = dir.icb;
        dirName  = dir.dirName;
        dirty    = dir.dirty;
        err      = dir.err;
        contents = dir.contents;
        return *this;
    }

    Error error() const { return err; }

    void markClean() { dirty = False; }

    Address icbAddress() const { return icb; }
    CONST_CHAR_PTR name() const { return (CONST_CHAR_PTR) dirName; }
    Boolean isDirty() const { return dirty; }


};

//----------------------------------------------------------------------
// CLASS: DirectoryStack     --JD
//
//  <terse_desc>
//  <long_desc_if_needed>
//
// PUBLIC DATA MEMBERS:
//  type   member   <descr>
//  type   member   <descr>
//
// PUBLIC METHODS:
//  methodName()
//      Description_of_method
//
// USAGE NOTES:
//  User notes go here.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  
//  <collaborating_class> (e.g., OWNS: and/or USES:)
//
// PRIVATE METHODS:
//  methodName()
//      Description_of_method
//
// IMPLEMENTATION NOTES:
//  Internal Notes go here (special cases, algorithm issues, etc.)
//
//----------------------------------------------------------------------
class DirectoryStack {

    Directory nullDirectory;

    Directory *element;
    unsigned size, count;

    Error err;

    void copy(const DirectoryStack &);
    void extend();
    void free() { delete [] element; element = 0; }
    int full() const { return count == size; }
    int empty() const { return count == 0; }

public:

    // Constructors
    DirectoryStack( unsigned sz = 10 )
    {
        element = new Directory[size = (sz) ? sz : 1 ];
        if (element)
            count = 0;
        else {
            err = error_No_free_store;
            return;
        }
    }

    DirectoryStack(const DirectoryStack &obj) { copy(obj); }

    // Destructor
    ~DirectoryStack() { free(); }

    Error error() const { return err; }

    // Modifiers
    DirectoryStack &operator=( const DirectoryStack &obj )
    {
        if (this != &obj) {
            free();
            copy(obj);
        }
        return *this;
    }

    // JKH:  changed argument to const Directory& rather than const Directory

    void push( const Directory& vol )
    { 
        if ( full() ) extend();
        element[count++] = vol;
    }

    Directory pop()
    {
        if ( empty() )
            return nullDirectory;
        else
            return element[--count];
    }

    // Selectors
    unsigned getCount() const { return count; }
};

#endif // _NSRIMP_
