//--*-c++-*-------------------------------------------------------------
//  nsrInt.C
//   Objects used by the Interface to the Archive object.
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//  CLASSES:
//   Address
//   Extent
//   ExtentList
//   FileSetInfo
//   FileSetList
//   VolumeInfo
//   VolumeInfoList
//   VolumeSetInfo
//  
//----------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfCore/ms/nsrInt.cpp,v $
// $Revision: 1.13 $    $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/04/30 22:56:03 $

#include "nsrLib.h"

//-----------------------------------------------------------------------------
// METHOD:  ExtentList::copy()                                     --JKH
//
// Performs a deep copy of one ExtentList to another.
//
// ARGUMENTS:
//
// const ExtentList&  obj   IN  Source list for copy.
//
// RETURNS:  None.
//
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be set equal to the source list.
//-----------------------------------------------------------------------------

void NSR::ExtentList::copy( const ExtentList &obj ) 
{
    if(&obj == this) {                          // Trying to overwrite myself!
        return;
    }
    clear();

    INT32       i;

    for(i = 0; i < obj.entries(); i++) {
        const Extent    *extp = obj.at(i);

        if(extp == NULL) {      // This is an error case.  Bail out.
            clear();
            return;
        }
        append(*extp);
    }
}

//-----------------------------------------------------------------------------
// METHOD:  ExtentList::free()                              --JKH
//
// Clears out the contents of an ExtentList.
//
// ARGUMENTS:  None.
//
// RETURNS:  None.
//
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be freed.
//-----------------------------------------------------------------------------

void NSR::ExtentList::free(void) 
{
    Extent      *extp;
    INT32       i;

    for(i = 0; i < entries(); i++) {
        extp = at(i);
        delete extp;
    }
    theList_.clear();
}

//-----------------------------------------------------------------------------
// METHOD:  ExtentList::remove()                                   --JKH
//
// Removes an item from the list.
//
// ARGUMENTS:
//
// UINT32       index   IN  Index of item to remove.
//
// RETURNS:  None.
//
// PRECONDITIONS:
//
// --The index is in the range [0,nentries)
//
// POSTCONDITIONS:
//
// --The item will be removed and its memory freed.
//-----------------------------------------------------------------------------

void NSR::ExtentList::remove( UINT32 index ) 
{
    if(index >= entries()) {
        err = error_Index_out_of_bounds;
        return;
    }
    Extent *extp = this->theList_.removeReference(this->theList_.at(index));
    delete extp;
}

//-----------------------------------------------------------------------------
// METHOD:  ExtentList::totalBytes()                            --JKH
//
// Returns the number of bytes in the entire list.
//
// ARGUMENTS:  None.
//
// RETURNS:  See above.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:  See above.
//-----------------------------------------------------------------------------

NSR::UINT32 NSR::ExtentList::totalBytes() const {

    UINT32 total = 0;
    for (unsigned i = 0; i < entries(); i++) {
        // check for overflow; ifso, return max UINT32
        if ( total + at(i)->len < total ) {
            total = 0xFFFFFFFF;
            break;
        }
        total += at(i)->len;
    }

    return total;

}

//----------------------------------------------------------------------
// METHOD: ExtentList::numSectors  --DVM
//      
//  Returns the number of sectors held in the extent list.
//
//----------------------------------------------------------------------
//  Uses totalBytes() method.
//----------------------------------------------------------------------

NSR::UINT32
NSR::ExtentList::numBlocks(UINT32 blockSize) const
{
    return (totalBytes() + blockSize -1) / blockSize;
}
    
//-----------------------------------------------------------------------------
// METHOD:  ExtentList::at()                                    --JKH
//
// Returns a pointer to the specified item given by the argument index
// (beginning at 0).
//
// ARGUMENTS:
//
// UINT32       index   IN  Index of desired list item.
//
// RETURNS:
//
// Returns a pointer to the desired item on success, NULL on error (such as
// an index out of range).
//
// PRECONDITIONS:
//
// --Index is in the range [0,# of items in list)
//
// POSTCONDITIONS:
//
// --See return values
//-----------------------------------------------------------------------------

NSR::Extent *
NSR::ExtentList::at(UINT32 index)
{
    if(index >= entries()) {
        err = error_Index_out_of_bounds;
        return NULL;
    }
    return theList_.at(index);
}

const NSR::Extent *
NSR::ExtentList::at(UINT32 index) const
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

NSR::Boolean
NSR::ExtentList::operator!=( const ExtentList & obj ) const
{
    if ( entries() != obj.entries() ) {
        return True;
    }
    INT32       i;
    for(i = 0; i < theList_.entries(); i++) {
        if (*theList_.at(i) != *obj.theList_.at(i) ) {
            return True;
        }
    }
    return False;
}
    
//-----------------------------------------------------------------------------
// METHOD:  ExtentList::insertAt()                              --JKH
//
// Adds a copy of the argument item to the list at the specified index.
//
// ARGUMENTS:
//
// UINT32           index IN  Where to add item
// const Extent&    ext   IN  Item to add to the list
//
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --A copy of the item will be added to the list.
//-----------------------------------------------------------------------------

void
NSR::ExtentList::insertAt(UINT32 index, const Extent& ext)
{
    Extent      *extp = new Extent;

    if(index >= entries()) {
        err = error_Index_out_of_bounds;
        return;
    }
    if(extp == NULL) {
        err = error_No_free_store;
        return;
    }
    *extp = ext;
    theList_.insertAt(index,extp);
}

//-----------------------------------------------------------------------------
// METHOD:  ExtentList::append()                                --JKH
//
// Adds a copy of the argument item to the list.
//
// ARGUMENTS:
//
// const Extent&    ext   IN  Item to add to the list
//
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --A copy of the item will be added to the list.
//-----------------------------------------------------------------------------

void
NSR::ExtentList::append(const Extent& ext)
{
    Extent      *extp = new Extent;

    if(extp == NULL) {
        err = error_No_free_store;
        return;
    }
    *extp = ext;
    theList_.append(extp);
}

void
NSR::ExtentList::fprint(FILE * fp) const
{
    INT32       i;

    for(i = 0; i < theList_.entries(); i++) {
        theList_.at(i)->fprint(fp);
    }
}
void
NSR::ExtentList::writeAsText(String& text) const
{
    INT32       i;

    for(i = 0; i < theList_.entries(); i++) {
        theList_.at(i)->writeAsText(text);
    }
}

void
NSR::ExtentList::print(void)
{
    INT32       i;
    Extent      *extptr;

    for(i = 0; i < theList_.entries(); i++) {
        extptr = theList_.at(i);
        printf("Extent %d, start = (%u,%u), length = %u\n",i+1,
                extptr->start.sid,extptr->start.sector,extptr->len);
    }
}


//-----------------------------------------------------------------------------
// METHOD:  NSR::FileSetList::append                            --JKH
//
// Appends a new FileSetInfo object to the list.
//
// ARGUMENTS:
//
// const FileSetInfo& fsi   IN  object to append.
// 
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --Memory will be allocated for a new FileSetInfo object; its values will
//   be set equal to the argument object; the new object will be appended to
//   the list.
//-----------------------------------------------------------------------------

void
NSR::FileSetList::append(const FileSetInfo& fsi)
{
    FileSetInfo *fsiptr = new FileSetInfo;

    if(fsiptr == NULL) {
        err = NSR::error_No_free_store;
        return;
    }

    *fsiptr = fsi;
    theList_.append(fsiptr);
}

//-----------------------------------------------------------------------------
// METHOD:  NSR::FileSetList::at                                --JKH
//
// Returns a pointer to the item at the given index, NULL if the index is
// out of the range [ 0, nentries )
//
// ARGUMENTS:
//
// UINT32  index   IN   Index of desired item.
// 
// RETURNS:  See above.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:  See return values.
//-----------------------------------------------------------------------------

NSR::FileSetInfo *
NSR::FileSetList::at(UINT32 index)
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

const NSR::FileSetInfo *
NSR::FileSetList::at(UINT32 index) const
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

//-----------------------------------------------------------------------------
// METHOD:  NSR::FileSetList::remove                            --JKH
//
// Removes the item at the given index and frees its memory
//
// ARGUMENTS:
//
// UINT32  index   IN   Index of desired item to remove.
// 
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:  See above.
//-----------------------------------------------------------------------------

void
NSR::FileSetList::remove(UINT32 index)
{
    if(index >= entries()) {
        return;
    }
    FileSetInfo *fsiptr = theList_.removeReference(theList_.at(index));
    delete fsiptr;
}

//-----------------------------------------------------------------------------
// METHOD:  NSR::FileSetList::operator==                        --JKH
//
// Defines the equality operator.
//
// ARGUMENTS:
//
// const FileSetList& viList  IN  List to compare against.
// 
// RETURNS:  non-zero if equal, 0 otherwise.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:  None.
//
// NOTES:  
//
// A list is equal to another list if each item in each list is identical,
// occurring in the same order.
//-----------------------------------------------------------------------------

NSR::INT32
NSR::FileSetList::operator==(const FileSetList& fsiList) const
{
    INT32       i;

    if(entries() != fsiList.entries()) {
        return 0;
    }
    for(i = 0; i < entries(); i++) {
        const FileSetInfo       *f1,*f2;

        f1 = at(i);
        f2 = fsiList.at(i);

        if(*f1 != *f2) {
            return 0;
        }
    }
    return 1;
}

//-----------------------------------------------------------------------------
// METHOD:  NSR::FileSetList::copy                              --JKH
//
// Copies data from one FileSetList to another.
//
// ARGUMENTS:
//
// const FileSetList& fsiList  IN  List to copy.
// 
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:  None.
//-----------------------------------------------------------------------------

void
NSR::FileSetList::copy(const FileSetList& fsiList)
{
    if(&fsiList == this) {          // Trying to overwrite myself!
        return;
    }

    clear();                    // Clear myself out.

    INT32       i;

    for(i = 0; i < fsiList.entries(); i++) {
        const FileSetInfo        *fsiptr = fsiList.at(i);

        if(fsiptr == NULL) {     // This is an error case.  Bail out.
            clear();
            return;
        }
        append(*fsiptr);
    }
}

//-----------------------------------------------------------------------------
// METHOD:  FileSetList::free()                              --JKH
//
// Clears out the contents of a FileSetList.
//
// ARGUMENTS:  None.
//
// RETURNS:  None.
//
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be freed.
//-----------------------------------------------------------------------------

void
NSR::FileSetList::free()
{
    FileSetInfo  *      fsiptr;
    INT32               i;

    for(i = 0; i < entries(); i++) {
        fsiptr = theList_.get();
        delete fsiptr;
    }
}

//-----------------------------------------------------------------------------
// METHOD:  NSR::VolumeInfoList::SearchBySid                    --JKH
//          NSR::VolumeInfoList::SearchByVolName
//          NSR::VolumeInfoList::SearchByVolNum
//                
//
// Searches the VolumeInfoList for an entry with the specified key.
//
// ARGUMENTS:
//
// sid_t            sid      IN  Sid of volume to find
//
// OR
//
// const RWCString& volname  IN  Name of volume to find.
// 
// RETURNS:
//
// Returns a pointer to the VolumeInfo object containing the desired record
// or NULL if a volume with the specified key is not found.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --See return values.
//-----------------------------------------------------------------------------

NSR::VolumeInfo * 
NSR::VolumeInfoList::searchBySid(sid_t sid)
{
    INT32       i;

    for(i = 0; i < entries(); i++) {
        VolumeInfo      *viptr = theList_.at(i);

        if(viptr == NULL) {
            return NULL;
        }
        if(viptr->sid == sid) {
            return viptr;
        }
    }
    return NULL;
}

NSR::VolumeInfo * 
NSR::VolumeInfoList::searchByVolName(const RWCString& volname)
{
    INT32       i;

    for(i = 0; i < entries(); i++) {
        VolumeInfo      *viptr = theList_.at(i);

        if(viptr == NULL) {
            return NULL;
        }
        if(viptr->volName == volname) {
            return viptr;
        }
    }
    return NULL;
}

NSR::VolumeInfo * 
NSR::VolumeInfoList::searchByVolNum(UINT16 volnum)
{
    INT32       i;

    for(i = 0; i < entries(); i++) {
        VolumeInfo      *viptr = theList_.at(i);

        if(viptr == NULL) {
            return NULL;
        }
        if(viptr->volNum == volnum) {
            return viptr;
        }
    }
    return NULL;
}

//-----------------------------------------------------------------------------
// METHOD:  NSR::VolumeInfoList::append                         --JKH
//
// Appends a new VolumeInfo object to the list.
//
// ARGUMENTS:
//
// const VolumeInfo& vinfo   IN  object to append.
// 
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --Memory will be allocated for a new VolumeInfo object; its values will
//   be set equal to the argument object; the new object will be appended to
//   the list.
//-----------------------------------------------------------------------------

void
NSR::VolumeInfoList::append(const VolumeInfo& vinfo)
{
    VolumeInfo  *viptr = new VolumeInfo;

    if(viptr == NULL) {
        return;
    }

    *viptr = vinfo;
    theList_.append(viptr);
}

//-----------------------------------------------------------------------------
// METHOD:  NSR::VolumeInfoList::at                             --JKH
//
// Returns a pointer to the item at the given index, NULL if the index is
// out of the range [ 0, nentries )
//
// ARGUMENTS:
//
// UINT32  index   IN   Index of desired item.
// 
// RETURNS:  See above.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:  See return values.
//-----------------------------------------------------------------------------

NSR::VolumeInfo *
NSR::VolumeInfoList::at(UINT32 index)
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

const NSR::VolumeInfo *
NSR::VolumeInfoList::at(UINT32 index) const
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

//-----------------------------------------------------------------------------
// METHOD:  NSR::VolumeInfoList::remove                         --JKH
//
// Removes the item at the given index and frees its memory
//
// ARGUMENTS:
//
// UINT32  index   IN   Index of desired item to remove.
// 
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:  See above.
//-----------------------------------------------------------------------------

void
NSR::VolumeInfoList::remove(UINT32 index)
{
    if(index >= entries()) {
        return;
    }
    VolumeInfo  *viptr = theList_.removeReference(theList_.at(index));
    delete viptr;
}

//-----------------------------------------------------------------------------
// METHOD:  NSR::VolumeInfoList::operator==                     --JKH
//
// Defines the equality operator.
//
// ARGUMENTS:
//
// const VolumeInfoList& viList  IN  List to compare against.
// 
// RETURNS:  non-zero if equal, 0 otherwise.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:  None.
//
// NOTES:  
//
// A list is equal to another list if each item in each list is identical,
// occurring in the same order.
//-----------------------------------------------------------------------------

NSR::INT32
NSR::VolumeInfoList::operator==(const VolumeInfoList& viList) const
{
    INT32       i;

    if(entries() != viList.entries()) {
        return 0;
    }
    for(i = 0; i < entries(); i++) {
        const VolumeInfo        *v1,*v2;

        v1 = at(i);
        v2 = viList.at(i);

        if(*v1 != *v2) {
            return 0;
        }
    }
    return 1;
}

//-----------------------------------------------------------------------------
// METHOD:  NSR::VolumeInfoList::copy                           --JKH
//
// Copies data from one VolumeInfoList to another.
//
// ARGUMENTS:
//
// const VolumeInfoList& viList  IN  List to copy.
// 
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:  None.
//-----------------------------------------------------------------------------

void
NSR::VolumeInfoList::copy(const VolumeInfoList& viList)
{
    if(&viList == this) {          // Trying to overwrite myself!
        return;
    }

    clear();                    // Clear myself out.

    INT32       i;

    for(i = 0; i < viList.entries(); i++) {
        const VolumeInfo        *viptr = viList.at(i);

        if(viptr == NULL) {     // This is an error case.  Bail out.
            clear();
            return;
        }
        append(*viptr);
    }
}

//-----------------------------------------------------------------------------
// METHOD:  VolumeInfoList::free()                              --JKH
//
// Clears out the contents of a VolumeInfoList.
//
// ARGUMENTS:  None.
//
// RETURNS:  None.
//
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be freed.
//-----------------------------------------------------------------------------

void
NSR::VolumeInfoList::free()
{
    VolumeInfo  *viptr;
    INT32       i;

    for(i = 0; i < entries(); i++) {
        viptr = theList_.get();
        delete viptr;
    }
}

//-----------------------------------------------------------------------------
// METHOD:  VolumeSetInfo::writeAsText()                              --TRP
//
// Fills a string buffer with formatted ascii text defining all class state.
//
// ARGUMENTS:
//   String&     IN/OUT  Buffer to be filled with text.
//
// RETURNS:  None.
//
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// -- The referred string buffer is filled with defining text.
//-----------------------------------------------------------------------------
void
NSR::VolumeSetInfo::writeAsText(String& text) const
{
    char buf[1024];

    // Vol set name
    text += "VolumeSet ";
    text += volSetName;
    text += " ( ";

    // media type, sector size, expected vols
    sprintf (buf, "type %s sector %ld expectedVols %d ",
             mediaType == media_Worm ? "worm" : "mo",
             sectorSize, expectedVols);
    text += buf;

    // root sid, isComplete
    sprintf (buf, "rootSid %ld ", rootSid);
    text += buf;
    text += "isComplete ";
    if (isComplete)
        text += "yes ";
    else
        text += "no ";

    // integrity type
    text += "integrityType ";
    switch (integrityType) {
    case integ_Open:
        text += "open ";
        break;
    
    case integ_Closed:
        text += "closed ";
        break;

    case integ_Unknown:
    default:
        text += "unKnown ";
        break;
    }

    // write protect type
    text += " writeProtectType ";
    switch (writeProtectType) {
    case  wrprot_None:
        text += "none)\n";
        break;
        
    case wrprot_Soft:
        text += "soft)\n";
        break;
    
    case wrprot_Hard:
        text += "hard)\n";
        break;

    default:
        text += "unKnown)\n";
        break;        
    }

    // Volume info list
    text += "VolumeInfoList {";
    volInfoList.writeAsText(text);
    text += " }";
    
    // File set list
    text += "\nFileSetList {";
    fileSetList.writeAsText(text);
    text += " }";

    // close vol set info scope
    text += "}\n";
}


void
NSR::VolumeInfoList::writeAsText(String &text) const
{
    UINT32 entries = theList_.entries();
    VolumeInfo* vol;
    for (int i = 0; i < entries; i++) {
        vol = theList_.at(i);
        vol->writeAsText(text);
    }   
}

void
NSR::VolumeInfo::writeAsText(String &text) const
{
    char buf[1024];
    
    text += "volName ";
    text += volName;
    text += "volNumber ";
    sprintf (buf, "sid %d volNumber %d labelLimit %ld ", sid, volNum, labelLimit);
    text += buf;

    text += "\nfreeSpace {";
    freeSpace.writeAsText(text);
    text += "\n\t}\n";

    text += "\nerasedSpace {";
    erasedSpace.writeAsText(text);
    text += "\n\t}\n";
}


