#ifndef _NSRINT_ // -*-c++-*-
#define _NSRINT_
//----------------------------------------------------------------------
//  nsrInt.h
//   Objects used in the interface to the Archive object.
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//  CLASSES:
//   Address
//   Extent
//   ExtentList
//   FileSetInfo
//   FileSetList
//   VolumeInfo
//   VolumeInfoList
//   VolumeSetInfo
//  
//----------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfCore/ms/nsrInt.h,v $
// $Revision: 1.17 $    $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/04/30 22:56:04 $

#define DEFAULT_FILE_DATA_START         300

//----------------------------------------------------------------------
// CLASS: Address     --JD
//
//  Simple class implementing abstraction of an address
//
// PUBLIC DATA MEMBERS:
//  sid_t   sid    Surface Identifier
//  UINT32  sector Absolute Sector Address
//
// PUBLIC METHODS:
//  Address()         - Default constructor: sid=0; sector=0 (inline)
//  Address(sid,sect) - Constructor (inline)
//  operator==()      - Returns non-zero if two Addresses match (inline)
//  print()           - Prints contents to supplied FILE *. (inline)
//
// USAGE NOTES:
//  This should be treated as a struct with an extra constructor, and
//  an eprint() method. The eprint() method is intended for internal
//  use only.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  
//  NONE
//
// PRIVATE METHODS:
//  NONE
//
// IMPLEMENTATION NOTES:
//  DVM: I removed the explicit copy constructor on 9/11/95.
//  The default memberwise copy should be fine.
//----------------------------------------------------------------------
class Address {

public:

    sid_t  sid;
    UINT32 sector;

    Address() { sid = 0; sector = 0; }
    Address(sid_t s_id, UINT32 sect) { sid = s_id; sector = sect; }

    int operator==(const Address &a) {
        return (( sid == a.sid) && (sector == a.sector));
    }
    Boolean operator!=(const Address &obj) const {
        return ((( sid != obj.sid) || (sector != obj.sector)) ?
                True : False );
    }
    void print(FILE * fp) const {
        fprintf(fp,"sid:%d   ",sid);
        fprintf(fp,"sector:%d\n",sector);
    }
};

//----------------------------------------------------------------------
// CLASS: Extent     --JD
//
//  This is a simple physical extent, with an address and a length.
//
// PUBLIC DATA MEMBERS:
//  Address start  Starting Address
//  UINT32  len    Length in bytes
//
// PUBLIC METHODS:
//  Extent()     - Default constructor: len=0. (inline)
//  print()      - Prints extent to supplied FILE *. (inline)
//
// USAGE NOTES:
//  User notes go here.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  
//  Address
//
// PRIVATE METHODS:
//  NONE.
//
// IMPLEMENTATION NOTES:
//  DVM: Removed Extent(const Extent *) constructor on 9/11/95, and
//  cookie data member. 
//  JKH:  Re-added cookie data member to accomodate compilation with
//        Satchmo (03/13/96)
//
//----------------------------------------------------------------------
class Extent {

public:

    Address start;
    UINT32 len;
        UINT32 cookie;

    Extent() : len(0) {}

    void fprint(FILE * fp) const
    {
        String buf;
        writeAsText(buf);
        fprintf(fp, "%s\n", (const char *)buf);
    }
    void writeAsText (String& text) const
    {
        char buf[1024];
        
        sprintf (buf,"sid %10d sector %10d len %10d ",
                 start.sid, start.sector, len);
        text += buf;
    }
    
    Boolean operator!= ( const Extent & obj ) const
    {
        return (((start != obj.start) || (len != obj.len) ) ?
                True : False);
    }
};

//----------------------------------------------------------------------
// CLASS: ExtentList     --JD
//
//  Simple container of Extents.
//
//  This supports operator[] so that it may be treated as an array,
//  but otherwise, only simple appending, plus some others. This is
//  a value-based implementation, which resizes itself as needed.
//
// PUBLIC DATA MEMBERS:
//  NONE.
//
// PUBLIC METHODS:
//  ExtentList(size) - Constructor (defaults to room for ten)
//  error()          - returns the error state of the list (inline)
//  clear()          - shrinks and deallocates the list (inline)
//  append(ext)      - appends an entry to the list
//  insertAt()       - inserts an entry at the specified location.
//  operator[](i)    - accesses the ith entry in the list
//  at()             - returns a pointer to the ith entry in the list
//  remove(n)        - removes the nth entry in the list
//  getCount()       - returns the number of entries in the list (inline)
//  totalBytes()     - returns the sum of the entry lengths
//
// USAGE NOTES:
//  Bounds are checked  on operator[], setting the state returned by
//  error(). If an out of bounds location is accessed, an empty
//  extent, with contents equal to the default Extent constructor) is
//  returned.
//
//  In general, extents with length > 2^30 should be avoided, because
//  UDF/ISO disallows them.
//
//----------------------------------------------------------------------
// COLLABORATORS:  
//  Extent
//  RWGSlist
//
// PRIVATE METHODS:
//  copy()   - Paul Anderson boilerplate
//  free()   - Paul Anderson boilerplate
//
// IMPLEMENTATION NOTES:
//
//----------------------------------------------------------------------

declare(RWGSlist,Extent)

class ExtentList {

    // !!! Remove the nullExtent field when operator []'s are removed

    Extent              nullExtent;
    RWGSlist(Extent)    theList_;
    Error   err;

    void copy(const ExtentList &);
    void free();

public:

    // Constructors

    ExtentList( UINT32 sz = 10 ) : err(error_None) { }

    ExtentList(const ExtentList &obj) { copy(obj); }

    // Destructor
    ~ExtentList() { free(); }

    ExtentList& operator=(const ExtentList& obj) {
        copy(obj);
        return *this;
    }

    Error error() const { return err; }

    // Modifiers

    void clear(void)  { free(); }
    Extent &operator[]( UINT32 i) {

        if (i < theList_.entries())
            return *(at(i));
        else {
            err = error_Index_out_of_bounds;
            return nullExtent;
        }
    }

    const Extent &operator[]( UINT32 i) const {

        if (i < entries())
            return *(at(i));
        else {
            // err = error_Index_out_of_bounds;
            return nullExtent;
        }
    }

    Boolean operator!=( const ExtentList & obj ) const;
    
    void insertAt(UINT32 index,const Extent& ext);
    void append( const Extent& ext );
    void remove( UINT32 );

    // Selectors
    unsigned getCount(void) const { return entries(); }
    Extent * at(UINT32 index);
    const Extent * at(UINT32 index) const;
    UINT32 entries(void) const { return theList_.entries(); }
    UINT32 totalBytes(void) const;
    UINT32 numBlocks(UINT32 blockSize) const;

    void fprint(FILE * fp) const;
    void writeAsText(String& text) const;
    void print(void);

};

//----------------------------------------------------------------------
// CLASS: FileSetInfo     --AP
//
//  This is a simple structure holding information about a file set.
//
// PUBLIC DATA MEMBERS:
//  RWCString        name            Name of the file set.
//  UINT32           num             Number associated with the file set.
//  UINT32           time            Time file set was written.
//  WriteProtectType fsdWriteProtect Type of protection:
//                                   wrprot_None,wrprot_Soft,wrprot_Hard
//                                                       
// PUBLIC METHODS:                                       
//  FileSet()   - Default constructor: name = "";num,time=0,wp=None
//
// USAGE NOTES:
//  User normally only inspects these as part of a FileSetList.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  
//  RWCString
//
// PRIVATE METHODS:
//  copy()  -  Copies one FileSetInfo object to another.
//
// IMPLEMENTATION NOTES:
//  Just a struct with a default constructor.
//  DVM: changed to initialize all elements, rather than just name
//  9/11/95.
//  JKH:  changed to use RWCString.  Added operator ==, !=, =,copy constructor
//  9/14/95
//
//----------------------------------------------------------------------
class FileSetInfo {

public:

    RWCString        name;
    UINT32           num;
    UINT32           time;
    WriteProtectType fsdWriteProtect;

    FileSetInfo():
            name(""),
            num(0),
            time(0),
            fsdWriteProtect(wrprot_None) {}
    FileSetInfo(const FileSetInfo& fsi) {
        copy(fsi);
    }
    INT32       operator==(const FileSetInfo& fsi) const {
        return ((name == fsi.name) && (num == fsi.num) &&
                (time == fsi.time) && (fsdWriteProtect == fsi.fsdWriteProtect));
    }
    INT32       operator!=(const FileSetInfo& fsi) const {
        return !(*this == fsi);
    }
    FileSetInfo& operator=(const FileSetInfo& fsi) {
        copy(fsi);
        return *this;
    }
    void writeAsText(String& text) const 
    {
        char buf[1024];

        text += "fileSetName ";
        text += name;

        sprintf (buf, " {fileSetNumber %ld time %lx ",
                 num, time);
        text += buf;

        // write protect type
        text += " writeProtectType ";
        switch (fsdWriteProtect) {
        case  wrprot_None:
            text += "none} ";
            break;
        
        case wrprot_Soft:
            text += "soft} ";
            break;
    
        case wrprot_Hard:
            text += "hard} ";
            break;

        default:
            text += "unKnown} ";
            break;        
        }
    }

private:

    void        copy(const FileSetInfo& fsi) {
        name = fsi.name;
        num = fsi.num;
        time = fsi.time;
        fsdWriteProtect = fsi.fsdWriteProtect;
    }
};

//----------------------------------------------------------------------
// CLASS: FileSetList     --AP (modified to use RWGSlist by JKH)
//
//  A simple list of FileSetInfo structs.
//
// PUBLIC DATA MEMBERS:
//  ExtentList   fsdSeqExtents   Where the FSD sequences are written
//                               (on the root volume).
//
// PUBLIC METHODS:
//  FileSetList(size) - Constructor (defaults to room for ten)
//  error()           - returns the error state of the list (inline)
//  clear()           - shrinks the list and deallocates the list (inline)
//  append(ext)       - appends an entry to the list
//  at(i)             - accesses the ith entry in the list
//  remove(n)         - removes the nth entry in the list
//  getCount()        - returns the number of entries in the list (inline)
//
// USAGE NOTES:
//  fsdSeqExtents is for internal use only.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  
//  FileSetInfo
//
// PRIVATE METHODS:
//  copy()   - Paul Anderson boilerplate
//  free()   - Paul Anderson boilerplate
//
// IMPLEMENTATION NOTES:
//  Note that this is built upon the RWGSlist class, but with value semantics.
//
//----------------------------------------------------------------------

declare(RWGSlist,FileSetInfo)

class FileSetList {

    RWGSlist(FileSetInfo)       theList_;

    Error err;


public:
    ExtentList  fsdSeqExtents;

    // Constructors
    FileSetList() : err(NSR::error_None) {} 
    FileSetList(const FileSetList &obj) { copy(obj); }
    // Destructor
    ~FileSetList() { free(); }

    Error error() const { return err; }

    // Modifiers

    void clear() { free(); }
    void append( const FileSetInfo& fsinfo );
    FileSetList &operator=( const FileSetList &obj ) {
        copy(obj);
        return *this;
    }
    INT32       operator==(const FileSetList &obj) const;
    INT32       operator!=(const FileSetList &obj) const {
        return !(*this == obj);
    }

    // Retained for backwards compatibility. . .
    // !!! DVM, kill these when they're removed from the Archive object.

    FileSetInfo& operator[](UINT32 i) {
        if(i >= entries()) {
            err = NSR::error_Index_out_of_bounds;
            return nullFSI;
        }
        return *(theList_.at(i));
    }
    const FileSetInfo& operator[](UINT32 i) const {
        if(i >= entries()) {
            return nullFSI;
        }
        return *(theList_.at(i));
    }


    void remove( UINT32 index );

    // Selectors
    unsigned            getCount() const { return entries(); }
    UINT32              entries() const { return theList_.entries(); }
    FileSetInfo *       at(UINT32 index);
    const FileSetInfo * at(UINT32 index) const;

    void writeAsText(String& text) const
    {
        text += "\n";
        text += "fsd {";
        fsdSeqExtents.writeAsText(text);
        text += "} ";

        text += "\n\tfileSetInfoList{ ";
        
        UINT32 entries = theList_.entries();
        FileSetInfo* info;
        for (int i = 0; i < entries; i++) {
            info = theList_.at(i);
            info->writeAsText(text);
        }
        text += "}\n";
    }

    private:

    // !!! DVM, kill this data member when brackets are removed from Arc obj.
    FileSetInfo nullFSI;        // For backwards compatibility;

    void copy(const FileSetList &);
    void free();
    
};

//----------------------------------------------------------------------
// CLASS: VolumeInfo     --DVM
//
//  A simple struct holding the information about a volume.
//
// PUBLIC DATA MEMBERS:
//  sid_t       sid            Surface Identifier
//  RWCString   volName        Volume Name
//  UINT16      volNum         Volume Number
//  UINT32      fileDataStart  First sector to hold file data
//  ExtentList  unerasedSpace  Available space requiring erasure
//  ExtentList  erasedSpace    Available space not requiring erasure
//
// PUBLIC METHODS:
//  VolumeInfo()  - Default constructor:
//                  sid=0;name="";num,
//                  fileDataStart=257;
//                  erased,unerased=empty 
//
// USAGE NOTES:
//  This is normally for inspection only.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  
//  String
//  ExtentList
//
// PRIVATE METHODS:
//  NONE
//
// IMPLEMENTATION NOTES:
//  This may have a bad value for the default fileDataStart.
//
//----------------------------------------------------------------------
class VolumeInfo {

public:

    sid_t       sid;
    RWCString   volName;
    UINT16      volNum;
    UINT32      labelLimit;
    ExtentList  freeSpace;
    ExtentList  erasedSpace;

    VolumeInfo()
            : sid(0),
              volName(""),
              volNum(0),
              labelLimit(DEFAULT_FILE_DATA_START) {}              

    operator==(const VolumeInfo& volinfo) const {
        return ((sid == volinfo.sid) &&
                (volName == volinfo.volName) &&
                (volNum == volinfo.volNum) &&
                (labelLimit == volinfo.labelLimit));
    }
    operator!=(const VolumeInfo& volinfo) const {
        return !(*this == volinfo);
    }
    void writeAsText(String& text) const;
};

//----------------------------------------------------------------------
// CLASS: VolumeInfoList                                --JKH
//
//  This is a simple list of information about volumes.
//
// PUBLIC DATA MEMBERS:
//  NONE
//
// PUBLIC METHODS:
//  searchBySid()  - search for a volume with given sid
//  searchByVolName() - search for a volume with a given volume name
//  searchByVolNum() - search for a volume with a given volume number
//  append()  - appends some VolumeInfo to the list.
//  at(i)     - returns a pointer to ith element.
//  remove(i) - removes the ith element.
//  entries() - returns the size of the list. (inline)
//  operator== - equality operator between two VolumeInfoLists.
//  operator= - assignment operator between two VolumeInfoLists.
//  clear()   - shrinks the list to size zero and deallocates all info. (inline)
//  
// USAGE NOTES:
//  User notes go here.
// 
//----------------------------------------------------------------------
// COLLABORATORS:
//  VolumeInfo
//  RWGSlist
//
// PRIVATE METHODS:
//  Uses RogueWave RWGSlist as foundation.
// copy() - copies one VolumeInfoList to another.
// free() - frees memory associated with a VolumeInfoList's contents.
//
// IMPLEMENTATION NOTES:
//
//----------------------------------------------------------------------

declare(RWGSlist, VolumeInfo)

class VolumeInfoList {

public:

    VolumeInfoList() {};
    VolumeInfoList(const VolumeInfoList& viList) { copy(viList); }
    ~VolumeInfoList() {};

    VolumeInfo          *searchBySid(sid_t sid);
    VolumeInfo          *searchByVolName(const RWCString& volname);
    VolumeInfo          *searchByVolNum(UINT16 volnum);
    void                append(const VolumeInfo& volinfo);
    VolumeInfo          *at(UINT32 index);
    const VolumeInfo    *at(UINT32 index) const;
    void                remove(UINT32 index);
    UINT32              entries() const { return theList_.entries(); }

    INT32               operator==(const VolumeInfoList& viList) const;
    INT32               operator!=(const VolumeInfoList& viList) const {
        return !(*this == viList);
    }
    VolumeInfoList&     operator=(const VolumeInfoList& viList) {
        copy(viList);
        return *this;
    }
    void                clear(void) { free(); }
    void                writeAsText (String& text) const;
    

private:

    RWGSlist(VolumeInfo)        theList_;       // Just the raw RW class

    void        copy(const VolumeInfoList& viList);
    void        free();

};

//----------------------------------------------------------------------
// CLASS: VolumeSetInfo     --DVM
//
//  Holds all information known about a volume set.
//
// PUBLIC DATA MEMBERS:
//   UINT16            expectedVols      Number of vols expected
//   UINT32            sectorSize        should be obvious
//   MediaType         mediaType         should be obvious
//   String            volSetName        Name given to Volume Set
//   Boolean           validRootVol      True iff valid root vol loaded
//   sid_t             rootSid           Sid for vol holding
//                                       FileSetInfo
//   IntegrityType     integrityType     Volume Set integrity
//   WriteProtectType  writeProtectType  Volume Set protection
//   VolumeInfoList    volList           List of volume info
//   FileSetList       fileSetList       List of file set info
//
// PUBLIC METHODS:
//  NONE.
//
// USAGE NOTES:
//  Normally, this is for inspection only.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  
//  String
//  VolumeInfoList
//  FileSetList
//
// PRIVATE METHODS:
//  NONE.
//
// IMPLEMENTATION NOTES:
//  Not sure expectedVols will remain.
//
//----------------------------------------------------------------------
class VolumeSetInfo {

public:

    UINT16            expectedVols;
    UINT32            sectorSize;
    MediaType         mediaType;
    String            volSetName;
    VolumeInfoList    volInfoList;
    Boolean           isComplete;
    sid_t             rootSid;
    IntegrityType     integrityType;
    WriteProtectType  writeProtectType;
    FileSetList       fileSetList;
    void              writeAsText(String& text) const;
};

#endif // _NSRINT_
