//-*-c++-*--------------------------------------------------------------
//  nsrArc.cpp
//   Archive object methods
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//  CLASSES:
//   Archive
//  
//----------------------------------------------------------------------
//
// $Source: /projects/HMS/nsr/udfCore/ms/nsrArc.cpp,v $
// $Revision: 1.101 $   $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/05/08 17:18:33 $
//

extern int trace;

#include "nsrLib.h"

// Archive object statics

NSR::UINT16  NSR::Archive::incrementalSeqSectors = 20;
NSR::Boolean NSR::Archive::embeddedEAs           = True;

NSR::NetwareTrusteeInfo     NSR::Archive::emptyTrustees;
NSR::ExtendedAttrHeaderDesc NSR::Archive::static_eahd;

#ifdef TRUSTEES
NSR::NovellTrusteeEA        NSR::Archive::static_ntea;
#endif // TRUSTEES

// _GENERAL_ Functions

//----------------------------------------------------------------------
// METHOD:  Archive::Archive
//      
//  Constructor and destructor for the archive object.
//
// ARGS: none.
//
// RETURNS: nothing.
// 
// PRE-CONDS: none.
// 
// POST-CONDS: none.
// 
// ERRORS: none.
// 
//----------------------------------------------------------------------
NSR::Archive::Archive() :

    partStartSector( LOC_PART_START ),
    uniqueID( 16 )  // start uniqueID at 16 for OSTA compliance
{
    err = error_None;
    icbTable = 0;
    newSeqExtent.len = 0;
    rootICBICBTable = False;
    volumeSetNameChanged = False;
    mode = mode_Closed;
    volumeSetName = "";
}

NSR::Archive::~Archive()
{
    delete icbTable;
}

//----------------------------------------------------------------------
// METHOD:  Archive::setDevice
//      
// ARGS:
//      MSDevice dev IN multi-surface device object
//
// RETURNS:
// 
// PRE-CONDS: 
// 
// POST-CONDS: 
//      The private data member "currentDevice" is set.
// 
// ERRORS:
// 
// NOTES: import & export
//----------------------------------------------------------------------
void
NSR::Archive::setDevice( MSDevice *msDev )
{
    Trace trace("Archive::setDevice()");

    if ( mode != mode_Closed ) {
        err = error_Invalid_function;
        return;
    }
    currentDevice = msDev;
    err = error_None;
    return;
}

//----------------------------------------------------------------------
// METHOD:  Archive::openArchive 
//      
// ARGS:
//    UINT16 maxNumLabeledVols IN   maximum number of labeled
//                                  volumes in set 
//    const char *volSetName   IN   volume set name
//    UINT32 sectorSize        IN   logical block size
//    MediaType mt             IN   Rewritable or WORM
//
// RETURNS:
// 
// PRE-CONDS: 
// 
// POST-CONDS: 
// -  volumeTable is populated with empty volumes. It will be filled
//    in later with real volume information via the function
//    addVolume(). Note that this implementation will likely change in
//    order to allow the 
// -  various other private member variables are set.
//
// ERRORS:
// 
// openArchive replaces the pre-alpha createArchive() and
// openArchive().  It requires a vsname, lvname, and a sectorSize, as
// well as the number of volumes in the volume set. The other two
// parameters from createArchive() are moved as follows:
//
// - fsname - beginExport(fsname)
// - fileDataStartSector - addUnlabeledVol(labelLimit)
//
// openArchive now also requires the media type, replacing
// setMediaType().  NOTE that the mediaType and sectorSize are then
// checked against the properties of the media in addVolume().
//
// NOTE that the vsname is simply stored off in a data member, and
// addVolume does not allow any volumes to be added whose vsname does
// not match it.
//
// NOTE ALSO that the sectorSize is used in the calculation of the
// overflow numbers below, inside of openArchive.
//
// - extentsPerICB
// - extentsPerBlock
// - extentsPerUSE
// - extentsPerAED
//
// CONCERNS:
//
// - How will other methods cope without the default openArchive()? Do
//   we need to have a private openArchive() with no parms? Surely it
//   should be named something else...?
//
//----------------------------------------------------------------------

void
NSR::Archive::openArchive( UINT16 maxNumLabeledVols,
                           const char *volSetName,
                           UINT32 sectorSize,
                           MediaType mt )
{
    Trace trace("Archive::openArchive()");

    // Only allow one open archive at a time...

    if (mode != mode_Closed) {
        err = error_Invalid_function;
        return;
    }

    // initialize state variables, reset flags

    vidCurrent     = 0;
    endICBTable    = False;
    endDirectory   = False;

    // Initialize volumeTable to nvols long

    volumeTable.clear();

    // Empty means: sid=0, name="", lbs=0.
    Volume emptyVolume( 0, "", 0);
    if (err=emptyVolume.error()) { return; }

    for (int j=0; j < (int)maxNumLabeledVols; j++)
        volumeTable.append( emptyVolume );

    if (err=volumeTable.error()) { return; }

    // Save off volumeSetName

    volumeSetName = String(volSetName);
    if ( err = volumeSetName.error() ) return;

    // Save off sectorSize
    
    logicalSectorSize = sectorSize;
    logicalBlockSize  = sectorSize;

    // Save off media type

    media = mt;

    // Initialize constants for calculating overflows, etc.

    LongAllocDesc   long_ad;
    FileEntry       fe(1);
    AllocExtentDesc aed(1,ad_Long);

    UINT32 long_ad_sz = long_ad.field.len();
    UINT32 fe_sz      = fe.descLen() - long_ad_sz;
    UINT32 aed_sz     = aed.descLen();

    // Note: The length of the ads are included in a FileEntry but not in
    //       a AllocExtentDesc.  Therefore, one long_ad needs to be subtracted
    //       from the descLen() returned by a FileEntry.

    extentsPerICB   = (logicalBlockSize - fe_sz - long_ad_sz)/long_ad_sz;
    extentsPerBlock = (logicalBlockSize - aed_sz - long_ad_sz)/long_ad_sz;

    UnallocatedSpaceEntry *use = new UnallocatedSpaceEntry(1);
    if (!use) { err = error_No_free_store; return; }
    extentsPerUSE = use->maxADTEnts(logicalBlockSize);
    delete use;

    AllocExtentDesc aed2(1,ad_Short);
    extentsPerAED = aed2.maxADTEnts(logicalBlockSize) - 1;

    // If success, set mode to 'mode_VolSet'
    mode = mode_VolSet;
    return;
}

//----------------------------------------------------------------------
// METHOD:  Archive::closeArchive
//      
//  Performs clean up operations on archive object.
//
// ARGS:
//  NONE.
//
// RETURNS:
// 
// PRE-CONDS:  
// 
// POST-CONDS: 
//    - The icb table is freed.
//    - mode = Closed.
//
// ERRORS:
//  This clears out any errors.
// 
// NOTES:
//----------------------------------------------------------------------
void
NSR::Archive::closeArchive()
{
    Trace trace("Archive::closeArchive()");

    // free any resources such as free store
    delete icbTable;
    icbTable = 0;

    // reset any variables, reset flags
    err = error_None;
    mode = mode_Closed;
}

// _LABELING_ Functions

//----------------------------------------------------------------------
// METHOD:  Archive::getVolumeInfo
//      
//    Return volume information to the archive user.
//
// ARGS:
//    sid_t sid,            IN    surface id of volume
//    UINT16 *nvols         OUT   # volumes. > 0 if is or has been root.
//    char *vsname          OUT   volume set name (132 bytes)
//    char *vname           OUT   volume name (32 bytes)
//    UINT16 *vid           OUT   ordinal volume number - 0,1,2,...
//    ExtentList *unerased  OUT   unerased space information
//    ExtentList *erased    OUT   erased space information
//    UINT32 *dataStart     OUT   where file data starts (end of fsd seq + 1)
//
// RETURNS:
//    Returns True if volume has a UDF label; otherwise, returns False.
// 
// PRE-CONDS:
// 
// POST-CONDS:
// 
// ERRORS:
// 
// NOTES:
//  This temporarily changes the mediaType and sectorSize states of
//  the Archive object, but then changes them back to what they were.
//
//----------------------------------------------------------------------
NSR::Boolean
NSR::Archive::getVolumeInfo( sid_t sid,
                             UINT16 *nvols,
                             char *vsname,
                             char *vname,
                             UINT16 *vid,
                             ExtentList *unerasedSpace,
                             ExtentList *erasedSpace,
                             UINT32 *dataStart )
{
    Trace trace("Archive::getVolumeInfo()");

    // Save off current archive settings for device, and then restore
    // them at end of this method. This is the ONLY method which needs
    // to allow access to media which does not match that specified in
    // the openArchive() arguments.
    MediaType savedMT  = media;
    UINT32    savedLSS = logicalSectorSize;
    UINT32    savedLBS = logicalBlockSize;
    
    media  = currentDevice->mediaType(sid);
    if ( err = currentDevice->error() ) {
        // Restore prior values.
        media             = savedMT ;
        logicalSectorSize = savedLSS;
        logicalBlockSize  = savedLBS;
        return False;
    }
    logicalSectorSize = currentDevice->sectorSize(sid);
    if ( err = currentDevice->error() ) {
        // Restore prior values.
        media             = savedMT ;
        logicalSectorSize = savedLSS;
        logicalBlockSize  = savedLBS;
        return False;
    }
    logicalBlockSize  = logicalSectorSize;

    // Need extents to call readvds
    Extent uICBAddr;
    Extent eICBAddr;

    if ( !readvds( sid, vid, vsname, vname,
                   NULL, NULL,nvols,
                   &uICBAddr, &eICBAddr,
                   NULL)) {
        // Restore prior values.
        media             = savedMT ;
        logicalSectorSize = savedLSS;
        logicalBlockSize  = savedLBS;
        return False;
    }

    // If given an unerasedSpace list to fill out, do it.
    if ( unerasedSpace != NULL ) {
        getSpace(uICBAddr,*unerasedSpace);
    }

    // If given an erasedSpace list to fill out, and not WORM, do
    // it. Otherwise, clear it.
    if ( erasedSpace != NULL) {
        if (media != media_Worm) {
            getSpace(eICBAddr,*erasedSpace);
        } else {
            erasedSpace->clear();
        }
    }

    // Dump out free space
    if(Trace::logAlgorithms()) {
        fprintf(Trace::logFile(),"getVolumeInfo() Unerased Space :\n");
        unerasedSpace->fprint(Trace::logFile());
        fprintf(Trace::logFile(),"getVolumeInfo() Erased Space :\n");
        erasedSpace->fprint(Trace::logFile());
    }
    
    // Now determine dataStart
    if (dataStart) {
        // Calculate from USEs whether root or not.
        PartitionAreaCalc partAreas(logicalBlockSize,
                                    media,
                                    uICBAddr,
                                    eICBAddr);
        *dataStart = partAreas.fileDataStart();
    }
    // Restore prior values.
    media             = savedMT ;
    logicalSectorSize = savedLSS;
    logicalBlockSize  = savedLBS;
    return True;
}

//----------------------------------------------------------------------
// METHOD: Archive::addLabeledVolume
//
//    Adds a volume to the Archive if, after checking its blockSize
//    and mediaType, as well as its vsname and its volume number, it
//    finds that it fits in the open VolumeSet.
//
//    Its mediaType, sectorSize, and volume set name must exactly
//    match that of the current volume set, and the volume number must
//    be less than or equal to the maxNumLabeledVols parameter
//    supplied in the openArchive() method.
//
// ARGS:
//    sid   IN  surface identifier
//
// RETURNS:
//    void
// 
// PRE-CONDS:  
// 
// POST-CONDS:
// 
// ERRORS:
//  This clears out any existent error before doing anything.
// 
// NOTES:
// 
//----------------------------------------------------------------------

void
NSR::Archive::addLabeledVolume( sid_t sid )
{
    Trace trace("Archive::addLabeledVolume()");
    
    // !!! Need to review error cases and handling
    // This does NO checking. This is PROBABLY right, since the mode
    // will not get changed if openArchive has an error. The only other
    // dependency would be if the supplied device does not exist.
    err = error_None;
    
    if ( mode != mode_VolSet ) {
        err = error_Invalid_function;
        return;
    }

    Volume vol( sid, "ShouldGetNameFromVol", logicalBlockSize );
    Boolean hasAVD ( False );
    vol.lvNumVols = 0;
    if ( err = vol.error() )
        return;
    
    // Error if media type does not match media type of volset.
    if ( currentDevice->mediaType(sid) != media ) {
        err = error_Wrong_media_type;
        return;
    }

    // Error if device.sectorSize(sid) does not match
    if ( logicalSectorSize != currentDevice->sectorSize(sid) ) {
        err = error_Wrong_sector_size;
        return;
    }

    char volSetName[128];
    char volName[32];
    Extent uICBAddr;
    Extent eICBAddr;
    // CLEAN: This could call getVolumeInfo() to do all of the space
    // and dataStart stuff, and then it would only need to call readvds
    // to get the fsdsExt and lvidsExt
    if (readvds(sid,&vol.vid,
                volSetName,volName,
                &vol.fsdsExt,&vol.lvidsExt,&vol.lvNumVols,
                &uICBAddr,&eICBAddr,
                &hasAVD,&vol.lvWriteProtect)) {
        // Err if not matching vsname.
        String vsn(volSetName);
        if ( volumeSetName != vsn ) {
            err = error_Wrong_volume_set;
            return;
        }
        // Err if already an entry in the volumeTable at that number.
        if ( volumeTable[vol.vid].logicalBlockSize != 0 ) {
            err = error_Duplicate_volume_number;
            return;
        }
        // Set the name field.
        vol.name(volName);

        // Fill in free space information
        getSpace(uICBAddr,vol.freeSpace);
        if (media != media_Worm) {
            getSpace(eICBAddr,vol.erasedFreeSpace);
        } else {
            vol.erasedFreeSpace.clear();
        }

        // Dump out free space
        if(Trace::logAlgorithms()) {
            fprintf(Trace::logFile(),"addLabeledVol() Unerased Space :\n");
            vol.freeSpace.fprint(Trace::logFile());
            fprintf(Trace::logFile(),"addLabeledVol() Erased Space :\n");
            vol.erasedFreeSpace.fprint(Trace::logFile());
        }

        // Now determine dataStart
        PartitionAreaCalc partAreas(logicalBlockSize,
                                    media,
                                    uICBAddr,
                                    eICBAddr);
        vol.fileDataStart = partAreas.fileDataStart();

        // If it has an LVID Sequence, read it in.
        if ( vol.lvNumVols != 0 ) {
            LogicalVolumeIntegrityDesc *lvid = NULL;
            int nFileSets = 0;
            // Read the LVID sequence.
            readLVIDSeq( vol, &lvid, &nFileSets );
            if (err) {
                delete lvid;
                return;
            }
            // Fill in volume fields.
            vol.lvIntegrity        = lvid->integrityType();
            vol.lvIntegImplementID = lvid->implementID.id();
            delete lvid;
        }
        
        // Add the volume into the table.
        volumeTable[vol.vid] = vol;
        if ( err = volumeTable.error() ) {
            return;
        }
    } else {
        if (hasAVD) {
            if (!err) {
                // volume not written by hp.
                err = error_Wrong_volume_set;
            }
            return;
        } else {
            err = error_Volume_not_labeled; // Use addUnlabeledVolume()
        }
    }
}

//----------------------------------------------------------------------
// METHOD: Archive::isCompleteVolSet  --DVM
//      
//  This returns True if all labeled volumes in the set have been
//  added. This means that there are no missing volumes (i.e., all
//  volumes up through the highest-numbered volume have been added),
//  and that the current highest-numbered volume is a valid root.
//
// ARGS:
//  NONE.
//
// RETURNS:
//  See above.
// 
// PRE-CONDS:
//  openArchive() must have been called.
// 
// POST-CONDS:
//  May set the error state.
// 
// ERRORS:
//  error_Empty_volume_set - Not complete if empty
//  error_Not_root_volume  - Highest-numbered vol is not a root
//  error_Missing_volume   - Holes in set.
// 
// NOTES:
//
//----------------------------------------------------------------------
NSR::Boolean
NSR::Archive::isCompleteVolSet( void )
{
    Trace trace("Archive::isCompleteVolSet()");
    
    // First, find the size of the volume set
    UINT16 size = sizeVolSet(volumeTable);
    err = error_None;

    // Error if no entries
    if ( size == 0 ) {
        err = error_Empty_volume_set;
        return False;
    }

    // Error if last vol is not root.
    if ( ! isValidRoot( size-1 ) ) {
        err = error_Not_root_volume; 
        return False;
    }

    // Error if missing volumes in set.
    if ( size != numContigVols(volumeTable)) {
        err = error_Missing_volume;
        return False;
    }
    return True;
}

//----------------------------------------------------------------------
// METHOD: Archive::beginGrow  --DVM
//      
//  This initiates the volume set growth process.
//
//  Complete VolSets and empty VolSets may be grown, by adding
//  unlabeled volumes to them. This must be called beforehand. See
//  completeGrow() also.
//
// ARGS: none.
//
// RETURNS: none.
// 
// PRE-CONDITIONS:
//  A complete volume set must be loaded, and the Archive object must
//  not be importing or exporting. The current volume set may be
//  either open or closed, as completeGrow() will propagate that state
//  onto the grown volume set. See markOpenIntegrity() and
//  checkOpenIntegrity().
// 
// POST-CONDITIONS:
//  The volume table is cleaned out of any empty entries and the
//  volume set is in groth mode, or an error state may be set.
// 
// ERRORS:
//  error_
//  error_
//  error_
// 
// NOTES:
//  This holds the current root information, so that the copying of
//  the fileset information may be performed by completeGrow().
//
//----------------------------------------------------------------------
void
NSR::Archive::beginGrow( void ) 
{
    Trace trace("Archive::beginGrow()");
    
    err = error_None;
    
    // Error if not in proper mode
    if ( mode != mode_VolSet ) {
        err = error_Invalid_function;
        return;
    }

    // Need to treat virgin media differently.
    Boolean emptySoFar(False);
    
    // Error if incomplete volume set.
    if ( ! isCompleteVolSet() ) {
        emptySoFar = (err == error_Empty_volume_set );
        if (emptySoFar) {
            err = error_None;   // Not an error if empty.
        } else {
            return;
        }
    }

    if ( ! emptySoFar ) {

        // Error if logical volume is write protected.
        if ( volumeTable.rootVol().lvWriteProtect != wrprot_None ) {
            err = error_Protected_logical_volume;
            return;
        }
    }

    // Shrink the volume Table.
    while ( volumeTable.getCount() > sizeVolSet(volumeTable) ) {
        volumeTable.remove(volumeTable.getCount()-1);
    }
    if ( err = volumeTable.error() ) {
        return;
    }
        
    //  Save info on current root.
    preGrowthNumVols = volumeTable.getCount();

    if ( preGrowthNumVols > 0 ) {
        FileSetList fsList;
        getFileSetInfo(& fsList);
        preGrowthNumFileSets = fsList.getCount();
        preGrowthRootVol = volumeTable.rootVol();
    } else {
        preGrowthNumFileSets = 0;
        preGrowthRootVol = Volume(0,"nullVolume",0); // Should not be looked at.
    }

    // Change modes
    mode = mode_GrowVS;
}

//----------------------------------------------------------------------
// METHOD: Archive::addUnlabeledVolume
//
//    Labels a volume and adds it to the volume set.
// 
//    Labels the supplied volume with the supplied name, and adds it
//    to the current volume set. The next available vid is assigned to
//    it, as is the current volume set name. Note that it must have
//    the same sectorSize and media type as the current volume set in
//    order to be added. Thus, it grows the volume set by one volume.
//
//    The volume name must be unique within the volume set, so if it
//    is not unique, an error will be returned.
//
//    The label contains free space information also, so the initial
//    free space must be supplied to this method as well. Note that
//    the free space information is resupplied to the Archive object
//    on every export, but also note that this is the free space
//    information which will be returned on getVolumeInfo() and
//    getVolSetInfo() until an export is performed.
//
//    The limit for how many sectors may be used by the label
//    information at the beginning of media must also be
//    supplied. The labeling process itself will not necessarily use
//    all of the available space, but rather round this limit down to
//    where the difference between it and 257 is a multiple of
//    four. This number must be greater than or equal to 297, or an
//    error will be returned.
//
//    For example:
//
//        labelLimit = 296 => ERROR
//        labelLimit = 297 => used = 297
//        labelLimit = 298 => used = 297
//        labelLimit = 299 => used = 297
//        labelLimit = 300 => used = 297
//        labelLimit = 301 => used = 301
//
// ARGS:
//    sid_t      sid           IN  surface identifier
//    char      *volName       IN  volume name
//    UINT32     labelLimit    IN  Maximum number of sectors allowed
//                                 for label information at beginning
//                                 of media.
//    ExtentList unerasedSpace IN  Free space after numLabeledSectors
//    ExtentList erasedSpace   IN  Free space after numLabeledSectors
//
// RETURNS:
//    void
// 
// PRE-CONDS:
//    beginGrow() must have been called before this is called.
// 
// POST-CONDS:
//    The volume set is grown by one, but the root information has not
//    yet been copied to it.
// 
// ERRORS:
//    Any of several device errors may be set, plus the following:
//
//      error_Invalid_function
//      error_Wrong_media_type
//      error_Wrong_sector_size
//      error_Volume_already_labeled
// 
// NOTES:
//    
//----------------------------------------------------------------------
void
NSR::Archive::addUnlabeledVolume ( sid_t sid,
                                   const char *volName,
                                   UINT32 labelLimit,
                                   const ExtentList & unerasedSpace,
                                   const ExtentList & erasedSpace )
{
    Trace trace("Archive::addUnlabeledVolume()");
    
    // !!! Need to review error cases and handling
    // This does NO checking. This is PROBABLY right, since the mode
    // will not get changed if openArchive has an error. The only other
    // dependency would be if the supplied device does not exist.

    // Clear out any errors.
    err = error_None;

    //  Error if not in proper mode
    if ( mode != mode_GrowVS ) {
        err = error_Invalid_function;
        return;
    }

    // Error if media type does not match media type of volset.
    if ( currentDevice->mediaType(sid) != media ) {
        err = error_Wrong_media_type;
        return;
    }

    // Error if device.sectorSize(sid) does not match
    if ( logicalSectorSize != currentDevice->sectorSize(sid) ) {
        err = error_Wrong_sector_size;
        return;
    }

    // Error if non-unique volumeName
    // !!! Check for unique volumeName. (UNIMPLEMENTED)
    
    // Error if already labeled.
    UINT16 nvols;
    char junkVolSetName[128];
    char junkVolName[32];
    UINT16 vid;
    Boolean hasAVD;
    if (readvds(sid,&vid,junkVolSetName,junkVolName,
                NULL,NULL,
                &nvols,NULL,NULL,
                &hasAVD,NULL) || hasAVD ) {
        err = error_Volume_already_labeled; // Use addLabeledVolume
        return;
    }

    // Label disk with volName, next available volNum.
    Volume newVolume( sid, volName, logicalBlockSize );
    newVolume.lvNumVols = 0;
    newVolume.vid = volumeTable.getCount();

    if (err=newVolume.error()) {
        return;
    }

    // Error if improper label limit supplied.
    if (labelLimit < LOC_FILE_DATA_START_MIN) {
        err = error_Invalid_labelLimit;
        return;
    }
    // Else, set limit.
    newVolume.fileDataStart = labelLimit;

    // Add free space lists into volumeTable entry
    newVolume.free(unerasedSpace,erasedSpace);

    // Dump out free space
    if(Trace::logAlgorithms()) {
        fprintf(Trace::logFile(),"addUnlabeledVol() Unerased Space :\n");
        newVolume.freeSpace.fprint(Trace::logFile());
        fprintf(Trace::logFile(),"addUnlabeledVol() Erased Space :\n");
        newVolume.erasedFreeSpace.fprint(Trace::logFile());
    }
        
    // Add to volumeTable
    volumeTable.append( newVolume );
    if (err=volumeTable.error()) {
        return;
    }

    // FINALLY, place the label down onto the disk.
    writeVolumeLabel( volumeTable.rootVol() );
}

//----------------------------------------------------------------------
// METHOD: Archive::completeGrow  --DVM
//      
//  This completes the volume set growth process.
//
//  This writes root volume information on the current
//  highest-numbered volume, including copying any fileset history
//  from the prior root to the new root. This must be called before
//  performing import or export, once beginGrow() has been
//  called. The grown volume set is marked the same as the volume set
//  was marked during beginGrow().
//
// ARGS: none.
//
// RETURNS: none.
// 
// PRE-CONDITIONS:
//  beginGrow() must have been called.
// 
// POST-CONDITIONS:
//  The Volume Set how has a valid root volume, although its integrity
//  is still OPEN.
// 
// ERRORS:
//  error_
//  error_
//  error_
// 
// NOTES:
//
//----------------------------------------------------------------------
void
NSR::Archive::completeGrow( void ) 
{
    Trace trace("Archive::completeGrow()");
    
    // Clear out any errors.
    err = error_None;

    // Error if not in proper mode
    if ( mode != mode_GrowVS ) {
        err = error_Invalid_function;
        return;
    }

    // This might make the code easier to read.
    Volume & rootVol = volumeTable.rootVol();
    
    // Calculate LVIDS location
    // This is valid whether root moves or not.
    VolAreaCalc volAreas(rootVol.sid(),
                         logicalSectorSize,
                         256);  // Bogus value, since not really
                                // needed.
    // Save in volume table entry
    rootVol.lvidsExt = volAreas.integSeqArea();

    // Calculate FSDS location
    // This is done ahead of time, so that moveRoot may use it if it
    // likes.
    PartitionAreaCalc partAreas(rootVol.sid(),
                                logicalBlockSize,
                                media,
                                rootVol.fileDataStart);

    // Update information in root volume.
    rootVol.fsdsExt  = partAreas.fsdSeqArea();

    if ( preGrowthNumVols == 0 || preGrowthNumFileSets == 0 ) {
        // New volume set OR one without an FSD to copy forward.

        // Write an empty FSD (Leave blank if Worm)
        if (media != media_Worm) {
            writeTD( rootVol.sid(),
                     rootVol.fsdsExt.start.sector - partStartSector,
                     rootVol.fsdsExt.start.sector );
            if (err) { return; }
        }
    } else {
        // There was an LVD, and one or more FSDs
        // Copy the FSDS
        copyFSDS(preGrowthRootVol);
        if (err) { return; }
    }
    
    // Write the LVD
    writeLVD();
    if (err) { return; }

    // Complete initialization of volumeTable entry (for root) by
    // setting lvNumVols properly. This is so that it will exactly
    // match the state after addUnlabeledVol().
    rootVol.lvNumVols = volumeTable.getCount();

    // Update the Space, in case the copyFSDS needed to allocate some.
    updateSpace( rootVol );

    // Dump out free space
    if(Trace::logAlgorithms()) {
        fprintf(Trace::logFile(),"completeGrow() Unerased Space :\n");
        rootVol.freeSpace.fprint(Trace::logFile());
        fprintf(Trace::logFile(),"completeGrow() Erased Space :\n");
        rootVol.erasedFreeSpace.fprint(Trace::logFile());
    }

    // New volume sets start out as OPEN.
    // Otherwise, propagate from previous Root.
    if ( preGrowthNumVols == 0 ) {
        rootVol.lvIntegrity = integ_Open;
    } else {
        rootVol.lvIntegrity = preGrowthRootVol.lvIntegrity;
    }
    writeIntegrityEntry( rootVol );
    if (err) { return; }
    
    // Change the mode
    mode = mode_VolSet;
    return;
}

//----------------------------------------------------------------------
// METHOD:  Archive::getVolSetInfo    --AP
//      
//  This returns a summary of the information currently held by the
//  archive object about the current volume set.
//
//  Note that if the isComplete field is False, the rest of the
//  structure is simply zeroed out. Thus:
//
//  Always filled in appropriately:
//      expectedVols
//      sectorSize
//      mediaType
//      volSetName
//      volList
//      isComplete
//  Zeroed if volume set incomplete:
//      rootSid
//      integrityType
//      writeProtectType
//      fileSetList
//  
//  
// ARGS:
//  VolumeSetInfo *vsip   OUT   all info about volume set
//
// RETURNS: void
// 
// PRE-CONDS:
//  NONE
// 
// POST-CONDS:
//  Error may be set, based upon whether or not a valid root volume
//  exists. It should return the same errors as isCompleteVolSet().
// 
// ERRORS:
//  error_
// 
// NOTES:
//  volInfoList will be fill in only for those volumes which have been
//  added to the Archive. Thus, volInfoList.entries() returns how many
//  volumes have been added so far.
//
//----------------------------------------------------------------------
void
NSR::Archive::getVolSetInfo( VolumeSetInfo *vsip ) 
{
    Trace trace("Archive::getVolSetInfo()");
    
    // Clear out any errors.
    err = error_None;
    
    if (!vsip)
        return;

    err = volumeTable.error();

    // vsip->clear(); // Make sure we've got a blank slate
    vsip->expectedVols = volumeTable.getCount();
    vsip->sectorSize   = logicalSectorSize;
    vsip->mediaType    = media;
    vsip->volSetName   = volumeSetName;
    vsip->volInfoList  .clear();
    VolumeInfo volInfo;

    UINT16 i;
    for ( i = 0; i < volumeTable.getCount(); i++ ) {
        if ( volumeTable[i].logicalBlockSize != 0 ) {
            volInfo.sid         = volumeTable[i].sid()           ;
            volInfo.volName     = volumeTable[i].name()          ;
            volInfo.volNum      = volumeTable[i].vid             ;
            volInfo.labelLimit  = volumeTable[i].fileDataStart   ;
            volInfo.freeSpace   = volumeTable[i].freeSpace       ;
            volInfo.erasedSpace = volumeTable[i].erasedFreeSpace ;
            vsip->volInfoList.append(volInfo);
        }
    }
    vsip->isComplete       = isCompleteVolSet();

    // Now fill in part which is unavailable unless volume set is complete
    if ( vsip->isComplete) {
        vsip->rootSid          = volumeTable.rootVol().sid();
        vsip->integrityType    = volumeTable.rootVol().lvIntegrity;
        vsip->writeProtectType = volumeTable.rootVol().lvWriteProtect;
        getFileSetInfo( &(vsip->fileSetList) );
    } else {
        vsip->rootSid          = 0;
        vsip->integrityType    = integ_Open;
        vsip->writeProtectType = wrprot_None;
        vsip->fileSetList      .clear();
    }
    return;
}

//----------------------------------------------------------------------
// METHOD: Archive::getRootVolume  --DVM
//      
//  Returns sid of root volume in Volume Set.
//
// ARGS:
//  NONE.
//
// RETURNS:
//  The Surface ID of the root volume in the volume table. If an error
//  occurs, it returns zero, and sets the Archive object's error.
// 
// PRE-CONDS: A complete volume set must be loaded.
// 
// POST-CONDS: The error may be set.
// 
// ERRORS:
// 
// NOTES:
//  getRootVolume() now is only informational, and need not be called
//  during each export (or import).
//
//  The root volume is BY DEFINITION the volume in the set with the
//  largest vid or volume number. This method may be used on import,
//  labeling, and/or export and simply returns the sid for that
//  volume.
//
//----------------------------------------------------------------------
NSR::sid_t
NSR::Archive::getRootVolume( void )
{
    Trace trace("Archive::getRootVolume()");
    
    if ( ! isCompleteVolSet() ) {
        // isCompleteVolSet() sets error itself.
        return 0;
    }
    return volumeTable.rootVol().sid();
}

//----------------------------------------------------------------------
// METHOD:  Archive::checkOpenIntegrity  --DVM
//      
//  This is returns True if the currently-loaded volume set is OPEN.
//
//  This may be used with a partially-loaded volume set (i.e., with
//  only the root volume loaded. To do this, use the REAL volume set
//  size for the maxLabeledVols argument in openArchive(), and then
//  load it. 
//
// ARGS: none.
//
// RETURNS:
//  True if OPEN, False if CLOSED.
// 
// PRE-CONDS:
//  The root volume must be loaded.
// 
// POST-CONDS:
//  This sets an error if the highest-numbered volume is not a root
//  volume.
// 
// ERRORS:
//  error_Not_root_volume (returns False)
// 
// NOTES:
//  This simply returns the value in the volumeTable for the root
//  volume.
//
//----------------------------------------------------------------------
NSR::Boolean
NSR::Archive::checkOpenIntegrity( void )
{
    Trace trace("Archive::checkOpenIntegrity()");
    
    // Clear out any errors.
    err = error_None;

    // Error if not root. 
    if ( ! isValidRoot( volumeTable.rootIndex()) ) {
        err = error_Not_root_volume;
        return False;
    }
    return ((volumeTable.rootVol().lvIntegrity == integ_Open) ?
            True : False);
}

//----------------------------------------------------------------------
//METHOD: Archive::markOpenIntegrity
//      
//    The last logical volume descriptor is marked open. This is a 
//    necessary step prior to writing any new data to the volume
//    set, or destroying a file set, or performing any exports.
//
//    This also marks as open any other previous root volumes
//    which are marked closed. This is so that if a complete-looking
//    subset of the current volume set is loaded, it too will be OPEN.
//
// ARGS:
//  NONE
//
// RETURNS:
//  void
// 
// PRE-CONDS:
//  - Must have a complete volume set.
//  - The Logical Volume must not be write-protected (hard OR soft)
//  - WORM - All file sets must be hard write-protected.
// 
// POST-CONDS: 
// 
// ERRORS:
// 
// NOTES:
//  This is all that is required before modifying WORM
//  media. Rewritable media, on the other hand, should also have its
//  file set removed via destroyFileSet(), so that the disk contains
//  no pointer into a partial load of metadata.
//
//  Since this toolkit ONLY supports writing out file sets, we go
//  ahead and deny opening the LV if any FileSets are not hard
//  write-protected. This is better than waiting till the data
//  and/or metadata has been written, although the same check is made
//  there as well..
// 
//  No non-Worm checking on this, since must openIntegrity before
//  destroying a file set.
//
//  Note that it, like markOpenIntegrity(), requires that the archive
//  object be open, with a complete volume set.
//
// BACKWARD COMPATIBILITY NOTE:
//  Original Maestro volume sets stored the ICB table in the Integrity
//  Descriptor. Therefore, calling this function on those volume sets
//  will make file sets on those volume sets unreadable. Recommended
//  practice, then, is to IMPORT the file set BEFORE marking the
//  volume set open.
//
//  This is NOT true for newer (Beta and beyond) volume sets, where
//  volume growth (which requires first that the volume set be marked
//  open) may be done before import is performed.
//
//----------------------------------------------------------------------
void
NSR::Archive::markOpenIntegrity( void ) 
{
    Trace trace("Archive::markOpenIntegrity()");
    
    // Clear out any errors.
    err = error_None;

    // Error if not in proper mode
    if ( mode != mode_VolSet ) {
        err = error_Invalid_function;
        return;
    }

    // Error if incomplete volume set.
    if ( ! isCompleteVolSet() ) {
        // isCompleteVolSet() sets error itself.
        return;
    }

    // Grab reference to root volume
    Volume & rootVol = volumeTable.rootVol();
    
    // Error if LV is write-protected (soft OR hard)
    if ( rootVol.lvWriteProtect ) {
        err = error_Protected_logical_volume;
    }

    // Error on Invalid WORM file sets.
    if (media == media_Worm) {
        // Grab file set information
        FileSetList fsList;
        getFileSetInfo(&fsList);
        if (err)
            return;
        // Error if ANY non-hard-write-protected.
        for (int idx = 0; idx < fsList.getCount(); idx++) {
            if (! fsList[idx].fsdWriteProtect == wrprot_Hard) {
                err = error_WORM_file_sets_not_protected;
                return;
            }
        }
    }
         
    // Trace volumeTable BEFORE
    UINT16 i;
    if(Trace::logAlgorithms()) {
        for (i = 0; i<volumeTable.getCount(); i++) {
            fprintf (Trace::logFile(),"markOpenIntegrity()[before]: vt[%d] = ",i);
            volumeTable[i].fprint(Trace::logFile());
        }
    }
    // Sequence through all volume table entries, looking for any
    //  prior root volumes which are marked CLOSED. Mark them OPEN.
    for (i = 0; i<volumeTable.getCount(); i++) {
        if ( (volumeTable[i].lvNumVols != 0 ) &&              // Prior root
             (volumeTable[i].lvIntegrity == integ_Closed) ) { // Still CLOSED
            volumeTable[i].lvIntegrity        = integ_Open;
            writeIntegrityEntry( volumeTable[i] );
            if (err) { return; }
        }
    }
    // Trace volumeTable AFTER
    if(Trace::logAlgorithms()) {
        for (i = 0; i<volumeTable.getCount(); i++) {
            fprintf (Trace::logFile(),"markOpenIntegrity()[after]: vt[%d] = ",i);
            volumeTable[i].fprint(Trace::logFile());
        }
    }
}

//----------------------------------------------------------------------
// METHOD: Archive::destroyFileSet  --DVM
//      
//  Overwrites the FileSetDescriptor.
//
//  This wipes out the file set descriptor on the supplied root volume
//  sid. It is used on rewritable when metadata space is re-used, in
//  addition to marking a logical volume open.
//
//  This is called in addition to markOpenIntegrity() on Rewritable.
//
// ARGS:
//  NONE
//
// RETURNS:
//  NONE.
// 
// PRE-CONDS:
//  - The volume set is complete.
//  - markOpenIntegrity() has been called.
//  - The Media type is NOT Worm.
//  - The Logical Volume must NOT be write protected (soft OR hard).
//  - The number of file sets found must exactly equal one.
//   
// 
// POST-CONDS:
// 
// ERRORS:
//  error_
// 
// NOTES:
//  This should be called after markOpenIntegrity() on rewritable
//  media. It overwrites the FSD with a TD, so that all knowledge of a
//  fileset is destroyed on the supplied volume.
//
//  Note that it, like markOpenIntegrity(), requires that the archive
//  object be open, with a complete volume set.
//
//  This returns an error if the media is WORM, and if there is no
//  file set descriptor?.
//
//----------------------------------------------------------------------
void
NSR::Archive::destroyFileSet( void )
{
    Trace trace("Archive::destroyFileSet()");
    
    if(Trace::logAlgorithms()) {
        fprintf (Trace::logFile(),"destroyFileSet(): rootVol = ");
        volumeTable.rootVol().fprint(Trace::logFile());
    }
    
    // Clear out any errors.
    err = error_None;

    // Error if not in proper mode
    if ( mode != mode_VolSet ) {
        err = error_Invalid_function;
        return;
    }

    // Error if incomplete volume set.
    if ( ! isCompleteVolSet() ) {
        // isCompleteVolSet() sets error itself.
        return;
    }

    // Error if WORM
    if ( media == media_Worm ) {
        err = error_Invalid_function_for_media;
        return;
    }

    // Create a reference to the root volume
    Volume & rootVol = volumeTable.rootVol();
    
    // Error if LogicalVolume is write protected
    if ( rootVol.lvWriteProtect ) {
        err = error_Protected_logical_volume;
        return;
    }

    // Error if integrity not marked open
    if ( rootVol.lvIntegrity != integ_Open ) {
        err = error_Volume_set_integrity_open;
        return;
    }

    // Grab file set information
    FileSetList fsList;
    getFileSetInfo(&fsList);
    if (err)
        return;

    // Set appropriate errors for too many or protected file sets.
    if (fsList.getCount() > 0) {
        if (fsList.getCount() > 1) {
            err = error_Too_many_file_sets;
            return;
        } else {
            if (fsList[0].fsdWriteProtect == wrprot_Hard) {
                err = error_Protected_file_set;
                return;
            }
        }
    }
    
    // Write a TD into the first sector of the FSDS.
    writeTD( rootVol.fsdsExt.start.sid,
             rootVol.fsdsExt.start.sector - partStartSector,
             rootVol.fsdsExt.start.sector );
    return;
}

//----------------------------------------------------------------------
// METHOD: Archive::destroyVolumeLabel  --DVM
//      
//  Nukes NSR information from supplied surface.
//
//  This overwrites the AVDPs on the supplied surface, if any
//  exist. This essentially makes the disk non-NSR, removing enough
//  information that it is no longer recognizable by
//  getVolumeInfo().
//
// ARGS:
//  sid   IN  surface identifier
//
// RETURNS:
//  NONE
// 
// PRE-CONDS:
//  The archive object must be OPEN, and no exports or imports may be
//  in process. Also, the media must not be WORM.
// 
// POST-CONDS:
//  Volume can now only be added to a volume set via
//  addUnlabeledVolume().
// 
// ERRORS:
//  error_Invalid_function
//  error_Wrong_media_type
//  error_Wrong_sector_size
//  error_Invalid_function_for_media
//
// NOTES:
//  This overwrites the AVDPs on the supplied surface, if any
//  exist. This essentially makes the disk non-NSR.
// 
//  This returns an error if the media is WORM.
// 
//  This must be used with extreme care, in order to prevent
//  destroying existing volumeSets. Also note that this is required
//  since the addUnlabeledVolume() function requires that a disk be
//  non-NSR if it is to label it.
//
//----------------------------------------------------------------------
void
NSR::Archive::destroyVolumeLabel ( sid_t sid )
{
    Trace trace("Archive::destroyVolumeLabel()");
    
    // Clear out any errors.
    err = error_None;
    
    //  Error if not in proper mode
    if ( mode != mode_VolSet &&
         mode != mode_GrowVS ) {
        err = error_Invalid_function;
        return;
    }

    // ONLY works on MO. Should compare with WORM....
    // Error if media type does not match media type of volset.
    if ( currentDevice->mediaType(sid) != media ) {
        err = error_Wrong_media_type;
        return;
    }

    // Error if device.sectorSize(sid) does not match
    if ( logicalSectorSize != currentDevice->sectorSize(sid) ) {
        err = error_Wrong_sector_size;
        return;
    }

    // Error if WORM
    if ( media == media_Worm ) {
        err = error_Invalid_function_for_media;
        return;
    }

    const UINT32 numSectors = currentDevice->numSectors(sid);
    const UINT32 lastSector = numSectors - 1;

    if ( err=currentDevice->error() ) {
        return;
    }

    // Nuke AVDs if valid AVDs held there.
    
    UINT16 numSectorsWritten = 0;

    numSectorsWritten += nukeDesc(tag_AVD,sid,
                                  LOC_AVD1,LOC_AVD1);
    if ( err )
        return;

    numSectorsWritten += nukeDesc(tag_AVD,sid,
                                  lastSector,lastSector);
    if ( err )
        return;

    numSectorsWritten += nukeDesc(tag_AVD,sid,
                                  numSectors-256,numSectors-256);
    if ( err )
        return;

    if ( numSectorsWritten > 0 ) { // had AVDs, so nuke BEA.

        // Volume Recognition Location:
    
        UINT32 volRecSeqPtr = ( 32767 / logicalSectorSize ) +
            (( 32767 % logicalSectorSize ) > 0 );

        // Calculate how many sectors to overwrite for BEA

        BegExtAreaDesc *bea = new BegExtAreaDesc;
        if (!bea) { err = error_No_free_store; return; }
        const UINT32 beaSectors = (( bea->descLen() / logicalSectorSize ) + 
                                   (( bea->descLen() % logicalSectorSize ) > 0 ));
        delete bea;

        // Allocate a sector-long buffer and zero it out.
        BYTE *bp;
        bp = new BYTE[logicalSectorSize];
        if (!bp) { err = error_No_free_store; return; }
        memset( bp, 0, logicalSectorSize);

        // Overwrite BEA
        for ( int i=0; i<beaSectors; i++ ) {
            writeSector( sid, volRecSeqPtr++, bp);
            if (err) { delete [] bp; return; }
        }
        delete [] bp;
    }
}

//----------------------------------------------------------------------
// METHOD: Archive::getWriteProtect  --DVM
//      
//  Returns the write protect status from the current volume.
//
// ARGS: none.
//
// RETURNS:
//  Write protect status of current LogicalVolume. Returns Hard
//  protect if there are any errors. Thus, should case on error if
//  this is the case.
// 
// PRE-CONDS: archive must be open.
// 
// POST-CONDS: error state may be set.
// 
// ERRORS:
//  error_Not_root_volume - either unlabelled or non-root volume.
//  error_Volume_set_not_supported - Can't handle non-HP??
// 
// NOTES:
//  Note that this will only work on the currently loaded volume set.
//  HOWEVER, this may be called on an incomplete volume set.
//
//----------------------------------------------------------------------
NSR::WriteProtectType
NSR::Archive::getWriteProtect(void)
{
    Trace trace("Archive::getWriteProtect()");
    
    // Clear out any errors.
    err = error_None;
    
    // Make a local alias for root volume
    Volume & rootVol = volumeTable.rootVol();

    // Error if not a valid root volume
    if (! isValidRoot(volumeTable.rootIndex())) {
        err = error_Not_root_volume;
        return wrprot_Hard;
    }
    return (volumeTable.rootVol().lvWriteProtect);
}

//----------------------------------------------------------------------
// METHOD: Archive::setWriteProtect  --DVM
//      
//  This sets the LogicalVolume's write protect to the supplied value.
//
//  setWriteProtect() enforces the limitations OSTA places on
//  hardWriteProtect never being altered, returning an appropriate
//  error if requests cannot be satisfied.
//
// ARGS:
//  wp       IN  desired write protect status
//
// RETURNS:
//  NONE.
// 
// PRE-CONDS: archive object must be open.
// 
// POST-CONDS: error state may be set.
// 
// ERRORS:
//  error_
//  error_
//  error_
// 
// NOTES:
//  Note that this operates upon the current volume set, which must
//  have its root volume loaded. It may or may not work unless all
//  volumes are loaded into the volume set.
//
//----------------------------------------------------------------------
void
NSR::Archive::setWriteProtect( WriteProtectType wp )
{
    Trace trace("Archive::setWriteProtect()");
    
    // Make sure we do not propagate an error through here.
    err = error_None;

    if ( mode != mode_VolSet ) {
        err = error_Invalid_function;
        return;
    }
    
    // Make an alias for the root volume.
    Volume & rootVol = volumeTable.rootVol();
    
    // Check to be sure we have a valid root.
    if (! isValidRoot(volumeTable.rootIndex())) {
        err = error_Not_root_volume;
        return;
    }

    // Make an alias for the current writeProtect type.
    WriteProtectType writeProtect = rootVol.lvWriteProtect;

    if (writeProtect != wp ) {
        if ( writeProtect == wrprot_Hard ) {
            err = error_Cant_change_hard_writeProtect;
        } else {
            writeProtect = wp;
            writeLVD();
        }
    }
    return;
}

// _EXPORT_ Functions

//----------------------------------------------------------------------
// METHOD:  Archive::overflowSectors(noExtents)
//      
//  Calculates the number of sectors required in addition to a file
//  ICB in order to record all its extents, or its path information.
//
//  Called during export to determine the number of overflow sectors
//  required when writing an FE. It is a f(number of extents) used in
//  storing the file data. The alternate form is an f(pathlist) used
//  in storing a symbolic link.
//
// ARGS:
//   UINT32           noExtents IN  The number of extents associated with
//                                  this file.
//   PathElementList  pList     IN  The path for the symbolic link. 
//
// RETURNS:
//  The number of additional sectors required for the file/symlink.
// 
// PRE-CONDS:
// 
// POST-CONDS:
// 
// ERRORS:
//  error_Invalid_function
//  error_No_free_store
// 
// NOTES:
//  
//----------------------------------------------------------------------
NSR::UINT32
NSR::Archive::overflowSectors(UINT32 noExtents)
{
    Trace trace("Archive::overflowSectors()");
    
    return overflowSectors( noExtents, emptyTrustees );
}

NSR::UINT32 
NSR::Archive::overflowSectors( const PathElementList &pList )
{
    Trace trace("Archive::overflowSectors()");
    
    return overflowSectors( pList, emptyTrustees );
}

//----------------------------------------------------------------------
// The following are private forms, differing only in the inclusion of
// another argument:
//
// ARGS:
//   NetwareTrusteeInfo &trustees IN  The netware trustee information.
//
// NOTES:
//  Algorithm:
//  Calculate number of sectors required to store the trustee information.
//  NetwareTrusteeInfo information is stored either in:
//  - The embedded file entry EA space
//    Every FE has space, L_EA for extended attributes. The amount of this
//    space is a function of the number of extents needed for the file data.
//  or
//  - In a dedicated FE, EAHdrDesc, IUEA combination.
//    Every FE has an EA ICB field which specifies the location of an 
//    ext attr file entry.
//----------------------------------------------------------------------

NSR::UINT32
NSR::Archive::overflowSectors( UINT32 noExtents,     // PRIVATE
                               const NetwareTrusteeInfo &trustees)
{
    UINT32 ovf = 0;
    Boolean embeddedEA = False;
    UINT32 nBytesForEA = 0;

    // Calculations are valid, regardless of mode.

    if ( ! trustees.empty() ) {
        nBytesForEA += bytesForEAHD();
        nBytesForEA += bytesForTrusteesEA(trustees);
    }
#ifdef NETWARE_ATTR
    nBytesForEA += bytesForNovellAttrEA();
#endif // NETWARE_ATTR

    nBytesForEA += (nBytesForEA % 4) ? 4 - (nBytesForEA % 4) : 0;

    // now, see if the EA stuff will fit in the embedded area
    if (embeddedEAs && (noExtents < extentsPerICB)) {
        FileEntry *fe = new FileEntry( noExtents );
        if (!fe) {
            err = error_No_free_store;
            return 0;
        }
        if (nBytesForEA <= fe->unusedBytes(logicalBlockSize)) {
            embeddedEA = True;
        }
        delete fe;
    }

    if ( !embeddedEA && nBytesForEA > 0 ) {
        ovf += 1;   // for the EA ICB
        ovf += B2S(nBytesForEA);
    }

    if ( noExtents > extentsPerICB ) {
        UINT32 extsLeft = noExtents - extentsPerICB;
        ovf += ( extsLeft / extentsPerBlock ) 
               + ( ( extsLeft % extentsPerBlock ) > 0 );
    }
    return ovf;
}

NSR::UINT32 
NSR::Archive::overflowSectors( const PathElementList &pList, // PRIVATE
                               const NetwareTrusteeInfo &trustees ) 
{
    // Calculations are valid, regardless of mode.

    UINT32 ovf = 0;
    UINT32 nBytesForEA = 0;
    // Boolean eaIsEmbedded = True; !!! Apparently not used
    
#ifdef NETWARE_ATTR
    nBytesForEA += bytesForEAHD();
    nBytesForEA += bytesForNovellAttrEA();
#endif // NETWARE_ATTR

    if ( ! trustees.empty() ) {
#ifndef NETWARE_ATTR
        nBytesForEA += bytesForEAHD();
#endif // NETWARE_ATTR
        nBytesForEA += bytesForTrusteesEA(trustees);
        // build a File ICB - assume only 1 ad is needed for the path list stuff
        FileEntry *fe = new FileEntry( 1, ad_Long );
        if (!fe) { err = error_No_free_store; return 0; }
        if (embeddedEAs && (fe->unusedBytes(logicalBlockSize)) >= nBytesForEA) {
            // ea can be embedded
        }
        else {
            ovf += 1;  // for the EA FE
            ovf += B2S(nBytesForEA);  // for the EA space
        }
        delete fe;
    }
    ovf += B2S(pList.totalBytes());
    return ovf;
}

//----------------------------------------------------------------------
// METHOD:  Archive::sizeICBTable
//      
// ARGS:
//    UINT32 nDirs    IN    number of directories, including root
//    UINT32 nFiles   IN    number of files
//    UINT32 ovf      IN    number of overflow sectors
//
// RETURNS:
//    number of sectors required to write the icb's (fileEntries).
// 
// PRE-CONDS:  
// 
// POST-CONDS: 
// 
// ERRORS:
// 
// NOTES:
//  Used during export.
//
// Calculations are valid, regardless of mode.
// ISSUE: should figure in some amount for the icb table ea in the
//        root ICB. It is a non-embedded EA, so for now, assume it is
//        2 sectors. That gives 1 sector for the EAFE and  1 sector for
//        the EA. Worst case for 512 byte media:
//            512
//           - 24     ea header
//           - 48     impl use ea fields
//           ====     
//            440     bytes for icb table 
//
//        Each extent (long_ad) is 16 bytes, so there is room for
//           440/16 = 27 extents. This should be enough.
//
//----------------------------------------------------------------------
NSR::UINT32 
NSR::Archive::sizeICBTable( UINT32 nDirs, UINT32 nFiles, UINT32 nOvfl )
{
    Trace trace("Archive::sizeICBTable()");
    
    return (nDirs + nFiles + nOvfl + 2);
}

//----------------------------------------------------------------------
// METHOD:  Archive::sizeDirectory
//      
//    Calculated sectors required to write a directory.
//
// ARGS:
//    UINT32 nentries,         IN    number of files and dirs in this directory
//    UINT32 sum_name_len      IN    sum of strlen() over all entries
//
// RETURNS:
//    number of sectors required to write the directory
// 
// PRE-CONDS:  
// 
// POST-CONDS: 
// 
// ERRORS:
// 
// NOTES:  export
//----------------------------------------------------------------------
NSR::UINT32 
NSR::Archive::sizeDirectory( UINT32 nEntries, UINT32 sumNameLen ) 
{
    Trace trace("Archive::sizeDirectory()");
    
    // Calculations are valid, regardless of mode.

    FileIdentifierDesc fid;
    const UINT32 fid_size = fid.descLen();
    UINT32 nblocks;
    nblocks = ((nEntries * ( 3 + fid_size )) + sumNameLen + logicalBlockSize 
               - 1) / logicalBlockSize;
    return nblocks;
}

//----------------------------------------------------------------------
// METHOD: Archive::setAvailSpace  --DVM
//      
//  Supplies partition parameters to the Archive object (for Export)
//
//  This must be called for all volumes in the set, before
//  allocateICBTable() are called.  may be called immediately after a
//  labeled or unlabeled volume is added, if desired.
//
// ARGS:
//  sid            IN  Surface Identifier
//  unerased       IN  list of unerased space (after labelLimit)
//  erased         IN  list of erased space (after labelLimit)
//
// RETURNS:
//  <explanation of different values>
// 
// PRE-CONDS:  (E.g., assumes)
//  List_preconds
// 
// POST-CONDS: (E.g., changes to class data members)
//  List_postconds
// 
// ERRORS:
//  error_
//  error_
//  error_
// 
// NOTES:
//
//  The PDs/USEs are not written until each volume is visited later in
//  the export process. This is because the USEs cannot be written
//  until after space used by NSR has been removed first.
//
//----------------------------------------------------------------------
void
NSR::Archive::setAvailSpace ( sid_t sid,
                              ExtentList &unerased,
                              ExtentList &erased )
{
    Trace trace("Archive::setAvailSpace()");
    
    // Clear out any errors
    err = error_None;
    
    // Error if wrong mode.
    if ( mode != mode_VolSet ) {
        err = error_Invalid_function;
        return;
    }
    // Error if any erased space on WORM
    if (media == media_Worm && erased.entries() != 0) {
        err = error_Invalid_parameter_for_media;
        return;
    }
    // Error if either Extent in error
    if (unerased.error()) {
        err = error_Invalid_extentlist;
        return;
    }
    if (erased.error()) {
        err = error_Invalid_extentlist;
        return;
    }
    // Find the appropriate entry in the volumeTable
    UINT16 volNum = volumeTable.vid(sid);
    if (volNum == 0xffff) {
        err = error_Invalid_surface_id;
    }
    Volume & vol = volumeTable[volNum];

    // Free the space into the volume.
    vol.free( unerased, erased );

    // Dump out free space
    if(Trace::logAlgorithms()) {
        fprintf(Trace::logFile(),"setAvailSpace() Unerased Space :\n");
        vol.freeSpace.fprint(Trace::logFile());
        fprintf(Trace::logFile(),"setAvailSpace() Erased Space :\n");
        vol.erasedFreeSpace.fprint(Trace::logFile());
    }
}

//----------------------------------------------------------------------
// METHOD:  Archive::allocateSeqSectors
//      
//  Allocates extents to be used during the export process for
//  extending sequences into the partition space. 
//
//  This may be used for extending the USE sequence or the file set
//  descriptor sequence. Space is allocated here and used later, as
//  needed, when writing the USE or FSD sequence.
//
// ARGS:
//
// RETURNS:
//     The total amount of free space left on all the volumes.
//
// PRE-CONDS:  
// 
// POST-CONDS: 
// 
// ERRORS:
//    error_End_of_Sequence_space
// 
// NOTES: export
//----------------------------------------------------------------------
NSR::UINT32
NSR::Archive::allocateSeqSectors()
{
    Trace trace("Archive::allocateSeqSectors()");
    
#ifdef SUPPORT_FOR_SEQ_IN_PARTITION
    UINT16 idx;
    UINT32 totalFree = 0;
    
    for (idx = 0; idx < volumeTable.getCount(); idx++)
    {
        newSeqExtent = volumeTable[idx].alloc( volumeTable[idx].neededSeqSectors, True );
    }
    if (newSeqExtent.len != S2B(nSectors)) {
        err = error_End_of_Sequence_space;
        return;
    }
    for (idx = 0; idx < volumeTable.getCount(); idx++)
    {
        totalFree += volumeTable[idx].totalFree();
    }
    return totalFree;
#else
    return 0;
#endif
}

//----------------------------------------------------------------------
// METHOD:  Archive::allocateICBTable
//      
//  Allocate space from the volume for use in writing the icb information.
//
//
// ARGS:
//    UINT32 nodirs  IN    number of directories to export (incl root)
//    UINT32 files   IN    number of files to export
//    UINT32 ovf     IN    number of overflow sectors calculated
//
// RETURNS: none.
// 
// PRE-CONDS:  
// 
// POST-CONDS: error may be set.
// 
// ERRORS:
// 
// NOTES:
//    Also allocates space for use in writing an "icb extent" table.
//    If there is any netware information associated with the root
//    icb, space for it is also calculated.
//
//    Allocates space for the ICB table.  The total amount of space
//    equals the number directories + files + overflow sectors + one
//    for each extent allocated for the ICB table.  For WORM, a
//    verification is done on the extents allocated to ensure that the
//    icb's will be located properly.
//----------------------------------------------------------------------
void
NSR::Archive::allocateICBTable( UINT32 nodirs, UINT32 files, UINT32 ovf)
{
    Trace trace("Archive::allocateICBTable()");
    
    allocateICBTable(nodirs,files,ovf,emptyTrustees);
}

//----------------------------------------------------------------------
// PRIVATE METHOD: Archive::allocateICBTable
// This is a private method.
//
// This differs from the public form only in that it has a fourth arg:
//
//    const NetwareTrusteeInfo &root_trustees 
//                          IN    netware trustee information
//----------------------------------------------------------------------

void
NSR::Archive::allocateICBTable( UINT32 nodirs,  //PRIVATE
                                UINT32 files,
                                UINT32 ovf,
                                const NetwareTrusteeInfo &root_trustees )
{
    // Clear out any errors
    err = error_None;

    // Error if mode is not volset
    if ( mode != mode_VolSet ) {
        err = error_Invalid_function;
        return;
    }
    
    numFiles = files;
    numDirs  = nodirs;

    UINT16 i;

    Boolean *used = new Boolean[volumeTable.getCount()];
    if (!used) { err = error_No_free_store; return; }
    for ( i=0; i<volumeTable.getCount(); i++ ) used[i] = False;
    
    Boolean valid_ID = True;
    UINT32 totalFree, t;
    UINT32 n = nodirs + files + ovf;
    UINT16 vid = volumeTable.rootVid(); // start with root.

    ExtentList elist;
    Extent e = volumeTable[vid].alloc( n, False );
    
    // for worm
    Boolean verified;
    
    // root ea stuff
    Boolean figuringEA = False;
    UINT32 rootEAExtents = 0;
    ICBTableEA *icbtea; 
    UINT32 maxExtraExtents = 0;
    UINT32 nBytesForEA = bytesForEAHD();
    nBytesForEA += bytesForTrusteesEA(root_trustees);
#ifdef NETWARE_ATTR
    nBytesForEA += bytesForNetwareAttr();
#endif // NETWARE_ATTR

    UINT32 nextents = 0;
    while( (n > 0) && (valid_ID) ) 
    {
        while( (e.len != 0) && (n > 0) ) 
        {
            verified = True;
            if (media == media_Worm) {
                if(Trace::logAlgorithms()) {
                    fprintf(Trace::logFile(),"allocateICBTable(): verifying Extent :");
                    e.fprint(Trace::logFile());
                }
                UINT32 nonBlankSector;
                if (currentDevice->verifyBlank(volumeTable[vid].sid(),
                    e.start.sector, B2S(e.len), &nonBlankSector ) == False) {
                    verified = False;
                }
            }
            if (verified) {
                elist.append( e );
                if ( err = elist.error() ) { delete used; return; }
                n -= B2S(e.len);
                nextents++;
                if (figuringEA) {
                    rootEAExtents++;
                    if ( rootEAExtents > extentsPerICB ) {
                        // This should really never happen. 
                        err = error_Root_EA_FileEntry_full;
                        delete used;
                        return;
                    }
                    if (rootEAExtents > maxExtraExtents) {
                        // will need to use additional sectors for ea space
                        n++;
                        maxExtraExtents += logicalBlockSize/16;
                    }
                }
            }
            if ( n <= 0 && !figuringEA) {
                //---------------------------------------------------
                // All the space has been allocated and verified.
                // See if any additional space is needed for the
                // icb table. The icb table is to be stored as an
                // extended attribute of the root icb.
                //---------------------------------------------------
                figuringEA = True;
                icbtea = new ICBTableEA(nextents+1);
                if (!icbtea) { err = error_No_free_store; delete used; return; }
                nBytesForEA += icbtea->attrLen();
                nBytesForEA += (nBytesForEA % 4) ? 4 - (nBytesForEA % 4) : 0;
                delete icbtea;
                n += 1;                  // one sector for the EA ICB 
                n += B2S(nBytesForEA);   // sectors for the EA
                root_ovf = 1 + B2S(nBytesForEA);
                maxExtraExtents = 1 + (logicalBlockSize-nBytesForEA)/16; // long_ad
            }
            if ( n > 0 )  {
                e = volumeTable[vid].alloc( n, False );
            }
        }

        used[ vid ] = True;

        if ( n > 0 ) {

            // pick next volume
            valid_ID = False; totalFree = 0;
            for ( i=0; i<volumeTable.getCount()-1; i++ ) {
                if ( !used[i] ) {
                    t = volumeTable[i].totalFree();
                    if ( t > totalFree ) { 
                        vid = i; totalFree = t; valid_ID = True;
                    }
                }
            }

            e = volumeTable[vid].alloc( n, False );

        }
    }

    delete [] used;

    if (!valid_ID) {
        err = error_No_free_space;
        return;
    }
    icbTable = new ICB_Table( elist, logicalBlockSize );
    if (!icbTable) {
        err = error_No_free_store;
        return;
    }
    endICBTable = False;
    vidCurrent = vid;
}

//----------------------------------------------------------------------
// METHOD: Archive::beginExport  --DVM
//      
//  This begins the export process, naming the fileset.
//
// ARGS:
//  fsname   IN  File Set Name
//
// RETURNS:
//  NONE.
// 
// PRE-CONDS:
//  - A complete volume set must be loaded.
//  - The Logical Volume must not be write-protected (soft OR hard).
//  - markOpenIntegrity() must have been called.
//  - setAvailSpace() must have been called for all vols with space.
//  - WORM - All file sets are hard write-protected.
//  - else - No file sets exist.
// 
// POST-CONDS:
//  Either the operation is successful, placing the Archive object
//  into export mode, or an error state will be set.
// 
// ERRORS:
//  error_
//  error_
//  error_
// 
// NOTES:
//
//----------------------------------------------------------------------
void
NSR::Archive::beginExport( const char *fsname )
{
    Trace trace("Archive::beginExport()");
    
    // Clear out any errors
    err = error_None;

    // Error if wrong mode
    if ( mode != mode_VolSet ) {
        err = error_Invalid_function;
        return;
    }

    // Error if incomplete volume set.
    if ( ! isCompleteVolSet() ){
        // isCompleteVolSet() sets error itself.
        return;
    }

    // Error if write protected
    if ( volumeTable.rootVol().lvWriteProtect != wrprot_None ) {
        err = error_Protected_logical_volume;
        return;
    }

    // Error if logical Volume not OPEN
    Boolean open = checkOpenIntegrity();
    if (err)
        return;
    
    if ( ! open ) {
        err = error_Volume_set_integrity_closed;
        return;
    }

    // Grab File set information
    FileSetList fsList;
    getFileSetInfo(&fsList);
    if (err)
        return;
    
    if (media == media_Worm) {
        // Error if not all file sets are hard
        for (int idx = 0; idx < fsList.getCount(); idx++) {
            if (! fsList[idx].fsdWriteProtect == wrprot_Hard) {
                err = error_WORM_file_sets_not_protected;
                return;
            }
        }
    } else {
        // Error if ANY file sets present
        if (fsList.getCount() != 0) {
            err = error_Too_many_file_sets;
            return;
        }
    }

    // Check to be sure there will be enough room in label area.
    if (media == media_Worm) {
        UINT32 neededSeqSectors = 0;
        checkRootVolumeSpace( neededSeqSectors );
        if (err) {
            return;
        }
    }

    fileSetName = fsname;
    if ( err = fileSetName.error() ) {
        return;
    }

    // Mark all volumes dirty, so that they WILL get their space
    // written again on completeExport().
    for (int j=0; j < volumeTable.getCount(); j++) {
        volumeTable[j].markDirty();
    }

    // Error if volumeTable in error.
    if (err=volumeTable.error()) {
        return;
    }

    // Change the MODE
    mode = mode_Export;

    // May have error, but ready to return nonetheless.
    return;
}

//----------------------------------------------------------------------
// METHOD:  Archive::createDirectory
//      
//  Creates a directory. 
//
//  Creates the subdirectory specified by dname for the current
//  directory.  The entry for the subdirectory must already exist by
//  using addDirectory().  Also, the current directory must already
//  have been written to disk.
//
//  If no current Directory exists, and no value is given for dname,
//  the root directory is created.
//
// ARGS:
//    char *dname    IN    The name of the directory.
//
// RETURNS:
// 
// PRE-CONDS:  
// 
// POST-CONDS: 
//
//    currentDirectory: newly created directory.
//
//    directoryStack:   | currentDirectory (prev)   |<-- top of stack
//                      |---------------------------|
//                      | currentDirectory (prev-1) |<-- top of stack - 1
// 
// ERRORS:
//    error_Directory_not_found: Directory to be created not found in 
//        currentDirectory.
// 
// NOTES: export
//----------------------------------------------------------------------
void
NSR::Archive::createDirectory( char *dname )
{
    Trace trace("Archive::createDirectory()");
    
    // Clear out any errors
    err = error_None;

    // Error if not in export mode.
    if (mode != mode_Export) {
        err = error_Invalid_function;
        return;
    }

    // Handle root specially
    if ( ( dname == 0 ) && ( directoryStack.getCount() == 0 ) ) {
        // this is root directory

        icbTable->resetPtr();
        Directory nDirectory( "", icbTable->currentAddr() );
        if (err = nDirectory.error()) return;

        // add parent directory entry
        DirectoryEntry entry;
        entry.dir     = True;
        entry.name    = String( "" );
        entry.icbAddr = nDirectory.icbAddress();
        if (err = entry.name.error()) return;

        // add entry to current directory
        nDirectory.contents.append( entry );
        if (err = nDirectory.contents.error()) return;

        // bump ICB table pointer
        endICBTable = False;
        icbTable->bumpPtr( 1 + root_ovf);

        if (icbTable->error() == error_End_of_ICB_table)
            endICBTable = True;
        else
            err = icbTable->error();

        if (err) return;

        // initialize current directory as new directory
        currentDirectory = nDirectory;

    } else {

        // lookup dname in current directory - it should be there
        unsigned i=0; Boolean notFound = True;

        String dn( dname );
        if (err=dn.error()) return;
        while( (i < currentDirectory.contents.getCount()) && notFound )
            if ( currentDirectory.contents[i++].name == dn ) 
                notFound = False;

        if ( notFound ) {
            err = error_Directory_not_found;
            return;
        }

        Directory nDirectory( currentDirectory.contents[i-1].name,
                              currentDirectory.contents[i-1].icbAddr );
        if (err = nDirectory.error()) return;

        // add parent directory entry
        DirectoryEntry entry;
        entry.dir     = True;
        entry.name    = String( "" );
        entry.icbAddr = currentDirectory.icbAddress();
        if (err = entry.name.error()) return;

        // add entry to current directory
        nDirectory.contents.append( entry );
        if (err = nDirectory.contents.error()) return;

        // push current directory on stack
        directoryStack.push( currentDirectory );
        if (err = directoryStack.error()) return;

        // initialize current directory as new directory
        currentDirectory = nDirectory;
    }
    err = currentDirectory.error();
}

//----------------------------------------------------------------------
// METHOD:  Archive::addDirEntry
//      
//  Add a file or subdirectory to the contents of currentDirectory.
//
// ARGS:
//    Boolean  isSubDir IN  Whether is a subdirectory or a file.
//    char    *name     IN  name of file or subdirectory
//    UINT32   ovf      IN  overflow sectors for file or subdirectory
//
// RETURNS:
// 
// PRE-CONDS:  
// 
// POST-CONDS: 
// 
// ERRORS:
//    error_Name_not_unique
// 
// NOTES:
//  Overflow sectors are normally only for files, but if
//  extendedAttributes are attached to a directory, ovfl will be
//  non-zero. 
//----------------------------------------------------------------------
void
NSR::Archive::addDirEntry( Boolean isSubDir, char *name, UINT32 ovf )
{
    Trace trace("Archive::addDirEntry()");
    
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Export) {
        err = error_Invalid_function;
        return;
    }

    String nm( name );
    if (err=nm.error())
        return;

    // verify name is unique in directory
    for ( unsigned i=0; i<currentDirectory.contents.getCount(); i++) {
        if ( currentDirectory.contents[i].name == nm ) { 
            err = error_Name_not_unique;
            return; 
        }
    }

    // create entry and get ICB table pointer location
    DirectoryEntry entry;
    entry.dir     = isSubDir;
    entry.name    = nm;
    entry.icbAddr = icbTable->currentAddr();
    if (err = entry.name.error())
        return;

    // add entry to current directory
    currentDirectory.contents.append( entry );
    if (err = currentDirectory.contents.error())
        return;

    // bump ICB table pointer
    endICBTable = False;
    icbTable->bumpPtr( 1 + ovf );

    if (icbTable->error() == error_End_of_ICB_table)
        endICBTable = True;
    else
        err = icbTable->error();

    return;
}

//----------------------------------------------------------------------
// METHOD:  Archive::writeDirectory
//      
//  Write currentDirectory to disk.
//
//  Dumps info about current directory, the sets current directory to
//  parent.
//
// ARGS:
//    Extent *ext    OUT    Location where currentDirectory was written.
//                          Used in writing the ICB.
//
// RETURNS:
// 
// PRE-CONDS:
// 
// POST-CONDS:
//    currentDirectory: marked clean
// 
// ERRORS:
// 
// NOTES: export
//----------------------------------------------------------------------
void
NSR::Archive::writeDirectory( Extent *ext )
{
    Trace trace("Archive::writeDirectory()");
    
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Export) {
        err = error_Invalid_function;
        return;
    }

    // calculate number of sectors required
    unsigned i;
    UINT32 totalBytes = 0;
    FileIdentifierDesc *fid;

    for ( i=0; i < currentDirectory.contents.getCount(); i++ ) 
    {
        fid = new FileIdentifierDesc( currentDirectory.contents[i].name );
        if (!fid) { err = error_No_free_store; return; }
        totalBytes += fid->descLen();
        delete fid;
    }

    UINT32 nBlocks = B2S(totalBytes);

    // if current volume is full, update to volume label and pick another
    if ( (volumeTable[vidCurrent].totalFree()<(nBlocks*logicalBlockSize)) ||
       (volumeTable[vidCurrent].largestExtent()<(nBlocks*logicalBlockSize)) ) {

        if(Trace::logAlgorithms()) {
            fprintf(Trace::logFile(),"writeDirectory(): full volume");
            fprintf(Trace::logFile(),"- updating space for volNum[%d]\n",
                    vidCurrent);
        }
        updateSpace( volumeTable[vidCurrent] ); // Write out free space.
        if (err) return;
        volumeTable[vidCurrent].markClean();

        Boolean valid_ID = True;
        UINT32 totalFree, t;
        UINT16 vid;

        valid_ID = False; totalFree = 0;
        for ( i=0; i<volumeTable.getCount()-1; i++ ) // all but last (root) vol.
        {
            t = volumeTable[i].totalFree();
            if ( t > totalFree ) { 
                vid = i; totalFree = t; valid_ID = True;
            }
        }

        if ( valid_ID )
            vidCurrent = vid;
        else
            err = error_No_free_space;
            return;
    }

    // allocate directory from free space 
    Boolean verified = False;
    while (!verified)
    {
        *ext = volumeTable[vidCurrent].alloc( nBlocks, True );
        ext->len = totalBytes;
        if (!(ext->len)) { err = error_No_free_space; return; }
        if(Trace::logAlgorithms()) {
            fprintf(Trace::logFile(),
                "writeDirectory  : at sid %3d sector %7d num_sectors %7d\n",
                ext->start.sid, ext->start.sector, nBlocks);
        }
        verified = True;
        if (media == media_Worm) {
            verified = False;
            UINT32 nonBlankSector;
            // do a verify on this extent to make sure it can be written
            if (currentDevice->verifyBlank(volumeTable[vidCurrent].sid(),
                            ext->start.sector,B2S(ext->len),
                            &nonBlankSector) == True ) {
                verified = True;
            }
            else {
                fprintf(Trace::logFile(),"writeDirectory: WARNING: non-blank sector:%d\n",nonBlankSector);
            }
        }
    }
    
    BYTE *bp = new BYTE[S2B(nBlocks)];
    if (!bp) { err = error_No_free_store; return; }
    memset( bp, 0, S2B(nBlocks));
    UINT32 bytePtr = 0;

    for ( i=0; i < currentDirectory.contents.getCount(); i++ ) 
    {
        fid = new FileIdentifierDesc( currentDirectory.contents[i].name );
        if (!fid) { err = error_No_free_store; delete [] bp; return; }

        fid->icbAddr.extType    ( ext_Recorded );
        fid->icbAddr.logBlockNo ( currentDirectory.contents[i]
                                  .icbAddr.sector - partStartSector );
        fid->icbAddr.partRefNo  ( volumeTable.vid( currentDirectory
                                                  .contents[i].icbAddr.sid ) );
        fid->icbAddr.extLength  ( logicalBlockSize );

        if ( currentDirectory.contents[i].hidden ) fid->markHidden();
        
        if ( currentDirectory.contents[i].dir ) fid->markDir();
        if ( i == 0 ) fid->markParent(); // parent is always first entry
        

        fid->toISO( bp + bytePtr, ext->start.sector + (bytePtr/logicalBlockSize) - partStartSector );
        if (err = fid->error()) { delete fid; delete [] bp; return; }
        bytePtr += fid->descLen();
        delete fid;
    }

    for ( i = 0; i < nBlocks; i++ )
    {
        writeBlock( volumeTable[vidCurrent].sid(), ext->start.sector + i, 
                     bp + (S2B(i)));
    }

    delete [] bp;

    if (err) return;

    // mark directory as clean
    currentDirectory.markClean();
    return;
}

//----------------------------------------------------------------------
// METHOD:  Archive::closeDirectory
//      
//  User is done with this directory. Pop the directory stack.
//
// ARGS:
//
// RETURNS:
// 
// PRE-CONDS:  
// 
// POST-CONDS: 
//    currentDirectory: directoryStack.pop();
//    
//    directoryStack:   | currentDirectory (prev-1) |<-- top of stack
//                      |---------------------------|
//                      | currentDirectory (prev-2) |<-- top of stack - 1
// 
// ERRORS:
//    error_Directory_dirty
// 
// NOTES: export
//----------------------------------------------------------------------
void
NSR::Archive::closeDirectory()
{
    Trace trace("Archive::closeDirectory()");
    
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Export) {
        err = error_Invalid_function;
        return;
    }

    // check if current directory is dirty
    if ( currentDirectory.isDirty() ) {
        err = error_Directory_dirty;
        return;
    }

    // pop parent directory off stack, and place as current directory
    currentDirectory = directoryStack.pop();
    err = directoryStack.error();
}

//----------------------------------------------------------------------
// METHOD:  Archive::writeRootICB
//      
//  Write the ICB for the root directory.
//
//  This will also write the ICB table - the list of extents used
//  in writing the meta-data information for this volume set. This
//  set of locations is later read during import of the volume set.
//
//  Writes the ICB for the root directory at the first
//  location on the ICB table.  It then sets the ICB pointer
//  to the next ICB.
//
// ARGS:
//    Attributes rootAttr  IN    Attributes for the root directory.
//    Extent rootExt       IN    Where the root directory is located.
//
// RETURNS: none.
// 
// PRE-CONDS:  
// 
// POST-CONDS: 
//  icbTable is bumped by 1 + #sectors used in writing the EA for the
//  Root ICB.
//
// ERRORS:
//  error_Invalid_function
//  error_No_free_store
// 
// NOTES: export
//----------------------------------------------------------------------
void
NSR::Archive::writeRootICB( Attributes &rootAttr,
                            const Extent &rootExt )
{
    Trace trace("Archive::writeRootICB()");
    
    writeRootICB( rootAttr, rootExt, emptyTrustees );
}

//----------------------------------------------------------------------
// PRIVATE METHOD: Archive::writeRootICB
//
// This is a private method, differing only in that it
// has an extra argument:
// 
// ARGS:
//    trustees             IN    Associated trustee information.
//----------------------------------------------------------------------

void
NSR::Archive::writeRootICB( Attributes &rootAttr,    //PRIVATE
                            const Extent &rootExt,
                            const NetwareTrusteeInfo &trustees )
{
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Export) {
        err = error_Invalid_function;
        return;
    }

    BYTE *bp = new BYTE[logicalBlockSize];
    if (!bp) { err = error_No_free_store; return; }
    
    // build ICB
    FileEntry *rfe = new FileEntry( 1 );
    if (!rfe) { err = error_No_free_store; delete [] bp; return; }

    // OLD: rootAttr.fileSize     = rootExt.len;
    rootAttr.fileType(file_Directory);
    rootAttr.toFileEntry(*rfe);   // was rfe->setAttributes ( rootAttr );
    rfe->infoLength = rootExt.len;  // new approach
    rfe->logicalBlocksRecorded = B2S(rootExt.len);

    // uniqueID is always set to 0 for root directory according to OSTA
    rfe->uniqueID.low( 0 );
    rfe->uniqueID.high( 0 );

    // resetPtr(), prior to writing any EA's
    icbTable->resetPtr();
    endICBTable = False;
    
    LongAllocDesc *adt = rfe->longADT();
    adt[0].extType    ( ext_Recorded );
    adt[0].logBlockNo ( rootExt.start.sector - partStartSector );
    adt[0].partRefNo  ( volumeTable.vid( rootExt.start.sid ) );
    adt[0].extLength  ( rootExt.len );

    // write the extended attributes
    Boolean isRootICB = True;
    UINT32 nEASectors;
    writeEAs(icbTable->currentAddr().sid,
             1,
             rfe,
             trustees,
             isRootICB,
             nEASectors,
             rootAttr);
    if (err) {
        delete rfe;
        delete [] bp;
        return;
    }
    memset( bp, 0, logicalBlockSize);
    rfe->toISO( bp, icbTable->currentAddr().sector - partStartSector );
    if (err = rfe->error()) { delete rfe; delete [] bp; return; }
    if(Trace::logAlgorithms())
        fprintf(Trace::logFile(),"writeRootICB   FE: at sid %3d sector %7d\n",
                icbTable->currentAddr().sid, icbTable->currentAddr().sector);
    writeBlock( icbTable->currentAddr().sid, 
                icbTable->currentAddr().sector, bp );
    delete rfe;
    delete [] bp;

    if (err) return;

    endICBTable = False;
    icbTable->bumpPtr( 1 + nEASectors );

    if (icbTable->error() == error_End_of_ICB_table)
        endICBTable = True;
    else
        err = icbTable->error();
}

//----------------------------------------------------------------------
// METHOD:  Archive::writeNextICB
//      
//  Write a non-root ICB.
//
//  Verifies that the file id matches next ICB to be written.  Writes
//  the next ICB and any required overflow sectors at the ICB pointer
//  location on the ICB table.  It then increments the ICB pointer to
//  the next available ICB.
//
// ARGS:
//    Attributes &attr            IN    Attributes for this file or dir.
//    const ExtentList &extList   IN    Locations of file data or directory.
//    PathElementList pList       IN    The symbolic link information.
//    Extent EAICBAddr            IN    Location of ICB for EAs associated
//                                      with file or directory.
//
// RETURNS:
// 
// PRE-CONDS: 
// 
// POST-CONDS:
//  The icb table address is bumped the appropriate amount.
//  endICBTable: may be set.
// 
// ERRORS:
// 
// NOTES: export
//----------------------------------------------------------------------
void
NSR::Archive::writeNextICB( Attributes &attr,
                            const ExtentList &extList )
{
    Trace trace("Archive::writeNextICB()");
    Extent nullExtent;          // Constructor makes it all empty
    writeNextICB( attr,
                  extList,
                  nullExtent);
}

void
NSR::Archive::writeNextICB( Attributes attr,
                            PathElementList pList ) 
{
    Trace trace("Archive::writeNextICB()");
    Extent nullExtent;          // Constructor makes it all empty
    writeNextICB( attr, pList, nullExtent );
}

void
NSR::Archive::writeNextICB( Attributes &attr,
                            const ExtentList &extList,
                            const Extent &EAICBAddr )
{
    Trace trace("Archive::writeNextICB()");
    writeNextICB( attr,
                  extList,
                  emptyTrustees,
                  EAICBAddr,
                  icbTable,   // Grabbed from Archive object space.
                  False );
}

void
NSR::Archive::writeNextICB( const Attributes &attr,
                            const PathElementList &pList,
                            const Extent &EAICBAddr )
{
    Trace trace("Archive::writeNextICB()");
    writeNextICB( attr, pList, emptyTrustees, EAICBAddr );
}

//----------------------------------------------------------------------
// PRIVATE METHOD: Archive::writeNextICB
// These are private methods, which support some extra arguments:
//
// ARGS:
//   NetwareTrusteeInfo &trustees IN    Associated extended attributes.
//   &trustees                    IN    Associated extended attributes.
//   ICB_Table *icbTbl            IN    Use this icb table to
//                                      determine where to write this
//                                      particular ICB. 
//   Boolean isEAfe               IN    If true, then mark ICB as such.
//----------------------------------------------------------------------

void
NSR::Archive::writeNextICB( Attributes &attr,      // PRIVATE
                            const ExtentList &extList,
                            const NetwareTrusteeInfo &trustees, 
                            const Extent &EAICBAddr,
                            ICB_Table *icbTbl,
                            Boolean isEAfe )
{
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Export) {
        err = error_Invalid_function;
        return;
    }
    // Error if end of ICB table
    if (!isEAfe && endICBTable) {
        err = error_End_of_ICB_table;
        return;
    }
    // Error if extList is in error.
    if (extList.error()) {
        err = error_Invalid_extentlist;
        return;
    }
    // Only supports Raw, Directory, and ExtendedAttributes
    if ((attr.fileType() != file_Raw) &&
        (attr.fileType() != file_Directory) &&
        (attr.fileType() != file_ExtendedAttributes)) {
        err = error_Improper_file_type;
        return;
    }
    // Error if say it's an EA AND say it's not.
    if (( isEAfe ) != ( attr.fileType() == file_ExtendedAttributes)) {
        err = error_Improper_file_type;
        return;
    }
    BYTE *bp   = new BYTE[logicalBlockSize];
    if (!bp) {
        err = error_No_free_store;
        return;
    }
    UINT32 N   = extList.getCount();
    UINT32 ovf = overflowSectors( N );
    UINT32 index = 0;
    UINT32 nEASectors = 0;
    
    FileEntry *fe;
    AllocExtentDesc *aed;

    UINT32 n;
    if ( ovf > 0 ) 
        n = extentsPerICB;
    else
        n = N;

    N -= n;

    // build the File ICB
    fe = new FileEntry( n + ( ovf > 0 ), ad_Long );
    if (!fe) {
        err = error_No_free_store;
        delete [] bp;
        return;
    }
    // was attr.fileSize     = extList.totalBytes();
    attr.toFileEntry(*fe);    // was fe->setAttributes( attr );
    fe->infoLength = extList.totalBytes();
    fe->logicalBlocksRecorded = B2S(extList.totalBytes());

    // Fill in the uniqueID
    fe->uniqueID.low( uniqueID.low() );
    fe->uniqueID.high( uniqueID.high() );
    uniqueID.next();

    // load long ads in file entry
    LongAllocDesc *adt = fe->longADT();
    
    for ( UINT32 j=0; j<n; j++ ) {
        adt[j].extType    ( ext_Recorded );
        adt[j].logBlockNo ( extList[index].start.sector - partStartSector );
        adt[j].partRefNo  ( volumeTable.vid( extList[index].start.sid ) );
        adt[j].extLength  ( extList[index].len );
        index++;
    }

    // if there is more, load a continued long ad
    if ( ovf > 0 ) {
        adt[n].extType    ( ext_Recorded );
        adt[n].logBlockNo ( icbTbl->nextAddr().sector - partStartSector );
        adt[n].partRefNo  ( volumeTable.vid( icbTbl->nextAddr().sid ) );
        adt[n].extLength  ( logicalBlockSize );
        adt[n].markContinued();
    }

    if (! isEAfe) {
        // write the extended attributes ... before the toISO()
        Boolean isRootICB = False;
        writeEAs(icbTbl->currentAddr().sid,
                 1+ovf,
                 fe,
                 trustees,
                 isRootICB,
                 nEASectors,
                 attr);

        // OR, Fill in the EAICBAddr field
        Extent nullExtent;      // constructor makes it all empty
        if (EAICBAddr != nullExtent) {
            fe->extendedAttrICB.extType    ( ext_Recorded );
            fe->extendedAttrICB.logBlockNo ( EAICBAddr.start.sector - partStartSector );
            fe->extendedAttrICB.partRefNo  ( volumeTable.vid( EAICBAddr.start.sid ) );
            fe->extendedAttrICB.extLength  ( EAICBAddr.len );
        }
    }

    // write the File ICB
    memset( bp, 0, logicalBlockSize );
    fe->toISO( bp, icbTbl->currentAddr().sector - partStartSector );
    if (err = fe->error()) {
        delete fe;
        delete [] bp;
        return;
    }
    if(Trace::logAlgorithms())
        fprintf(Trace::logFile(),"writeNextICB   FE: at sid %3d sector %7d\n",
            icbTbl->currentAddr().sid, icbTbl->currentAddr().sector);

    writeBlock( icbTbl->currentAddr().sid, 
                 icbTbl->currentAddr().sector, bp );
    
    delete fe;

    if (err) {
        delete [] bp;
        return;
    }

    icbTbl->bumpPtr( 1 );

    if (!isEAfe) {
        endICBTable = False;
        if (icbTbl->error() == error_End_of_ICB_table)
            endICBTable = True;
        else
            err = icbTbl->error();
    }

    if (err) {
        delete [] bp;
        return;
    }

    // write required AEDs to hold remaining extents
    for ( UINT32 i=0; i<ovf; i++ ) {

        if ( N > extentsPerBlock ) 
            n = extentsPerBlock;
        else
            n = N;

        N -= n;

        aed = new AllocExtentDesc( (n + ( ovf != (i+1) )), ad_Long );
        if (!aed) {
            err = error_No_free_store;
            delete [] bp;
            return;
        }
        memset( bp, 0, logicalBlockSize );

        // load long ads in allocation extent desc block
        adt = aed->longADT();
        for ( UINT32 j=0; j<n; j++ ) {
            adt[j].extType    ( ext_Recorded );
            adt[j].logBlockNo ( extList[index].start.sector - partStartSector );
            adt[j].partRefNo  ( volumeTable.vid( extList[index].start.sid ) );
            adt[j].extLength  ( extList[index].len );
            index++;
        }

        // if there are more, load a continued long ad
        if ( ovf != (i+1) ) {
            adt[n].extType    ( ext_Recorded );
            adt[n].logBlockNo ( icbTbl->nextAddr().sector - partStartSector );
            adt[n].partRefNo  ( volumeTable.vid( icbTbl->nextAddr().sid ) );
            adt[n].extLength  ( logicalBlockSize );
            adt[n].markContinued();
        }

        aed->toISO( bp, icbTbl->currentAddr().sector - partStartSector );
        if (err = aed->error()) {
            delete aed;
            delete [] bp;
            return;
        }
        delete aed;

        // write the AED block
        if(Trace::logAlgorithms())
            fprintf(Trace::logFile(),"writeNextICB  AED: at sid %3d sector %7d\n",
                icbTbl->currentAddr().sid, icbTbl->currentAddr().sector);
        writeBlock( icbTbl->currentAddr().sid, 
                    icbTbl->currentAddr().sector, bp );

        if (err) {
            delete [] bp;
            return;
        }

        endICBTable = False;
        icbTbl->bumpPtr( 1 );

        if (!isEAfe) {
            if (icbTbl->error() == error_End_of_ICB_table)
                endICBTable = True;
            else
                err = icbTbl->error();
        }
        if (err) {
            delete [] bp;
            return;
        }
        
    }
    delete [] bp;
    
    if (!isEAfe) {
        icbTbl->bumpPtr( nEASectors );
        if (icbTbl->error() == error_End_of_ICB_table) {
            endICBTable = True;
        }
        else {
            err = icbTbl->error();
        }
    }
}

void
NSR::Archive::writeNextICB( Attributes attr,     // PRIVATE
                            PathElementList pList,
                            const NetwareTrusteeInfo &trustees,
                            const Extent &EAICBAddr) 
{
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Export) {
        err = error_Invalid_function;
        return;
    }
    // Error if end of ICB Table
    if (endICBTable) {
        err = error_End_of_ICB_table;
        return;
    }
    // Error if PathElementList is in error.
    if (pList.error()) {
        err = error_Invalid_pathelementlist;
        return;
    }
    // Error if file type is not symlink.
    if (attr.fileType() != file_SymbolicLink) {
        err = error_Improper_file_type;
        return;
    }
    BYTE *bp   = new BYTE[logicalBlockSize];
    if (!bp) {
        err = error_No_free_store;
        return;
    }
    UINT32 ovf = overflowSectors( pList );

    FileEntry *fe;
    PathComponent *pc;

    // build the File ICB
    fe = new FileEntry( ovf, ad_Long );
    if (!fe) {
        err = error_No_free_store;
        delete [] bp;
        return;
    }
    // was attr.fileSize     = pList.totalBytes();
    attr.toFileEntry(*fe);  // was fe->setAttributes( attr );
    fe->infoLength = pList.totalBytes();
    fe->logicalBlocksRecorded = ovf;

    // Fill in the uniqueID
    fe->uniqueID.low( uniqueID.low() );
    fe->uniqueID.high( uniqueID.high() );
    uniqueID.next();

    Address feAddr = icbTable->currentAddr();
    endICBTable = False;
    icbTable->bumpPtr( 1 );  // for the fe

    if (icbTable->error() == error_End_of_ICB_table)
        endICBTable = True;
    else
        err = icbTable->error();

    if (err) {
        delete fe;
        delete [] bp;
        return;
    }
    // Now grab an extentlist from the ICBTable to hold the plist.
    ExtentList elist;
    getICBExtentList( 0, elist, pList.totalBytes(), False);
    if (err) {
        delete fe;
        delete [] bp;
        return;
    }
    LongAllocDesc *adt = fe->longADT();
    for ( unsigned i=0; i<ovf; i++ ) {
        adt[i].extType    ( ext_Recorded );
        adt[i].logBlockNo ( elist[i].start.sector-partStartSector );
        adt[i].partRefNo  ( volumeTable.vid( elist[i].start.sid ) );
        adt[i].extLength  ( elist[i].len );
    }
    // now that the icbTable ptr has been incremented, can write the EA's 
    Boolean isRootICB = False;
    UINT32 nEASectors = 0;
    writeEAs(icbTable->currentAddr().sid,
             0,
             fe,
             trustees,
             isRootICB,
             nEASectors,
             attr);
    
    // Fill in the EAICBAddr field ONLY if supplied
    Extent nullExtent;          // Constructor makes it all empty
    if (EAICBAddr != nullExtent) {
        fe->extendedAttrICB.extType    ( ext_Recorded );
        fe->extendedAttrICB.logBlockNo ( EAICBAddr.start.sector - partStartSector );
        fe->extendedAttrICB.partRefNo  ( volumeTable.vid( EAICBAddr.start.sid ) );
        fe->extendedAttrICB.extLength  ( EAICBAddr.len );
    }
    // write the File ICB
    memset( bp, 0, logicalBlockSize );
    fe->toISO( bp, feAddr.sector - partStartSector );
    if (err = fe->error()) {
        delete fe;
        delete [] bp;
        return;
    }
    if(Trace::logAlgorithms())
        fprintf(Trace::logFile(),"writeNextICB   FE: at sid %3d sector %7d\n",
            feAddr.sid, feAddr.sector);
    writeBlock( feAddr.sid, feAddr.sector, bp );
    delete fe;
    delete [] bp;

    if (err)
        return;

    // Create another buffer to hold the ISO version of the plist.
    bp = new BYTE[S2B(elist.numBlocks(logicalBlockSize))];
    if (!bp) {
        err = error_No_free_store;
        return;
    }
    // Zero it out
    memset( bp, 0, S2B(elist.numBlocks(logicalBlockSize)));

    // Now write the overflow sectors
    // For each path element, convert to pathComponent
    UINT32 bytePtr = 0;
    for ( i=0; i<pList.getCount(); i++ ) {
        pc = new PathComponent();
        *pc = pList[i];
        if (!pc) {
            err = error_No_free_store;
            delete [] bp;
            return;
        }
        pc->toISO( bp + bytePtr );
        if (err = pc->error()) {
            delete pc;
            delete [] bp;
            return;
        }
        bytePtr += pc->descLen();
        delete pc;
    }
    for ( i=0; i<elist.getCount(); i++ ) {
        if(Trace::logAlgorithms()) {
            fprintf(Trace::logFile(),"writeNextICB path: at sid %3d sector %7d\n",
                    elist[i].start.sid, elist[i].start.sector);
        }
        writeBlock( elist[i].start.sid, elist[i].start.sector, bp + (S2B(i)));
    }

    delete [] bp;
    icbTable->bumpPtr( nEASectors );
    if (err)
        return;
    return;
}

//----------------------------------------------------------------------
// METHOD:  Archive::completeExport
//      
//  Performs all actions appropriate for end of export.
//
//  Writes nsr volume information on all volumes. Writes a closed
//  integrity entry.
//
// ARGS: none.
//
// RETURNS: none.
// 
// PRE-CONDS: 
// 
// POST-CONDS: 
// 
// ERRORS:
// 
// NOTES:
//
//  Updates each volume still marked dirty with space
//  information. Then, writes FSD, LVID on root, and closes the
//  logical volume. 
//
//----------------------------------------------------------------------
void
NSR::Archive::completeExport()
{
    Trace trace("Archive::completeExport()");
    
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Export) {
        err = error_Invalid_function;
        return;
    }
    // Write out space information on each volume in set.
    if ( volumeTable.getCount()>=1 )
        for (int i=0; i<=volumeTable.rootIndex(); i++)
            if ( volumeTable[i].isDirty() ) {
                updateSpace( volumeTable[i] ); // Write out free space.
                if (err) {
                    return;
                }
                volumeTable[i].markClean(); // just for cleanliness
            }
    if (err) {
        return;
    }

    // Now and only now write the FSD for the file set.
    writeFSD();
    if (err) {
        return;
    }

    // Close the Volume
    volumeTable.rootVol().lvIntegrity = integ_Closed;
    writeIntegrityEntry( volumeTable.rootVol() );
    if (err) {
        return;
    }

    // FINALLY done with Export
    mode = mode_VolSet;
}

// _IMPORT_ Functions

//----------------------------------------------------------------------
// METHOD: Archive::getImplementInfo  --DVM
//      
//  This reads the ImplementationRegID from the root ICB, and returns
//  that information, so that the user may know for sure which user of
//  this code wrote the file set.
//
//  If fsNum is not supplied, the default file set information is
//  provided. 
//
// ARGS:
//  fsNum                IN   Which file set being checked.
//  implementationInfo   OUT  Information about implementation
//
// RETURNS:
//  void
// 
// PRE-CONDS:
//  Must have a complete volume set
// 
// POST-CONDS:
//  NONE
// 
// ERRORS:
//  error_Invalid_function - not in proper mode.
//  various device errors are possible as well.
// 
// NOTES:
//  <Notes_go_here>
//
//----------------------------------------------------------------------
void
NSR::Archive::getImplementInfo( ImplementationInfo * implInfo ) 
{
    getImplementInfo(-1,implInfo);
}

void
NSR::Archive::getImplementInfo( INT32 fsnum, ImplementationInfo * implInfo ) 
{
    Trace trace("Archive::getImplementInfo()");
    
    // Clear out any errors
    err = error_None;

    // Error if wrong mode.
    if (mode != mode_VolSet) {
        err = error_Invalid_function;
        return;
    }
    // Get the address for the Root ICB
    getRootICB(fsnum,rootICB);  // Archive::rootICB...
    if (err) { 
        delete icbTable;        // !!! Should this be done here?
        return; 
    }

    // Read implementation info from rootICB
    NetwareTrusteeInfo trustees;
    readRootICB(NULL,NULL,implInfo,trustees,True);
    // This sets the (err) appropriately

}

//----------------------------------------------------------------------
// METHOD:  Archive::beginImport
//    
//  Begin the import process.
//
// ARGS:
//
//  UINT16      fsnum    IN    Import this file set number. If this is
//                             not supplied, the last fileSet found
//                             will be imported. Note that File sets
//                             are numbered  0,1,2,...
//  ExtentList *icbSpace OUT   The space occupied by the fileset's
//                             ICBs. This may be spread over many
//                             volumes. This is made available, so
//                             that in the case of Rewritable, this
//                             space may be reclaimed. This does NOT
//                             include extents used in writing the
//                             directories.
//
// RETURNS: none
// 
// PRE-CONDS: A valid complete volume set is loaded.
// 
// POST-CONDS: The error state may be set.
// 
// ERRORS:
//  error_
// 
// NOTES: import 
//----------------------------------------------------------------------
void
NSR::Archive::beginImport( ExtentList *icbSpace )
{
    Trace trace("Archive::beginImport()");
    
    if ( media != media_Worm ) {
        beginImport( -1, icbSpace );
    } else {
        err = error_Invalid_function;
    }
}

void
NSR::Archive::beginImport( INT32 fsnum )
{
    Trace trace("Archive::beginImport()");
    
    if ( media == media_Worm ||
         fsnum == -1 ) {
        beginImport( fsnum, NULL );
    } else {
        err = error_Invalid_function;
    }
}

void
NSR::Archive::beginImport( void )
{
    beginImport( -1 );
}

void
NSR::Archive::beginImport(INT32 fsnum,           // PRIVATE
                          ExtentList *icbSpace )
{
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_VolSet) {
        err = error_Invalid_function;
        return;
    }
    // Error if fails import test.
    if (! okToImport()) {
        return;
    }
    // Read in old ICBTable from LVID (if there)
    Boolean LVIDICBTable(getICBTableFromLVID());
    if (err) {
        return;
    }
    // Get the address for the Root ICB
    getRootICB(fsnum,rootICB);
    if (err) { 
        delete icbTable;        // !!! Should this be done here?
        return; 
    }
    // Grab the ICBTable if needed.
    if (! LVIDICBTable ) {
        // read it from the root icb.
        Attributes attr;
        Extent ext;
        NetwareTrusteeInfo trustees;
        readRootICB(&attr,&ext,NULL,trustees,True);     
        if (err)
            return;
    }

    // Check for valid rootICB
    if (!(rootICB == icbTable->currentAddr())) {
        err = error_Bad_volume_label;
        return;
    }
    if (icbSpace) {
        *icbSpace = icbTable->location();
    }
    // Set mode
    mode = mode_Import;
}

NSR::Boolean
NSR::Archive::okToImport() 
{
    Boolean ok;
    ok = isCompleteVolSet();
    // If had problems, don't even try next piece.
    if (err)
        return ok;

    if ( media != media_Worm ) {
        // Rewriteable
        // Error if not closed.
        if ( volumeTable.rootVol().lvIntegrity != integ_Closed ) {
            err = error_Volume_set_integrity_open;
            return False;
        }

        // Error if LVID implementation is not HP
        if (( volumeTable.rootVol().lvIntegImplementID !=
              NSR::ImplRegID::HPID ) && 
            ( volumeTable.rootVol().lvIntegImplementID !=
              NSR::ImplRegID::HPEarlyID )) { 
            err = error_UDF_volume_set_not_supported;
            return False;
        }
    }
    // Worm AND MO - at least one file set.
    FileSetList fsList;
    getFileSetInfo(&fsList);
    if (err) {
        return False;
    }
    ok = (fsList.getCount() > 0) ? True : False;
    return ok;
}

//----------------------------------------------------------------------
// METHOD:  Archive::readRootICB
//      
//  Read the root ICB.
//
//  This must be the first ICB read, as the icb table is located here.
//
// ARGS:
//    Attributes *rootDirAttr      OUT  Attributes for root directory.
//    Extent *rootDirExt           OUT  Location of root directory.
//    ImplementationInfo *implInfo OUT  Implementation Information.
//    ExtentList *EAExtList        OUT  Locations holding EAs.
//
// RETURNS:
// 
// PRE-CONDS: 
//    rootICB must be set to the location of the root ICB. This is done in
//    beginImport().
//
// POST-CONDS:
//  Sets the ICB pointer to the next ICB.
// 
// ERRORS:
//  error_
// 
// NOTES:
//
//----------------------------------------------------------------------
void
NSR::Archive::readRootICB( Attributes *rootDirAttr,
                           Extent *rootDirExt )
{
    Trace trace("Archive::readRootICB()");
    
    readRootICB( rootDirAttr, rootDirExt, NULL, emptyTrustees, False );
}

//----------------------------------------------------------------------
// PRIVATE METHOD: Archive::readRootICB
//
// This private form supports reading the root ICB in order to obtain
// the ICBTable, and supports NetwareTrusteeInfo:
//
// ARGS:
//    NetwareTrusteeInfo &trustees OUT  Associated trustee information.
//    Boolean readingICBTable      IN   True if this was an archive-generated
//                                      call to read the icb table. This is
//                                      done during beginImport().
//
//----------------------------------------------------------------------

#ifdef TRUSTEES
void
NSR::Archive::readRootICB( Attributes *rootDirAttr,   // PRIVATE
                           Extent *rootDirExt,
                           NetwareTrusteeInfo &nti )
{
    readRootICB( rootDirAttr, rootDirExt, NULL, nti, False );
}
#endif //TRUSTEES

void
NSR::Archive::readRootICB( Attributes *rootDirAttr,    //PRIVATE
                           Extent *rootDirExt,
                           ImplementationInfo *implInfo, 
                           NetwareTrusteeInfo &trustees, 
                           Boolean readingICBTable )
{
    // Clear out any errors
    err = error_None;

    // readingICBTable is ONLY set to true when it is called
    // internally (i.e., by the archive object itself. Therefore, we
    // only check the mode stuff if ! readingICBTable.

    if( ! readingICBTable && mode != mode_Import ) {
        err = error_Invalid_function;
        return;
    }

#ifdef TRUSTEES
    trustees.clear();
#endif // TRUSTEES

    BYTE *bp;
    FileEntry *rfe;

    bp = new BYTE[logicalBlockSize];
    if (!bp) { err = error_No_free_store; return; }

    if ( ! readingICBTable )
        icbTable->resetPtr();

    // read root dir ICB   
    rfe = new FileEntry;
    if (!rfe) { 
        err = error_No_free_store; 
        delete [] bp; 
        return; 
    }

    if(Trace::logAlgorithms())
        fprintf(Trace::logFile(),"readRootICB   FE: at sid %3d sector %7d\n",
                rootICB.sid, rootICB.sector);
    readBlock( rootICB.sid, rootICB.sector, bp);
    if (err) {
        delete rfe;
        delete [] bp;
        return;
    }
    rfe->fromISO( bp, rootICB.sector - partStartSector );
    if (err=rfe->error()) {
        delete rfe;
        delete [] bp;
        return;
    }
    // First, get the Trustees
    // (Also fills in Archive::icbTable)
    UINT32 nEASectors = 0; // Needed later
    loadEAs( rootICB, rfe, trustees, nEASectors);
    if (err) {
        delete rfe;
        delete [] bp;
        return;
    }
    // If no icbTable by this point, BAD NEWS!
    if (!icbTable) {
        err = error_Invalid_ICB_table;
        delete rfe;
        delete [] bp;
        return;
    }
    // Fill in rootDirAttr
    if (rootDirAttr != NULL) {
        rootDirAttr-> fromFileEntry(*rfe);
    }
    // Fill in rootDirExt
    if (rootDirExt != NULL) {
        LongAllocDesc *adt = rfe->longADT();
        rootDirExt->start.sid    = volumeTable[adt[0].partRefNo()].sid();
        rootDirExt->start.sector = adt[0].logBlockNo() + partStartSector;
        rootDirExt->len          = adt[0].extLength();
    }
    // Fill in implInfo
    if (implInfo != NULL) {
        rfe->implementID.getImplementInfo(*implInfo);
    }                                               
    delete rfe;
    delete [] bp;

    if (! readingICBTable ) {
        // set the ICB pointer to the next sector after root ICB   
        endICBTable = False;
        icbTable->bumpPtr( 1 + nEASectors );
        if (icbTable->error() == error_End_of_ICB_table)
            endICBTable = True;
        else 
            err = icbTable->error() ;
    }
    return;
}

//----------------------------------------------------------------------
// METHOD:  Archive::readNextICB
//      
//  Read the next ICB.
//
//  Reads the next ICB from the ICB table. Also increments the 
//  ICB pointer to the next ICB.  
//
//  If extList returns NULL, the next ICB was not on the current
//  volume.  The next volume is then specified by addr.
//
// ARGS:
//    Attributes *attr           OUT    Attributes for next directory.
//    ExtentList *extList        OUT    Location of next directory.
//    Extent     *EAICBAddr      OUT    Location of ICB for EAs.
//
// RETURNS:
//      -1    The end of the ICB table has been reached.
//       0    The end of the ICB table has not been reached.
//
// PRE-CONDS: 
//
// POST-CONDS:
// 
// ERRORS:
//  error_
// 
// NOTES:
//    import
//
//----------------------------------------------------------------------
int
NSR::Archive::readNextICB( Address    *addr,
                           Attributes *attr,
                           ExtentList *extList )
{
    Trace trace("Archive::readNextICB()");
    return readNextICB(addr,
                       attr,
                       extList,
                       NULL);
}

int
NSR::Archive::readNextICB( Address    *addr,
                           Attributes *attr,
                           ExtentList *extList,
                           Extent     *EAICBAddr )
{
    Trace trace("Archive::readNextICB()");
    return readNextICB(addr,
                       attr,
                       extList,
                       EAICBAddr,
                       emptyTrustees);
}

//----------------------------------------------------------------------
// PRIVATE METHOD: Archive::readNextICB
//
// This private form supports reading the root ICB in order to obtain
// the ICBTable, and supports NetwareTrusteeInfo:
//
// ARGS:
//    NetwareTrusteeInfo &trustees OUT  Associated trustee information.
//    Boolean readingICBTable      IN   True if this was an archive-generated
//                                      call to read the icb table. This is
//                                      done during beginImport().
//
//----------------------------------------------------------------------

#ifdef TRUSTEES
int
NSR::Archive::readNextICB( Address    *addr,     // PRIVATE
                           Attributes *attr,
                           ExtentList *extList, 
                           NetwareTrusteeInfo &trustees )
{
    return readNextICB(addr,
                       attr,
                       extList,
                       NULL,
                       trustees);
}
#endif

int
NSR::Archive::readNextICB( Address    *addr,     // PRIVATE
                           Attributes *attr,
                           ExtentList *extList,
                           Extent     *EAICBAddr,
                           NetwareTrusteeInfo &trustees )
{
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Import) {
        err = error_Invalid_function;
        return 0;
    }

#ifdef TRUSTEES
    trustees.clear();
#endif // TRUSTEES

    if (endICBTable) {
        err = error_End_of_ICB_table;
        return 0;
    }

    BYTE *bp;

    FileEntry *fe = 0;

    bp = new BYTE[logicalBlockSize];
    if (!bp) {
        err = error_No_free_store;
        return 0;
    }
    // read next ICB
    extList->clear();   // make sure list is empty
    *addr = icbTable->currentAddr();

    if(Trace::logAlgorithms())
        fprintf(Trace::logFile(),"readNextICB   FE: at sid %3d sector %7d\n",
                icbTable->currentAddr().sid,
                icbTable->currentAddr().sector);

    readBlock( icbTable->currentAddr().sid,
               icbTable->currentAddr().sector,
               bp );
    if (err) {
        delete [] bp;
        return 0;
    }
    fe = new FileEntry;
    if (!fe) {
        err = error_No_free_store;
        delete [] bp;
        return 0;
    }
    fe->fromISO( bp, icbTable->currentAddr().sector - partStartSector );
    if (err = fe->error()) {
        delete fe;
        delete [] bp;
        return 0;
    }
    // load attributes
    attr->fromFileEntry(*fe); // was *attr = fe->getAttributes();
    // check - should not be ea !
    if (attr->fileType() == file_ExtendedAttributes) {
        err = error_Unexpected_EA_icb;
        delete fe;
        delete [] bp;
        return 0;
    }
    getICBExtents( fe, extList );
    if (err) {
        delete fe;
        delete [] bp;
        return 0;
    }
    // Fill in EAICBAddr
    if (EAICBAddr != NULL) {
        const LongAllocDesc & eaicba = fe->extendedAttrICB;
        EAICBAddr->start.sid    = volumeTable[eaicba.partRefNo()].sid();
        EAICBAddr->start.sector = eaicba.logBlockNo() + partStartSector;
        EAICBAddr->len          = eaicba.extLength();
    }
    
    // load extended attributes
    UINT32 nEASectors = 0;
#ifdef TRUSTEES
    loadEAs( icbTable->currentAddr(),
             fe,
             trustees,
             nEASectors);
    if (err) {
        delete fe;
        delete [] bp;
        return 0;
    }
#endif //TRUSTEES
    delete fe;
    delete [] bp;

    // set the ICB pointer to the next ICB
    endICBTable = False;
    icbTable->bumpPtr(1 + nEASectors);

    if (attr->fileType() == file_SymbolicLink) {                                            
        UINT32 nBlocks = B2S(extList->totalBytes());
        icbTable->bumpPtr( nBlocks );
    }                                                                    

    if (icbTable->error() == error_End_of_ICB_table)
        endICBTable = True;
    else 
        err = icbTable->error();

    return (endICBTable) ? -1 : 0;
}

//----------------------------------------------------------------------
// METHOD:  Archive::readPath
//
//    Read the path information for a symbolic link.
//
// ARGS:
//    ExtentList &elist,       IN    Locations for path information
//    PathElementList *pList   OUT   The path information.
//
// RETURNS:
// 
// PRE-CONDS: 
// 
// POST-CONDS:
// 
// ERRORS:
//  error_
// 
// NOTES:
//    import 
//----------------------------------------------------------------------
void
NSR::Archive::readPath( ExtentList &elist,
                        PathElementList *pList ) 
{
    Trace trace("Archive::readPath()");
    
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Import) {
        err = error_Invalid_function;
        return;
    }
    // Make a buffer big enough for all of it.
    BYTE *bp = new BYTE[S2B(elist.numBlocks(logicalBlockSize))];
    if (!bp) {
        err = error_No_free_store;
        return;
    }
    // Zero it all out.
    memset( bp, 0, S2B(elist.numBlocks(logicalBlockSize)));
    UINT32 bytePtr = 0;

    for ( unsigned i=0; i<elist.getCount(); i++ ) {
        if(Trace::logAlgorithms())
            fprintf(Trace::logFile(),"readNextICB path: at sid %3d sector %7d\n",
                    elist[i].start.sid, elist[i].start.sector);
        readBlock( elist[i].start.sid, elist[i].start.sector, bp + (S2B(i)));
    }
    PathElement pe;
    PathComponent *pc;
    pList->clear();
    while ( bytePtr < elist.totalBytes() ) {

        pc = new PathComponent();
        if (!pc) {
            err = error_No_free_store;
            delete [] bp;
            return;
        }
        pc->fromISO( bp + bytePtr );
        if (err = pc->error()) {
            delete pc;
            delete [] bp;
            return;
        }
        pe = PathElement(*pc);
        bytePtr += pc->descLen();
        delete pc;
        pList->append( pe );
        if (err = pList->error()) {
            delete [] bp;
            return;
        }
    }
    delete [] bp;
    return;
}

//----------------------------------------------------------------------
// METHOD: Archive::endOfICBTable  --AP
//      
//  Returns True if at end of ICB Table.
//
// ARGS:
//  NONE
//
// RETURNS:
// 
// PRE-CONDS:
// 
// POST-CONDS:
// 
// ERRORS:
// 
// NOTES:
//
//----------------------------------------------------------------------
NSR::Boolean 
NSR::Archive::endOfICBTable() 
{
    Trace trace("Archive::endOfICBTable()");
    
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Import) {
        err = error_Invalid_function;
        return True;
    }
    return endICBTable;
}

//----------------------------------------------------------------------
// METHOD:  Archive::locateDirectory
//
//  Attempts to find the specified directory on the current volume in
//  the device.
//
//  If found, sets the current directory to that specified 
//
// ARGS:
//    Extent ext    IN    Location of the directory.
//
// RETURNS:
// 
// PRE-CONDS: 
// 
// POST-CONDS:
// 
// ERRORS:
//  error_
// 
// NOTES:
//    import
//
//  Locates a directory and populates currentDirectory with its
//  contents. 
//----------------------------------------------------------------------
void
NSR::Archive::locateDirectory( Extent ext )
{
    Trace trace("Archive::locateDirectory()");
    
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Import) {
        err = error_Invalid_function;
        return;
    }
    unsigned nBlocks = B2S(ext.len);
    BYTE *bp = new BYTE[S2B(nBlocks)];
    if (!bp) {
        err = error_No_free_store;
        return;
    }
    readExtent(ext, bp);
    if (err) {
        delete [] bp;
        return;
    }
    FileIdentifierDesc fid;
    DirectoryEntry entry;
    Directory tmp;
    currentDirectory = tmp;

    UINT32 ptr=0;
    while ( ptr < ext.len ) 
    {
        fid.fromISO( bp+ptr, ext.start.sector + (ptr / logicalBlockSize) - partStartSector );
        if (err=fid.error()) {
            delete [] bp;
            return;
        }
        entry.dir            = fid.isDir();
        entry.hidden         = fid.isHidden();
        entry.name           = fid.fileID();
        entry.icbAddr.sid   = volumeTable[fid.icbAddr.partRefNo()].sid();
        entry.icbAddr.sector = fid.icbAddr.logBlockNo() + partStartSector;
        if (err=entry.name.error()) {
            delete [] bp;
            return;
        }
        if ( ! fid.isParent() ) {
            currentDirectory.contents.append( entry );
            if (err=currentDirectory.contents.error()) {
                delete [] bp;
                return;
            }
        }
        ptr += fid.descLen();
    }

    delete [] bp;
    entryPtr=0;
    endDirectory = False;
}

//----------------------------------------------------------------------
// METHOD: Archive::numDirEntries  --AP
//      
//  Returns the number of entries in the current directory.
//
// ARGS:
//  NONE
//
// RETURNS:
//  See above.
// 
// PRE-CONDS:
//  Assumes that a directory has been properly found using
//  locateDirectory(). 
// 
// POST-CONDS:
//  NONE.
// 
// ERRORS:
//  error_
//  error_
// 
// NOTES:
//----------------------------------------------------------------------
NSR::UINT32
NSR::Archive::numDirEntries()
{
    Trace trace("Archive::numDirEntries()");
    
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Import) {
        err = error_Invalid_function;
        return 0;
    }
    return currentDirectory.contents.getCount();
}

//----------------------------------------------------------------------
// METHOD: Archive::getFirstDirEntry
//      
//  Reads the first entry in the current directory. Sets the entry
//  pointer to the next directory entry.
//
// ARGS:
//  Boolean *dir   OUT  Entry is a subdirectory
//  char    *name  OUT  name of the dir entry.
//  Address *addr  OUT  ICB addr of the entry.
//
// RETURNS:
//  void
// 
// PRE-CONDS:
//  Assumes that a directory has been properly found using
//  locateDirectory(). 
// 
// POST-CONDS:
//  Pointer set to second entry, ready for getNextDirEntry().
// 
// ERRORS:
//  error_
//  error_
// 
// NOTES:
//  <Notes_go_here>
//
//----------------------------------------------------------------------
void
NSR::Archive::getFirstDirEntry( Boolean *dir,
                                char *name,
                                Address *addr )
{
    Trace trace("Archive::getFirstDirEntry()");
    
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Import) {
        err = error_Invalid_function;
        return;
    }
    if ( currentDirectory.contents.getCount() == 0 ) {
        err = error_Invalid_function;
        return;
    }

    // set entry pointer at first entry
    entryPtr=0;

    // load first entry info into parameter pointers   
    *dir = currentDirectory.contents[entryPtr].dir;
    currentDirectory.contents[entryPtr].name.toMem( name );
    *addr = currentDirectory.contents[entryPtr].icbAddr;

    err = currentDirectory.contents.error();
}

//----------------------------------------------------------------------
// METHOD: Archive::getNextDirEntry
//      
//  Reads the next entry in the current directory. Sets the entry
//  pointer to the next directory entry.
//
// ARGS:
//  Boolean *dir   OUT  Entry is a subdirectory
//  char    *name  OUT  name of the dir entry.
//  Address *addr  OUT  ICB addr of the entry.
//
// RETURNS:
//   The return value should probably be ignored. Look instead at the
//   error() return value after calling this.
// 
// PRE-CONDS:
//  Assumes that a directory has been properly found using
//  locateDirectory(). 
// 
// POST-CONDS:
//  Pointer set to next entry.
// 
// ERRORS:
//  error_
//  error_
// 
// NOTES:
//  <Notes_go_here>
//
//----------------------------------------------------------------------
int
NSR::Archive::getNextDirEntry( Boolean *dir,
                               char *name,
                               Address *addr )
{
    Trace trace("Archive::getNextDirEntry()");
    
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Import) {
        err = error_Invalid_function;
        return -1;
    }
    if ( currentDirectory.contents.getCount() == 0 ) {
        err = error_Invalid_function;
        return -1;
    }
    // if at end of buffer, return error   
    if ( entryPtr >= (currentDirectory.contents.getCount()-1) ) {
        err = error_End_of_directory;
        endDirectory = True;
        return 0;
    }

    // otherwise, bump pointer   
    entryPtr++;

    // load info from entry pointer into parameter pointers   
    *dir = currentDirectory.contents[entryPtr].dir;
    currentDirectory.contents[entryPtr].name.toMem( name );
    *addr = currentDirectory.contents[entryPtr].icbAddr;

    if (err = currentDirectory.contents.error())
        return 0;

    if ( entryPtr == (currentDirectory.contents.getCount()-1) )
        endDirectory = True;
    else
        endDirectory = False;

    return (endDirectory) ? -1 : 0;
}

//----------------------------------------------------------------------
// METHOD: Archive::getDirEntry
//      
//  Reads the specified entry in the current directory. Sets the entry
//  pointer to the next directory entry.
//
// ARGS:
//  Boolean *dir   OUT  Entry is a subdirectory
//  char    *name  OUT  name of the dir entry.
//  Address *addr  OUT  ICB addr of the entry.
//
// RETURNS:
//  void
// 
// PRE-CONDS:
//  Assumes that a directory has been properly found using
//  locateDirectory(). 
// 
// POST-CONDS:
//  Pointer set to next entry.???
// 
// ERRORS:
//  error_
//  error_
// 
// NOTES:
//  <Notes_go_here>
//
//----------------------------------------------------------------------
void
NSR::Archive::getDirEntry( unsigned i,
                           Boolean *dir,
                           char *name,
                           Address *addr )
{
    Trace trace("Archive::getDirEntry()");
    
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Import) {
        err = error_Invalid_function;
        return;
    }
    if ( currentDirectory.contents.getCount() == 0 ) {
        err = error_Invalid_function;
        return;
    }

    // if at end of buffer, return error   
    if ( i >= currentDirectory.contents.getCount() ) {
        err = error_End_of_directory;
        endDirectory = True;
        return;
    } else
        endDirectory = False;

    // load info from entry pointer into parameter pointers   
    *dir = currentDirectory.contents[i].dir;
    currentDirectory.contents[i].name.toMem( name );
    *addr = currentDirectory.contents[i].icbAddr;

    err = currentDirectory.contents.error();
    return;
}

//----------------------------------------------------------------------
// METHOD: Archive::endOfDirectory -- AP
//      
//  Returns True if all entries in the current directory are
//  exhausted. 
//
// ARGS:
//
// RETURNS:
//  See above.
// 
// PRE-CONDS:
//  Assumes that a directory has been properly found using
//  locateDirectory(). 
// 
// POST-CONDS:
//  NONE.
// 
// ERRORS:
//  error_
//  error_
// 
// NOTES:
//  <Notes_go_here>
//
//----------------------------------------------------------------------
NSR::Boolean 
NSR::Archive::endOfDirectory() 
{
    Trace trace("Archive::endOfDirectory()");
    
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Import) {
        err = error_Invalid_function;
        return False;
    }
    return endDirectory;
}

//----------------------------------------------------------------------
// METHOD:  Archive::completeImport
//  Cleans up the state of the archive object.
//
// ARGS:
//  NONE.
//
// RETURNS:
//  void
// 
// PRE-CONDS: 
//  NONE.
// 
// POST-CONDS:
//  The archive object is left ready to grow a volume group, export,
//  or import.
// ERRORS:
//  error_
// 
// NOTES:
//    import
//----------------------------------------------------------------------
void
NSR::Archive::completeImport()
{
    Trace trace("Archive::completeImport()");
    
    // Clear out any errors
    err = error_None;

    // Error if not in export mode
    if (mode != mode_Import) {
        err = error_Invalid_function;
        return;
    }
    mode = mode_VolSet;
}

// _DEBUGGING_ Methods
//----------------------------------------------------------------------
// METHOD: Archive::logFile  --DVM
//      
//  Sets the file to send logging output to.
//
// ARGS:
//  FILE * lf  IN   File pointer.
//
// RETURNS:
//  void
// 
// PRE-CONDITIONS:
//  none.
// 
// POST-CONDITIONS:
//  Any logging will go to the supplied file.
// 
// ERRORS:
//  none.
// 
// NOTES:
//  This affects the static members of Trace.
//
//----------------------------------------------------------------------
// Private comments go here - will not be extracted into HTML.
//----------------------------------------------------------------------
void
NSR::Archive::logFile(FILE * lf) {
    Trace::logFile(lf);
}

//----------------------------------------------------------------------
// METHOD: Archive::logErrors  --DVM
//      
//  Enables or disables the logging of the Archive Object's error
//  state. Whenever a public method is exited, the error, if set, will
//  be printed to the log, if a logFile has been specified.
//
// ARGS:
//  Boolean   le  IN   True -> enables; False -> disables.
//
// RETURNS:
//  void
// 
// PRE-CONDITIONS:
//  none.
// 
// POST-CONDITIONS:
//  Errors will or will not be logged to logFile.
// 
// ERRORS:
//  none.
// 
// NOTES:
//  This affects the static members of Trace.
//
//----------------------------------------------------------------------
// Private comments go here - will not be extracted into HTML.
//----------------------------------------------------------------------
void
NSR::Archive::logErrors(Boolean le) {
    if (le) {
        Trace::defaultError(err);
    }
    Trace::logErrors(le);
}
//----------------------------------------------------------------------
// METHOD: Archive::logAlgorithms  --DVM
//      
//  Enables or disables the logging of the Archive Object's
//  miscellaneous algorithmic output.
//
// ARGS:
//  Boolean   la  IN   True -> enables; False -> disables.
//
// RETURNS:
//  void
// 
// PRE-CONDITIONS:
//  none.
// 
// POST-CONDITIONS:
//  Miscellaneous printf()s will or will not be logged to logFile.
// 
// ERRORS:
//  none.
// 
// NOTES:
//  This affects the static members of Trace.
//
//----------------------------------------------------------------------
// Private comments go here - will not be extracted into HTML.
//----------------------------------------------------------------------
void
NSR::Archive::logAlgorithms(Boolean la) {
    Trace::logAlgorithms(la);
}

//----------------------------------------------------------------------
// METHOD: Archive::logCalls  --DVM
//      
//  Enables or disables the logging of the entry and exit of all of
//  the Archive Object's public methods.
//
// ARGS:
//  Boolean   lc  IN   True -> enables; False -> disables.
//
// RETURNS:
//  void
// 
// PRE-CONDITIONS:
//  none.
// 
// POST-CONDITIONS:
//  Entries and Exits to/from Archive Object's public methods will or
//  will not be logged to the logFile.
// 
// ERRORS:
//  none.
// 
// NOTES:
//  This affects the static members of Trace.
//
//----------------------------------------------------------------------
// Private comments go here - will not be extracted into HTML.
//----------------------------------------------------------------------
void
NSR::Archive::logCalls(Boolean lc) {
    Trace::logCalls(lc);
}
    
// _PRIVATE_ Methods
//----------------------------------------------------------------------
// METHOD:  Archive::parseEA
//
//    Decode any extended attributes associated with an ICB.
//
// ARGS:
//    UINT32 eaLogicalSector,   IN    Logical block address of this EA.
//                                    Needed in fromISO().
//    UINT32 nBytesInEA,        IN    Number of bytes in EA.
//    BYTE *ea,                 IN    Buffer containing ea info.
//    trustees,                 OUT   Where to place ea info.
//
// RETURNS:
// 
// PRE-CONDS: 
// 
// POST-CONDS:
// 
// ERRORS:
//  error_
// 
// NOTES:
//    import 
//----------------------------------------------------------------------
void
NSR::Archive::parseEA( UINT32 eaLogicalSector, UINT32 nBytesInEA, BYTE *ea, 
                       NetwareTrusteeInfo &trustees)
{
#ifdef TRUSTEES
    NovellTrusteeEA *ntea = NULL;
#endif // TRUSTEES

    ICBTableEA *icbea = NULL;
    LongAllocDesc *icbt = NULL;
    Extent ext;
    ExtentList elist;
    UINT16 idx;
    
    ExtendedAttrHeaderDesc *eahd = new ExtendedAttrHeaderDesc;
    if (!eahd) {
        err = error_No_free_store;
        return;
    }
    SearchEA *sea = new SearchEA;
    if (!sea) {
        err = error_No_free_store;
        delete eahd;
        return;
    }
    eahd->fromISO( ea, eaLogicalSector );
    if (err=eahd->error()) {
        delete sea;
        delete eahd;
        return;
    }
    // just worry about implementation attributes ??
    ea += eahd->implAttrLocation();
    nBytesInEA -= eahd->descLen();
    
    while (nBytesInEA) 
    {
        sea->fromISO( ea );
        switch (sea->eaType()){
        case sea_HPNovellTrustee:
#ifdef TRUSTEES
            ntea = new NovellTrusteeEA;
            if (!ntea) {
                err = error_No_free_store;
                delete sea;
                delete eahd;
                return;
            }
            ntea->fromISO( ea );
            if (err) {
                delete sea;
                delete eahd;
                delete ntea;
                return;
            }
            ea += ntea->attrLen();
            nBytesInEA -= ntea->attrLen();
            
            ntea->getTrusteeInfo(trustees);
            delete ntea;
#endif
            break;
        case sea_HPICBTable: 
            icbea = new ICBTableEA(1);  // fromISO() will resize it
            if (!icbea) {
                err = error_No_free_store;
                delete sea;
                delete eahd;
                return;
            }
            icbea->fromISO( ea );
            if (err) {
                delete sea;
                delete eahd;
                delete icbea;
                return;
            }
            ea += icbea->attrLen();
            nBytesInEA -= icbea->attrLen();

            if (!rootICBICBTable) {
                // read in the extents for the icb information
                icbt = icbea->icbTbl();
                for ( idx = 0; idx < icbea->numEnts(); idx++ )
                {
                    ext.start.sid    = volumeTable[icbt[idx].partRefNo()].sid();
                    ext.start.sector = icbt[idx].logBlockNo() + partStartSector;
                    ext.len          = icbt[idx].extLength();
                    elist.append( ext );
                    if (err=elist.error()) {
                        delete icbea;
                        delete sea;
                        delete eahd;
                        return;
                    }
                }
            
                delete icbTable;
                icbTable = new ICB_Table( elist, logicalBlockSize );
                if (!icbTable) {
                    err = error_No_free_store;
                    delete icbea;
                    delete sea;
                    delete eahd;
                    return;
                }
                rootICBICBTable = True;
            }
            delete icbea;
            break;
        default:
            ea += sea->attrLen();
            nBytesInEA -= sea->attrLen();
        }
    }
    delete eahd;
    delete sea;
    return;
}


//----------------------------------------------------------------------
// METHOD:  Archive::loadEAs
//      
//    Load the extended attributes for this ICB. 
//
//
// ARGS:
//    const Address &addr,   IN    The icbtable->currentAddr().
//    FileEntry *fe,         IN    The file entry to get the ea's for.
//    &trustees,             IN    Where to put the ea.
//    UINT32 &nEASectors     OUT   Number of EA sectors.
//
// RETURNS:
// 
// PRE-CONDS: 
// 
// POST-CONDS:
// 
// ERRORS:
//  error_
// 
// NOTES:
//    import
//----------------------------------------------------------------------
void
NSR::Archive::loadEAs( const Address &addr,
                       FileEntry *fe, 
                       NetwareTrusteeInfo &trustees,
                       UINT32 &nEASectors )
{
    BYTE *ea = NULL;
    UINT32 nBytesForEAs = 0;
    nEASectors = 0;

    if ( fe->extendedAttrICB.extLength() > 0 ) {
        // There is an extended attribute fe. Read it.
        nEASectors += 1;
        FileEntry *eafe;
        BYTE *bp = new BYTE[logicalBlockSize];
        if (!bp) {
            err = error_No_free_store;
            return;
        }
        readBlock( addr.sid,
                   fe->extendedAttrICB.logBlockNo()+partStartSector,
                   bp );
        if (err) {
            delete [] bp;
            return;
        }
        eafe = new FileEntry;
        if (!eafe) {
            err = error_No_free_store;
            delete [] bp;
            return;
        }
        eafe->fromISO( bp, fe->extendedAttrICB.logBlockNo());
        if (err = eafe->error()) {
            delete eafe;
            delete [] bp;
            return;
        }
        ExtentList extList;
        getICBExtents( eafe, &extList );
        if (err) {
            delete eafe;
            delete [] bp;
            return;
        }
        // total size of EA space is in the eafe
        Attributes eafeattr;
        eafeattr.fromFileEntry(*eafe);
        nBytesForEAs = extList.totalBytes();
        nEASectors += B2S(nBytesForEAs);

        // read the EA's
        ea = new BYTE[S2B(B2S(nBytesForEAs))];
        if (!ea) {
            err = error_No_free_store;
            delete [] bp;
            delete eafe;
            return;
        }
        BYTE *eaptr = ea;
        for (UINT32 idx = 0; idx < extList.getCount(); idx++ )
        {
            readExtent(extList[idx],ea);
            eaptr += extList[idx].len;
        }
        
        parseEA( extList[0].start.sector - partStartSector,
                 nBytesForEAs,
                 ea,
                 trustees);
        delete eafe;
        delete ea;
        delete [] bp;
        if (err)
            return;
    }
    else {
        if (fe->lenExtendedAttr) {
            ea = (BYTE *)fe->embeddedEAp();
            nBytesForEAs = fe->lenExtendedAttr;
            parseEA( addr.sector - partStartSector, nBytesForEAs, ea, trustees);
            if (err)
                return;
        } else {
#ifdef TRUSTEES
            trustees.clear();
#endif // TRUSTEES
        }
    }
}

//----------------------------------------------------------------------
// METHOD:  Archive::getICBExtents
// 
//    Get the extents associated with this ICB.
//
// ARGS:
//    FileEntry *fe,         IN    The ICB to find the extents for.
//    ExtentList *extList    OUT   The location of data for this ICB. For
//                                 directories, it will be just one extent.
//                                 A file can have a bunch of extents.
//
// RETURNS:
// 
// PRE-CONDS: 
// 
// POST-CONDS:
// 
// ERRORS:
//  error_
// 
// NOTES:
//    import
//----------------------------------------------------------------------
void
NSR::Archive::getICBExtents(FileEntry *fe, ExtentList *extList )
{
    Extent ext;
    LongAllocDesc *adt = fe->longADT();
    AllocExtentDesc *aed = NULL;
    Address nextAED;
    
    BYTE *bp = new BYTE[logicalBlockSize];
    if (!bp) {
        err = error_No_free_store;
        return;
    }
    Attributes attr;            // Just to get the file type
    attr.fromFileEntry(*fe);
    
    if ( attr.fileType() == file_Directory ) {
        ext.start.sid   = volumeTable[adt[0].partRefNo()].sid();
        ext.start.sector = adt[0].logBlockNo() + partStartSector;
        ext.len          = adt[0].extLength();
        extList->append( ext );
    } else {
        UINT32 noExtents = fe->numADTEnts();
        UINT32 i = 0;
        adt = fe->longADT();
        while( i<noExtents )
        {
            if ( adt[i].isContinued() ) {
                nextAED = icbTable->nextAddr();
                if ( (adt[i].logBlockNo() + partStartSector != nextAED.sector)
                   || (volumeTable[adt[i].partRefNo()].sid() != nextAED.sid)) {
                    err = error_Bad_ICB_table;
                    delete [] bp;
                    return;
                }

                icbTable->bumpPtr( 1 );
                if (err = icbTable->error()) {
                    delete [] bp;
                    return;
                }
                if(Trace::logAlgorithms())
                    fprintf(Trace::logFile(),"getICBExtents AED: at sid %3d sector %7d\n",
                        icbTable->currentAddr().sid, 
                        icbTable->currentAddr().sector);

                readBlock( icbTable->currentAddr().sid, 
                           icbTable->currentAddr().sector, bp );
                if (err) {
                    delete [] bp;
                    return;
                }
                delete aed;
                aed = new AllocExtentDesc( 1, ad_Long);
                if (!aed) {
                    err = error_No_free_store;
                    delete [] bp;
                    return;
                }
                aed->fromISO( bp, icbTable->currentAddr().sector - partStartSector );
                if (err = aed->error()) { 
                    delete aed; 
                    delete [] bp; 
                    return; 
                }
                noExtents = aed->numADTEnts();
                adt       = aed->longADT();
                i         = 0;
            } else {
                ext.start.sid   = volumeTable[adt[i].partRefNo()].sid();
                ext.start.sector = adt[i].logBlockNo() + partStartSector;
                ext.len          = adt[i].extLength();
                extList->append( ext );
                if (err = extList->error()) {
                    delete aed;
                    delete [] bp;
                    return;
                }
                i++;
            }
        }
    }
    delete aed; 
    delete [] bp; 
}

void
NSR::Archive::readSector( sid_t sid, UINT32 a, BYTE *bp )
{

    if (currentDevice != 0) {
        currentDevice->errorClear();
        currentDevice->read( sid, a, logicalBlockSize, bp );
        err = currentDevice->error();
    } else
        err = error_Invalid_function;

// ISSUE: ?? the device dev is not returning the correct error code ??
    if (err==error_Drive_read) {
        err = error_Drive_blank_check;
    }
}

void
NSR::Archive::readBlock( sid_t sid, UINT32 a, BYTE *bp )
{

    UINT16 nBlocks = logicalBlockSize/logicalSectorSize;

    for (unsigned i=0; i<nBlocks; i++)
        readSector( sid, a, bp + (S2B(i)));

}

void 
NSR::Archive::writeSector( sid_t sid, UINT32 a, BYTE *bp ) 
{

    if ( currentDevice != 0 ) {
        currentDevice->errorClear();
        currentDevice->write( sid, a, logicalSectorSize, bp );
        err = currentDevice->error();
    } else
        err = error_Invalid_function;
}

void 
NSR::Archive::writeBlock( sid_t sid, UINT32 a, BYTE *bp ) 
{

    UINT16 nBlocks = logicalBlockSize/logicalSectorSize;

    for (unsigned i=0; i<nBlocks; i++)
        writeSector( sid, a, bp + (S2B(i)));

}

//----------------------------------------------------------------------
// METHOD:  Archive::writeIndirectEntry
//
//    Write an indirect entry so the USE sequence can be continued.
//
// ARGS:
//    sid_t sid,                 IN    Surface ID.
//    UINT32 indirectEntryAddress, IN    Address for indirect entry.
//    Extent &e,                   IN    Where indirect entry points.
//    UINT32 &priorDirects,        IN    Prior number of indirect entries.
//    UINT16 &maxEntries,          IN    Max number of indirect entries.
//    UINT16 &strategyParm         IN    Strategy used in writing ie.
//
// RETURNS:
// 
// PRE-CONDS: 
// 
// POST-CONDS:
// 
// ERRORS:
//  error_
// 
// NOTES: export
//    This only is used for the USE sequence and only if partitionSeqAllowed
//    is set True.
//
//    The fsd sequence is continued via a different mechanism.
//----------------------------------------------------------------------
void 
NSR::Archive::writeIndirectEntry(sid_t sid, UINT32 indirectEntryAddress, 
                                 Extent &e, UINT32 &priorDirects, 
                                 UINT16 &maxEntries, UINT16 &strategyParm ) 
{
    IndirectEntry *ie = new IndirectEntry;
    if (!ie) {
        err = error_No_free_store;
        return;
    }
    BYTE *bp = new BYTE[logicalSectorSize];
    if (!bp) {
        err = error_No_free_store;
        delete ie;
        return;
    }
    // fill in the indirect entry
    ie->icbTag.icbType(icb_IndirectEntry);      // same for all sequences
    ie->icbTag.strategyType = 2;                // same for all sequences
    ie->icbTag.strategyParm = strategyParm;
    ie->icbTag.priNoDirectEntries = priorDirects;
    ie->icbTag.maxEntries = maxEntries;

    ie->indirectICB.extType    ( ext_Recorded );
    ie->indirectICB.partRefNo  ( volumeTable.vid(sid) );
    ie->indirectICB.logBlockNo ( e.start.sector - partStartSector );
    ie->indirectICB.extLength  ( e.len );
    
    // write the indirect entry
    ie->toISO(bp,indirectEntryAddress);
    if (err=ie->error()) {
        delete ie;
        delete [] bp;
        return;
    }
    writeSector( sid, indirectEntryAddress, bp );
    if (err) { 
        if(Trace::logAlgorithms()) {
            fprintf(Trace::logFile(),"can't write indirect entry icb at 0x%x\n",indirectEntryAddress);
        }
    }
    delete ie; 
    delete [] bp; 
}

//----------------------------------------------------------------------
// METHOD:  Archive::getSeqExtent
//      
//    Get an extent for use in extending a USE or FSD sequence.
//
// Use up the previously allocated space. The allocation had to occur
// before the areas for the icb's was allocated so that the check for
// available space was accurate.
// ISSUE: this should be an extent list ??
//
// ARGS:
//    sid_t sid,       IN    Surface ID.
//    UINT32 nSectors,   IN    Number of sectors.
//    Extent &newExtent  OUT   The extent.
//
// RETURNS:
// 
// PRE-CONDS: 
// 
// POST-CONDS:
// 
// ERRORS:
//  error_
// 
// NOTES:
//    used only for WORM and only if sequences can grow > fileDataStart
//----------------------------------------------------------------------
void 
NSR::Archive::getSeqExtent(sid_t sid, UINT32 nSectors, Extent &newExtent)
{
    if (nSectors < B2S(newSeqExtent.len)) {
        newExtent.start.sector = newSeqExtent.start.sector; 
        newExtent.len = S2B(nSectors);
        newSeqExtent.start.sector += nSectors;
        newSeqExtent.len -= B2S(nSectors);
    }
    else {
        err = error_No_free_space;
        return; 
    }
    err = error_None;
}


//----------------------------------------------------------------------
// METHOD:  Archive::checkRootVolumeSpace
//
//    Checks that there is enough volume space left for an export to be 
//    successful.
//
// ARGS:
//    UINT32 &neededSectors    OUT   Number of sectors in partition needed for
//                                   continuing USE, FSD sequences.
//
// RETURNS:
// 
// PRE-CONDS: 
// 
// POST-CONDS:
// 
// ERRORS:
//  error_
// 
// NOTES:
//    import
//----------------------------------------------------------------------
void
NSR::Archive::checkRootVolumeSpace(UINT32 &neededSectors)
{
    UINT32 nextBlankAddr;
    UINT32 lastRecordedAddr;
    UINT16 nVolsInSet;
    UINT16 nSectors;
    Extent extent;
    UINT32 nRead;
    UINT32 lastDescNum;
    Boolean needsIndirect;
    UINT32 rootFDS = volumeTable.rootVol().fileDataStart;
    sid_t  rootSid = volumeTable.rootVol().sid();
    
    neededSectors = 0;
    
    nVolsInSet = volumeTable.getCount();
    
    //-------------------------------------------------------------
    // check space for LVID. Size of LVID is a f(nVolsInSet). 
    //-------------------------------------------------------------
    LogicalVolumeIntegrityDesc *lvid = new LogicalVolumeIntegrityDesc(nVolsInSet,0);
    if (!lvid) {
        err = error_No_free_store;
        return;
    }
    // export: open and close, import: open and close
    nSectors = 4 * B2S(lvid->descLen());  
    delete lvid;

    extent.start.sid = rootSid;
    extent.start.sector = (logicalSectorSize == 512) ? LOC_LVID_512 : LOC_LVID_1024;
    extent.len = S2B(LOC_AVD1 - extent.start.sector);
    nRead = 0;
    lastDescNum = 0;
    nextBlankAddr = 0;
    lastRecordedAddr = 0;
    needsIndirect = False;
    getNextSeqAddress(rootSid,extent,nSectors,&nRead,&nextBlankAddr,
                      &lastDescNum,&lastRecordedAddr,&needsIndirect,
                      &bogusICBInfo);
    if (err) {
        if (err == error_Drive_blank_check && media == media_Worm) {
            err = error_None;
        }
    }
    if (needsIndirect) {
        //----------------------------------------------------------------
        // can't extend LVID past 255. It is confined to volume space.
        // Also note that the 1st 32k is disallowed.
        //----------------------------------------------------------------
        err = error_End_of_Sequence_space;
        return;
    }

    //------------------------------------------------------------------
    // check space for USE. USE is always one sector, but may need AED.
    //------------------------------------------------------------------
    extent.start.sector = LOC_USE;
    extent.len = S2B((rootFDS-partStartSector)>>1);
    nSectors = 1;
    nRead = 0;
    lastDescNum = 0;
    nextBlankAddr = 0;
    lastRecordedAddr = 0;
    needsIndirect = False;
    getNextSeqAddress(rootSid,extent,nSectors,&nRead,&nextBlankAddr,&lastDescNum,
                      &lastRecordedAddr,&needsIndirect,&bogusICBInfo);
    if (err) {
        if (err == error_Drive_blank_check && media == media_Worm) {
            err = error_None;
        }
    }
    if (needsIndirect) {
        if (partitionSeqAllowed) {
            // space for a new use extent
            neededSectors += incrementalSeqSectors;
        }
        else {
            err = error_End_of_Sequence_space; 
            return;
        }
    }

    //-------------------------------------------------------------------
    // Check space for use aed's - this comes from partition space.
    // For WORM, this must come from above fileDataStart.
    //-------------------------------------------------------------------
    UINT32 N = volumeTable.rootVol().freeSpace.getCount();
    if (N > extentsPerUSE) {
        if (partitionSeqAllowed) {
            UINT32 extsLeft = N - extentsPerUSE - 1;
            while ( extsLeft > 0 )
            {
                neededSectors++;
                if ( extsLeft > extentsPerAED) {
                    extsLeft -= (extentsPerAED - 1);  // last one pts to next aed
                }
                else {
                    extsLeft = 0;
                }
            }
        }
        else {
            fprintf(Trace::logFile(),"checkRootVolumeSpace: aed's for use won't be written\n");
        }
    }

    //-------------------------------------------------------------
    // check space for FSD. FSD is always one sector.
    //-------------------------------------------------------------
    nSectors = 1;
    nRead = 0;
    lastDescNum = 0;
    nextBlankAddr = 0;
    lastRecordedAddr = 0;
    needsIndirect = False;
    getNextSeqAddress(rootSid,
                      volumeTable.rootVol().fsdsExt,
                      nSectors,&nRead,&nextBlankAddr,&lastDescNum,
                      &lastRecordedAddr,&needsIndirect,&bogusICBInfo);
    if (err) {
        if (err == error_Drive_blank_check && media == media_Worm) {
            err = error_None;
        }
    }
    if (needsIndirect) {
        if (!partitionSeqAllowed) {err = error_End_of_Sequence_space; return;}
        neededSectors += incrementalSeqSectors;
    }
    err = error_None;
}

NSR::UINT32
NSR::Archive::writeExtents(sid_t sid, ICB_Table *icbTbl, UINT32 n, BYTE *bp)
{
    UINT32 nwritten = 0;

    for (int idx = 0; idx < n; idx++)
    {
        if(Trace::logAlgorithms())
            fprintf(Trace::logFile(),"write extent:%d\n",icbTbl->currentAddr().sector);
        writeSector(sid, icbTbl->currentAddr().sector, bp);
        bp += logicalBlockSize;
        icbTbl->bumpPtr(1);
        nwritten++;
    }
    return nwritten;
}

NSR::UINT32
NSR::Archive::readExtent(const Extent &e, BYTE *bp)
{
    UINT32 nread = 0;
    Boolean done = False;
    UINT32 addr1 = e.start.sector;
    UINT32 addr2 = addr1 + B2S(e.len);
    for ( UINT32 addr = addr1; addr < addr2 && !done; addr++)
    {
        readSector(e.start.sid, addr, bp);
        if ( err ) {
            done = True;
        }
        else {
            nread++;
        }
        bp += logicalBlockSize;
    }
    return nread;
}

//----------------------------------------------------------------------
// METHOD: Archive::copyFSDS  --DVM
//      
//  Copy FSDS from an old Root to the current root volume.
//
//  This works for WORM and MO. It is used solely by completeGrow(),
//  and propagates old FileSetDescriptors from the old root volume to
//  the new one.
//
// ARGS:
//  Volume & oldRootVol  IN   Volume to copy FSDS from.
//
// RETURNS: none.
// 
// PRE-CONDITIONS:
//  The volumeTable must hold a valid FSDS extent in both the volumes.
// 
// POST-CONDITIONS:
//  The FSDS is copied entirely. The current root Volume has its FSDS
//  extent updated to first reference the old FSDS, and then chain to
//  any new ones.
// 
// ERRORS:
//  error_
//  error_
//  error_
// 
// NOTES:
//  This replaces moveRoot().
//
//----------------------------------------------------------------------
void
NSR::Archive::copyFSDS(const Volume & prevRootVol)
{
    // Make constant references to previous root.
    const UINT16   prevRootSid = prevRootVol.sid();
    const Extent & prevRootFSD = prevRootVol.fsdsExt;

    // Make non-const references to new root.
    Volume & newRootVol = volumeTable.rootVol();
    UINT16   newRootSid = newRootVol.sid();
    Extent & newRootFSD = newRootVol.fsdsExt;

    // We'll need a File Set Descriptor.
    FileSetDesc *fsd = new FileSetDesc;
    if (!fsd) {
        err = error_No_free_store;
        return;
    }
    // We'll need a buffer
    BYTE *bp;

    // Dump the previous Root's FSDS Extent
    if(Trace::logAlgorithms()) {
        fprintf(Trace::logFile(),"copyFSDS(): prevRootFSD:\n");
        prevRootFSD.fprint(Trace::logFile());
    }

    if ( media == media_Worm ) {
        //-----------------------------------------------------------------
        // Must find how long the previous FSD is. It must be moved to the
        // new root. This will dictate the space requirements on new root
        // volume for the old FSDS.
        //----------------------------------------------------------------

        Extent fsdHistoryExt;   // Where new copies go.
        Extent lastFsdNextExt;  // For chaining algorithm

        UINT32 nFSDSectors      = 0;
        UINT32 nSectors         = 1;
        UINT32 nextBlankAddr    = 0;
        UINT32 lastRecordedAddr = 0;
        UINT32 lastDescNum      = 0;
        Boolean needsIndirect   = False;
        getNextSeqAddress(prevRootSid,prevRootFSD,nSectors,&nFSDSectors,
                          &nextBlankAddr,&lastDescNum,
                          &lastRecordedAddr,&needsIndirect,&bogusICBInfo);
        if (err) {
            if (err == error_Drive_blank_check && media == media_Worm) {
                err = error_None;
            }
        }
        if(Trace::logAlgorithms()) {
            fprintf(Trace::logFile(),"copyFSDS: fsd sequence length: %d\n",nFSDSectors);
            fprintf(Trace::logFile(),"copyFSDS: last fsd number: %d\n",lastDescNum);
            fprintf(Trace::logFile(),"copyFSDS: lastRecordedAddr:: %d\n",lastRecordedAddr);
        }

        //------------------------------------------------------------
        // Allocate space for previous fsd's. Also verify it is blank.
        // If there is not enough space, the whole export will fail.
        //------------------------------------------------------------
        Boolean verified = False;
        UINT32 nonBlankSector;
        while ( !verified ) 
        {
            // Allocate it.
            fsdHistoryExt = newRootVol.alloc( nFSDSectors, True );
            // Error if not enough room
            if (!(fsdHistoryExt.len)) {
                err = error_No_free_space;
                delete fsd;
                return;
            }
            // Verify that it is blank
            if (currentDevice->
                verifyBlank( newRootSid, 
                             fsdHistoryExt.start.sector, 
                             B2S(fsdHistoryExt.len), 
                             &nonBlankSector) == True) {
                verified = True;
            }
            else {
                fprintf(Trace::logFile(),"copyFSDS: WARNING: non-blank sector %d\n",
                        nonBlankSector);
            }
        }

        //------------------------------------------------------------
        // Move previous fsd's, in chunks for efficiency.
        // Each extent of old FSDS is read, and then copied to the new
        // root.
        //------------------------------------------------------------
        UINT32 nSectorsRead;
        UINT32 nWritten = 0;
        Extent prevfsde = prevRootFSD;
        Boolean done = False;
        while (!done) 
        {
            //-------------------------------------------
            // Allocate a buffer for the extent.
            //-------------------------------------------
            bp = new BYTE[prevfsde.len];
            if (!bp) {
                err = error_No_free_store;
                delete fsd;
                return;
            }
            //-------------------------------------------
            // Read in the extent.
            //-------------------------------------------
            nSectorsRead = readExtent(prevfsde, bp);
            if (err) {
                if ( err == error_Drive_blank_check ) {
                    done = True; 
                }
                else {
                    delete fsd;
                    delete [] bp;
                    return;
                }
            }
            if(Trace::logAlgorithms()) {
                fprintf(Trace::logFile(),"copyFSDS:read %d sectors of old sequence\n",
                        nSectorsRead);
            }

            //-------------------------------------------
            // write the extent (to the new volume).
            //-------------------------------------------
            for (int j = 0; j < nSectorsRead; j++) 
            {
                //-------------------------------------------
                // First, convert to a valid FSD
                //-------------------------------------------
                fsd->fromISO( bp + (j*logicalSectorSize), 
                              prevfsde.start.sector + j - partStartSector);
                if (err=fsd->error()) { 
                    delete fsd;
                    delete [] bp;
                    return;
                }
                if (j == (nSectorsRead - 1)) {
                    lastFsdNextExt.start.sector = (fsd->nextExtent.logBlockNo() +
                                                   partStartSector );
                    lastFsdNextExt.len = fsd->nextExtent.extLength();
                }
                // If this was last extent of FSDS, set to continue in
                //  normal FSDS location (held in newRootFSD).
                if (done && (j == (nSectorsRead - 1))) {
                    if(Trace::logAlgorithms()) { fprintf(Trace::logFile(),"setting fsd->nextExtent\n"); }
                    fsd->nextExtent.partRefNo(volumeTable.rootVid());
                    fsd->nextExtent.logBlockNo(newRootFSD.start.sector
                                               -partStartSector);
                    fsd->nextExtent.extLength( newRootFSD.len );
                } else {
                    // Otherwise, set nextExtent field to zeros,
                    // indicating that the sequence continues with the
                    // next sector (normal FSDS).
                    if(Trace::logAlgorithms()) { fprintf(Trace::logFile(),"fsd->nextExtent len = 0\n"); }
                    fsd->nextExtent.partRefNo(0);
                    fsd->nextExtent.logBlockNo(0);
                    fsd->nextExtent.extLength(0);
                }

                //-------------------------------------------
                // Translate FSD back to ISO in buffer.
                //-------------------------------------------
                fsd->toISO( bp + (j*logicalSectorSize),
                            fsdHistoryExt.start.sector + nWritten 
                            - partStartSector );
                writeSector(newRootSid,
                            fsdHistoryExt.start.sector + nWritten, 
                            bp + (j*logicalSectorSize));
                if (err) {
                    delete fsd;
                    delete [] bp;
                    return;
                }
                nWritten++;
                if(Trace::logAlgorithms()) { 
                    fprintf(Trace::logFile(),"wrote new fsd at sid:%d\n",newRootSid); 
                    fprintf(Trace::logFile(),"               sector:%d\n",fsdHistoryExt.start.sector+j); 
                }
            }

            //-----------------------------------------------------------
            // Extent was full - no blank sectors, so look at nextExtent
            // field of last FSD for where to continue.
            //-----------------------------------------------------------
            if (!done) {
                //-----------------------------------------------------------
                // look for possible next extent identified in last fsd read
                //-----------------------------------------------------------
                if(Trace::logAlgorithms()) { 
                    fprintf(Trace::logFile(),"lastFsdNextExt.len:%d\n",
                            lastFsdNextExt.len);
                }
                prevfsde = lastFsdNextExt;
                //-----------------------------------------------------------
                // If there was no continuation, then it was the last FSD
                // that exists. Therefore, copy is complete. NOTE that
                // nextExtent field of FSD was already filled in inside
                // the loop.
                //-----------------------------------------------------------
                if (prevfsde.len == 0) {
                    done = True;
                }
            }
        }
        // Update information in root volume.
        newRootFSD = fsdHistoryExt;

        // Delete buffer
        delete [] bp;
        
    } else {                    // media_Rewritable

        // Simply need to copy the single FSD from previous root to
        // new root.

        // Error if a longer FSDS than a single sector
        if ( (prevRootFSD.len != logicalBlockSize) ||
             (newRootFSD.len  != logicalBlockSize) ) {
            err = error_Wrong_FSDS_extent_length;
            delete fsd;
            return;
        }
        
        // Allocate a buffer for the extent.
        bp = new BYTE[prevRootFSD.len];
        if (!bp) {
            err = error_No_free_store;
            delete fsd;
            return;
        }

        // Read in the extent.
        readExtent(prevRootFSD, bp);
        if (err) {
            delete fsd;
            delete [] bp;
            return;
        }

        // Dump out where read.
        if(Trace::logAlgorithms()) {
            fprintf(Trace::logFile(),"copyFSDS: read old FSDS at ");
            prevRootFSD.fprint(Trace::logFile());
        }

        // Translate the FSD to match new address, etc.
        fsd->fromISO( bp, (prevRootFSD.start.sector
                           - partStartSector));
        if (err=fsd->error()) { 
            delete fsd;
            delete [] bp;
            return;
        }
        
        // Set nextExtent field to zeros, indicating that the sequence
        // continues with the next sector (normal FSDS).
        // These should already be set, but...
        fsd->nextExtent.partRefNo(0);
        fsd->nextExtent.logBlockNo(0);
        fsd->nextExtent.extLength(0);

        // Translate FSD back to ISO in buffer.
        fsd->toISO( bp, (newRootFSD.start.sector 
                         - partStartSector ));
        // Write it out.
        writeSector(newRootSid,
                    newRootFSD.start.sector, 
                    bp);
        if (err) {
            delete fsd;
            delete [] bp;
            return;
        }
        // Dump out where written
        if(Trace::logAlgorithms()) { 
            fprintf(Trace::logFile(),"copyFSDS(): wrote new fsd at: ");
            newRootFSD.fprint(Trace::logFile());
        }
#ifdef _FSDS_TD_
        // Write a single TD to terminate sequence
        writeTD( newRootSid,
                 newRootVol.fsdsExt.start.sector - partStartSector +1,
                 newRootVol.fsdsExt.start.sector +1 );
        // No check for errors, because this is the last action
#endif // _FSDS_TD_
        
        // newRootFSD remains unchanged.

        // Delete buffer
        delete [] bp;
    }
    // delete FSD
    delete fsd;
}

//----------------------------------------------------------------------
// METHOD:  Archive::readBlind
//
//  This method reads a sector at the supplied address, and then if
//  the sector contains a valid descriptor, reallocates the supplied
//  buffer and then reads in the entire descriptor. The
//  passed-in-by-reference SearchDescriptor is left reflecting the
//  state of what it found on the disk.
//
// ARGS:
//    sid_t sid,            IN      Surface ID
//    UINT32 addr,          IN      Sector
//    SearchDescriptor & sd IN      Search Descriptor
//    BYTE **bpp            IN/OUT  Pointer to data
//
// RETURNS:
// 
// PRE-CONDS: 
// 
// POST-CONDS:
// 
// ERRORS:
//  error_
// 
// NOTES:
//    import
//----------------------------------------------------------------------
void
NSR::Archive::readBlind(sid_t sid,
                        UINT32 addr,
                        SearchDescriptor & sd, 
                        BYTE **bpp)
{
    BYTE *bp = *bpp;
    UINT32 nSectors;
    
    readBlock( sid, addr, bp );
    if (err) {
        return;
    }
    sd.fromISOMyopic(bp);
    if (err = sd.error()) {
        // This returns with type == Unknown, and error set appropriately.
        return;
    }
    // All is OK, so go ahead and trust descLen().
    nSectors = (sd.descLen() + logicalSectorSize - 1)/logicalSectorSize;

    if ( nSectors > 1 ) {
        delete [] bp;
        bp = new BYTE[logicalSectorSize * nSectors];
        if (!bp) {
            err = error_No_free_store;
            return;
        }
        for (int j=0; j<nSectors; j++) 
        {
            readSector( sid, addr+j, bp + (j*logicalSectorSize));
            if (err) {
                return;
            }
        }
        sd.fromISOMyopic( bp );
        err = sd.error();
    }
}

//----------------------------------------------------------------------
//  PROCEDURE NAME:
//   lastInExtent()                     --AP
//
//  DESCRIPTION:
//   Find the start address of the last recorded ICB and the address
//   of the next unrecorded sector. This is useful in re-export of
//   WORM volumes when it is necessary to rewrite the USE, FSD, LVID.
//   
//   sid_t sid                  volume identifier
//   Extent &extent             extent to search
//   UINT32 *nRead              number of non-blank sectors read
//   UINT32 *lastRecordedAddr   put last recorded icb here
//                              if none found, this is 0. 
//   UINT32 *nextBlankAddr      put next blank address here
//                              if none found, this is 0. 
//
// ----------------------------------------------------------------------
void
NSR::Archive::lastInExtent(const sid_t sid,
                           const Extent &extent, 
                           UINT32 *nRead,
                           UINT32 *lastRecordedAddr, 
                           UINT32 *nextBlankAddr)
{
    SearchDescriptor sd; 
    UINT32 nSectors;

    // Hold extent information in two local variables.
    UINT32 address = extent.start.sector;
    UINT32 max_address = address + B2S(extent.len) - 1;

    // Holds the first descriptor's type
    TagType firstDescType;

    // Allocate a single-sector buffer.
    BYTE *bp = new BYTE[logicalSectorSize];
    if (!bp) {
        err = error_No_free_store;
        return;
    }

    // Loop through entire range of supplied extent.
    while ( address <= max_address )
    {
        readBlind( sid, address, sd, &bp );
        if ( err ) {
            if ( err == error_Drive_blank_check ) {
                *nextBlankAddr = address;
                err = error_None;
            }
            delete [] bp;
            return;
        }
        // Grab type of first descriptor read.
        if ( address == extent.start.sector ) {
            firstDescType = sd.type();
        }
        // Error if this descriptor type does not match the first.
        if (firstDescType != sd.type()) {
            err = error_Bad_tag_descriptor_type;
            // Special handling for USE/IE combination
            if ( (sd.type() == tag_IE) &&
                 (firstDescType == tag_USE ||
                  firstDescType == tag_IE) ) {
                // Clear out error.
                err = error_None;
            }
            // If error not cleared out, bail.
            if (err) {
                delete [] bp;
                return;
            }
            
        }
        // Treat USE/IE combination specially.
        if ( sd.type() == tag_IE ) {
            if (firstDescType != tag_USE &&
                firstDescType != sd.type()) {
                err = error_Bad_tag_descriptor_type;
                delete [] bp;
                return;
            }
        } else if (firstDescType != sd.type()) {
            err = error_Bad_tag_descriptor_type;
            delete [] bp;
            return;
        }
        *lastRecordedAddr = address;
        nSectors = (sd.descLen() + logicalSectorSize - 1)/logicalSectorSize;
        address += nSectors;
        *nRead += nSectors;
    }
    delete [] bp;
    return;
}

// ----------------------------------------------------------------------
//
//  PROCEDURE NAME:
//   getNextSeqAddress
//
//  DESCRIPTION:
//   Get next candidate sector to be used for extending an
//   LVID,USE,FSD sequence.  Follow the "nextExtent" field in FSD or
//   LVID if necessary. Does NOT create an indirect entry if runs out
//   of space.
//   
//   sid_t sid                  Surface Identifier
//   Extent &extent             the extent to start searching in
//   UINT32 nSectors            number of sectors calling f() needs to write
//   UINT32 * nRead             number of sectors read
//   UINT32 * nextBlankAddr     next blank sector
//   UINT32 * lastDescNum       last (FSD) descriptor number 
//   UINT32 * lastRecordedAddr  last recorded address
//   Boolean * needsIndirect    True if will be last item in extent
//   ICBInfo * lastICBInfo      caller may need info in last ICB tag
// 
//  RETURNED VALUES:
//
//  AUTHOR: Andy Purcell
// ----------------------------------------------------------------------
void
NSR::Archive::getNextSeqAddress(const sid_t sid,
                                const Extent &extent, 
                                UINT32 nSectors,
                                UINT32 *nRead, 
                                UINT32 *nextBlankAddr,
                                UINT32 *lastDescNum, 
                                UINT32 *lastRecordedAddr, 
                                Boolean *needsIndirect,
                                ICBInfo *lastICBInfo)
{
    UINT32 max_address;
    BYTE *bp = new BYTE[logicalSectorSize];
    if (!bp) {
        err = error_No_free_store;
        return;
    }
    Extent next;
    
    max_address = extent.start.sector + B2S(extent.len) - 1;
    *needsIndirect = False;
    UINT32 lastRecorded = 0;
    UINT32 nSectorsRead = 0;
    lastInExtent(sid,
                 extent,
                 &nSectorsRead,
                 &lastRecorded,
                 nextBlankAddr);
    if (err) {
        delete [] bp;
        return;
    }
    if (lastRecorded) {
        *lastRecordedAddr = lastRecorded;
        *nRead += nSectorsRead;
        //-------------------------------------------------------
        // check if last recorded has a "nextExtent" identified or
        // is an indirect entry. If so, follow that extent.
        //-------------------------------------------------------
        SearchDescriptor sd;
        
        readBlind( sid, *lastRecordedAddr, sd, &bp );   
        if (err) {
            delete [] bp;
            return;
        }
            
        switch (sd.type()) {
        case tag_IE: {
            IndirectEntry *ie = new IndirectEntry;
            if (!ie) {
                err = error_No_free_store;
                delete [] bp;
                return;
            }
            ie->fromISO( bp, *lastRecordedAddr - partStartSector );
            if (err=ie->error()) {
                delete ie;
                delete [] bp;
                return;
            }
            next.start.sector = ie->indirectICB.logBlockNo() + partStartSector;
            next.len = ie->indirectICB.extLength();

            // Hold icbTag info.
            lastICBInfo->priNoDirectEntries = ie->icbTag.priNoDirectEntries;
            lastICBInfo->strategyParm = ie->icbTag.strategyParm;
            lastICBInfo->maxEntries = ie->icbTag.maxEntries;

            getNextSeqAddress(sid,next,nSectors,nRead,nextBlankAddr,
                              lastDescNum,lastRecordedAddr,needsIndirect,
                              lastICBInfo);
            delete ie;
            break;
        }
        case tag_LVID: {
            LogicalVolumeIntegrityDesc *lvid = new LogicalVolumeIntegrityDesc(1,1);
            if (!lvid) {
                err = error_No_free_store;
                delete [] bp;
                return;
            }
            lvid->fromISO( bp, *lastRecordedAddr );
            if (err=lvid->error()) {
                delete lvid;
                delete [] bp;
                return;
            }
            if (lvid->nextIntegrityExtent.length) {
                next.len = lvid->nextIntegrityExtent.length;
                next.start.sector = lvid->nextIntegrityExtent.location;
                delete lvid;
                delete [] bp;   // !!! This will get deleted TWICE
                                // (here and at end of method)
                bp = NULL;      // delete at null pointer OK.
                getNextSeqAddress(sid,next,nSectors,nRead,nextBlankAddr,
                                  lastDescNum,lastRecordedAddr,needsIndirect,
                                  lastICBInfo);
            } else {
                delete lvid;
            }
            break;
        } 
        case tag_FSD: {
            FileSetDesc *fsd = new FileSetDesc;
            if (!fsd) {
                err = error_No_free_store;
                delete [] bp;
                return;
            }
            fsd->fromISO( bp, *lastRecordedAddr - partStartSector );
            if (err=fsd->error()) {
                delete fsd;
                delete [] bp;
                return;
            }
            *lastDescNum = fsd->fileSetDescNo;
            if (fsd->nextExtent.extLength()) {
                next.len = fsd->nextExtent.extLength();
                next.start.sector = fsd->nextExtent.logBlockNo() + LOC_PART_START;
                delete fsd;
                delete [] bp;   // !!! This will get deleted TWICE
                                // (here and at end of method)
                bp = NULL;      // delete at null pointer OK.
                getNextSeqAddress(sid,next,nSectors,nRead,nextBlankAddr,
                                  lastDescNum,lastRecordedAddr,needsIndirect,
                                  lastICBInfo);
            } else {
                delete fsd;
            }
            break;
        }
        case tag_USE: {
            /* the use sequence ends with an IE */
            break;
        }
        default:
            err = error_Bad_tag_descriptor_type;
            delete [] bp;
            return;
        }
    }
    //--------------------------------------------------------
    // If we get here, we know that there has been something
    // recorded in the extent and that it is not an indirect.
    // Check to see if there is enough space.
    //--------------------------------------------------------
    if (*nextBlankAddr) {
        if (*nextBlankAddr + nSectors - 1 > max_address ) {
            err = error_End_of_Sequence_space;
        }
        //--------------------------------------------------------
        // check if this needs to be indirect - assumes next one written
        // will be of same size as this one.
        //--------------------------------------------------------
        if (*nextBlankAddr + nSectors + nSectors - 1 > max_address ) {
            *needsIndirect = True;
        }
    }
    delete [] bp;
    return;
}

//----------------------------------------------------------------------
// METHOD:  Archive::writeSpaceEntry
//      
//  Write out an unallocated space entry for the extentlist provided. 
//
// ARGS:
//  sid IN      Surface id.
//  spcType     IN      ext_Recorded,...
//  space       IN      the extent list describing the space
//  e           IN      the extent for the use,aed's.
//
// RETURNS: none.
// 
// PRE-CONDS:
// 
// POST-CONDS: may set error state in Archive object.
// 
// ERRORS:
//  error_
// 
// NOTES:
// !!! USE/AEDs don't look like they are properly written
//     We are supplying some space here which is to hold both USEs and
//     AEDs. AEDs cannot be mixed into ICBs alongside USEs, so this
//     looks very dubious.
//----------------------------------------------------------------------
void
NSR::Archive::writeSpaceEntry( sid_t sid,
                               ExtType spcType,
                               const ExtentList &space, 
                               const Extent &e)
{
    UINT32      addr = e.start.sector;
    UINT32      nextBlankAddr;
    BYTE        *bp;
    int         j;
    // int      volDescSeqNo = 1; !!! Apparently not needed
    Extent      extent;
    UINT32      lastDescNum = 0;
    UINT32      lastRecordedAddr = 0;
    Boolean     needsIndirect = False;
    UINT32      nRead = 0;
    ICBInfo     lastUSEICBInfo;
    UINT16      vid(volumeTable.vid(sid));

    bp = new BYTE[logicalSectorSize];
    if (!bp) {
        err = error_No_free_store;
        return;
    }
    UINT32 N   = space.getCount();
    UINT32 ovf = 0;
    if ( N > extentsPerUSE ) {
        UINT32 extsLeft = N - extentsPerUSE - 1;
        while ( extsLeft > 0 )
        {
            ovf++;
            if ( extsLeft > extentsPerAED) {
                extsLeft -= (extentsPerAED - 1);  // last one pts to next aed
            }
            else {
                extsLeft = 0;
            }
        }
    }

    UINT32 index = 0;
    AllocExtentDesc *aed;
    UnallocatedSpaceEntry *use;

    if (ovf) {
        if ( e.start.sector + ovf > e.start.sector + B2S(e.len) ) {
            fprintf(Trace::logFile(),"writeSpaceEntry: no room to write all extents: truncating\n");
            ovf = B2S(e.len) - 1;
            N = extentsPerUSE + (ovf * extentsPerAED);
        }
    }
    
    UINT32 n;
    if ( ovf > 0 )  {
        n = extentsPerUSE - 1;
    }
    else {
        n = N;
    }
    use = new UnallocatedSpaceEntry(n + (ovf > 0));
    if (!use) {
        err = error_No_free_store;
        delete [] bp;
        return;
    }
    N -= n;

    // Initialize lastUSEICBInfo for getNextSeqAddress()
    lastUSEICBInfo.priNoDirectEntries = 0;
    lastUSEICBInfo.maxEntries = (B2S(e.len))/(ovf+1);
    lastUSEICBInfo.strategyParm = (B2S(e.len))/(ovf+1) - 1;

    if (media == media_Worm) {
        extent.start.sector = e.start.sector;
        extent.len = e.len;
        nRead = 0;
        lastDescNum = 0;
        lastRecordedAddr = 0;
        nextBlankAddr = 0;
        needsIndirect = False;
        getNextSeqAddress(sid,extent,1,&nRead,&nextBlankAddr,&lastDescNum,
                          &lastRecordedAddr,&needsIndirect,&lastUSEICBInfo);
        if(Trace::logAlgorithms()) {
            fprintf(Trace::logFile(),
                    "writeSpaceEntry(): nextBlankAddr = %d;lastRecordedAddr = %d.\n",
                    nextBlankAddr,lastRecordedAddr);
        }
        if (err) {
            if (err == error_Drive_blank_check && media == media_Worm) {
                err = error_None;
            }
        }
        if (needsIndirect) {
            if(Trace::logAlgorithms()) {
                fprintf(Trace::logFile(),"writeSpaceEntry(): needs Indirect\n");
            }
            if (partitionSeqAllowed) {
                getSeqExtent(sid,incrementalSeqSectors,extent); 
                if (err) {
                    delete [] bp;
                    delete use;
                    return;
                }
                UINT16 strategyParm = incrementalSeqSectors - 1;
                if(Trace::logAlgorithms()) {
                    fprintf(Trace::logFile(),
                            "writeSpaceEntry(): writing Indirect at %d.\n",
                            nextBlankAddr);
                }
                writeIndirectEntry(sid, nextBlankAddr,extent,lastDescNum,
                                   incrementalSeqSectors,strategyParm);
                if (err) {
                    delete [] bp;
                    delete use;
                    return;
                }
                nextBlankAddr = extent.start.sector;
            }
            else {
                if (vid == volumeTable.rootVid()) {
                    // need new root volume
                    err = error_End_of_Sequence_space;
                    delete [] bp;
                    delete use;
                    return;
                }
                else { 
                    // non-root volume - write a use indicating no space left.
                    N = 0;
                    n = 0;
                    ovf = 0;
                }
            }
        }
        addr = nextBlankAddr;
    }
    use->icbTag.icbType(icb_UnallocatedSpaceEntry);   // never changes
    // !!! The following logic may not be correct. It ensures that MO
    // conforms, but has not been checked for WORM conformance.
    // - DVM 4/30/96
    if (media == media_Worm) {
        if (use->icbTag.priNoDirectEntries) {
            use->icbTag.priNoDirectEntries =
                lastUSEICBInfo.priNoDirectEntries + 1;
        }
        use->icbTag.strategyType   = 2;
        use->icbTag.maxEntries     = lastUSEICBInfo.maxEntries;
        use->icbTag.strategyParm   = lastUSEICBInfo.strategyParm;
        use->icbTag.parentICB.logicalBlockNo = e.start.sector - partStartSector;
        use->icbTag.parentICB.partitionRefNo = vid;
    } else {
        use->icbTag.priNoDirectEntries = 0;
        use->icbTag.strategyType       = 4;
        use->icbTag.maxEntries         = 1;
        use->icbTag.strategyParm       = 0;
        use->icbTag.parentICB.logicalBlockNo = 0;
        use->icbTag.parentICB.partitionRefNo = 0;
    }

    ShortAllocDesc *adt = use->shortADT();
    
    for ( j=0; j<n; j++ ) {
        adt[j].extType    ( spcType );
        adt[j].logBlockNo ( space[index].start.sector - partStartSector );
        adt[j].extLength  ( space[index].len );
        index++;
    }

    // if there is more, load a continued short ad
    if ( ovf > 0 ) {
        if (media != media_Worm) {
            adt[n].extType    ( ext_Recorded );
            adt[n].logBlockNo ( addr + 1 - partStartSector );
            adt[n].extLength  ( logicalBlockSize );
            adt[n].markContinued();
        }
        else {
            if (partitionSeqAllowed) {
                // get space for aed's
                getSeqExtent(sid,ovf,extent); 
                if (err) {
                    delete use;
                    delete [] bp;
                    return;
                }
                addr = extent.start.sector;
            }
            else {
                fprintf(Trace::logFile(),"writeSpace: aed's can't be written\n");
                adt[index].extType    ( spcType );
                adt[index].logBlockNo ( space[index].start.sector - partStartSector );
                adt[index].extLength  ( space[index].len );
                ovf = 0;
            }
        }
    }

    // write the unallocated space entry
    memset( bp, 0, logicalBlockSize );
    use->toISO( bp, addr - partStartSector );
    if (err = use->error()) {
        delete use;
        delete [] bp;
        return;
    }
    if(Trace::logAlgorithms()) {
        fprintf(Trace::logFile(),
                "writeSpaceEntry(): writing USE to sid %d sector %d\n",
                sid,addr);
    }
    writeBlock( sid, addr, bp );
    delete use;
    if (err) {
        delete [] bp;
        return;
    }
    // write required AEDs to hold remaining extents
    if (media != media_Worm) {
        addr++;
    }
    else {
        addr = extent.start.sector;
    }
    for ( UINT32 i=0; i<ovf; i++, addr++ ) 
    {
        if ( N > extentsPerAED ) 
            n = extentsPerAED - 1;
        else
            n = N;

        N -= n;

        aed = new AllocExtentDesc( (n + ( ovf != (i+1) )), ad_Short );
        if (!aed) {
            err = error_No_free_store;
            delete [] bp;
            return;
        }
        memset( bp, 0, logicalBlockSize );

        // load short ads in allocation extent desc block
        adt = aed->shortADT();
        for ( j=0; j<n; j++ ) {
            adt[j].extType    ( spcType );
            adt[j].logBlockNo ( space[index].start.sector - partStartSector );
            adt[j].extLength  ( space[index].len );
            index++;
        }

        // if there are more, load a continued short ad
        if ( ovf != (i+1) ) {
            adt[n].extType    ( ext_Recorded );
            adt[n].logBlockNo ( addr + 1 - partStartSector );
            adt[n].extLength  ( logicalBlockSize );
            adt[n].markContinued();
        }

        aed->toISO( bp, addr - partStartSector );
        if (err = aed->error()) {
            delete aed;
            delete [] bp;
            return;
        }
        delete aed;

        // Dump it out if tracing
        if(Trace::logAlgorithms()) {
            fprintf(Trace::logFile(),
                    "writeSpaceEntry(): writing AED to sid %d sector %d\n",
                    sid,addr);
        }

        // write the AED block
        writeBlock( sid, addr, bp );

        if (err) {
            delete [] bp;
            return;
        }
    }
    delete [] bp;
    return;
}

//----------------------------------------------------------------------
// METHOD: Archive::searchForDesc  --DVM
//      
//  Searches a single VDS for a specific descriptor type.
//
//  If the descriptor is found, the vdsNum is returned, along with a
//  modified extent. The address of the extent where the descriptor
//  was found.
//
// ARGS:
//  TagType descrType IN      Descriptor to search for
//  Extent  seqExtent IN/OUT  Location to be searched
//  UINT32  vdsNum    OUT     Volume Descriptor Sequence Number
//
// RETURNS:
//  True if found; False if not.
// 
// PRE-CONDITIONS: open archive object.
// 
// POST-CONDITIONS:
// 
// ERRORS:
//  error_No_free_store
//  misc device or fromISO errors
// 
// NOTES:
//  Does not currently follow VDS continuations. Could be added later.
//
//----------------------------------------------------------------------
NSR::Boolean
NSR::Archive::searchForDesc( TagType searchDescType,
                             Extent & seqExtent,
                             UINT32 & vdsNum ) 
{
    // Local variables
    SearchDescriptor sd;

    // Allocate a single sector buffer
    UINT32 bufLen = logicalSectorSize;// hold this to enable resizing.
    BYTE *bp = new BYTE[bufLen];
    if (!bp) {
        err = error_No_free_store;
        return False;
    }

    // Holders for info about the descriptor inside of the loop.
    TagType thisDescType  = tag_Unknown;
    UINT32  thisDescSeqNo = 0;
    UINT16  thisDescSectors = 1;// Holds numSectors in this desc

    // Booleans for exit conditions
    Boolean blankWORM(False);   // blank check on WORM media
    Boolean thisIsTD(False);    // this == TD
    Boolean thisMatches(False); // this == search
    Boolean foundMatch(False);  // at least one match
    
    // State variable for iterating is an Extent.
    // It shrinks on the front end by the appropriate number of
    // sectors at the end of the loop.
    Extent remainingExtent = seqExtent;

    // Loop until exhaust remainingExtent
    while (remainingExtent.len >= logicalSectorSize) {
        
        // Read the first sector in remainingExtent
        readSector( remainingExtent.start.sid,
                    remainingExtent.start.sector,
                    bp );

        // Don't count blank on WORM as error.
        blankWORM = ((err == error_Drive_blank_check) &&
                     (media == media_Worm)) ? True : False;
        if ( blankWORM ) {
            err = error_None;
        }
        // Exit on error
        if (err) {
            delete [] bp;
            return False;
        }

        // Treat blank WORM the same as TD.
        if ( blankWORM ) {
            thisDescType    = tag_TD;
            thisDescSectors = 1;
        } else {
            sd.fromISO( bp,remainingExtent.start.sector );
            if (err=sd.error()) {
                delete [] bp;
                return False;
            }
            thisDescType    = sd.type();
            thisDescSectors = B2S(sd.descLen());
        }

        // Grow buffer if needed and read in entire descriptor if
        // needed. This is solely for the LVD, but should work for all.
        if ( S2B(thisDescSectors) > bufLen ) {
            bufLen = S2B(thisDescSectors);
            delete [] bp;
            bp = new BYTE[bufLen];
            if (!bp) {
                err = error_No_free_store;
                return False;
            }
            // Read in entire descriptor.
            for (int j=0; j < thisDescSectors; j++) {
                readSector( remainingExtent.start.sid,
                            remainingExtent.start.sector+j,
                            bp + (j*logicalSectorSize));
                if (err) {
                    delete [] bp;
                    return False;
                }
            }
            // No need to do another sd.fromISO(). Better to have
            // error in following switch statement.
        }
        
        // ONLY purpose of switch is to grab the volDescSeqNo from the
        // descriptor. It must also maintain descSectors
        switch (thisDescType) {
        case tag_PVD:
        {
            // Allocate PVD
            PrimaryVolumeDesc *pvd = new PrimaryVolumeDesc;
            if (!pvd) { 
                err = error_No_free_store; 
                delete [] bp;
                return False; 
            }
            // Extract PVD from buffer
            pvd->fromISO( bp, remainingExtent.start.sector );
            if ( err = pvd->error() ) {
                delete pvd;
                delete [] bp;
                return False;
            }
            // Error if not HP implementation
            if ( (pvd->implementID.id() != NSR::ImplRegID::HPID ) &&
                 (pvd->implementID.id() != NSR::ImplRegID::HPEarlyID )) {
                err = error_UDF_volume_set_not_supported;
                delete pvd;
                delete [] bp;
                return False;
            }
            // Grab volDescSeqNo.
            thisDescSeqNo = pvd->volDescSeqNo;

            // Destroy PVD
            delete pvd;
            break;
        }
        case tag_IUVD :
        {
            // Allocate IUVD
            
            OSTAVolumeDesc *iuvd = new OSTAVolumeDesc;
            if (!iuvd) { 
                err = error_No_free_store; 
                delete [] bp;
                return False; 
            }
            // Extract IUVD from buffer
            iuvd->fromISO( bp, remainingExtent.start.sector );
            if (err=iuvd->error()) { 
                delete iuvd; 
                delete [] bp; 
                return False; 
            }
            // Error if not HP implementation
            if ( (iuvd->implementID.id() != NSR::ImplRegID::HPID ) &&
                 (iuvd->implementID.id() != NSR::ImplRegID::HPEarlyID ) ) {
                err = error_UDF_volume_set_not_supported;
                delete iuvd;
                delete [] bp;
                return False;
            }
            // Grab volDescSeqNo.
            thisDescSeqNo = iuvd->volDescSeqNo;

            // Destroy IUVD
            delete iuvd;
            break;
        }
        case tag_USD:
        {
            // Allocate USD
            UnallocatedSpaceDesc *usd = new UnallocatedSpaceDesc;
            if (!usd) { 
                err = error_No_free_store; 
                delete [] bp;
                return False; 
            }
            // Extract USD from buffer
            usd->fromISO( bp, remainingExtent.start.sector );
            if (err=usd->error()) { 
                delete usd; 
                delete [] bp; 
                return False; 
            }
            // No implementation ID field.
            
            // Grab volDescSeqNo.
            thisDescSeqNo = usd->volDescSeqNo;

            // Destroy USD
            delete usd;
            break;
        }
        case tag_PD:
        {
            // Allocate PD
            PartitionDesc *pd = new PartitionDesc;
            if (!pd) { 
                err = error_No_free_store; 
                delete [] bp;
                return False; 
            }
            // Extract PD from buffer
            pd->fromISO( bp, remainingExtent.start.sector );
            if (err=pd->error()) { 
                delete pd; 
                delete [] bp; 
                return False; 
            }
            // Error if not HP implementation
            if ( (pd->implementID.id() != NSR::ImplRegID::HPID ) &&
                 (pd->implementID.id() != NSR::ImplRegID::HPEarlyID ) ) {
                err = error_UDF_volume_set_not_supported;
                delete pd;
                delete [] bp;
                return False;
            }
            // Grab volDescSeqNo.
            thisDescSeqNo = pd->volDescSeqNo;

            // Destroy PD
            delete pd;
            break;
        }
        case tag_LVD:
        {
            // Allocate LVD
            LogicalVolumeDesc *lvd = new LogicalVolumeDesc;
            if (!lvd) { 
                err = error_No_free_store; 
                delete [] bp;
                return False; 
            }
            // Extract LVD from buffer
            lvd->fromISO( bp, remainingExtent.start.sector );
            if (err=lvd->error()) { 
                delete lvd; 
                delete [] bp; 
                return False; 
            }
            // Error if not UDF or HP near-UDF
            if ( ! lvd->domainID.isUDF() &&
                 ! lvd->domainID.isEarlyHP() ) {
                err = error_Non_UDF_volume_set;
                delete lvd;
                delete [] bp;
                return False;
            }
            // Error if not HP implementation
            if ( (lvd->implementID.id() != NSR::ImplRegID::HPID ) &&
                 (lvd->implementID.id() != NSR::ImplRegID::HPEarlyID ) ) {
                err = error_UDF_volume_set_not_supported;
                delete lvd;
                delete [] bp;
                return False;
            }
            // Grab volDescSeqNo.
            thisDescSeqNo = lvd->volDescSeqNo;

            // Destroy LVD
            delete lvd;
            break;
        }
        case tag_TD:
            thisDescSeqNo ++;
            break;
            
        default:
            // Not one of ours
            err = error_UDF_volume_set_not_supported;
            delete [] bp;
            return False;
        }

        // Note that both may be True
        thisIsTD    = (thisDescType == tag_TD) ? True : False ;
        thisMatches = (thisDescType == searchDescType) ? True : False ;

        // Set foundMatch as soon as a match is found.
        if (thisMatches) foundMatch = True;

        // Modify Extent, set vdsNum if match OR if no match, but at end.
        if ( thisMatches || (thisIsTD && !foundMatch)) {
            seqExtent = remainingExtent;
            vdsNum    = thisDescSeqNo;
        }

        // Return after reaching a TD
        if ( thisIsTD ) {
            delete [] bp;
            return foundMatch;
        }
        
        // Increment counter (shorten extent)
        remainingExtent.start.sector += thisDescSectors;
        remainingExtent.len          -= S2B(thisDescSectors);
    }

    // If we exited the loop, we are in error.
    err = error_End_of_Sequence_space;
    delete [] bp;
    return False;
}

//----------------------------------------------------------------------
// METHOD: Archive::searchVolDescSeq  --DVM
//      
//  Searches for a specific descriptor in VDS
//  
//  This uses the AVDP to find the VDS and then searches it for a
//  specific descriptor. It returns True if the descriptor is
//  found. Whether found or not, it fills in the supplied VDSInfo
//  struct with the appropriate information. NOTE that it reads both
//  the main VDS and the reserve VDS.
//
//  Also NOTE that it returns information about the PREVAILING
//  descriptor, and not simply the first one it finds.
//
//  It terminates on a TD or a blank sector, in which case, the VDS
//  info struct it returns refers to the TD or the blank sector. WORM
//  users may search for a TD, which matches a blank sector as well as
//  a valid TD.
//
//  Currently, it does NOT support extensions to the VDS via the
//  VDP. This may be added in the future.
//
// args:
//  sid_t   sid       IN   Surface Identifier
//  TagType descrType IN   Descriptor to search for
//  VDSInfo vdsi      OUT  VDSInfo struct at location
//
// RETURNS:
//  True if found. False if not.
// 
// PRE-CONDITIONS: The archive object must be opened.
// 
// POST-CONDITIONS: No change to internal state of Archive object.
// 
// ERRORS:
//  error_No_free_store
//  misc. Drive errors.
// 
// NOTES:
//  This is used for writing more than for reading VDSes. It ignores
//  whether or not HP wrote the other descriptors, or even the one it
//  finds.
//
//----------------------------------------------------------------------

NSR::Boolean
NSR::Archive::searchVolDescSeq( sid_t sid,
                                TagType searchDescType,
                                VDSInfo & vdsi )
{
    // Local variables
    SearchDescriptor sd;

    // Fill in return information if Error.
    vdsi.mainAddr     = 0;
    vdsi.resvAddr     = 0;
    vdsi.volDescSeqNo = 0;

    const UINT32 lastSector = currentDevice->numSectors(sid) - 1;

    // err = error_None handled by following line.
    if (err = currentDevice->error())
        return False;

    // Allocate a single sector buffer 
    BYTE *bp = new BYTE[logicalSectorSize];
    if (!bp) {
        err = error_No_free_store;
        return False;
    }

    // AVD Stuff
    // Get AVD if there.
    AnchorVolumeDesc *avdp = new AnchorVolumeDesc;
    if (!avdp) {
        err = error_No_free_store;
        delete [] bp;
        return False;
    }
    readSector( sid, LOC_AVD1, bp);
    if (err) {
        delete avdp;
        delete [] bp;
        return False;
    }
    avdp->fromISO( bp, LOC_AVD1);
    if (err=avdp->error()) {
        readSector( sid, lastSector, bp);
        if (err) {
            delete avdp;
            delete [] bp;
            return False;
        }
        avdp->fromISO( bp, lastSector );
        if (err=avdp->error()) {
            delete avdp;
            delete [] bp;
            return False;
        }
    }

    // Grab VDS information as Extents
    Extent  mainVDSExt;
    Extent  resvVDSExt;

    mainVDSExt.start.sid    = sid;
    mainVDSExt.start.sector = avdp->mainVolDescSeqExt.location;
    mainVDSExt.len          = avdp->mainVolDescSeqExt.length;

    resvVDSExt.start.sid    = sid;
    resvVDSExt.start.sector = avdp->resvVolDescSeqExt.location;
    resvVDSExt.len          = avdp->resvVolDescSeqExt.length;  

    // Nuke AVD and buffer
    delete avdp;
    delete [] bp;

    // Make sure main and reserve match exactly!
    if ( mainVDSExt.len != resvVDSExt.len ) {
        err = error_VDS_mirror_bad;
        return False;
    }

    // Instantiate vars to hold info for both sequences.
    UINT32  mainVolDescSeqNo = 0;
    UINT32  resvVolDescSeqNo = 0;

    Boolean foundMain;
    Boolean foundResv;

    // Search both sequences
    foundMain = searchForDesc(searchDescType,mainVDSExt,mainVolDescSeqNo);
    if ( err )
        return False;
    foundResv = searchForDesc(searchDescType,resvVDSExt,resvVolDescSeqNo);
    if ( err )
        return False;

    // Make sure they are the same
    if (( mainVolDescSeqNo != resvVolDescSeqNo ) ||
        ( foundMain != foundResv )) {
        err = error_VDS_mirror_bad;
        return False;
    }

    // Fill in return information.
    vdsi.mainAddr     = mainVDSExt.start.sector;
    vdsi.resvAddr     = resvVDSExt.start.sector;
    vdsi.volDescSeqNo = mainVolDescSeqNo;

    // Return whether found or not.
    return foundMain;
}

//----------------------------------------------------------------------
// METHOD: Archive::writeVolRecogSeq  --DVM
//      
//  Writes the BEA/NSRD/TEA Volume Recognition Sequence.
//
//  This method is WORM-intolerant. That is, if the sectors have
//  already been written on a WORM surface, this will set the error,
//  rather than assuming that it has already been written correctly.
//
// ARGS:
//  sid_t   sid     IN  Surface Identifier
//
// RETURNS: none.
// 
// PRE-CONDITIONS: Must have opened the Archive.
// 
// POST-CONDITIONS: none. No changes are made to the state of the
//  Archive object beyond its error possibly being set.
// 
// ERRORS:
//  error_No_free_store
//  any errors coming from the device.
// 
// NOTES:
//
//----------------------------------------------------------------------

void
NSR::Archive::writeVolRecogSeq(sid_t sid)
{
    // Initialize error state, rather than propagating it.
    err = error_None;

    // Set proper location
    VolAreaCalc volAreas(sid,logicalSectorSize,256);// Bogus value .. not needed.
    
    UINT32 sector = volAreas.volRecogSeqLoc().sector;

    // Initialize nBytes, nSectors.
    VolStructDesc *vsd = new VolStructDesc;
    if (!vsd) {
        err = error_No_free_store;
        return;
    }

    const UINT32 nBytes   = vsd->descLen();
    const UINT32 nSectors = B2S(nBytes);

    delete vsd;

    // Allocate a buffer nBytes long.
    BYTE *bp = new BYTE[nBytes];
    if ( bp == NULL ) {
        err = error_No_free_store;
        return;
    }

    // Write the BEA
    BegExtAreaDesc *bea = new BegExtAreaDesc;
    if (!bea) {
        err = error_No_free_store;
        delete [] bp;
        return;
    }
    memset( bp, 0, nBytes);
    bea->toISO( bp );
    if (err=bea->error()) {
        delete bea;
        delete [] bp;
        return;
    }
    for ( UINT16 i=0; i<nSectors; i++ ) {
        writeSector( sid, sector++, bp + (i*logicalSectorSize));
        if (err) {
            delete bea;
            delete [] bp;
            return;
        }
    }
    delete bea;

    // Write the NSRD
    NSRDescriptor *nsrd = new NSRDescriptor;
    if (!nsrd) {
        err = error_No_free_store;
        delete [] bp;
        return; }
    memset( bp, 0, nBytes);
    nsrd->toISO( bp );
    if (err=nsrd->error()) {
        delete nsrd;
        delete [] bp;
        return;
    }
    for ( i=0; i<nSectors; i++ ) {
        writeSector( sid, sector++, bp + (i*logicalSectorSize) );
        if (err) {
            delete nsrd;
            delete [] bp;
            return;
        }
    }
    delete nsrd;

    // Write the TEA
    TermExtAreaDesc *tea = new TermExtAreaDesc;
    if (!tea) {
        err = error_No_free_store;
        delete [] bp;
        return;
    }
    memset( bp, 0, nBytes);
    tea->toISO( bp );
    if (err=tea->error()) {
        delete tea;
        delete [] bp;
        return;
    }
    for ( i=0; i<nSectors; i++ ) {
        writeSector( sid, sector++, bp + (i*logicalSectorSize) );
        if (err) {
            delete tea;
            delete [] bp;
            return;
        }
    }
    delete tea;
    delete [] bp;
    return;
}

//----------------------------------------------------------------------
// METHOD: Archive::readVolRecogSeq  --DVM
//      
//  Reads the VRS. Error indicates if OK
//
//  Tries to read the VRS, and stops with ANY error. error_None
//  indicates that the entire VRS was found.
//
// ARGS:
//  sid_t  sid   IN    Surface Identifier
//
// RETURNS:
//  True if all found. Otherwise False.
// 
// PRE-CONDITIONS: none.
// 
// POST-CONDITIONS: Sets the error member with its results
// 
// ERRORS:
//  misc device errors.
// 
// NOTES:
//
//----------------------------------------------------------------------
NSR::Boolean
NSR::Archive::readVolRecogSeq(sid_t sid)
{
    // Initialize error state, rather than propagating it.
    err = error_None;

    // Set proper location
    VolAreaCalc volAreas(sid,logicalSectorSize,256);// Bogus value .. not needed.
    
    UINT32 sector = volAreas.volRecogSeqLoc().sector;

    // Allocate a VSD
    VolStructDesc *vsd = new VolStructDesc;
    if (!vsd) {
        err = error_No_free_store;
        return False;
    }

    // Initialize nBytes, nSectors.
    const UINT32 nBytes   = vsd->descLen();
    const UINT32 nSectors = B2S(nBytes);

    // Allocate a buffer nBytes long.
    BYTE *bp = new BYTE[nBytes];
    if ( bp == NULL ) {
        err = error_No_free_store;
        delete vsd;
        return False;
    }

    // BEA Stuff
    // Read the BEA
    for ( UINT16 i=0; i<nSectors; i++ ) {
        readSector( sid, sector++, bp + (i*logicalSectorSize));
        if (err) {
            delete vsd;
            delete [] bp;
            return False;
        }
    }
    // Decode the BEA
    vsd->fromISO( bp );
    if (err=vsd->error()) {
        delete vsd;
        delete [] bp;
        return False;
    }
    // Check to be sure it is a BEA
    if ( vsd->standardId() != String("BEA01") ) {
        err = error_Bad_volume_recog_sequence;
        delete vsd;
        delete [] bp;
        return False;
    }

    // NSRD Stuff
    // Read the NSRD
    for ( i=0; i<nSectors; i++ ) {
        readSector( sid, sector++, bp + (i*logicalSectorSize));
        if (err) {
            delete vsd;
            delete [] bp;
            return False;
        }
    }
    // Decode the NSRD
    vsd->fromISO( bp );
    if (err=vsd->error()) {
        delete vsd;
        delete [] bp;
        return False;
    }
    // Check to be sure it is a NSRD
    if ( vsd->standardId() != String("NSR02") ) {
        err = error_Bad_volume_recog_sequence;
        delete vsd;
        delete [] bp;
        return False;
    }

    // TEA Stuff
    // Read the TEA
    for ( i=0; i<nSectors; i++ ) {
        readSector( sid, sector++, bp + (i*logicalSectorSize));
        if (err) {
            delete vsd;
            delete [] bp;
            return False;
        }
    }
    // Decode the TEA
    vsd->fromISO( bp );
    if (err=vsd->error()) {
        delete vsd;
        delete [] bp;
        return False;
    }
    // Check to be sure it is a EA
    if ( vsd->standardId() != String("TEA01") ) {
        err = error_Bad_volume_recog_sequence;
        delete vsd;
        delete [] bp;
        return False;
    }
    
    // Nuke the VSD and buffer
    delete vsd;
    delete [] bp;

    // You win!! If got here, found valid VRS.
    return True;
}

//----------------------------------------------------------------------
// METHOD: Archive::ensureVolRecogSeq  --DVM
//      
//  Makes sure that a volume has a Volume Recognition Sequence
//
//  If the media is Rewritable, it simply writes a volume recognition
//  sequence. If it is Worm, it tries to read the VRS. If it succeeds,
//  it returns. If it does not, and the error was Blank_check, it
//  tries to write it.
//
// ARGS:
//  Volume   vol  IN   Volume to write to.
//
// RETURNS: none.
// 
// PRE-CONDITIONS: none.
// 
// POST-CONDITIONS: error state is set if any problems occurred.
// 
// ERRORS:
//  error_No_free_store
//  misc device errors.
// 
// NOTES:
//
//----------------------------------------------------------------------
void
NSR::Archive::ensureVolRecogSeq( const Volume & vol ) 
{
    // WORM is biggest issue here.
    if ( media == media_Worm ) {
        // Try reading VRS and return if OK
        if ( readVolRecogSeq(vol.sid()) ) {
            return;
        }
        // Clear blank check err, otherwise return.
        if ( err == error_Drive_blank_check ) {
            err = error_None;
        } else {
            return;
        }
    }
    // Write it.
    writeVolRecogSeq(vol.sid());
    return;
}


//----------------------------------------------------------------------
// METHOD: Archive::writeTD()  --DVM
//      
//  Writes a Terminating Descriptor at the supplied address.
//
// ARGS:
//  sid_t   sid      IN  Surface Identifier
//  UINT32  logAddr  IN  address to be encoded into descriptor tag
//  UINT32  addr     IN  physical sector in which to write the TD.
//
// RETURNS: none.
// 
// PRE-CONDITIONS: none.
// 
// POST-CONDITIONS: none.
// 
// ERRORS:
//  error_No_free_store
//  any drive Error
// 
// NOTES:
//  This allocates space on the heap for the one-sector-long buffer,
//  and for the TD itself. Both are freed before it returns.
//
//  Both addresses are required, so that it may be used in the VDS
//  (part 3) and in the FSDS (part 4).
//
//----------------------------------------------------------------------
void
NSR::Archive::writeTD(sid_t sid,UINT32 logAddr,UINT32 addr)
{
    // Allocate and Zero buffer one sector long
    BYTE_PTR bp = new BYTE[logicalSectorSize];
    if (!bp) {
        err = error_No_free_store;
        return;
    }
    memset( bp, 0, logicalSectorSize );

    // Allocate and fill in in-core TD
    TerminatingDesc *td = new TerminatingDesc;
    if (!td) {
        err = error_No_free_store;
        delete [] bp;
        return;
    }
    // Write Main TD
    td->toISO( bp, logAddr );
    if (err=td->error()) {
        delete td;
        delete [] bp;
        return;
    }
    writeSector( sid, addr, bp );
    if (err) {
        delete td;
        delete [] bp;
        return;
    }
    // Nuke in-core TD
    delete td;
    // Free buffer
    delete [] bp;
    return;
}

//----------------------------------------------------------------------
// METHOD: Archive::writeVDS     --DVM
//      
//  This writes the Volume label which goes onto each volume.
//
//  This writes the following:
//  - 2 AVDPs
//  - 2 PVDs
//  - 2 PDs
//    - 1 USE (WORM) or 2 USEs (MO)
//  - 2 OVDs
//  - 2 USDs
//  - 2 TDs (if not WORM)
//
// ARGS:
//  UINT16  volNum IN  Number of volume to be labeled.
//
// RETURNS: none.
// 
// PRE-CONDITIONS: none.
// 
// POST-CONDITIONS: none.
// 
// ERRORS:
//  error_No_free_store
//  drive errors.
// 
// NOTES:
//  This does no checking to see if a label already exists. If the
//  sectors have already been written on WORM, it will set an error.
//
//  This only writes to the HP pre-defined locations. It does NOT
//  append to existing VDSes or even look at the AVDPs on the media,
//  if any.
//
//  COULD pass in fileDataStart, plus whatever else is needed, rather
//  than giving access to entire vol. This would allow vol to change
//  without impacting this method. 
//----------------------------------------------------------------------
void
NSR::Archive::writeVDS( Volume & vol )
{
    // !!! writeVDS: Needs to verifyBlank() if WORM

    // Define some useful local variables
    UINT32 mainAddr;            // sector number
    UINT32 resvAddr;            // sector number
    UINT32 volDescSeqNo = 0;    // Very important.

    // --------------------------------------------
    // Sort out positions:

    // --------------------------------------------
    //  First: volume areas:
    VolAreaCalc volAreas(vol.sid(),
                         logicalSectorSize,
                         currentDevice->numSectors(vol.sid()));
    
    // --------------------------------------------
    // Next: Partition areas
    PartitionAreaCalc partAreas(vol.sid(),
                                logicalBlockSize,
                                media,
                                vol.fileDataStart);

    // - Define Partition itself (Based upon volAreas)
    vol.setPartition( vol.vid,
                      volAreas.partitionArea().start.sector,
                      volAreas.partitionArea().len / logicalBlockSize);


    // Allocate a buffer one sector long
    BYTE * bp = new BYTE[logicalSectorSize];
    if ( bp == NULL ) {
        err = error_No_free_store;
        return;
    }

    // --------------------------------------------
    // AVD Stuff
    // Allocate and fill in in-core AVD
    AnchorVolumeDesc *avd = new AnchorVolumeDesc;
    if (!avd) {
        err = error_No_free_store;
        delete [] bp;
        return;
    }
    avd->mainVolDescSeqExt.location = volAreas.mainVDSArea().start.sector;
    avd->mainVolDescSeqExt.length   = volAreas.mainVDSArea().len;
    avd->resvVolDescSeqExt.location = volAreas.resvVDSArea().start.sector;
    avd->resvVolDescSeqExt.length   = volAreas.resvVDSArea().len;

    // Set up main and reserve addresses.
    mainAddr = volAreas.mainAVDLoc().sector;
    resvAddr = volAreas.resvAVDLoc().sector;

    memset( bp, 0, logicalSectorSize );

    // Write the Main AVD
    avd->toISO( bp, mainAddr );
    if (err=avd->error()) {
        delete avd;
        delete [] bp;
        return;
    }
    writeSector( vol.sid(), mainAddr, bp );
    if (err) {
        delete avd;
        delete [] bp;
        return;
    }
    // Write the Reserve AVD 
    avd->toISO( bp, resvAddr );
    if (err=avd->error()) {
        delete avd;
        delete [] bp;
        return;
    }
    writeSector( vol.sid(), resvAddr, bp );
    if (err) {
        delete avd;
        delete [] bp;
        return;
    }
    // Nuke the in-core AVD
    delete avd;

    // --------------------------------------------
    // PVD Stuff
    // Allocate and fill in the in-core PVD
    PrimaryVolumeDesc *pvd = new PrimaryVolumeDesc;
    if (!pvd) {
        err = error_No_free_store;
        delete [] bp;
        return;
    }
    pvd->volDescSeqNo     = volDescSeqNo++;
    pvd->priVolDescNo     = 0;
    pvd->volumeIdentifier = vol.name();
    pvd->volumeSeqNo      = vol.vid + 1;
    pvd->maxVolumeSeqNo   = volumeTable.getCount();
    pvd->volumeSetID      = volumeSetName;
    pvd->recordDateTime   = time(0); // now

    // Set up main and reserve addresses.
    mainAddr = volAreas.mainVDSArea().start.sector;
    resvAddr = volAreas.resvVDSArea().start.sector;
    
    memset( bp, 0, logicalSectorSize );

    // Write the Main PVD 
    pvd->toISO( bp, mainAddr );
    if (err=pvd->error()) {
        delete pvd;
        delete [] bp;
        return;
    }
    writeSector( vol.sid(), mainAddr, bp );
    if (err) {
        delete pvd;
        delete [] bp;
        return;
    }
    // Write the Reserve PVD 
    pvd->toISO( bp, resvAddr );
    if (err=pvd->error()) {
        delete pvd;
        delete [] bp;
        return;
    }
    writeSector( vol.sid(), resvAddr, bp );
    if (err) {
        delete pvd;
        delete [] bp;
        return;
    }
    // Nuke in-core PVD
    delete pvd;

    // Bump both addresses.
    mainAddr++;
    resvAddr++;

    // --------------------------------------------
    // PD Stuff
    // Allocate and fill in the in-core PD
    PartitionDesc *pd = new PartitionDesc;
    if (!pd) {
        err = error_No_free_store;
        delete [] bp;
        return; }

    pd->volDescSeqNo              = volDescSeqNo++;
    pd->partitionFlags.allocation = True;
    pd->partitionNo               = vol.vid;
    
    if ( media != media_Worm ) {
        // Pre-erased free space
        pd->partitionHdrDesc.
            unallocSpaceTable.logBlockNo ( partAreas.erasedICBAddr().start.sector
                                           - partStartSector );
        pd->partitionHdrDesc.
            unallocSpaceTable.extLength  ( partAreas.erasedICBAddr().len );
        pd->partitionHdrDesc.
            unallocSpaceTable.extType    ( ext_Recorded );

        pd->partitionHdrDesc.
            freedSpaceTable.logBlockNo ( partAreas.unerasedICBAddr().start.sector
                                         - partStartSector );
        pd->partitionHdrDesc.
            freedSpaceTable.extLength  ( partAreas.unerasedICBAddr().len );
        pd->partitionHdrDesc.
            freedSpaceTable.extType    ( ext_Recorded );
    } else {
        pd->partitionHdrDesc.
            unallocSpaceTable.logBlockNo ( partAreas.unerasedICBAddr().start.sector
                                           - partStartSector );
        pd->partitionHdrDesc.
            unallocSpaceTable.extLength  ( partAreas.unerasedICBAddr().len );
        pd->partitionHdrDesc.
            unallocSpaceTable.extType    ( ext_Recorded );
    }

    pd->accessType   = ( media == media_Worm ? WriteOnce : Rewritable );
    pd->partStartLoc = volAreas.partitionArea().start.sector;
    pd->partLength   = volAreas.partitionArea().len / logicalBlockSize;

    // Zero out buffer
    memset( bp, 0, logicalSectorSize );

    // Write Main PD
    pd->toISO( bp, mainAddr );
    if (err=pd->error()) {
        delete pd;
        delete [] bp;
        return;
    }
    writeSector( vol.sid(), mainAddr, bp );
    if (err) {
        delete pd;
        delete [] bp;
        return;
    }
    // Write Reserve PD
    pd->toISO( bp, resvAddr );
    if (err=pd->error()) {
        delete pd;
        delete [] bp;
        return;
    }
    writeSector( vol.sid(), resvAddr, bp );
    if (err) {
        delete pd;
        delete [] bp;
        return;
    }
    // Nuke in-core PD
    delete pd;

    // Bump both addresses
    mainAddr++;
    resvAddr++;

    // --------------------------------------------
    // OVD Stuff
    // Allocate an fill in OVD
    OSTAVolumeDesc *iuvd = new OSTAVolumeDesc;
    if (!iuvd) {
        err = error_No_free_store;
        delete [] bp;
        return;
    }
    iuvd->volDescSeqNo = volDescSeqNo++;
    iuvd->logicalVolumeID = volumeSetName;
    
    // Zero out buffer
    memset( bp, 0, logicalSectorSize );

    // Write the main OVD
    iuvd->toISO( bp, mainAddr );
    if (err=iuvd->error()) {
        delete iuvd;
        delete [] bp;
        return;
    }
    writeSector( vol.sid(), mainAddr, bp );
    if (err) {
        delete iuvd;
        delete [] bp;
        return;
    }
    // Write the reserve OVD
    iuvd->toISO( bp, resvAddr );
    if (err=iuvd->error()) {
        delete iuvd;
        delete [] bp;
        return;
    }
    writeSector( vol.sid(), resvAddr, bp );
    if (err) {
        delete iuvd;
        delete [] bp;
        return;
    }
    // Nuke in-core OVD
    delete iuvd;

    // Bump both addresses
    mainAddr++;
    resvAddr++;

    // --------------------------------------------
    // USD Stuff
    // Allocate and fill in in-core USD
    UnallocatedSpaceDesc *usd = new UnallocatedSpaceDesc;
    if (!usd) {
        err = error_No_free_store;
        delete [] bp;
        return;
    }
    usd->volDescSeqNo = volDescSeqNo++;

    // Zero buffer.
    memset( bp, 0, logicalSectorSize );

    // Write main USD
    usd->toISO( bp, mainAddr );
    if (err=usd->error()) {
        delete usd;
        delete [] bp;
        return;
    }
    writeSector( vol.sid(), mainAddr, bp );
    if (err) {
        delete usd;
        delete [] bp;
        return;
    }

    // Write Reserve USD
    usd->toISO( bp, resvAddr );
    if (err=usd->error()) {
        delete usd;
        delete [] bp;
        return;
    }
    writeSector( vol.sid(), resvAddr, bp );
    if (err) {
        delete usd;
        delete [] bp;
        return;
    }
    // Nuke in-core USD
    delete usd;

    // Bump both addresses
    mainAddr++;
    resvAddr++;

    // Free buffer
    delete [] bp;

    // --------------------------------------------
    // TD Stuff
    // Write Both Terminating Descriptors
    if (media != media_Worm) {
        // logicalAddr == physicalAddr
        writeTD(vol.sid(),mainAddr,mainAddr);
        writeTD(vol.sid(),resvAddr,resvAddr);
        if (err) {
            return;
        }
    }

    // --------------------------------------------
    // USE Stuff

#ifdef DEBUG_SPACE              // !!! Apparently for Andy's testing purposes
    Boolean debugSpace = False;
    if (debugSpace) {
        for (int i = 0; i < 1000; i++) 
        {
            Extent crap;
            crap.start.sector = partStartSector + i;
            crap.len = logicalSectorSize;
            vol.freeSpace.append(crap);
        }
    }
#endif
    // Write out the free space information.
    if(Trace::logAlgorithms()) {
        fprintf(Trace::logFile(),"writeVDS(): calling writeSpaceEntry() on UnerasedSpace.\n");
    }
    writeSpaceEntry(vol.sid(),ext_Allocated,
                    vol.freeSpace,
                    partAreas.unerasedSpcTblArea());
    
    if (media != media_Worm) {
        if(Trace::logAlgorithms()) {
            fprintf(Trace::logFile(),"writeVDS(): calling writeSpaceEntry() on ErasedSpace.\n");
        }
        writeSpaceEntry(vol.sid(),ext_Allocated,
                        vol.erasedFreeSpace,
                        partAreas.erasedSpcTblArea());
    }
    return;
}

//----------------------------------------------------------------------
// METHOD:  Archive::writeVolumeLabel
//      
//  Writes the nsr volume information.
//
//  This does NOT check to see if the volume has a label. This should
//  not be called if the volume is already labeled.
//
//  This information is recorded below "fileDataStart". The information
//  consists of anchor volume descriptors, main and reserve volume 
//  descriptor sequences, and free space information.
//
//  This does NOT write the root volume label information, and ALWAYS
//  uses hard-coded addresses for the locations of its descriptors,
//  etc.
//
// ARGS:
//  Volume vol  IN      the volume to write nsr information to.
//
// RETURNS: none.
// 
// PRE-CONDS: open archive object, and volumeTable[vid].fileDataStart
//  must be filled in.
// 
// POST-CONDS: error state may be set.
// 
// ERRORS:
//  error_
// 
// NOTES:
//  This is used solely by addUnlabeledVolume().
//
//  LEGACY DOCUMENTATION:
//  If this volume is the root, an integrity descriptor sequence and
//  file set descriptor sequence will be written. In addition, the
//  volume descriptor sequence will have a logical volume descriptor.
//
//----------------------------------------------------------------------
void
NSR::Archive::writeVolumeLabel( Volume & vol )
{
    // Local references to vol information.
    sid_t  sid = vol.sid();

    if (err=vol.error()) {
        return;
    }

    // Make sure that a VolRecogSequence is on the disk.
    // NOTE that since destroyVolumeLabel does not erase this, we had
    // to check for it before writing anything.
    ensureVolRecogSeq(vol);
    if ( err ) {
        return;
    }

    // Write the main label information.
    writeVDS(vol);
    if (err) {
        return;
    }
}

//----------------------------------------------------------------------
// METHOD: Archive::writeLVD  --DVM
//      
//  Writes the LVD, using the info in the Archive class.
//
//  This supports WORM, in that it merely appends one whether or not
//  one already exists, or overwrites the one already there.
//
//  This does NOT write an FSDS, or an IntegrityEntry.
//
//  It gets the information to write from what is in Archive class,
//  which is filled in by the readvds() call, and thus matches the
//  contents of the prevailing LVD.
//
// ARGS:
//  Volume &  vol  IN  Volume Information to use.
//
// RETURNS: void
// 
// PRE-CONDITIONS: open Archive object
// 
// POST-CONDITIONS: error state may be set.
// 
// ERRORS:
//  error_
// 
// NOTES:
//  Mostly extracted from pre-alpha code.
//
//----------------------------------------------------------------------
void
NSR::Archive::writeLVD( void ) 
{
    // !!! Must be in export or labeling state (and complete).
    // Declare some local variables.
    BYTE    *bp;                // Simple buffer
    VDSInfo vdsi;               // Holds sequence info
    int i = 0;                  // counter
    Boolean hasLVD(False);      // Logic boolean
    
    // Create local references to root volume.
    Volume &vol = volumeTable.rootVol();
    sid_t   sid = vol.sid();

    // --------------------------------------------
    // Search for LVD
    // This sets vdsi to TD if no LVD... Just right!
    hasLVD = searchVolDescSeq( sid,tag_LVD,vdsi );
    if (err) {
        return;
    }
    
    // If found one on WORM, then needs to seek to end of sequence, so
    // that new one will get written to proper location.
    if ( hasLVD && (media == media_Worm) ) {
        Boolean hasTD = searchVolDescSeq( sid,tag_TD,vdsi );
        if ( ! hasTD ) {
            err = error_No_room_in_VDS;
            return;
        }
    }
    // Should be in proper location.

    // !!! Andyp - does FSD-writing part of writeVolumeLabel()
    // properly extend sequence if needed, or error if no room.

    // --------------------------------------------
    // LVD Stuff
    // Allocate and fill in in-core LVD
    LogicalVolumeDesc *lvd = new LogicalVolumeDesc(volumeTable.getCount());
    if (!lvd) {
        err = error_No_free_store;
        return;
    }
    lvd->volDescSeqNo           = vdsi.volDescSeqNo++;
    lvd->logicalVolumeID        = volumeSetName;
    lvd->logicalBlockSize       = logicalBlockSize;
    lvd->fileSetDesc.extType    ( ext_Recorded );

    lvd->fileSetDesc.partRefNo  (vol.vid);
    lvd->fileSetDesc.logBlockNo (vol.fsdsExt.start.sector - partStartSector);
    lvd->fileSetDesc.extLength  (vol.fsdsExt.len);

    lvd->integritySeqExtent.location = vol.lvidsExt.start.sector;
    lvd->integritySeqExtent.length   = vol.lvidsExt.len;

    PartitionMap *pm = lvd->partitionMapTable.entry;
    lvd->mapTableLength = (UINT32) (volumeTable.getCount() *
                                    pm[0].field.len());
    lvd->noPartMaps     = (UINT32) volumeTable.getCount();
    for (i=0; i<volumeTable.getCount(); i++) {
        pm[i].volumeSeqNo = i + 1;
        pm[i].partitionNo = i;
    }
    lvd->writeProtect(vol.lvWriteProtect);

    // Allocate and zero ISO-length buffer.
    UINT32 nSectors = B2S(lvd->descLen());
    bp = new BYTE[S2B(nSectors)];
    if (!bp) {
        err = error_No_free_store;
        delete lvd;
        return;
    }
    memset( bp, 0, S2B(nSectors));

    // Write Main LVD
    lvd->toISO( bp, vdsi.mainAddr );
    if (err=lvd->error()) {
        delete lvd;
        delete [] bp;
        return;
    }
    for (i=0; i < nSectors; i++) {
        writeSector( sid, vdsi.mainAddr+i, bp + (i*logicalSectorSize) );
        if (err) {
            delete lvd;
            delete [] bp;
            return;
        }
    }
    // Write Reserve LVD
    lvd->toISO( bp, vdsi.resvAddr );
    if (err=lvd->error()) {
        delete lvd;
        delete [] bp;
        return;
    }
    for (i=0; i < nSectors; i++) {
        writeSector( sid, vdsi.resvAddr+i, bp + (i*logicalSectorSize) );
        if (err) {
            delete lvd;
            delete [] bp;
            return;
        }
    }

    // Nuke in-core LVD and free buffer
    delete lvd;
    delete [] bp;

    // Bump both addresses
    vdsi.mainAddr += nSectors;
    vdsi.resvAddr += nSectors;

    // Write a Both Terminating Descriptors
    if (media != media_Worm) {
        writeTD(sid,vdsi.mainAddr,vdsi.mainAddr);
        writeTD(sid,vdsi.resvAddr,vdsi.resvAddr);
        if (err) {
            return;
        }
    }
    return;
}

//----------------------------------------------------------------------
// METHOD: Archive::writeFSD  --DVM
//      
//  This writes an FSD into a possibly empty FSDS.
//
//  It uses the information held in the Archive Object's VolumeTable
//  object.
//
// ARGS: none.
//
// RETURNS: none.
// 
// PRE-CONDITIONS:
//  icbTable must be in valid. currentAddr() is used for rootDirICB
//  field. 
// 
// POST-CONDITIONS: error state may be set.
// 
// ERRORS:
//  error_
//  error_
//  error_
// 
// NOTES:
//  This is pretty much lifted from the pre-alpha code.
//
//----------------------------------------------------------------------
void
NSR::Archive::writeFSD(void)
{
    // --------------------------------------------
    // Local Variables
    Volume & vol=volumeTable.rootVol();

    if (icbTable == NULL) {
        err = error_Invalid_ICB_table;
        return;
    }

    // --------------------------------------------
    // Allocate and Zero a buffer one-sector long
    BYTE_PTR bp = new BYTE[logicalSectorSize];
    if (!bp) {
        err = error_No_free_store;
        return;
    }
    memset( bp, 0, logicalSectorSize );

    // --------------------------------------------
    // FSD Stuff
    // Root ICB is in icbTable, so reset its internal state. Then,
    // it can supply an address for the root ICB.
    icbTable->resetPtr();

    // Allocate and fill in an in-core FSD
    FileSetDesc *fsd = new FileSetDesc;
    if (!fsd) {
        err = error_No_free_store;
        delete [] bp;
        return;
    }

    fsd->logicalVolumeID        = volumeSetName;
    fsd->fileSetID              = fileSetName;
    fsd->rootDirICB.extType    ( ext_Recorded );
    fsd->rootDirICB.extLength  ( logicalBlockSize );
    fsd->rootDirICB.logBlockNo ( icbTable->currentAddr().sector
                                 - partStartSector );
    fsd->rootDirICB.partRefNo  ( volumeTable.vid( icbTable->currentAddr().sid ) );

    UINT32 address;

    if (media != media_Worm) {
        address            = vol.fsdsExt.start.sector;
        fsd->fileSetNo     = 0;
        fsd->fileSetDescNo = 1;
        fsd->writeProtect(wrprot_Soft);
    } else {
        Extent fsdsExt = vol.fsdsExt;

        UINT32 nRead            = 0;
        UINT32 lastDescNum      = 0;
        UINT32 lastRecordedAddr = 0;
        UINT32 nextBlankAddr    = 0;
        Boolean needsIndirect   = False;
        if(Trace::logAlgorithms()) {
            fprintf(Trace::logFile(),"writeFSD: fsd search at %d len:%d\n",fsdsExt.start.sector,fsdsExt.len);
        }
        getNextSeqAddress(vol.sid(),fsdsExt,1,&nRead,&nextBlankAddr,&lastDescNum,
                          &lastRecordedAddr,&needsIndirect,&bogusICBInfo);
        if (err) {
            if (err == error_Drive_blank_check && media == media_Worm) {
                err = error_None;
            }
        }
        if (needsIndirect) {
            if (!partitionSeqAllowed) {
                err = error_End_of_Sequence_space;
                delete fsd;
                delete [] bp;
                return;
            }
            else {
                getSeqExtent(vol.sid(),incrementalSeqSectors,fsdsExt); 
                if (err) {
                    delete fsd;
                    delete [] bp;
                    return;
                }
                fsd->nextExtent.extLength(fsdsExt.len);
                fsd->nextExtent.logBlockNo(fsdsExt.start.sector - partStartSector);
            }
        }
        address            = nextBlankAddr;
        fsd->fileSetNo     = lastDescNum;
        fsd->fileSetDescNo = lastDescNum + 1;
        fsd->writeProtect(wrprot_Hard);
    }

    if(Trace::logAlgorithms()) {
        fprintf(Trace::logFile(),
                "writeFSD(): writing to address %d, fileSetNo %d, fileSetDescNum %d\n",
                address, (UINT32)fsd->fileSetNo, (UINT32)fsd->fileSetDescNo);
    }
    
    // Write FSD
    fsd->toISO( bp, address - partStartSector );
    if (err=fsd->error()) {
        delete fsd;
        delete [] bp;
        return;
    }
    writeBlock( vol.sid(), address, bp );
    if (err) {
        delete fsd;
        delete [] bp;
        return;
    }
    // Nuke in-core FSD
    delete fsd;
#ifdef _FSDS_TD_
    // !!! Shouldn't need this, since fsdsExt is only one sector long on Rewritable.
    // Write a Terminating Descriptor
    if (media != media_Worm) {
        address++;
        writeTD( vol.sid(),
                 address - partStartSector,
                 address );
        if (err) {
            delete [] bp;
            return;
        }
    }
#endif // _FSDS_TD_
    // Free buffer.
    delete [] bp;
    return;
}

//----------------------------------------------------------------------
// METHOD: Archive::updateSpace  --DVM
//      
//  This updates the free space information on the supplied volume.
//
// ARGS:
//  Volume vol IN   The volume to have its space updated. 
//
// RETURNS: none.
// 
// PRE-CONDITIONS: The media must be already labeled. This is only
//  valid in export mode.
// 
// POST-CONDITIONS: This may set the error state.
// 
// ERRORS:
//  error_
//  error_
//  error_
// 
// NOTES:
//  This does NOT close the logical volume at all. It ONLY updates the
//  free space on the surface.
//
//----------------------------------------------------------------------
void
NSR::Archive::updateSpace( const Volume & vol )
{
    // !!! check to be sure it is in the export mode. (or GrowVS?)
    // !!! Don't perform anything if certain errors???

    // Need extents to call readvds
    Extent uICBAddr;
    Extent eICBAddr;

    char volSetName[132];       // throw-away
    char volName[32];           // throw-away
    UINT16 nvols;               // throw-away
    UINT16 vid;                 // throw-away

    // Read the VDS in order to obtain the ICBAddresses of the USEs.
    if ( !readvds( vol.sid(), &vid, volSetName, volName,
                   NULL, NULL,&nvols,
                   &uICBAddr, &eICBAddr, NULL)) {
        err = error_Invalid_volume;
        return;
    }

    // Recalculate UnallocatedSpaceTable areas
    PartitionAreaCalc partAreas(vol.sid(),
                                media,
                                uICBAddr,
                                eICBAddr);

    // Grab current space recorded on disk.
    ExtentList unerasedOnDisk;
    ExtentList erasedOnDisk;
    
    getSpace(uICBAddr,unerasedOnDisk);
    if (media != media_Worm) {
        getSpace(eICBAddr,erasedOnDisk);
    }

    // Dump out free space
    if(Trace::logAlgorithms()) {
        fprintf(Trace::logFile(),"updateSpace() Unerased Space (on disk):\n");
        unerasedOnDisk.fprint(Trace::logFile());
        fprintf(Trace::logFile(),"updateSpace() Unerased Space (in memory):\n");
        vol.freeSpace.fprint(Trace::logFile());
        fprintf(Trace::logFile(),"updateSpace() Erased Space (on disk):\n");
        erasedOnDisk.fprint(Trace::logFile());
        fprintf(Trace::logFile(),"updateSpace() Erased Space (in memory):\n");
        vol.erasedFreeSpace.fprint(Trace::logFile());
    }

    // Write out space table entries only if they have changed.
    if ( vol.freeSpace != unerasedOnDisk ) {
        if(Trace::logAlgorithms()) {
            fprintf(Trace::logFile(),"Unerased space changed -- calling writeSpaceEntry()\n");
        }
        writeSpaceEntry(vol.sid(),ext_Allocated,
                        vol.freeSpace,
                        partAreas.unerasedSpcTblArea());
    }
    
    if ( (media != media_Worm) &&
         (vol.erasedFreeSpace != erasedOnDisk) ) {
        if(Trace::logAlgorithms()) {
            fprintf(Trace::logFile(),"Erased space changed -- calling writeSpaceEntry()\n");
        }
        writeSpaceEntry(vol.sid(),ext_Allocated,
                        vol.erasedFreeSpace,
                        partAreas.erasedSpcTblArea());
    }
    return;
}

//----------------------------------------------------------------------
// METHOD:  Archive::writeIntegrityEntry
//      
//  Writes an integrity entry to the specified volume.
//
//  IntegrityEntries are only valid for root volumes. This is offered
//  to write to any volume because as volume sets are grown, the old
//  roots may be marked CLOSED, and should be marked as OPEN as soon
//  as the new root is marked as OPEN.
//
//  LVIDs vary in size based on how many partitions are in their
//  logical volume. This code sets the appropriate size, so that it
//  matches the logical volume which references it.
//
// ARGS:
//  Volume & vol              IN  Volume to write integrity entry to.
//
// RETURNS: none.
// 
// PRE-CONDS:
//  The volume must have a valid lvidsExt for this to work.
//
// POST-CONDS:
//  An error may be set.
//
// ERRORS:
//  error_
// 
// NOTES: 
// 
//----------------------------------------------------------------------
void
NSR::Archive::writeIntegrityEntry( Volume & vol )
{
    BYTE *bp = new BYTE[logicalSectorSize];
    if (!bp) {
        err = error_No_free_store;
        delete [] bp;
        return;
    }
    UINT32  nextBlankAddr;
    UINT32  lastRecordedAddr;
    UINT32  lastDescNum;
    UINT32  nRead;
    Boolean needsIndirect = False;

    // First, create an appropriately-sized LVID.
    UINT32 nvols = vol.lvNumVols;
    LogicalVolumeIntegrityDesc *lvid;
    lvid = new LogicalVolumeIntegrityDesc( nvols, 0 );
    if (!lvid) {
        err = error_No_free_store;
        delete [] bp;
        return;
    }

    // First, fill in all the fields which are the same.
    lvid->integrityType(vol.lvIntegrity);
    for ( unsigned i=0; i<nvols; i++ ) {
        // freeSpaceTable holds Blocks(Sectors), not Bytes.
        lvid->freeSpaceTable[i] = B2S(volumeTable[i].freeSpace.totalBytes());
        lvid->sizeTable[i]      = volumeTable[i].partLength();
    }
    lvid->uniqueID.low(  uniqueID.low() );
    lvid->uniqueID.high( uniqueID.high() );
    uniqueID.next();        // !!! Don't think this is right...

    // Next, fill in fields differently for CLOSED vs. OPEN
    if ( vol.lvIntegrity == integ_Closed ) {
        if ( nvols != volumeTable.getCount() ) {
            err = error_Not_root_volume;
            delete [] bp;
            delete lvid;
            return;
        }
        // Fill in fields
        lvid->numFiles       = numFiles;
        lvid->numDirectories = numDirs;

    } else {
        // Zero fields
        lvid->numFiles       = 0;
        lvid->numDirectories = 0;
    }

    // Update the volume table entry with the new implementID
    vol.lvIntegImplementID = lvid->implementID.id();
    
    // Write the LVID
    Extent lvidsExt = vol.lvidsExt;
    UINT32 address = lvidsExt.start.sector;
    if (media == media_Worm) {
        nRead = 0;
        lastDescNum = 0;
        lastRecordedAddr = 0;
        nextBlankAddr = 0;
        needsIndirect = False;
        getNextSeqAddress(lvidsExt.start.sid,lvidsExt,1,
                          &nRead,&nextBlankAddr,&lastDescNum,
                          &lastRecordedAddr,&needsIndirect,&bogusICBInfo);
        if (err) {
            if (err == error_Drive_blank_check && media == media_Worm) {
                err = error_None;
            } else {
                delete lvid;
                delete [] bp;
                return;
            }
        }
        if (needsIndirect) {
            //-------------------------------------------------
            // this is the last one we can write - lvid can't extend
            // into partition space.
            //-------------------------------------------------
            lvid->nextIntegrityExtent.length = 0;
            lvid->nextIntegrityExtent.location = 0;
        }
        address = nextBlankAddr;
    }
    // Zero buffer, and then translate to ISO.
    memset( bp, 0, logicalSectorSize );
    lvid->toISO( bp, address );
    if (err=lvid->error()) {
        delete lvid;
        delete [] bp;
        return;
    }
    // Write the sector
    writeSector( lvidsExt.start.sid, address, bp );
    currentDevice->flush();
    err = currentDevice->error();

    // Free the heap
    delete lvid;
    delete [] bp;
    return;
}

//----------------------------------------------------------------------
// METHOD: Archive::nukeDesc  --DVM
//      
//  Overwrites descriptor at supplied address if it is of the type
//  specified.
//
// ARGS:
//  tag    IN  Type of descriptor to nuke.
//  sid    IN  surface identifier
//  sector IN  Sector to read/write
//  addr   IN  Sector or block address of descriptor
//
// RETURNS:
//  number of sectors written (1 or 0)
// 
// PRE-CONDS:  (E.g., assumes)
//  List_preconds
// 
// POST-CONDS: (E.g., changes to class data members)
//  List_postconds
// 
// ERRORS:
//  Device errors may be set.
// 
// NOTES:
//
//----------------------------------------------------------------------
NSR::UINT16
NSR::Archive::nukeDesc( TagType descType,
                        sid_t sid,
                        UINT32 sector,
                        UINT32 addr ) 
{
    UINT16 ns = 0;               // number of sectors written

    if ( media == media_Worm ||
         currentDevice->mediaType(sid) == media_Worm ) {
        err = error_Invalid_function_for_media;
        return ns;
    }

    const UINT16 sectorSize = currentDevice->sectorSize(sid);

    // Allocate a sector-long buffer.
    BYTE *rbp = new BYTE[sectorSize];
    if(rbp == NULL) {
        err = error_No_free_store;
        return ns;
    }
    memset( rbp, 0, sectorSize);

    SearchDescriptor sd;

    readSector( sid, sector, rbp);

    // If blank, everything is OK!
    if (err == error_Drive_blank_check) {
        err = error_None;
        delete rbp;
        return ns;
    }

    // If some other problem it is a real error -- bail.
    if (err) {
        delete rbp;
        return ns;
    }

    sd.fromISO( rbp, addr );
    if ( sd.type() == descType ) { // If we are SURE it a
                                    // bona-fide descriptor of the
                                    // right type, ...
        // Make it invalid, but leave descriptor type
        memmove( rbp+4, "Nuked!", 6 ); // Overwrite cksum, serialno, CRC
        memset( rbp+10, 0, 6 );        // axe CRCLen, Loc
        writeSector( sid, sector, rbp );
        ns ++;
    }
    delete [] rbp;
    return ns;
}

//----------------------------------------------------------------------
// METHOD:  Archive::getFileSetInfo
//      
//  Grabs information about the file sets on the root volume.
//
// ARGS:
//   fslst      OUT     ptr to filesetlist.
//
// RETURNS:
// 
// PRE-CONDS:
// 
// POST-CONDS:
// 
// ERRORS:
//  error_No_free_store
//  error_Bad_tag_descriptor_type
// 
// NOTES:
//    This makes the most sense on WORM, where there can be many
//    file sets.
//----------------------------------------------------------------------
void
NSR::Archive::getFileSetInfo(FileSetList *fslst)
{
    // Grab a reference to the root Volume
    Volume & rootVol = volumeTable.rootVol();

    // Grabs fsdsExt from root volume, rather than using readvds.
    Extent fsdsExt = rootVol.fsdsExt;

    fslst->fsdSeqExtents.append(fsdsExt);

    FileSetDesc *fsd = new FileSetDesc;
    if (!fsd) {
        err = error_No_free_store;
        return;
    }
    BYTE *bp = new BYTE[logicalBlockSize];
    if (!bp) {
        err = error_No_free_store;
        delete fsd; 
        return;
    }
    
    UINT32 sector = fsdsExt.start.sector;
    Boolean done = False;
    FileSetInfo fileSetInfo;
    SearchDescriptor sd; 
    int j = 0;
    // Terminates on end of FSDS Extent
    while (!done &&
           (j*logicalBlockSize < fsdsExt.len) )
    {
        readSector( rootVol.sid(), sector+j, bp );
        if (err) {
            if ( err == error_Drive_blank_check ) { 
                // Terminates on Blank Sector
                err = error_None;
                done = True; 
            } else {
                // Terminates on read error
                delete fsd;
                delete [] bp;
                return;
            }
        } else {
            sd.fromISO( bp, sector+j-partStartSector );
            if (err=sd.error()) { 
                delete fsd;
                delete [] bp; 
                return; 
            }
            switch( sd.type() ) {
            case tag_FSD:
                // Terminates on ISO error on FSD
                fsd->fromISO( bp, sector+j-partStartSector);
                if (err=fsd->error()) { 
                    delete fsd;
                    delete [] bp; 
                    return; 
                }
                // Terminates on non-UDF FSD
                if ( ! fsd->domainID.isUDF() &&
                     ! fsd->domainID.isEarlyHP() ) {
                    err = error_Non_UDF_volume_set;
                    delete fsd;
                    delete [] bp;
                    return;
                }
                fileSetInfo.name = (const char *)(String)fsd->fileSetID;
                fileSetInfo.num  = fsd->fileSetNo;
                fileSetInfo.time = fsd->recordDateTime;
                fileSetInfo.fsdWriteProtect = fsd->writeProtect();
                fslst->append(fileSetInfo);
                if (err) {
                    delete fsd;
                    delete [] bp;
                    return;
                }
                if (fsd->nextExtent.extLength()) {
                    sector = fsd->nextExtent.logBlockNo()+partStartSector;
                    j = 0;
                    // Does NOT follow to another surface
                    fsdsExt.start.sid = rootVol.sid();
                    fsdsExt.start.sector = sector;
                    fsdsExt.len = fsd->nextExtent.extLength();
                    fslst->fsdSeqExtents.append(fsdsExt);
                }
                else {
                    j++;
                }
                break;
            case tag_TD:
                // Terminates on TD
                done = True;
                break;
            default:
                // Terminates on unknown Descriptor type
                err  = error_Bad_tag_descriptor_type;
                done = True;
                delete fsd;
                delete [] bp; 
                return;
            }
        }
    }
    delete fsd;
    delete [] bp;
    return;
}

//----------------------------------------------------------------------
// METHOD:  Archive::bytesForEAHD
//      
//  Returns bytes required for an extended attribute header descriptor.
//
// ARGS:
//
// RETURNS:
//  Returns bytes required for an extended attribute header descriptor.
// 
// ERRORS:
// 
// NOTES: import & export
//----------------------------------------------------------------------
NSR::UINT32
NSR::Archive::bytesForEAHD(void)
{
    return static_eahd.descLen();
}

//----------------------------------------------------------------------
// METHOD:  Archive::bytesForTrusteesEA
//      
//  Calculates number of bytes required for writing a netware trustee EA.
//
// ARGS:
//    trustees    IN
//
// RETURNS:
// 
// PRE-CONDS:  
// 
// POST-CONDS: 
// 
// ERRORS:
// 
// NOTES: export
//----------------------------------------------------------------------
NSR::UINT32
NSR::Archive::bytesForTrusteesEA(const NetwareTrusteeInfo &trustees)
{
#ifdef TRUSTEES
    return static_ntea.attrLen();
#else
    return 0;
#endif // TRUSTEES
}

NSR::UINT32
NSR::Archive::bytesForNovellAttrEA()
{
#ifdef NETWARE_ATTR
    static NovellAttrEA naea;
    return naea.ISOLen();
#else
    return 0;
#endif // NETWARE_ATTR
}

//----------------------------------------------------------------------
// METHOD:  Archive::getICBExtentList
//      
//    Attempts to provide the caller an extent list for use in writing ICB's.
//
//    The need for this arises when writing extended attributes before the 
//    fe itself is written. The archive object must get a set of appropriate
//    addresses to use in writing a non-embedded EA. This may or may not
//    be contiguous, hence the extent "list". This is also used by
//    writeNextICB() to get locations for overflow sectors.
//
//    There is really no allocation going on here. Sectors are taken from
//    the icb table.
//
// ARGS:
//    UINT32 offset,         IN     See explanation for "forEA".
//    ExtentList &elist,     OUT    Where to write.
//    UINT32 nBytes,         OUT    Total number of bytes required .
//    Boolean forEA          IN     If getting an extent list for an EA,
//                                  use icb table addresses offset by 
//                                  "offset". The current icb table address
//                                  must be preserved for writing the ICB.
//
// RETURNS:
// 
// PRE-CONDS: 
//      icbTable must be valid.
// 
// POST-CONDS:
//      icbTable        If not getting space for an EA, the icbTable will
//                      get bumped.
//      endICBTable     Set to True if reach end of icb table.
// 
// ERRORS:
//  error_End_of_ICB_table
// 
// NOTES: export
//----------------------------------------------------------------------
void
NSR::Archive::getICBExtentList( UINT32 offset, ExtentList &elist, UINT32 nBytes,
                                Boolean forEA )
{
    Extent ext;
    UINT32 nSectors = B2S(nBytes);
    
    ext.len = logicalBlockSize;

    if (!forEA) endICBTable = False;
    for ( unsigned i=0; i<nSectors; i++ ) 
    {
        if (forEA) {
            ext.start = icbTable->addrAtOffset(offset+i);
        }
        else {
            ext.start = icbTable->currentAddr();
        }
        if (err = icbTable->error()) {
            if (!forEA && err == error_End_of_ICB_table) {
                endICBTable = True;
            }
            return;
        }
        if ( i == (nSectors-1) ) {
            if ( nBytes % logicalBlockSize ) {
                ext.len = nBytes % logicalBlockSize;
            }
            else {
                ext.len = logicalBlockSize;
            }
        }
        elist.append(ext);
        if (!forEA) icbTable->bumpPtr( 1 );
    }
}


//----------------------------------------------------------------------
// METHOD:  Archive::writeEAs
//      
//  Write the extended attributes associated with the file or directory.
//
//  Write the extended attributes. Will try to write all of the EA's in the
//  file entry itself if possible, otherwise will write an extended attribute
//  fe and then an ea.
//
//      embedded EA:
//      *-----------------------------*
//      | FE                          |
//      |    eahd                     |
//      |    ea1 (icb table)          |
//      |    ea2 (netware trustee)    |
//      *-----------------------------*
//
//
//      non-embedded EA:
//      *------------*----------*--------------------*
//      | FE         |   EA FE  | EA space           |
//      |            |   ad --->| eahd,ea1,ea2,...   |
//      *------------*----------*--------------------*
//
//
//  For 512 byte media, non-embedded EA's will be used whenever there are more 
//  than 320 bytes for extended attributes.
//
// ARGS:
//    sid_t sid             IN    The surface identifier for a volume.
//    UINT32 eafe_offset    IN    Sector offset for non-embedded EA - must
//                                leave room for ICB and overflow sectors.
//    FileEntry *fe         IN    file or directory icb that may have an 
//                                associated extended attribute.
//    trustees              IN    The netware trustee information to place in EA
//    Boolean isRootICB     IN    If true, write the icb table as an EA.
//    UINT32 nEASectors     IN    Number of sectors used up in writing the ea.
//                                This is needed in order to know how many 
//                                sectors to bump the icb table pointer.
//
// RETURNS:
// 
// PRE-CONDS:  
// 
// POST-CONDS: 
// 
// ERRORS:
// 
// NOTES: export
//----------------------------------------------------------------------
void
NSR::Archive::writeEAs(sid_t sid,
                       UINT32 eafe_offset,
                       FileEntry *fe, 
                       const NetwareTrusteeInfo &trustees,
                       Boolean &isRootICB,
                       UINT32 &nEASectors,
                       Attributes attr)
{
    nEASectors = 0;
    
    if (trustees.empty() && !isRootICB ) return;
    
    Boolean embeddedEA = True;
    UINT32 nBytesForEA = 0;
    FileEntry *eafe = NULL;
    ICBTableEA *icbea = NULL;
    ICB_Table *eaicbtbl = NULL;
    BYTE *eabp = NULL;
    BYTE *eap = NULL;
    ExtentList elist;
    ExtentList elistea;
#ifdef TRUSTEES
    NovellTrusteeEA *ntea = NULL;
#endif // TRUSTEES

    // We know there will be some ea's - calculate nBytesForEA.
    // Start with the extended attribute header descriptor
    ExtendedAttrHeaderDesc *eahd = new ExtendedAttrHeaderDesc();
    if (!eahd) {
        err = error_No_free_store;
        return;
    }
    nBytesForEA += eahd->descLen();
    eahd->implAttrLocation(eahd->descLen());
    
#ifdef TRUSTEES
    // create and fill in object to hold netware trustee info
    if ( ! trustees.empty() ) {
        ntea = new NovellTrusteeEA();
        if (!ntea) {
            err = error_No_free_store;
            delete eahd;
            return;
        }
        ntea->setTrusteeInfo(trustees);
        if (err) {
            delete eahd;
            delete ntea; 
            return;
        } 
        nBytesForEA += ntea->attrLen();
    }
#endif // TRUSTEES

    // create and fill in object to hold icb table if necessary
    if ( isRootICB ) {
        // add in the EA bytes for the icb table
        ExtentList icbTblelist = icbTable->location();
        icbea = new ICBTableEA(icbTblelist.getCount());
        if (!icbea) { 
            err = error_No_free_store; 
#ifdef TRUSTEES
            delete ntea; 
#endif // TRUSTEES
            delete eahd; 
            return; 
        }
        nBytesForEA += icbea->attrLen();
        
        // fill in the icb table.
        LongAllocDesc *icbt = icbea->icbTbl();
        for (int i=0; i<icbTblelist.getCount(); i++) 
        {
            icbt[i].partRefNo  ( volumeTable.vid( icbTblelist[i].start.sid ) );
            icbt[i].logBlockNo ( icbTblelist[i].start.sector - partStartSector );
            icbt[i].extLength  ( icbTblelist[i].len );
            icbt[i].extType    ( ext_Recorded );
            if(Trace::logAlgorithms()) {
                fprintf(Trace::logFile(),"icb table elist[%d] partRefNo:%d logBlockNo:%d sectors:%d\n",
                        i, volumeTable.vid( icbTblelist[i].start.sid), 
                        icbTblelist[i].start.sector - partStartSector,
                        B2S(icbTblelist[i].len));
            }
        }
    }

    nBytesForEA += (nBytesForEA % 4) ? 4 - (nBytesForEA % 4) : 0;
    eahd->applAttrLocation(nBytesForEA);

    if (embeddedEAs && !isRootICB && (nBytesForEA <= fe->unusedBytes(logicalBlockSize))) {
        embeddedEA = True;
        if (err = fe->setEALen(logicalBlockSize,nBytesForEA)) {
#ifdef TRUSTEES
            delete ntea; 
#endif // TRUSTEES
            delete icbea;
            delete eahd; 
            return; 
        }
    }
    else {
        embeddedEA = False;
        nEASectors = 1 + B2S(nBytesForEA);
        
        // get an extent to be used in writing the eafe
        getICBExtentList(eafe_offset, elist, logicalBlockSize, True);
        // get an extent list to be used in writing the easpace 
        getICBExtentList(eafe_offset+1, elistea, nBytesForEA, True);
        for (int idx = 0; idx < elistea.getCount(); idx++)
        {
            elist.append(elistea[idx]);
        }
        if (elist.error()) {
            err = error_Invalid_extentlist;
#ifdef TRUSTEES
            delete ntea; 
#endif // TRUSTEES
            delete icbea;
            delete eahd; 
            return; 
        }
        eaicbtbl = new ICB_Table( elist, logicalBlockSize );
        if (!eaicbtbl) {
            err = error_No_free_store;
#ifdef TRUSTEES
            delete ntea; 
#endif // TRUSTEES
            delete icbea;
            delete eahd; 
            return; 
        }
        eaicbtbl->resetPtr();
        
        eafe = new FileEntry(elistea.getCount());
        if (!eafe) {
            err = error_No_free_store;
#ifdef TRUSTEES
            delete ntea; 
#endif // TRUSTEES
            delete eaicbtbl;
            delete icbea;
            delete eahd; 
            return; 
        }
        // fill in fe->extendedAttrICB field (LongAllocDesc)
        fe->extendedAttrICB.partRefNo(volumeTable.vid(sid));
        fe->extendedAttrICB.logBlockNo(eaicbtbl->currentAddr().sector - partStartSector);
        if(Trace::logAlgorithms())
            fprintf(Trace::logFile(),"fe->extendedAttrICB.logBlockNo:%d\n",fe->extendedAttrICB.logBlockNo());
        fe->extendedAttrICB.extLength(logicalBlockSize);
        fe->extendedAttrICB.extType(ext_Recorded);
       
        attr.fileType(file_ExtendedAttributes);

        // write the eafe (recursion - be careful)
        // side effects of this call:
        //   - the eaicbtbl->bumpPtr() will be called as necessary
        Extent nullExtent;      // Constructor makes it all empty
        writeNextICB(attr,elistea,emptyTrustees,nullExtent,eaicbtbl,True);
        if (err) {
#ifdef TRUSTEES
            delete ntea; 
#endif // TRUSTEES
            delete eaicbtbl;
            delete icbea;
            delete eahd; 
            return; 
        }
    }

    // set up pointer to the EA space
    if (embeddedEA) {
        eabp = (BYTE *)fe->embeddedEAp();
    }
    else {
        eabp = new BYTE[S2B(B2S(nBytesForEA))];
        if (!eabp) {
            err = error_No_free_store;
#ifdef TRUSTEES
            delete ntea; 
#endif // TRUSTEES
            delete eaicbtbl;
            delete icbea;
            delete eahd; 
            return; 
        }
        memset( eabp, 0, S2B(B2S(nBytesForEA)));
    }
    eap = eabp;

    // fill in the extended attribute header descriptor
    if (embeddedEA) {
        eahd->toISO( eap, icbTable->currentAddr().sector - partStartSector );
    }
    else {
        eahd->toISO( eap, eaicbtbl->currentAddr().sector - partStartSector );
    }
    eap += eahd->descLen();

    if (icbea) {
        icbea->toISO( eap );
        eap += icbea->attrLen();
    }

#ifdef TRUSTEES
    if (ntea) {
        ntea->toISO( eap );
        eap += ntea->attrLen();
    }
#endif // TRUSTEES

    if (!embeddedEA) {
        writeExtents( sid, eaicbtbl, nEASectors-1, eabp);
    }

#ifdef TRUSTEES
    delete ntea; 
#endif // TRUSTEES
    delete [] eabp;
    delete eaicbtbl;
    delete icbea;
    delete eahd; 
    return; 
}

//----------------------------------------------------------------------
// METHOD:  Archive::readLVIDSeq
//      
//  Reads the logical volume integrity descriptor sequence.
//
// ARGS:
//    Volume & vol             IN    volume to be checked
//    LogicalVolumeIntegrityDesc **lvidpp,
//                             OUT   Holds last recorded lvid.
//    int *nFileSets           OUT   Number of open->close transitions.
//
// RETURNS:
// 
// PRE-CONDS:
// 
// POST-CONDS:
// 
// ERRORS:
//  error_
// 
// NOTES:
//    import & export
//
//----------------------------------------------------------------------
void
NSR::Archive::readLVIDSeq(const Volume & vol,
                          LogicalVolumeIntegrityDesc **lvidpp,
                          int *nFileSets)
{
    sid_t sid = vol.sid();
    Address integloc = vol.lvidsExt.start;
    int j=0;
    UINT32 nSectors;
    UINT32 sector = integloc.sector;
    SearchDescriptor sd;
    UINT32 lastInteg = integ_Open;
    LogicalVolumeIntegrityDesc *lvid = NULL;
   
    if (nFileSets) *nFileSets = 0;

    BYTE *bp = new BYTE[logicalBlockSize];
    if (!bp) {
        err = error_No_free_store;
        return;
    }
    if (*lvidpp)
        delete *lvidpp;
    
    lvid = new LogicalVolumeIntegrityDesc(1,0);
    if (!lvid) {
        err = error_No_free_store;
        delete [] bp;
        return;
    }
    Boolean done = False;
    do {
        readSector( sid, sector+j, bp );
        if (err)
            break;

        sd.fromISO( bp, sector+j );
        if (err=lvid->error()) {
            delete lvid;
            delete [] bp;
            return;
        }
        nSectors = B2S(sd.descLen());

        if ( nSectors > 1 ) {
            delete [] bp;
            bp = new BYTE[S2B(nSectors)];
            if (!bp) {
                err = error_No_free_store;
                delete lvid;
                return;
            }
            for (int i=0; i<nSectors; i++) {
                readSector(sid,sector+i+j,bp+(S2B(i)));
                if (err)
                    break;
            }
        }
        if (err) break;

        switch (sd.type()) {
        case tag_LVID:
            lvid->fromISO( bp, sector+j );
            if (err=lvid->error()) {
                delete lvid;
                delete [] bp;
                return;
            }
            if (lvid->integrityType() == integ_Closed) {
                if ((lastInteg == integ_Open) && nFileSets ) {
                    (*nFileSets)++;
                }
            }
            lastInteg = lvid->integrityType();
        
            if (lvid->nextIntegrityExtent.location) {
                sector = lvid->nextIntegrityExtent.location;
                j = 0;
            }
            else {
                j += nSectors;
            }
            break;
        case tag_TD:
            done = True;
            break;
        default:
            err = error_Bad_tag_descriptor_type;
            done = True;
            break;
        }
    } while (media == media_Worm && !done);

    if ((err == error_Drive_blank_check) && (media == media_Worm)) 
        err = error_None;

    // Don't delete lvid -- stuff it into lvidpp (which was already deleted).
    *lvidpp = lvid;
    delete [] bp;
    return;
}


//----------------------------------------------------------------------
// METHOD:  Archive::getSpace
//      
//  Returns the extent information from an unallocated space entry.
//
// ARGS:
//  Extent     spcTblExt  IN   location of the unallocated space table.
//  ExtentList space      OUT  extent information gets placed in here.
//
// RETURNS: none.
// 
// PRE-CONDS:
// 
// POST-CONDS: may set the error state.
// 
// ERRORS:
//  error_
// 
// NOTES:
//----------------------------------------------------------------------
void
NSR::Archive::getSpace(const Extent & spcTblExt,
                       ExtentList & space)
{
    // Clear out any space already on list.
    space.clear();
    
    // get the freed space on this volume from the use
    sid_t sid = spcTblExt.start.sid;
    UINT32 addr = spcTblExt.start.sector;

    if (media == media_Worm) {
        UINT32 nRead = 0;
        UINT32 nextBlankAddr = 0;
        UINT32 lastRecordedAddr = 0;
        UINT32 lastDescNum = 0;
        Boolean needsIndirect = False;
        getNextSeqAddress(sid,spcTblExt,0,&nRead,&nextBlankAddr,
                          &lastDescNum,&lastRecordedAddr,&needsIndirect,
                          &bogusICBInfo);
        if (err) {
            if (err == error_Drive_blank_check) {
                err = error_None;
            }
            return; 
        }
        addr = lastRecordedAddr;
    }
    
    BYTE *bp = new BYTE[logicalSectorSize];
    if (!bp) {
        err = error_No_free_store;
        return;
    }
    readSector( sid, addr, bp);
    if (err) {
        delete [] bp;
        return;
    }
    UnallocatedSpaceEntry *use = new UnallocatedSpaceEntry(1);
    if (!use) {
        delete [] bp;
        return;
    }
    use->fromISO( bp, addr - partStartSector);
    if (err=use->error()) {
        delete use;
        delete [] bp;
        return;
    }
    ShortAllocDesc *adt = use->shortADT();
    UINT32 noExtents = use->numADTEnts();
    AllocExtentDesc *aed = NULL;
    UINT32 i = 0;
    Address nextAED;
    Extent ext;
    
    while( i<noExtents )
    {
        if ( adt[i].isContinued() ) {
            nextAED.sector = adt[i].logBlockNo() + partStartSector;
            readBlock( sid, nextAED.sector, bp );
            if (err) {
                delete aed; 
                delete use;     // OK, since is NULL or valid.
                delete [] bp; 
                return; 
            }
            delete aed;
            aed = new AllocExtentDesc( 1, ad_Short);
            if (!aed) { err = error_No_free_store; delete [] bp; delete use; return; }
            aed->fromISO( bp, nextAED.sector - partStartSector );
            if (err = aed->error()) { 
                delete aed; 
                delete use;
                delete [] bp; 
                return; 
            }
            noExtents = aed->numADTEnts();
            adt       = aed->shortADT();
            i         = 0;
        } else {
            ext.start.sid   = sid;
            ext.start.sector = adt[i].logBlockNo() + partStartSector;
            ext.len          = adt[i].extLength();
            space.append( ext );
            if (err = space.error()) {
                delete aed;
                delete use;
                delete [] bp;
                return;
            }
            i++;
        }
    }
    delete aed;
    delete use;
    delete [] bp;
    return;
}

//----------------------------------------------------------------------
// METHOD:  Archive::readvds
//      
//  Read and return everything one might want to know about
//  a volume descriptor sequence.
//
// ARGS:
//   sid_t sid,                  IN    Surface ID.
//   UINT16 *vid,                OUT   archive ID for this surface ID.
//   char *volSetName,           OUT   volume set name 
//   char *volName,              OUT   volume name.
//   Extent *fsdsExt             OUT   Location of file set descriptor seq.
//   Address *integDescSeq,      OUT   Location of lvid sequence.
//   UINT16 *no_vols,            OUT   Number of volumes in volume set.
//   Extent *unerasedICBAddr     OUT   Locations of free space on volume.
//   Extent *erasedICBAddr       OUT   Locations of erased space on volume.
//   Boolean *hasAVD             OUT   True if avd is recorded at 256.
//   WriteProtectType *lvWriteProtect OUT  Hard/Soft/None
//
// RETURNS: True if valid volume.
// 
// PRE-CONDS:
// 
// POST-CONDS:
// 
// ERRORS:
// 
// NOTES:
//----------------------------------------------------------------------
NSR::Boolean
NSR::Archive::readvds(sid_t sid, UINT16 *vid, 
                      char *volSetName, char *volName,
                      Extent *fsdsExt, 
                      Extent *lvidsExt,
                      UINT16 *no_vols, 
                      Extent *unerasedICBAddr,
                      Extent *erasedICBAddr,
                      Boolean *hasAVD,
                      WriteProtectType *lvWriteProtect)
{
    // Fill in all non-null pointed-to structs with zeroes.
    if (fsdsExt)          fsdsExt->len           = 0;
    if (unerasedICBAddr)  unerasedICBAddr->len   = 0;
    if (erasedICBAddr)    erasedICBAddr->len     = 0;
    if (lvidsExt)         lvidsExt->start.sector = 0;
    if (no_vols)         *no_vols                = 0;
    if (hasAVD)          *hasAVD                 = False;
    if (lvWriteProtect)  *lvWriteProtect         = wrprot_None;

    // Make sure that if it is a non-root volume, it looks OK.
    // The logic here is to set to True the variables saying
    // descriptors are valid for all the optional descriptors.
    Boolean validAVD  = False,
            validPVD  = False,
            validPD   = False,
            validLVD  = True,   // Need not be root. 
            validIUVD = True,   // Compatible with older Archive
            validTD   = False,
            isHP_NSR  = False;
    
    UINT32 nSectors;

    // Instantiate pointers to the different descriptor types.
    SearchDescriptor     sd;
    AnchorVolumeDesc   *avd = NULL;
    PrimaryVolumeDesc  *pvd = NULL;
    LogicalVolumeDesc  *lvd = NULL;
    PartitionDesc       *pd = NULL;
    OSTAVolumeDesc    *iuvd = NULL;

    // Grab lastSector from Device.
    const UINT32 lastSector = currentDevice->numSectors(sid) - 1;
    if (err = currentDevice->error())
        return False;

    // Allocate a buffer one sector long.
    BYTE *bp = new BYTE[logicalSectorSize];
    if (!bp) {
        err = error_No_free_store;
        return False;
    }
    // Zero out buffer.
    memset(bp, 0, logicalSectorSize);

    // ---- Check to see if it has an AVD.
    // Instantiate an AVD
    avd = new AnchorVolumeDesc;
    // Error if failed.
    if (!avd) {
        err = error_No_free_store;
        delete [] bp;
        return False;
    }
    // Read the first AVD location
    readSector( sid, LOC_AVD1, bp);
    // Hold special case for WORM, and clear the error. (Blank sectors
    // error is zeroed out on WORM, but passed through as errors on
    // MO. Both paths are fine!
    Boolean wormBlank = ((err == error_Drive_blank_check) &&
                         (media == media_Worm))?True:False;
    if (wormBlank) {
        err = error_None;
    }

    // Bail on any other errors.
    if ( err ) {
        delete avd;
        delete [] bp;
        return False;
    }
    // Do FromISO if not wormBlank
    if (!wormBlank) {
        avd->fromISO( bp, LOC_AVD1);
        err = avd->error();
    }

    // Read the second location if first was blank or had ISO error.
    if (err || wormBlank) {
        // Do the read
        readSector( sid, lastSector, bp);

        // Hold special case for WORM, and clear the error.
        wormBlank = ((err == error_Drive_blank_check) &&
                     (media == media_Worm))?True:False;
        if (wormBlank) {
            err = error_None;
        }
        // Bail on any other errors.
        if ( err ) {
            delete avd;
            delete [] bp;
            return False;
        }
        // Do FromISO if not wormBlank
        if (!wormBlank) {
            avd->fromISO( bp, lastSector);
            err = avd->error();
        }
        // Bail on ISO error.
        if ( err ) {
            delete avd;
            delete [] bp;
            return False;
        }
    }
    // Has returned if any error. Also should return if wormBlank.
    if (wormBlank){
        delete avd;
        delete [] bp;
        return False;
    }
    // No errors, wormBlank False, so if hasAVD non-NULL, set to True,
    // and grab the fields from the descriptor.
    if (hasAVD)
        *hasAVD = True;
    
    UINT32 volDescSeqLoc = avd->mainVolDescSeqExt.location;
    UINT32 volDescSeqLen = avd->mainVolDescSeqExt.length;
    delete avd;
    validAVD = True;
    String label;
    
    for ( int i=0; i < B2S(volDescSeqLen); i++ ) 
    {
        delete [] bp;
        bp = new BYTE[logicalSectorSize];
        if (!bp) {
            err = error_No_free_store;
            return False;
        }
        // read next Volume Descriptor
        readSector( sid, volDescSeqLoc + i, bp );
        if (err) { 
            if ((err == error_Drive_blank_check) && (media == media_Worm)) {
                err = error_None;
                if (validAVD && validPVD && validPD && validLVD && validIUVD)
                    validTD = True;
                break;
            } else {
                delete [] bp;
                return False;
            }
        }

        sd.fromISO( bp, volDescSeqLoc + i );
        if ( err = sd.error() ) {
            delete [] bp;
            return False;
        }
        nSectors = B2S(sd.descLen());

        if ( nSectors > 1 ) {
            delete [] bp;
            bp = new BYTE[S2B(nSectors)];
            if (!bp) {
                err = error_No_free_store;
                return False;
            }
            for (int j=0; j < nSectors; j++) {
                readSector( sid, volDescSeqLoc+i+j,
                            bp + (j*logicalSectorSize));
                if (err) {
                    delete [] bp;
                    return False;
                }
            }
            sd.fromISO( bp, volDescSeqLoc + i );
            if ( err = sd.error() ) {
                delete [] bp;
                return False;
            }
        }

        switch( sd.type() ) {
        case tag_PVD:
            pvd = new PrimaryVolumeDesc;
            if (!pvd) { 
                err = error_No_free_store; 
                delete [] bp;
                return False; 
            }
            pvd->fromISO( bp, volDescSeqLoc + i );
            if (err=pvd->error()) { 
                delete pvd; 
                delete [] bp; 
                return False; 
            }
    
            label = pvd->volumeIdentifier;
            label.toMem( volName );
            label = pvd->volumeSetID;
            label.toMem( volSetName );
    
            *vid = pvd->volumeSeqNo - 1;
            if ( (pvd->implementID.id() == NSR::ImplRegID::HPID ) ||
                 (pvd->implementID.id() == NSR::ImplRegID::HPEarlyID ) ) {
                validPVD = True;
            }
            delete pvd;
            break;
        
        case tag_PD:
            pd = new PartitionDesc;
            if (!pd) { 
                err = error_No_free_store; 
                delete [] bp;
                return False; 
            }
            pd->fromISO( bp, volDescSeqLoc + i );
            if (err=pd->error()) { 
                delete pd; 
                delete [] bp; 
                return False; 
            }
            if ( (pd->implementID.id() == NSR::ImplRegID::HPID ) ||
                 (pd->implementID.id() == NSR::ImplRegID::HPEarlyID ) ) {
                validPD = True;
            }
            if (unerasedICBAddr) {
                if (media != media_Worm) {
                    unerasedICBAddr->start.sid = sid;
                    unerasedICBAddr->start.sector = pd->partitionHdrDesc.freedSpaceTable.logBlockNo()+partStartSector;
                    unerasedICBAddr->len = pd->partitionHdrDesc.freedSpaceTable.extLength();
                } else {
                    unerasedICBAddr->start.sid = sid;
                    unerasedICBAddr->start.sector = pd->partitionHdrDesc.unallocSpaceTable.logBlockNo()+partStartSector;
                    unerasedICBAddr->len = pd->partitionHdrDesc.unallocSpaceTable.extLength();
                }
            }
            
            if (erasedICBAddr && media != media_Worm) {
                erasedICBAddr->start.sid = sid;
                erasedICBAddr->start.sector = pd->partitionHdrDesc.unallocSpaceTable.logBlockNo()+partStartSector;
                erasedICBAddr->len = pd->partitionHdrDesc.unallocSpaceTable.extLength();
            }
            
            delete pd;
            break;

        case tag_LVD:
            lvd = new LogicalVolumeDesc;
            if (!lvd) { 
                err = error_No_free_store; 
                delete [] bp; 
                return False; 
            }
            lvd->fromISO( bp, volDescSeqLoc + i );
            if (err=lvd->error()) {
                delete lvd;
                delete [] bp;
                return False;
            }

            // Not valid if logical block size is not correct either.
            if (lvd->logicalBlockSize != logicalBlockSize) {
                err = error_Wrong_block_size_in_LVD;
                delete lvd;
                delete [] bp;
                return False;
            }
            
            // Capture the extent holding the LVID sequence
            if (lvidsExt) {
                lvidsExt->len          = lvd->integritySeqExtent.length;
                lvidsExt->start.sector = lvd->integritySeqExtent.location;
                lvidsExt->start.sid    = sid;
            }

            // Grab the number of volumes (== number of partition maps)
            *no_vols = lvd->noPartMaps;

            // Fill in the FileSetDescriptor Sequence Extent
            if (fsdsExt) {
                fsdsExt->start.sid = sid;
                fsdsExt->start.sector = lvd->fileSetDesc.logBlockNo()+partStartSector;
                fsdsExt->len = lvd->fileSetDesc.extLength();
                if(Trace::logAlgorithms()) {
                    fprintf(Trace::logFile(),"readvds: sid: %d\n fsdsExt:\n",sid);
                    fsdsExt->fprint(Trace::logFile());
                }
            }

            // Only valid if we wrote it.
            if ( (lvd->implementID.id() != NSR::ImplRegID::HPID ) &&
                 (lvd->implementID.id() != NSR::ImplRegID::HPEarlyID ) ) {
                validLVD = False;
            }
            // Grab whether the volume is write protected or not.
            if (lvWriteProtect) {
                *lvWriteProtect = lvd->writeProtect();
            }
            
            delete lvd;
            break;
        
        case tag_IUVD:
            iuvd = new OSTAVolumeDesc();
            if (!iuvd) { 
                err = error_No_free_store; 
                delete [] bp;
                return False; 
            }
            iuvd->fromISO( bp, volDescSeqLoc + i );
            if (err=iuvd->error()) { 
                delete iuvd; 
                delete [] bp; 
                return False; 
            }
            if ( (iuvd->implementID.id() != NSR::ImplRegID::HPID ) &&
                 (iuvd->implementID.id() != NSR::ImplRegID::HPEarlyID ) ) {
                validIUVD = False;
            }
            delete iuvd;
            break;

        case tag_TD:
            validTD = True;     
            break;
        
        default:
            // Simply ignore any descriptors of other types.
            // NOT being picky -- e.g., even an FSD in the sequence
            // would not cause any problems 
            break;
        }
        if ( sd.type() == tag_TD ) {
            break;
        }
    }

    delete [] bp;
    isHP_NSR = ( validAVD 
              && validPVD 
              && validPD 
              && validLVD 
              && validIUVD 
              && validTD ) ? True : False;

    return isHP_NSR;
}

//----------------------------------------------------------------------
// METHOD: Archive::getICBTableFromLVID  --DVM
//      
//  Returns True if the LVID has an ICBTable.
//  Sets the Archive object's error ONLY if it has trouble reading or
//  constructing objects.
//
// ARGS: none
//
// RETURNS: True if LVID has a valid ICBTable in it.
// 
// PRE-CONDITIONS: Must have a complete volume set loaded.
// 
// POST-CONDITIONS: error state may be set.
// 
// ERRORS:
//  error_
//  error_
//  error_
// 
// NOTES:
//
//----------------------------------------------------------------------
NSR::Boolean
NSR::Archive::getICBTableFromLVID(void) {
    // Clear out any errors.
    err = error_None;
    
    // Not there if on Worm (we never supported it.)
    if (media == media_Worm) {
        return False;
    }
    
    // Create an LVID.
    LogicalVolumeIntegrityDesc *lvid = NULL;
    int nFileSets = 0;
    readLVIDSeq( volumeTable.rootVol(),&lvid, &nFileSets );
    if (err) {
        delete lvid;
        return False;
    }

    // Only closed ones count
    if (lvid->integrityType() != integ_Closed) {
        delete lvid;
        return False;
    }

    // Only HP LVIDs count
    if ( ( lvid->implementID.id() != NSR::ImplRegID::HPID ) &&
         ( lvid->implementID.id() != NSR::ImplRegID::HPEarlyID ) ) {
        delete lvid;
        return False;
    }

    // Fill in the ICB Table if present in LVID.
    Boolean foundIt( lvid->hasICBTable() );
    if ( foundIt ) {
        if(Trace::logAlgorithms()) fprintf(Trace::logFile(),"lvid icb table\n");
        ExtentList elist;
        Extent ext;
        LongAllocDesc *icbt = lvid->icbTbl();
        for (int j=0; j<(int)lvid->numEnts(); j++) 
        {
            ext.start.sid   = volumeTable[icbt[j].partRefNo()].sid();
            ext.start.sector = icbt[j].logBlockNo() + partStartSector;
            ext.len          = icbt[j].extLength();
            elist.append( ext );

            if (err=elist.error()) {
                delete lvid;
                return False;
            }
        }
        // Reallocate ICBTable
        delete icbTable;
        icbTable = new ICB_Table( elist, logicalBlockSize );
        // Error if out of memory.
        if (!icbTable) {
            err = error_No_free_store;
            delete lvid;
            return foundIt;
        }
        // Error if icbTable in error.
        if (err=icbTable->error()) {
            delete lvid;
            return foundIt;
        }
    }
    delete lvid;
    return foundIt;
}

//----------------------------------------------------------------------
// METHOD:  Archive::getRootICB
//      
//    Gets the location of the root ICB. This works for Worm and
//    Rewritable
//
// ARGS:
//    INT32 fsnum       IN    File set number to find the root ICB of.
//                            File sets are numbered 0,1,2,...
//    Address &rootICB  OUT   Location of root ICB for fsnum.
//
// RETURNS: False if unable to obtain a valid address.
// 
// PRE-CONDS: 
// 
// POST-CONDS:
// 
// ERRORS:
//  The Archive::err state variable will be set if false, and NOT if
//  True.
// 
// NOTES:
//    import only. Could leave out the return value if we wanted,
//    since it is NOT checked, and always sets the error.
//
//----------------------------------------------------------------------
NSR::Boolean
NSR::Archive::getRootICB( INT32 fsnum,Address &rootICB )
{
    if(Trace::logAlgorithms()) {
        fprintf(Trace::logFile(),"getRootICB(fsnum = %d)\n",fsnum);
    }
    // Use a local variable for return value
    Boolean foundRootICB(False);
    
    // Instantiate some local variables
    // fsdsExt NEEDS to be value, since it will be changed.
    Extent fsdsExt = volumeTable.rootVol().fsdsExt;
    BYTE *bp;

    // Allocate the FSD
    FileSetDesc *fsd = new FileSetDesc;
    if (!fsd) {
        return False;
    }

    if (media != media_Worm) {
        // Allocate sector-size buffer
        bp = new BYTE[logicalBlockSize];
        if (!bp) {
            err = error_No_free_store;
            delete fsd;
            return False;
        }

        // Read the fsds extent
        readBlock( fsdsExt.start.sid,
                   fsdsExt.start.sector,
                   bp );
        if (err) {
            delete fsd;
            delete [] bp;
            return False;
        }
        // Translate it from ISO.
        fsd->fromISO( bp, fsdsExt.start.sector - partStartSector );
        if (err=fsd->error()) { 
            delete fsd; 
            delete [] bp; 
            return False;
        }
        // Error if not UDF or HP near-UDF
        if ( ! fsd->domainID.isUDF() &&
             ! fsd->domainID.isEarlyHP() ) {
            err = error_Non_UDF_volume_set;
            delete fsd;
            delete [] bp;
            return False;
        }
        // Save off GLOBAL state:

        // Set character set to be used by FIDs and PCs
        // (NOTE that WORM need not do this, since it never used OCU)
        NSR::FileIdentifierDesc::readCSType(fsd->fileSetCharSet.charSetType());
        NSR::PathComponent::readCSType(fsd->fileSetCharSet.charSetType());

        // Set rootICB, to be used by readRootICB()
        rootICB.sid = volumeTable[fsd->rootDirICB.partRefNo()].sid();
        rootICB.sector = fsd->rootDirICB.logBlockNo() + partStartSector;

        // Set the return value:
        foundRootICB = True;

        // Free the buffer.
        delete [] bp;
    } else { // Now for WORM:
        // Used to test for valid address at end
        rootICB.sector = 0;
        // Need to hold the rootICB in case default fsnum is used.
        Address lastRootICB;
        lastRootICB.sector = 0;

        //----------------------------------------------
        // Read previous fsd's, in chunks for efficiency
        //----------------------------------------------
        UINT32 nSectorsRead;
        Boolean done = False;
        while (!done)
        {
            //---------------------------------------------
            // Read an extent worth of FSD Sequence
            //---------------------------------------------
            bp = new BYTE[fsdsExt.len];
            if (!bp) {
                err = error_No_free_store;
                delete fsd;
                return False;
            }
            nSectorsRead = readExtent(fsdsExt, bp);
            if (err) {
                if ( err == error_Drive_blank_check ) {
                    err = error_None;
                    done = True; 
                }
                else {
                    delete fsd;
                    delete [] bp;
                    return False;
                }
            }

            //---------------------------------------------
            // convert fsd's, looking for match with fsnum 
            //---------------------------------------------
            for (int j = 0; j < nSectorsRead; j++)
            {
                fsd->fromISO( bp + (j*logicalSectorSize), 
                              fsdsExt.start.sector + j - partStartSector);
                if (err=fsd->error()) {
                    delete fsd;
                    delete [] bp;
                    return False;
                }
                // Error if not UDF or HP near-UDF
                if ( ! fsd->domainID.isUDF() &&
                     ! fsd->domainID.isEarlyHP() ) {
                    err = error_Non_UDF_volume_set;
                    delete fsd;
                    delete [] bp;
                    return False;
                }
                if(Trace::logAlgorithms()) {
                    fprintf(Trace::logFile(),"getRootICB(): saw fsnum=%d at sid=%d, sector=%d\n",
                            (UINT32)fsd->fileSetNo,
                            fsdsExt.start.sid,
                            fsdsExt.start.sector + j - partStartSector);
                }
                // Hold on to rootICB if default fileSetNo chosen (-1)
                if ( fsnum == -1) {
                    lastRootICB.sid = volumeTable[fsd->rootDirICB.partRefNo()].sid();
                    lastRootICB.sector = fsd->rootDirICB.logBlockNo() + partStartSector;
                    if(Trace::logAlgorithms()) {
                        fprintf(Trace::logFile(),
                                "getRootICB(): Held-root ICB at sid=%d; sect=%d\n",
                                (UINT32)lastRootICB.sid,
                                (UINT32)lastRootICB.sector);
                    }
                } else if (fsnum == fsd->fileSetNo) { 
                    // Break out with done==True if match            
                    rootICB.sid = volumeTable[fsd->rootDirICB.partRefNo()].sid();
                    rootICB.sector = fsd->rootDirICB.logBlockNo() + partStartSector;
                    if(Trace::logAlgorithms()) {
                        fprintf(Trace::logFile(),
                                "getRootICB(): FOUND-root ICB at sid=%d; sect=%d\n",
                                (UINT32)rootICB.sid,
                                (UINT32)rootICB.sector);
                    }
                    done = True;
                    break;
                }
                // Break out with done==False if Continued
                if (fsd->nextExtent.extLength()) {
                    if(Trace::logAlgorithms()) {
                        fprintf(Trace::logFile(),"getRootICB(): Continues");
                    }
                    done = False;
                    break;
                }
                // Break out with done==True if end of FSDSExtent
                if ( done && (fsnum == -1) && (j==(nSectorsRead-1))) {
                    rootICB.sid = volumeTable[fsd->rootDirICB.partRefNo()].sid();
                    rootICB.sector = fsd->rootDirICB.logBlockNo() + partStartSector;
                    if(Trace::logAlgorithms()) {
                        fprintf(Trace::logFile(),
                                "getRootICB(): FSDSEnd-rootICB at sid=%d,sect=%d\n",
                                (UINT32)rootICB.sid,
                                (UINT32)rootICB.sector);
                    }
                    done = True;
                    break;
                }
            }
            // If Not done, then grab continuation extent.
            if (!done) {
                //------------------------------
                // look for possible next extent
                //------------------------------
                // !!! This is a faulty algorithm -- doesn't support
                // !!! non-zero nextExtents in middle of sequence. 
                fsd->fromISO( bp + ((nSectorsRead-1)*logicalSectorSize), 
                             fsdsExt.start.sector + nSectorsRead - 1 - partStartSector);
                if (err=fsd->error()) {
                    delete fsd; 
                    delete [] bp; 
                    return False; 
                }
                // Error if not UDF or HP near-UDF
                if ( ! fsd->domainID.isUDF() &&
                     ! fsd->domainID.isEarlyHP() ) {
                    err = error_Non_UDF_volume_set;
                    delete fsd;
                    delete [] bp;
                    return False;
                }
                fsdsExt.start.sid = volumeTable[fsd->nextExtent.partRefNo()].sid();
                fsdsExt.start.sector = fsd->nextExtent.logBlockNo() + partStartSector;
                fsdsExt.len = fsd->nextExtent.extLength();
                if(Trace::logAlgorithms()) {
                    fprintf(Trace::logFile(),
                            "getRootICB(): Continuing at sid=%d, sector=%d, len=%d\n",
                            (UINT32)fsdsExt.start.sid,
                            (UINT32)fsdsExt.start.sector,
                            (UINT32)fsdsExt.len);
                }
                if (!fsdsExt.len) {
                    done = True;
                }
            }
            delete [] bp;
        }
        if (fsnum == -1) {
            rootICB = lastRootICB;
        }
        foundRootICB = (rootICB.sector == 0) ? False : True;
    }
    // Free the FSD
    delete fsd;

    // Set the proper error if no other error encountered, but file
    // set not found...
    if ( !foundRootICB && !err ) {
        err = error_File_set_not_found;
    }
    
    // Return
    return foundRootICB;
}

// Misc utilities to sort out volume sets...
// Returns the size of the volume set, independent of how many are
// missing.
NSR::UINT16
NSR::Archive::sizeVolSet( VolumeTable & vt )
{
    UINT16 i;                   // Index
    
    // Find the last occupied space in the table.
    // ( i is used as a base-one index here )
    for ( i = vt.getCount(); i > 0;i-- ) {
        // Stop search as soon as vt[i-1] is non-empty
        if ( vt[i-1].logicalBlockSize != 0 ) {
            break;
        }
    }
    return i;
}
    
NSR::UINT16
NSR::Archive::numContigVols( VolumeTable & vt )
{
    UINT16 numVols = 0;
    UINT16 i;
    // Search through vt, stopping at the first missing volume.
    for ( i=0; i < vt.getCount(); i++ ) {
        if ( vt[i].logicalBlockSize != 0 ) {
            numVols ++;
        } else {
            break;
        }
    }
    return numVols;
}
    
NSR::Boolean
NSR::Archive::isValidRoot( UINT16 volIndex ) 
{
    // Returns True if it has an LVD with a partition table the size
    // of the index so far.
    return ((volumeTable[volIndex].lvNumVols == volIndex + 1) ?
            True : False );
}
    
